/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2018 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Moritz Gelb, Felix Metzner                               *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#include <analysis/modules/FSRCorrection/ConservativeFSRCorrectionModule.h>

#include <framework/gearbox/Const.h>
#include <framework/logging/Logger.h>
#include <framework/datastore/RelationArray.h>

#include <analysis/dataobjects/Particle.h>
#include <mdst/dataobjects/MCParticle.h>
#include <mdst/dataobjects/Track.h>

#include <analysis/DecayDescriptor/ParticleListName.h>
#include <analysis/utility/PCmsLabTransform.h>

#include <cmath>
#include <algorithm>
#include <TMatrixFSym.h>

namespace Belle2 {

  REG_MODULE(ConservativeFSRCorrection)

  ConservativeFSRCorrectionModule::ConservativeFSRCorrectionModule() :
    Module()

  {
    setDescription("Takes the particles from the given lepton list copies them to the output list "
                   "and adds the 4-vector of the closest photon (considered as radiative) to the "
                   "lepton, if the given criteria for maximum angle and energy are fulfilled.");
    setPropertyFlags(c_ParallelProcessingCertified);

    addParam("leptonListName", m_leptonListName,
             "The lepton list containing the leptons to correct, should already exists.");
    addParam("gammaListName", m_gammaListName,
             "The gammas list containing possibly radiative gammas, should already exist.");
    addParam("correctedElectronListName", m_correctedElectronListName,
             "The newly created electron list where the corrected electrons are stored",
             m_correctedElectronListName);
    addParam("angleThreshold", m_angleThreshold,
             "The maximum angle (in degrees) between the lepton and the (radiative) gamma to be accepted.",
             m_angleThreshold);
    addParam("thetaThreshold", m_thetaThreshold,
             "The maximum theta (in degrees) between the lepton and the (radiative) gamma to be accepted."
             "Used only for arc method.",
             m_thetaThreshold);
    addParam("energyThreshold", m_energyThreshold,
             "The maximum energy of the (radiative) gamma to be accepted.",
             m_energyThreshold);
    addParam("momentumFactor", m_momentumFactor,
             "The maximum momentum of the (radiative) gamma with respect to the lepton momentum: "
             "p_gamma < momentumFactor * p_lepton.",
             m_momentumFactor);
    addParam("method", m_method_param,
             "The method which is used to look for potential photons. Can be either centeredCone, "
             "offsetCone or arc.",
             m_method_param);
    addParam("takeAllGammas", m_takeAllGammas,
             "Decides if all gammas which fulfill the criteria are used or only the best matching one.",
             m_takeAllGammas);
  }


  void ConservativeFSRCorrectionModule::initialize()
  {
    if (m_method_param == "centeredCone") {
      m_method = c_CenteredCone;
    } else if (m_method_param == "offsetCone") {
      m_method = c_OffsetCone;
    } else if (m_method_param == "arc") {
      m_method = c_Arc;
    } else {
      B2FATAL("The provided method " << m_method_param << " is not valid. "
              "Must be one of 'centeredCone', 'offsetCone' or 'arc'!");
    }

    StoreObjPtr<ParticleList>().isRequired(m_leptonListName);
    StoreObjPtr<ParticleList>().isRequired(m_gammaListName);

    //create Particle List where the corrected electrons are stored
    StoreObjPtr<ParticleList> eCorrList(m_correctedElectronListName);
    eCorrList.registerInDataStore();

    m_correctedPositronListName = ParticleListName::antiParticleListName(m_correctedElectronListName);
    StoreObjPtr<ParticleList> pCorrList(m_correctedPositronListName);
    pCorrList.registerInDataStore();

    m_maxAngle = cos(m_angleThreshold * M_PI / 180.0);
    m_maxTheta = cos(m_thetaThreshold * M_PI / 180.0);
  }


  void ConservativeFSRCorrectionModule::event()
  {
    StoreArray<Particle> particles;

    const StoreObjPtr<ParticleList> leptonList(m_leptonListName);
    const StoreObjPtr<ParticleList> gammaList(m_gammaListName);

    //Particle list for corrected electrons
    StoreObjPtr<ParticleList> eCorrList(m_correctedElectronListName);
    eCorrList.create();
    eCorrList->initialize(11, eCorrList.getName());

    StoreObjPtr<ParticleList> pCorrList(m_correctedPositronListName);
    pCorrList.create();
    pCorrList->initialize(-11, pCorrList.getName());

    eCorrList->bindAntiParticleList(*pCorrList);

    std::vector<int> usedGammas;

    // loop over leptons, and find suitable photons as candidates for FSR
    for (unsigned i = 0; i < leptonList->getListSize(); i++) {
      Particle* lepton = leptonList->getParticle(i);

      TLorentzVector lepton4Vector = lepton->get4Vector();
      TLorentzVector correctedLepton4Vec = lepton->get4Vector();
      TVector3 pLepton = lepton->getMomentum();

      double cosMax = -1.0;
      std::vector<Particle*> selectedGammas;

      // look for a possible FSR gamma
      for (unsigned j = 0; j < gammaList->getListSize(); j++) {
        Particle* gamma = gammaList->getParticle(j);

        // check if gamma energy is below threshold
        if (gamma->getEnergy() > m_energyThreshold) {
          continue;
        }

        // check if gamma momentum is valid:
        if (gamma->getMomentumMagnitude() > m_momentumFactor * lepton->getMomentumMagnitude()) {
          continue;
        }

        bool candidateFound = false;
        if (m_method == c_CenteredCone) {
          candidateFound = checkGammaWithCenteredCone(lepton, gamma, cosMax, m_takeAllGammas);
        } else if (m_method == c_OffsetCone) {
          candidateFound = checkGammaWithOffsetCone(lepton, gamma, cosMax, m_takeAllGammas);
        } else if (m_method == c_Arc) {
          candidateFound = checkGammaWithArc(lepton, gamma, cosMax, m_takeAllGammas);
        }

        if (candidateFound) {
          // check if gamma was already used
          if (std::find(usedGammas.begin(), usedGammas.end(), gamma->getMdstArrayIndex()) != usedGammas.end()) {
            continue;
          }
          selectedGammas.push_back(gamma);
        }
      }

      const TMatrixFSym& lepErrorMatrix = lepton->getMomentumVertexErrorMatrix();
      TMatrixFSym corLepMatrix(lepErrorMatrix);

      if (selectedGammas.size() > 0) {
        for (Particle* fsrGamma : selectedGammas) {
          correctedLepton4Vec += fsrGamma->get4Vector();
          usedGammas.push_back(fsrGamma->getMdstArrayIndex());
          fsrGamma->addExtraInfo("hasFSRCorrection", float(1));

          const TMatrixFSym& gammaErrorMatrix = fsrGamma->getMomentumVertexErrorMatrix();
          for (int irow = 0; irow <= 3; irow++) {
            for (int icol = irow; icol <= 3; icol++) corLepMatrix(irow, icol) += gammaErrorMatrix(irow, icol);
          }
        }
        Particle correctedLepton(correctedLepton4Vec, lepton->getPDGCode(), Particle::EFlavorType::c_Flavored,
                                 Particle::c_Track, lepton->getTrack()->getArrayIndex());


        correctedLepton.setMomentumVertexErrorMatrix(corLepMatrix);
        correctedLepton.setVertex(lepton->getVertex());
        correctedLepton.setPValue(lepton->getPValue());

        lepton->addExtraInfo("hasFSRCorrection", float(selectedGammas.size()));
        correctedLepton.addExtraInfo("fsrCorrected", float(selectedGammas.size()));

        correctedLepton.appendDaughter(lepton, false);
        for (Particle* fsrGamma : selectedGammas) {
          correctedLepton.appendDaughter(fsrGamma, false);
        }

        Particle* correctedElectron = particles.appendNew(correctedLepton);
        const MCParticle* mcLepton = lepton->getRelated<MCParticle>();
        if (mcLepton != nullptr) {
          correctedElectron->addRelationTo(mcLepton);
        }
        eCorrList->addParticle(correctedElectron);
      }
    }
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithCenteredCone(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax,
    bool takeAllGammas
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();
    TVector3 coneAxis = pLepton;
    double cosAngle = cos(coneAxis.Angle(pGamma));

    // check if angle is current best
    if (cosAngle > currentCosMax || takeAllGammas) {
      currentCosMax = cosAngle;
      // check if this angle fulfills the maxAngle bound
      if (cosAngle > m_maxAngle) {
        return true;
      }
    }
    return false;
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithOffsetCone(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax,
    bool takeAllGammas
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();
    TVector3 coneAxis = pLepton;
    coneAxis.SetPhi(pLepton.Phi() + lepton->getCharge() * m_angleThreshold * M_PI / 180.0);
    double cosAngle = cos(coneAxis.Angle(pGamma));

    // check if angle is current best
    if (cosAngle > currentCosMax || takeAllGammas) {
      currentCosMax = cosAngle;
      // check if this angle fulfills the maxAngle bound
      if (cosAngle > m_maxAngle) {
        return true;
      }
    }
    return false;
  }


  bool ConservativeFSRCorrectionModule::checkGammaWithArc(
    const Particle* lepton,
    const Particle* gamma,
    double& currentCosMax,
    bool takeAllGammas
  )
  {
    TVector3 pGamma = gamma->getMomentum();
    TVector3 pLepton = lepton->getMomentum();

    TVector3 arcAxis = pLepton;
    arcAxis.SetPhi(pLepton.Phi() + lepton->getCharge() * 1. / 3. * m_angleThreshold * M_PI / 180.0);

    double cosAngle = cos(arcAxis.Angle(pGamma));

    TVector3 arcAxisInPhiPlane = arcAxis;
    arcAxisInPhiPlane.SetTheta(0.);
    TVector3 arcAxisInThetaPlane = arcAxis;
    arcAxisInThetaPlane.SetPhi(0.);

    TVector3 pGammaInPhiPlane = pGamma;
    pGammaInPhiPlane.SetTheta(0.);
    TVector3 pGammaInThetaPlane = pGamma;
    pGammaInThetaPlane.SetPhi(0.);

    double cosPhi = cos(arcAxisInPhiPlane.Angle(pGammaInPhiPlane));
    double cosTheta = cos(arcAxisInThetaPlane.Angle(pGammaInThetaPlane));

    // check if this angle fulfills the maxAngle and maxTheta bound
    if (cosPhi > m_maxAngle and cosTheta > m_maxTheta) {
      // check if angle is current best
      if (cosAngle > currentCosMax || takeAllGammas) {
        currentCosMax = cosAngle;
        return true;
      }
    }
    return false;
  }
} // end Belle2 namespace

