"""
Constants and measured values from PDG or HFLAV used in the branching fraction calculations

Felix Metzner 2021
"""

import numpy as np
from uncertainties import ufloat, UFloat

from rdstar.branching_fractions.bf_utilities import BFMeasurement


__all__ = [
    "tau_p0",
    "f_pi",
    "two_body_iso_spin_ratio",
    "f_d1",
    "f_pipi",
    "three_body_iso_spin_ratio",
    # Theory R(D) and R(D*)
    "sm_r_d",
    "sm_r_d_st",
    # Old 2015 R(D) and R(D*)
    "belle_mc_r_d",
    "belle_mc_r_d_st",
    "huschle_r_d",
    "huschle_r_d_st",
    # Theory R(D**) values calculated by Florian
    "theo_r_dss_twost",
    "theo_r_dss_one",
    "theo_r_dss_oneprime",
    "theo_r_dss_zero",
    # Additional required measurement for D*_2 meson case
    "ratio_f_dst_2",
    "ratio_f_dst_2_to_D_pi_pi",
    # Unused constants
    "ratio_fpm_over_f00",
    "ratio_fpm",
    "ratio_f00",
    # Measured BF values (taken from the PDG Live Database on 4th May 2021)
    "bf_b0_to_d_l_nu_pdg_eval",
    "bf_bp_to_d_l_nu_pdg_eval",
    "bf_b0_to_d_st_l_nu_pdg_eval",
    "bf_bp_to_d_st_l_nu_pdg_eval",
    "bf_bp_to_d0_0_st_l_nu_with_d0_0_st_to_dp_pim",
    "bf_bp_to_dprime10_l_nu_with_dprime10_to_dstp_pim",
    "bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip",
    "bf_bp_to_d0_1_l_nu_with_d0_1_to_d_stp_pim",
    # Gap
    # Measurements to B -> D(*) pi l nu (resonant and non-resonant)
    "bf_b0_to_d_pi_l_nu_hfalv_eval",
    "bf_bp_to_d_pi_l_nu_hflav_eval",
    "bf_b0_to_d_st_pi_l_nu_hflav_eval",
    "bf_bp_to_d_st_pi_l_nu_hflav_eval",
    # Overall B -> D(*) pi pi l nu (resonant and non-resonant) from non-resonant BaBar measurement
    "ratio_Dpipi_to_D_babar",
    "ratio_Dstpipi_to_Dst_babar",
    "bf_b_to_d_pi_pi_l_nu_babar_non_res",
    "bf_b_to_d_st_pi_pi_l_nu_babar_non_res",
    # Other modes: B+ -> D_s(*) K l+ nu
    "bf_b_to_d_s_kaon_l_nu_measurement",
    "bf_b_to_d_st_s_kaon_l_nu_measurement",
    # Inclusive:
    "bf_b_to_xc_l_nu_hflav_combined",
    "bf_b0_to_xc_l_nu_inclusive",
    "bf_bp_to_xc_l_nu_inclusive",
    # B -> Xu l nu
    "bf_bp_to_pi0_l_nu_pdg",
    "bf_bp_to_eta_l_nu_pdg",
    "bf_bp_to_eta_prime_l_nu_pdg",
    "bf_bp_to_omega_l_nu_pdg",
    "bf_bp_to_rho_l_nu_pdg",
    "bf_b0_to_pi_l_nu_pdg",
    "bf_b0_to_rho_l_nu_pdg",
    "bf_bp_to_xu_l_nu_inclusive_lu_cao",
    "bf_b0_to_xu_l_nu_inclusive_lu_cao",
]


# Constants
# fmt: off
tau_p0 = ufloat(1.076, 0.004)  # type: UFloat  # lifetime ratio tau_B+ / tau_B0 (https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/)
f_pi = 2.0 / 3.0  # type: float  # f_pi = BF(D** -> D(*)- pi+) / BF(D** -> D(*) pi) = 2/3
two_body_iso_spin_ratio = 1.0 / f_pi  # type: float  # = 1 / f_pi
f_d1 = ufloat(2.32, 0.54)  # type: UFloat  # f_D1 = BF(D_1 -> D*- pi+) / BF(D_1 -> D0 pi+ pi-)
f_pipi = ufloat(1.0 / 2.0, 1.0 / 6.0)  # type: UFloat  # BF(D** -> D(*)- pi+ pi-) / BF(D** -> D(*) pi pi)
three_body_iso_spin_ratio = 1.0 / f_pipi  # type: UFloat  # = 1 / f_pipi

# Theory R(D) and R(D*)
sm_r_d = ufloat(nominal_value=0.299, std_dev=0.003)  # type: UFloat  # R(D) from HFLAV Eq. 224: https://arxiv.org/pdf/1909.12524
sm_r_d_st = ufloat(nominal_value=0.258, std_dev=0.005)  # type: UFloat  # R(D*) from HFLAV Eq. 225: https://arxiv.org/pdf/1909.12524

# Old R(D) and R(D*) values
belle_mc_r_d = ufloat(nominal_value=0.333, std_dev=0.01)  # type: UFloat # R(D) from Belle MC Dec File with ARBITRARY uncertainty
belle_mc_r_d_st = ufloat(nominal_value=0.266, std_dev=0.01)  # type: UFloat # R(D*) from Belle MC Dec File with ARBITRARY uncertainty

huschle_r_d = ufloat(nominal_value=0.375, std_dev=0.064)  # type: UFloat # R(D) from Huschle's 2015 Belle Result with stat. uncertainty
huschle_r_d_st = ufloat(nominal_value=0.293, std_dev=0.038)  # type: UFloat # R(D*) from Huschle's 2015 Belle Result with stat. uncertainty


# Predicted R(D**) ratios from Florian's Paper https://arxiv.org/pdf/1606.09300.pdf (Eq. 38)
theo_r_dss_twost = ufloat(0.07, 0.01)  # type: UFloat  # R(D*_2) from Eq. 38 in https://arxiv.org/pdf/1606.09300.pdf
theo_r_dss_one = ufloat(0.10, 0.02)  # type: UFloat  # R(D_1) from Eq. 38 in https://arxiv.org/pdf/1606.09300.pdf
theo_r_dss_oneprime = ufloat(0.06, 0.02)  # type: UFloat  # R(D'_1) = R(D*_1) from Eq. 38 in https://arxiv.org/pdf/1606.09300.pdf
theo_r_dss_zero = ufloat(0.08, 0.04)  # type: UFloat  # R(D_0) from Eq. 38 in https://arxiv.org/pdf/1606.09300.pdf

# Additional required measurement for D*_2 meson case (see https://confluence.desy.de/pages/viewpage.action?pageId=202394372 for details):
ratio_f_dst_2 = ufloat(1.54, 0.15)  # type: UFloat  # ratio f_D*_2 = BF(D*_2 - > D- pi+) / BF(D*_2 - > D*- pi+)
ratio_f_dst_2_to_D_pi_pi = ufloat(2.79, 1.47)  # type: UFloat  # ratio f_D*_2_to_Dpipi = BF(D*_2 - > D*- pi+) / BF(D*_2 - > D0 pi+ pi-)

# Unused constants
ratio_fpm_over_f00 = ufloat(nominal_value=1.058, std_dev=0.024)  # type: UFloat  # f+- / f00 from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/
ratio_fpm = ufloat(nominal_value=0.514, std_dev=0.006)  # type: UFloat  # f+- from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/
ratio_f00 = ufloat(nominal_value=0.486, std_dev=0.006)  # type: UFloat  # f00 from https://hflav-eos.web.cern.ch/hflav-eos/osc/PDG_2020/

# Measured BF values (taken from the PDG Live Database on 4th May 2021)
bf_b0_to_d_l_nu_pdg_eval = BFMeasurement(central_value=2.31e-2, stat_uncert=0.04e-2, sys_uncert=0.09e-2)  # B0 -> D+ l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=37&parCode=S042&home=MXXX045
bf_bp_to_d_l_nu_pdg_eval = BFMeasurement(central_value=2.35e-2, stat_uncert=0.03e-2, sys_uncert=0.09e-2)  # B- -> D0 l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=145&parCode=S041&home=MXXX045
# bf_b0_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.06e-2, stat_uncert=0.02e-2, sys_uncert=0.12e-2)  # B0 -> D*+ l- nu as used by Florian and Stephan
bf_b0_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.05e-2, stat_uncert=0.02e-2, sys_uncert=0.14e-2)  # B0 -> D*+ l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=181&parCode=S042&home=MXXX045
bf_bp_to_d_st_l_nu_pdg_eval = BFMeasurement(central_value=5.66e-2, stat_uncert=0.07e-2, sys_uncert=0.21e-2)  # B- -> D*0 l- nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=146&parCode=S041&home=MXXX045

bf_bp_to_d0_0_st_l_nu_with_d0_0_st_to_dp_pim = BFMeasurement(central_value=0.28e-2, stat_uncert=0.03e-2, sys_uncert=0.04e-2)  # B- -> D*_0^0 (-> D+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 77
bf_bp_to_dprime10_l_nu_with_dprime10_to_dstp_pim = BFMeasurement(central_value=0.28e-2, stat_uncert=0.06e-2, sys_uncert=0.0)  # B- -> D'1^0 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 76 w/o Belle measurement!
bf_bp_to_dst0_2_l_nu_with_dst0_2_to_dstm_pip = BFMeasurement(central_value=0.077e-2, stat_uncert=0.006e-2, sys_uncert=0.004e-2)  # B- -> D*_2^0 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 75
bf_bp_to_d0_1_l_nu_with_d0_1_to_d_stp_pim = BFMeasurement(central_value=0.281e-2, stat_uncert=0.01e-2, sys_uncert=0.015e-2)  # B- -> D0_1 (-> D*+ pi-) l- nu from https://arxiv.org/pdf/1909.12524 Table 74

# For Gap Calculations
# Measurements to B -> D(*) pi l nu (resonant and non-resonant)
bf_b0_to_d_pi_l_nu_hfalv_eval = BFMeasurement(central_value=0.406e-2, stat_uncert=0.047e-2, sys_uncert=0.0)  # B0 -> D0 pi+ l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_bp_to_d_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.443e-2, stat_uncert=0.037e-2, sys_uncert=0.0)  # B- -> D+ pi- l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_b0_to_d_st_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.565e-2, stat_uncert=0.061e-2, sys_uncert=0.0)  # B0 -> D*0 pi+ l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_bp_to_d_st_pi_l_nu_hflav_eval = BFMeasurement(central_value=0.589e-2, stat_uncert=0.044e-2, sys_uncert=0.0)  # B- -> D*+ pi- l- nu from Table 73 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf

# Overall B -> D(*) pi pi l nu (resonant and non-resonant) from non-resonant BaBar measurement
ratio_Dpipi_to_D_babar = BFMeasurement(central_value=0.067, stat_uncert=0.010, sys_uncert=0.008)  # R_pipi from https://arxiv.org/abs/1507.08303 Table 2
ratio_Dstpipi_to_Dst_babar = BFMeasurement(central_value=0.019, stat_uncert=0.005, sys_uncert=0.004)  # R*_pipi from https://arxiv.org/abs/1507.08303 Table 2

bf_b_to_d_pi_pi_l_nu_babar_non_res = BFMeasurement(central_value=152e-5, stat_uncert=23e-5, sys_uncert=np.sqrt(18e-5**2 + 7e-5**2))  # B -> D pi+ pi- l- nu from https://arxiv.org/abs/1507.08303 Table 2
bf_b_to_d_st_pi_pi_l_nu_babar_non_res = BFMeasurement(central_value=108e-5, stat_uncert=28e-5, sys_uncert=np.sqrt(23e-5**2 + 4e-5**2))  # B -> D* pi+ pi- l- nu from https://arxiv.org/abs/1507.08303 Table 2

# Other modes: B+ -> D_s(*) K l+ nu
# These modes are here not available in the Belle I MC and furthermore don't have large contributions (BF ~ 0.03e-2)
# Thus, they are not considered further.
bf_b_to_d_s_kaon_l_nu_measurement = BFMeasurement(central_value=3.0e-4, stat_uncert=0.9e-4, sys_uncert=0.5e-4 * (1.1 + 0.8))  # B+ -> D_s- K+ l+ nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=600&parCode=S041&home=MXXX045
bf_b_to_d_st_s_kaon_l_nu_measurement = BFMeasurement(central_value=2.9e-4, stat_uncert=1.6e-4, sys_uncert=0.5e-4 * (1.1 + 1.0))  # B+ -> D*_s- K+ l+ nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=571&parCode=S041&home=MXXX045

# Inclusive:
bf_b_to_xc_l_nu_hflav_combined = BFMeasurement(central_value=10.65e-2, stat_uncert=0.16e-2, sys_uncert=0.0)  # Inclusive B -> X_c l nu from Eq. 206 in HFLAV 2020 https://arxiv.org/pdf/1909.12524.pdf
bf_b0_to_xc_l_nu_inclusive = BFMeasurement(central_value=10.08e-2, stat_uncert=0.30e-2, sys_uncert=0.22e-2)  # Inclusive B0 -> X_c l nu from PDG https://pdglive.lbl.gov/BranchingRatio.action?desig=375&parCode=S042&home=MXXX045
bf_bp_to_xc_l_nu_inclusive = BFMeasurement(central_value=10.79e-2, stat_uncert=0.25e-2, sys_uncert=0.27e-2)  # Inclusive B+ -> X_c l nu from PDG https://pdglive.lbl.gov/BranchingRatio.action?desig=473&parCode=S041&home=MXXX045

# CKM suppressed semileptonic B decays
# B -> Xu l nu
bf_bp_to_pi0_l_nu_pdg = BFMeasurement(central_value=0.780e-4, stat_uncert=0.027e-4, sys_uncert=0.0)  # B+ -> pi0 l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=417&parCode=S041&home=MXXX045
bf_bp_to_eta_l_nu_pdg = BFMeasurement(central_value=0.39e-4, stat_uncert=0.05e-4, sys_uncert=0.0)  # B+ -> eta l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=327&parCode=S041&home=MXXX045
bf_bp_to_eta_prime_l_nu_pdg = BFMeasurement(central_value=0.23e-4, stat_uncert=0.08e-4, sys_uncert=0.0)  # B+ -> eta' l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=479&parCode=S041&home=MXXX045
bf_bp_to_omega_l_nu_pdg = BFMeasurement(central_value=1.19e-4, stat_uncert=0.09e-4, sys_uncert=0.0)  # B+ -> omega l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=173&parCode=S041&home=MXXX045
bf_bp_to_rho_l_nu_pdg = BFMeasurement(central_value=1.58e-4, stat_uncert=0.11e-4, sys_uncert=0.0)  # B+ -> rho0 l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=174&parCode=S041&home=MXXX045

bf_b0_to_pi_l_nu_pdg = BFMeasurement(central_value=1.50e-4, stat_uncert=0.06e-4, sys_uncert=0.0)  # B0 -> pi l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=150&parCode=S042&home=MXXX045
bf_b0_to_rho_l_nu_pdg = BFMeasurement(central_value=2.94e-4, stat_uncert=0.21e-4, sys_uncert=0.0)  # B0 -> rho l nu from https://pdglive.lbl.gov/BranchingRatio.action?desig=95&parCode=S042&home=MXXX045

bf_bp_to_xu_l_nu_inclusive_lu_cao = BFMeasurement(central_value=2.2e-3, stat_uncert=0.3e-3, sys_uncert=0.0)  # B+ -> Xu l nu from Table I in https://arxiv.org/pdf/2102.00020.pdf
bf_b0_to_xu_l_nu_inclusive_lu_cao = BFMeasurement(central_value=2.0e-3, stat_uncert=0.3e-3, sys_uncert=0.0)  # B0 -> Xu l nu from Table I in https://arxiv.org/pdf/2102.00020.pdf

# fmt: on
