"""
Classes handling Data MC Comparisons.

Felix Metzner 2021
"""


import os
import copy
import traceback

import numpy as np
import pandas as pd

from inspect import signature
from dataclasses import dataclass
from matplotlib import pyplot as plt
from typing import Optional, Union, List, Tuple, Dict, Any, TypeVar, Type

from templatefitter.plotter.plot_utilities import export, AxesType
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.binned_distributions.binning import Binning, BinsInputType
from templatefitter.binned_distributions.systematics import SystematicsInputType
from templatefitter.plotter.histogram_plots import DataMCHistogramPlot, DataMCComparisonOutputType

from analysistools.offline_analysis.sample_weights import belle_published_integrated_luminosity as belle_one_luminosity

import rdstar.offline_analysis.rdstar_plots as rdstar_plot_tools

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.systematics import SysColsInfo, get_systematics
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.validation.correlation_calculator import CorrCalcInfo, corr_eval_info_additional_info_key

from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollections

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeCollection

from rdstar.comparison_tools.scale_utilities import SampleScaleInfo, MCPreScaler
from rdstar.comparison_tools.select_utilities import SimpleCutInfo, ComplexCutInfo


__all__ = [
    "DataMCComparerInfo",
    "DataMCComparer",
]


logger = log_to_default(name=__name__)


DataMCComparerInfoType = TypeVar("DataMCComparerInfoType", bound="DataMCComparerInfo")


@dataclass(frozen=True)
class DataMCComparerInfo:
    key: str
    hist_var: HistVariable
    title: str
    info_text: str
    reco_mode_info_collection: RecoModeCollection = BSigRecoModes.main_modes
    mc_matching_collection: MCMatchingCollection = MCMatchingCollections.original__base__collection
    styles: Tuple[str, str] = ("summed", "stacked")
    text_h_pos: Union[float, Tuple[float, float]] = 0.66
    text_v_pos: Optional[Union[float, Tuple[float, float]]] = (0.88, 0.55)
    legend_loc: Optional[Union[int, str, Tuple[Union[int, str], Union[int, str]]]] = None
    legend_cols: Optional[Union[int, Tuple[int, int]]] = None
    fig_size: Tuple[float, float] = (5.5, 5.0)
    height_ratio: Tuple[float, float] = (3.5, 1.0)
    y_scale: Optional[Union[float, Tuple[float, float]]] = None
    split_chi_info_text: Union[bool, Tuple[bool, bool]] = True
    systematics: Optional[Tuple[SysColsInfo, ...]] = None
    sys_name_tag: Optional[str] = None
    file_name_appendix: Optional[str] = None
    requires_selection: Optional[Tuple[str, ...]] = None
    valid_sample_setups: Tuple[str, ...] = SampleSetupHandler.on_resonance_sample_setup_names
    data_component_label: str = "Data"
    lumi_label: str = r"$\int \mathcal{L} \,dt=" + f"{belle_one_luminosity}" + r"\,\mathrm{fb}^{-1}$"
    lumi_label_appendix: Optional[str] = None
    modes_to_skip: Optional[List[str]] = None
    use_short_titles: bool = False
    add_scaling_factor_to_info_text: bool = False
    proportional_rescaling_for_all_modes_case: bool = True
    reset_off_to_on_res_scaling: bool = False
    mc_pre_scaling: Optional[MCPreScaler] = None
    add_new_fei_calibration_factors: bool = False
    additional_selection_info: Optional[Union[SimpleCutInfo, Tuple[SimpleCutInfo], ComplexCutInfo]] = None
    additional_info: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        if not all(v in SampleSetupHandler.sample_setup_mapping for v in self.valid_sample_setups):
            raise ValueError(
                f"Values of parameter 'valid_sample_setups' must be keys of"
                f"SampleSetupHandler.sample_setup_mapping "
                f"(on of {tuple(SampleSetupHandler.sample_setup_mapping.keys())}), "
                f"but the provided values "
                f"{[k for k in self.valid_sample_setups if k not in SampleSetupHandler.sample_setup_mapping]}"
                f" are not!"
            )

    @classmethod
    def create_modified_comparer_info(
        cls: Type[DataMCComparerInfoType],
        comparer_info: DataMCComparerInfoType,
        **kwargs,
    ) -> DataMCComparerInfoType:
        comparer_info_signature = signature(cls)
        ci_param_names: List[str] = list(comparer_info_signature.parameters.keys())
        assert all(k in ci_param_names for k in kwargs.keys()), [k for k in kwargs.keys() if k not in ci_param_names]

        return cls(**{kw: copy.copy(kwargs.get(kw, getattr(comparer_info, kw))) for kw in ci_param_names})

    @classmethod
    def reset_systematics_in_comparer_info(
        cls: Type[DataMCComparerInfoType],
        comparer_info: DataMCComparerInfoType,
        new_systematics: Optional[Tuple[SysColsInfo, ...]],
        sys_name_tag: Optional[str],
    ) -> DataMCComparerInfoType:
        assert isinstance(comparer_info, DataMCComparerInfo), type(comparer_info)
        return cls.create_modified_comparer_info(
            comparer_info=comparer_info,
            systematics=copy.copy(new_systematics),
            sys_name_tag=sys_name_tag,
        )

    @classmethod
    def create_comparer_infos_with_systematics(
        cls: Type[DataMCComparerInfoType],
        comparer_info_list: List[DataMCComparerInfoType],
        systematics_collection: Tuple[SysColsInfo, ...],
        sys_name_tag: str,
    ) -> List[DataMCComparerInfoType]:
        assert isinstance(systematics_collection, tuple), type(systematics_collection)
        assert all(isinstance(sys, SysColsInfo) for sys in systematics_collection)
        return [
            cls.reset_systematics_in_comparer_info(
                comparer_info=c_info,
                new_systematics=systematics_collection,
                sys_name_tag=sys_name_tag,
            )
            for c_info in comparer_info_list
        ]

    @classmethod
    def get_default_lumi_label(cls) -> str:
        return r"$\int \mathcal{L} \,dt=" + f"{belle_one_luminosity}" + r"\,\mathrm{fb}^{-1}$"

    def get_sig_id_mapping_for(self, sig_id_col: str) -> Dict[str, Union[int, Tuple[int, ...]]]:
        return self.mc_matching_collection.get_sig_id_mapping_for(sig_id_column=sig_id_col)

    def get_reco_mode_title(self, mode_name: str, use_short_titles: bool) -> str:
        return self.reco_mode_info_collection.get_reco_mode_title(
            mode_name=mode_name,
            use_short_titles=use_short_titles,
        )

    @property
    def additional_info_dict(self) -> Dict[str, Any]:
        return dict() if self.additional_info is None else self.additional_info

    def get_additional_setting_with_default(self, setting_key: str, default_setting: Any) -> Any:
        return self.additional_info_dict.get(setting_key, default_setting)


class DataMCComparer:
    def __init__(
        self,
        comparer_info: DataMCComparerInfo,
    ) -> None:
        self._comparer_info: DataMCComparerInfo = comparer_info

    @property
    def comparer_info(self) -> DataMCComparerInfo:
        return self._comparer_info

    def plot_data_mc_comparison(
        self,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        weight_col: str,
        data_weight_col: Optional[str] = None,
        reco_mode_col: Optional[str] = None,
        lumi_label: bool = True,
        draw_legend: bool = True,
        legend_inside: bool = True,
        export_path: Optional[PathType] = None,
        use_short_titles: bool = False,
        style: str = "summed",
        ratio_type: str = "normal",
        fig_size: Optional[Tuple[float, float]] = None,
        height_ratio: Optional[Tuple[float, float]] = None,
        normalize_mc_to_data: bool = True,
        gof_check_method: Optional[str] = None,
        adaptive_binning: bool = False,
        dry_run: bool = False,
        show_plots: bool = False,
    ) -> Optional[Dict[str, PathType]]:
        if show_plots:
            assert not dry_run

        assert style in ("summed", "stacked"), style

        _use_short_titles: bool = use_short_titles

        systematics_cols: SystematicsInputType = None
        if self.comparer_info.systematics is not None and not dry_run:
            systematics_cols = get_systematics(df=mc_df, sys_cols_infos=self.comparer_info.systematics)

        appendix = (
            f"{self.comparer_info.file_name_appendix}_" if self.comparer_info.file_name_appendix is not None else ""
        )
        tag_info = f"{self.comparer_info.key}_" if self.comparer_info.key is not None else ""
        sys_tag = f"{self.comparer_info.sys_name_tag}_" if self.comparer_info.sys_name_tag is not None else ""
        file_name_base = f"data_vs_mc_{style}_{self.comparer_info.hist_var.df_label}_{appendix}{tag_info}{sys_tag}"

        output_dict: Dict[str, PathType] = {}
        excepted_errors: List[Tuple[Exception, str]] = []

        _fig_size: Tuple[float, float] = self.comparer_info.fig_size if fig_size is None else fig_size
        _height_ratio: Tuple[float, float] = self.comparer_info.height_ratio if height_ratio is None else height_ratio

        if reco_mode_col is None:
            _reco_mode_col: str = Cols.decay_mode_column_name
        else:
            _reco_mode_col = reco_mode_col

        if not dry_run:
            assert _reco_mode_col in mc_df.columns, _reco_mode_col
            assert _reco_mode_col in data_df.columns, _reco_mode_col

        for mode_name, mode_ids in self.comparer_info.reco_mode_info_collection.reco_mode_info_mapping_as_tuple:
            if self.comparer_info.modes_to_skip is not None:
                if mode_name in self.comparer_info.modes_to_skip:
                    continue

            _used_weight_col: str = weight_col
            if not dry_run and self.comparer_info.proportional_rescaling_for_all_modes_case:
                if data_weight_col is not None:
                    raise NotImplementedError(
                        f"Sample rescaling (parameter proportional_rescaling_for_all_modes_case = True) is not \n"
                        f"implemented if a data_weight_col is used (you use data_weight_col = '{data_weight_col}')\n"
                        f"Additional Info: The weight_col = '{weight_col}' is used."
                    )

                if mode_name in self.comparer_info.reco_mode_info_collection.special_reco_mode_id_group_names:
                    sample_scaler = SampleScaleInfo(
                        weight_column_name=weight_col,
                        reco_mod_column_name=_reco_mode_col,
                        comparison_hist_var=self.comparer_info.hist_var,
                    )

                    try:
                        sample_scaler.fill_from(
                            mc_df=mc_df,
                            data_df=data_df,
                            reco_mode_info_collection=self.comparer_info.reco_mode_info_collection,
                        )
                    except AssertionError as e:
                        raise AssertionError(
                            f"Error for observable {self.comparer_info.hist_var.df_label} "
                            f"and DataMCComparerInfo with key {self.comparer_info.key} "
                            f"and title {self.comparer_info.title}"
                        ).with_traceback(tb=e.__traceback__) from e
                    _used_weight_col = sample_scaler.rescale_sample_for_all_modes_case_only(
                        mc_sample=mc_df,
                        new_weight_column_name=SampleScaleInfo.default_tmp_weight_column_name,
                        correct_overall_normalization=False,
                    )

            try:
                output_entry: Optional[Tuple[str, PathType]] = self._plot_data_mc_comparison_for_mode(
                    mode_name=mode_name,
                    mode_ids=mode_ids,
                    mc_df=mc_df,
                    data_df=data_df,
                    weight_col=_used_weight_col,
                    data_weight_col=data_weight_col,
                    reco_mode_col=_reco_mode_col,
                    file_name_base=file_name_base,
                    systematics_cols=systematics_cols,
                    legend_info=(draw_legend, legend_inside),
                    use_short_titles=_use_short_titles or self.comparer_info.use_short_titles,
                    lumi_label=lumi_label,
                    export_path=export_path,
                    style=style,
                    ratio_type=ratio_type,
                    fig_size=_fig_size,
                    height_ratio=_height_ratio,
                    normalize_mc_to_data=normalize_mc_to_data,
                    gof_check_method=gof_check_method,
                    adaptive_binning=adaptive_binning,
                    dry_run=dry_run,
                    show_plots=show_plots,
                )

                if output_entry is not None:
                    file_name, output_path = output_entry  # type: str, PathType
                    assert file_name not in output_dict.keys(), file_name
                    output_dict.update({file_name: output_path})

            except Exception as e:
                logger.error(
                    f"Exception occurred for mode {mode_name} and variable {self.comparer_info.hist_var.df_label} "
                    f"with the following stack trace:\n{traceback.format_exc()}"
                )
                excepted_errors.append((e, traceback.format_exc()))

            if not dry_run and _used_weight_col != weight_col:
                del mc_df[_used_weight_col]

        if excepted_errors:
            raise Exception([type(e)(str(e) + "\n" + tbi) for e, tbi in excepted_errors])

        if export_path:
            return output_dict
        return None

    def _plot_data_mc_comparison_for_mode(
        self,
        mode_name: str,
        mode_ids: Tuple[int, ...],
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        weight_col: str,
        data_weight_col: Optional[str],
        reco_mode_col: str,
        file_name_base: str,
        systematics_cols: SystematicsInputType,
        legend_info: Tuple[bool, bool],
        use_short_titles: bool = False,
        lumi_label: bool = True,
        export_path: Optional[PathType] = None,
        style: str = "summed",
        ratio_type: str = "normal",
        fig_size: Tuple[float, float] = (5.5, 5.0),
        height_ratio: Tuple[float, float] = (3.5, 1.0),
        normalize_mc_to_data: bool = True,
        gof_check_method: Optional[str] = None,
        adaptive_binning: bool = False,
        dry_run: bool = False,
        show_plots: bool = False,
    ) -> Optional[Tuple[str, PathType]]:

        output_info: Optional[Tuple[str, PathType]] = None
        file_name: str = f"{file_name_base}{mode_name}"
        if export_path is not None:
            output_path: PathType = os.path.join(export_path, f"{file_name}.pdf")
            output_info = (file_name, output_path)

        if dry_run:
            return output_info

        draw_legend, legend_inside = legend_info  # type: bool, bool
        info_text_pos: Tuple[float, float] = self.get_info_text_position(style=style, draw_legend=draw_legend)
        legend_loc, legend_cols = self.get_legend_info(style=style)  # type: Union[int, str, None], Union[int, None]
        y_scale: float = self.get_y_scale(style=style)
        split_chi_info_text: bool = self.get_chi_text_split_info(style=style)

        mc_mode_match: np.ndarray = mc_df[reco_mode_col].isin(mode_ids)
        data_mode_match: np.ndarray = data_df[reco_mode_col].isin(mode_ids)

        corr_info_text: Optional[str] = self.get_correlation_info_text(
            mc_df=mc_df.loc[mc_mode_match],
            data_df=data_df.loc[data_mode_match],
            weight_col_name=weight_col,
        )

        dmc_hp = DataMCHistogramPlot(variable=self.comparer_info.hist_var)

        special_binning = self.comparer_info.get_additional_setting_with_default(
            setting_key="special_binning",
            default_setting=None,
        )

        if isinstance(special_binning, dict):
            assert mode_name in special_binning.keys(), (mode_name, list(special_binning.keys()))
            _this_special_binning: Union[None, BinsInputType, Binning] = special_binning[mode_name]
        else:
            _this_special_binning = special_binning

        dmc_hp.add_data_component(
            label=self.comparer_info.data_component_label,
            data=data_df.loc[data_mode_match],
            special_binning=_this_special_binning,
            data_weight_column=data_weight_col,
        )

        sig_id_column_name: str = self.comparer_info.mc_matching_collection.get_sig_id_column_name_for_df(df=mc_df)
        sig_id_mapping: Dict[str, Union[int, Tuple[int, ...]]] = self.comparer_info.get_sig_id_mapping_for(
            sig_id_col=sig_id_column_name
        )
        sig_id_group_name_to_label: Dict[
            str, str
        ] = self.comparer_info.mc_matching_collection.get_group_name_to_latex_label_mapping()
        sig_id_group_name_to_color: Dict[
            str, str
        ] = self.comparer_info.mc_matching_collection.get_group_name_to_color_mapping()

        for comp in self.comparer_info.mc_matching_collection.plot_order:
            component_sig_id: Union[int, Tuple[int, ...]] = sig_id_mapping[comp]

            if isinstance(component_sig_id, int):
                mc_sig_id_match: np.ndarray = mc_df[sig_id_column_name] == component_sig_id
            elif isinstance(component_sig_id, tuple):
                mc_sig_id_match = mc_df[sig_id_column_name].isin(component_sig_id)
            else:
                raise TypeError(f"Unexpected type for component_sig_id: {type(component_sig_id)} = {component_sig_id}")

            dmc_hp.add_mc_component(
                label=sig_id_group_name_to_label[comp],
                data=mc_df.loc[mc_mode_match & mc_sig_id_match],
                weights=weight_col,
                systematics=systematics_cols,
                color=sig_id_group_name_to_color[comp],
                special_binning=_this_special_binning,
            )

        fig, ax = DataMCHistogramPlot.create_hist_ratio_figure(
            fig_size=fig_size,
            height_ratio=height_ratio,
        )

        adaptive_binning_setting = self.comparer_info.get_additional_setting_with_default(
            setting_key="adaptive_binning",
            default_setting=adaptive_binning,
        )
        assert isinstance(adaptive_binning_setting, bool), type(adaptive_binning_setting).__name__
        gof_check_method_setting: Optional[str] = self.comparer_info.get_additional_setting_with_default(
            setting_key="gof_check_method",
            default_setting=gof_check_method,
        )

        try:
            gof_output: DataMCComparisonOutputType = dmc_hp.plot_on(
                ax1=ax[0],
                ax2=ax[1],
                style=style,
                ratio_type=ratio_type,
                draw_legend=draw_legend,
                legend_inside=legend_inside,
                legend_cols=legend_cols,
                legend_loc=legend_loc,
                y_scale=y_scale,
                normalize_to_data=normalize_mc_to_data,
                include_sys=True if systematics_cols else False,
                gof_check_method=gof_check_method_setting,
                adaptive_binning=adaptive_binning_setting,
                plot_outlier_indicators=False if ratio_type == "vs_uncert" else True,
            )
        except Exception as e:
            logger.error(
                f"Exception occurred for variable {self.comparer_info.hist_var.df_label} "
                f"with info text {self.comparer_info.info_text} and plot function called with\n"
                f"\tgof_check_method = {gof_check_method_setting},\n"
                f"\tadaptive_binning = {adaptive_binning_setting} and\n"
                f"\tnormalize_mc_to_data = {normalize_mc_to_data}!"
            )
            raise e

        reco_mode_label: str = self.comparer_info.get_reco_mode_title(
            mode_name=mode_name,
            use_short_titles=use_short_titles,
        )
        self.add_title_to_plot(axis=ax[0], reco_id_label=reco_mode_label)

        self.add_info_text_to_plot(
            axis=ax[0],
            style=style,
            info_text_position=info_text_pos,
            lumi_label=lumi_label,
            split_chi_info_text=split_chi_info_text,
            gof_info=gof_output,
            corr_info_text=corr_info_text,
        )

        fig.canvas.draw()
        x_ticks: Union[Any, List[plt.text.Text]] = ax[1].get_xticklabels()
        if x_ticks[0].get_text():
            if len(x_ticks[0].get_text().split("{")[1].split("}")[0]) > 4:
                new_x_ticks = x_ticks[1::2]
                if all(tick.get_text().split("{")[1].split("}")[0][-1] == "0" for tick in new_x_ticks):
                    for tick in x_ticks:
                        original_text = tick.get_text()
                        original_value = original_text.split("{")[1].split("}")[0]
                        tick.set_text(original_text.replace(original_value, original_value[:-1]))
                ax[1].set_xticklabels(x_ticks)
                plt.setp(x_ticks[::2], visible=False)
                plt.setp(new_x_ticks, visible=True)

        if export_path is not None:
            export(fig=fig, filename=file_name, target_dir=export_path, close_figure=not show_plots)

        if gof_output is not None:
            if gof_output.toy_output is not None:
                _, bin_edges, toys = gof_output.toy_output
                chi2_fig, _ = rdstar_plot_tools.plot_chi_squared_toys(
                    toys=toys,
                    bin_edges=bin_edges,
                    ndf=gof_output.ndf,
                    observed_chi2=gof_output.chi2,
                    full_output_path=None if export_path is None else os.path.join(export_path, file_name),
                    tag="chi2_distribution",
                )
                plt.close(chi2_fig)

        return output_info

    def add_title_to_plot(self, axis: AxesType, reco_id_label: str) -> None:
        axis.set_title(reco_id_label, loc="right")
        if self.comparer_info.title is not None:
            use_title = self.comparer_info.title
            if axis.get_ylim()[1] > 0.85e4:
                padding = " " * 9
                use_title = padding + self.comparer_info.title
            axis.set_title(use_title, loc="left")

    def add_info_text_to_plot(
        self,
        axis: AxesType,
        style: str,
        info_text_position: Tuple[float, float],
        lumi_label: bool,
        split_chi_info_text: bool,
        gof_info: DataMCComparisonOutputType,
        corr_info_text: Optional[str],
    ) -> None:
        info_text: str = self.comparer_info.lumi_label if lumi_label else ""
        if lumi_label and self.comparer_info.lumi_label_appendix is not None:
            info_text += self.comparer_info.lumi_label_appendix

        if self.comparer_info.info_text is not None and self.comparer_info.info_text:
            if info_text and self.comparer_info.info_text:
                info_text += "\n"
            info_text += self.comparer_info.info_text
        if gof_info is not None:
            if info_text:
                info_text += "\n"

            if self.comparer_info.add_scaling_factor_to_info_text:
                if gof_info.mc_scale_factor is not None and style == "stacked":
                    mc_scaling_info_text: str = r"MC $\times$ " + f"{gof_info.mc_scale_factor:.2f}\n"
                    info_text += mc_scaling_info_text

            info_text += r"$p$-Value = " + f"{gof_info.p_val:.2f}"

            chi_txt: str = "\n" + r"$\chi^2_\mathrm{" + gof_info.test_method_id + r"} / n_\mathrm{dof}$ = "

            if split_chi_info_text:
                if gof_info.chi2 >= 1000:
                    info_text += chi_txt + f"{gof_info.chi2:.0f}/{int(gof_info.ndf)}"
                else:
                    info_text += chi_txt + f"{gof_info.chi2:.1f}/{int(gof_info.ndf)}"
                info_text += "\n" + r"$\qquad\quad\,\,$ = " + f"{gof_info.chi2 / gof_info.ndf:.2f}"
            else:
                if gof_info.chi2 >= 1000:
                    info_text += chi_txt + f"{gof_info.chi2:.0f}/{int(gof_info.ndf)} = {gof_info.chi2 / gof_info.ndf:.2f}"
                else:
                    info_text += chi_txt + f"{gof_info.chi2:.1f}/{int(gof_info.ndf)} = {gof_info.chi2 / gof_info.ndf:.2f}"

        if corr_info_text is not None:
            info_text += "\n" + corr_info_text

        if info_text:
            axis.text(
                x=info_text_position[0],
                y=info_text_position[1],
                s=info_text,
                transform=axis.transAxes,
                va="top",
                linespacing=1.5,
            )

    def get_info_text_position(
        self,
        style: str,
        draw_legend: bool,
    ) -> Tuple[Union[float, int], Union[float, int]]:
        info_text_pos: List[float] = [0.66, 0.88]

        if isinstance(self.comparer_info.text_h_pos, tuple):
            assert all(isinstance(element, (int, float)) for element in self.comparer_info.text_h_pos)
            info_text_pos[0] = self.comparer_info.text_h_pos[0] if style == "summed" else self.comparer_info.text_h_pos[1]
        else:
            assert isinstance(self.comparer_info.text_h_pos, (int, float))
            info_text_pos[0] = self.comparer_info.text_h_pos

        if self.comparer_info.text_v_pos is not None:
            if isinstance(self.comparer_info.text_v_pos, tuple):
                assert all(isinstance(element, (int, float)) for element in self.comparer_info.text_v_pos)
                info_text_pos[1] = (
                    self.comparer_info.text_v_pos[0] if style == "summed" else self.comparer_info.text_v_pos[1]
                )
            else:
                assert isinstance(self.comparer_info.text_v_pos, (int, float))
                info_text_pos[1] = self.comparer_info.text_v_pos

        if info_text_pos[0] == 0.66 and info_text_pos[1] == 0.55:
            if style == "stacked" and not draw_legend:
                info_text_pos[1] = info_text_pos[1] + 0.3

        return info_text_pos[0], info_text_pos[1]

    def get_legend_info(self, style: str) -> Tuple[Optional[Union[int, str]], Optional[int]]:
        legend_loc, legend_cols = (None, None)
        if self.comparer_info.legend_loc is not None:
            if isinstance(self.comparer_info.legend_loc, tuple):
                legend_loc = self.comparer_info.legend_loc[0] if style == "summed" else self.comparer_info.legend_loc[1]
            else:
                legend_loc = self.comparer_info.legend_loc
        if self.comparer_info.legend_cols is not None:
            if isinstance(self.comparer_info.legend_cols, tuple):
                legend_cols = (
                    self.comparer_info.legend_cols[0] if style == "summed" else self.comparer_info.legend_cols[1]
                )
            else:
                legend_cols = self.comparer_info.legend_cols
        return legend_loc, legend_cols

    def get_y_scale(self, style: str) -> float:
        y_scale = 1.1
        if self.comparer_info.y_scale is not None:
            if isinstance(self.comparer_info.y_scale, tuple):
                y_scale = 1.1 * self.comparer_info.y_scale[0] if style == "summed" else self.comparer_info.y_scale[1]
            else:
                y_scale = self.comparer_info.y_scale
        return y_scale

    def get_chi_text_split_info(self, style: str) -> bool:
        if isinstance(self.comparer_info.split_chi_info_text, tuple):
            return (
                self.comparer_info.split_chi_info_text[0]
                if style == "summed"
                else self.comparer_info.split_chi_info_text[1]
            )
        else:
            assert isinstance(self.comparer_info.split_chi_info_text, bool)
            return self.comparer_info.split_chi_info_text

    def get_correlation_info_text(
        self,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        weight_col_name: str,
    ) -> str:
        corr_setting: Optional[CorrCalcInfo] = self.comparer_info.get_additional_setting_with_default(
            setting_key=corr_eval_info_additional_info_key,
            default_setting=None,
        )

        if corr_setting is None:
            return ""
        else:
            hv: HistVariable = self.comparer_info.hist_var
            mc_cond: pd.Series = (mc_df[hv.df_label] >= hv.scope[0]) & (mc_df[hv.df_label] <= hv.scope[1])
            data_cond: pd.Series = (data_df[hv.df_label] >= hv.scope[0]) & (data_df[hv.df_label] <= hv.scope[1])

            mc_info: Optional[str] = self.evaluate_correlation(
                corr_setting=corr_setting,
                df=mc_df.loc[mc_cond],
                weight_col_name=weight_col_name,
                is_data=False,
            )
            data_info: Optional[str] = self.evaluate_correlation(
                corr_setting=corr_setting,
                df=data_df.loc[data_cond],
                weight_col_name=weight_col_name,
                is_data=True,
            )
            assert mc_info is not None and data_info is not None, (mc_info, data_info)
            return f"{data_info}\n{mc_info}"

    def evaluate_correlation(
        self,
        corr_setting: CorrCalcInfo,
        df: pd.DataFrame,
        is_data: bool,
        weight_col_name: str,
    ) -> str:
        corr: float = corr_setting.calc_corr(
            df=df,
            is_data=is_data,
            weight_col_name=weight_col_name,
            base_hist_var=self.comparer_info.hist_var.df_label,
        )

        var_label: str = corr_setting.var_latex_label
        super_str: str = r"\mathrm{Data}" if is_data else r"\mathrm{MC}"
        sub_str: str = self.comparer_info.hist_var.variable_name[1:-1] + r",\ " + var_label[1:-1]

        # corr_symbol: str = r"\mathrm{Corr}"
        corr_symbol: str = r"\rho"
        return r"$" + corr_symbol + "_{" + sub_str + "}^{" + super_str + "}" + rf" = {corr:.2f}$"

    def apply_mc_sample_pre_scaling(
        self,
        df: pd.DataFrame,
        new_weight_column_name: str,
    ) -> None:
        if self.comparer_info.mc_pre_scaling is None:
            return

        self.comparer_info.mc_pre_scaling.apply_scaling(
            df=df,
            new_weight_column_name=new_weight_column_name,
        )

    def apply_additional_selection(self, df: pd.DataFrame) -> pd.DataFrame:
        sel_info: Union[
            None, SimpleCutInfo, Tuple[SimpleCutInfo], ComplexCutInfo
        ] = self.comparer_info.additional_selection_info
        if self.comparer_info.additional_selection_info is None:
            return df

        _df: pd.DataFrame = copy.deepcopy(df)
        if isinstance(sel_info, SimpleCutInfo):
            sel_info.apply_cut(df=_df)
        elif isinstance(sel_info, tuple) and all(isinstance(v, SimpleCutInfo) for v in sel_info):
            SimpleCutInfo.apply_multiple_cuts(simple_cut_infos=sel_info, df=_df)
        elif isinstance(sel_info, ComplexCutInfo):
            sel_info.apply_cuts(df=_df)
        else:
            raise TypeError(f"Don't know what to do with additional_selection_info of type {type(sel_info)}!")

        return _df
