"""
Utility functions to apply cuts for distribution comparison.

Felix Metzner 2021
"""


import pandas as pd

from dataclasses import dataclass
from typing import Union, Tuple, List, Sequence, Dict

from rdstar.cut_information_objects.cut_info import CutInfo


__all__ = [
    "SimpleCutInfo",
    "ComplexCutInfo",
]


@dataclass(frozen=True)
class SimpleCutInfo:
    cut_str: str
    cut_latex_label: str
    cut_var_col_name: str

    def __post_init__(self) -> None:
        assert self.cut_var_col_name in self.cut_str, (self.cut_var_col_name, self.cut_str)

    @staticmethod
    def from_cut_info(cut_info: CutInfo) -> "SimpleCutInfo":
        return SimpleCutInfo(
            cut_str=cut_info.cut_str,
            cut_latex_label=cut_info.cut_label,
            cut_var_col_name=cut_info.cut_var,
        )

    def apply_cut(self, df: pd.DataFrame) -> None:
        df.query(self.cut_str, inplace=True)

    def get_df_with_cut(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.query(self.cut_str).reset_index(drop=True)

    @staticmethod
    def apply_multiple_cuts(
        simple_cut_infos: Sequence["SimpleCutInfo"],
        df: pd.DataFrame,
    ) -> None:
        cut_str = " & ".join([sci.cut_str for sci in simple_cut_infos])
        df.query(cut_str, inplace=True)

    @staticmethod
    def get_df_with_multiple_cuts(
        simple_cut_infos: Sequence["SimpleCutInfo"],
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        cut_str = " & ".join([sci.cut_str for sci in simple_cut_infos])
        return df.query(cut_str).reset_index(drop=True)


@dataclass(frozen=True)
class ComplexCutInfo:
    secondary_var_col_name: str
    conditions_simple_cut_infos_pairs: Dict[Tuple[Union[int, float, str], ...], Tuple[SimpleCutInfo, ...]]

    def apply_cuts(self, df: pd.DataFrame) -> None:
        assert self.secondary_var_col_name in df.columns, self.secondary_var_col_name
        for conditions, simple_cut_infos in self.conditions_simple_cut_infos_pairs.items():
            assert all(sci.cut_var_col_name in df.columns for sci in simple_cut_infos), (
                conditions,
                [sci.cut_var_col_name for sci in simple_cut_infos if sci.cut_var_col_name not in df.columns],
            )

        cut_str_list = list()  # type: List[str]
        all_considered_conditions = list()  # type: List[Union[int, float, str]]
        for conditions, simple_cut_infos in self.conditions_simple_cut_infos_pairs.items():
            all_considered_conditions.extend([c for c in conditions])
            this_cut_str_list = f"({self.secondary_var_col_name} in {list(conditions)}) & "  # type: str
            this_cut_str_list += " & ".join(f"({sci.cut_str})" for sci in simple_cut_infos)
            cut_str_list.append(this_cut_str_list)

        # Taking into account all values of the secondary variable which are not covered by the provided conditions:
        cut_str_list.append(f"{self.secondary_var_col_name} not in {list(all_considered_conditions)}")

        cut_str = " | ".join([f"({c})" for c in cut_str_list])
        df.query(cut_str, inplace=True)
