"""
Definition of Continuum Suppression MVA Training

Felix Metzner 2019
"""

import numpy as np

from PyFastBDT import FastBDT
from types import SimpleNamespace
from typing import List, Dict, Union, Optional

from templatefitter.plotter.histogram_variable import HistVariable

from analysistools.dataframe_tools.hdf5_utils import load_hdf5

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.mva_methods.mva_methods import MVABaseClass, MVAVarInfo
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.latex_labels.latex_particle_names import LatexParticleNames


__all__ = [
    "ContinuumSuppression",
    "get_cs_variable_dict",
    "get_cs_variable_infos",
    "cs_validation_variables",
    "cs_additional_variables",
    "cs_vars_tag_side_name_space",
    "run_cs_training",
]


# fmt: off
base_cs_variable_infos = [
    MVAVarInfo(var="spR2__bc", bins=25, min=0.0, max=1.0, latex=LatexObsNames.cs_base_r2),
    MVAVarInfo(var="spcosTBTO__bc", bins=25, min=0.0, max=1.0, latex=LatexObsNames.cs_base_cos_theta_tbto),
    MVAVarInfo(var="spcosTBz__bc", bins=25, min=0.0, max=1.0, latex=LatexObsNames.cs_base_cos_theta_tbz),
    MVAVarInfo(var="spthrustBm__bc", bins=25, min=0.5, max=1.0, latex=LatexObsNames.cs_base_thrust_b),
    MVAVarInfo(var="spthrustOm__bc", bins=25, min=0.5, max=1.0, latex=LatexObsNames.cs_base_thrust_o),
    MVAVarInfo(var="spKSFWVariables__boet__bc__bc", bins=25, min=2.0, max=12.0, latex=LatexObsNames.cs_base_energy_t),
    MVAVarInfo(var="spKSFWVariables__bomm2__bc__bc", bins=25, min=-10.0, max=10.0, latex=LatexObsNames.cs_base_mm2),
    MVAVarInfo(var="spKSFWVariables__bohso00__bc__bc", bins=25, min=0, max=1.5, latex=LatexObsNames.cs_base_h_so_c_0),
    # MVAVarInfo(var="spKSFWVariables__bohso01__bc__bc", bins=25, min=0, max=1.5, latex=LatexObsNames.cs_base_h_so_c_1),  # Not considered atm, but should we include it?  # noqa: E501, E950 # fmt: skip
    MVAVarInfo(var="spKSFWVariables__bohso02__bc__bc", bins=25, min=-0.6, max=1.1, latex=LatexObsNames.cs_base_h_so_c_2),
    # MVAVarInfo(var="spKSFWVariables__bohso03__bc__bc", bins=25, min=0, max=1.5, latex=LatexObsNames.cs_base_h_so_c_3),  # Not considered atm, but should we include it?  # noqa: E501, E950 # fmt: skip
    MVAVarInfo(var="spKSFWVariables__bohso04__bc__bc", bins=25, min=-0.4, max=0.6, latex=LatexObsNames.cs_base_h_so_c_4),
    MVAVarInfo(var="spKSFWVariables__bohso10__bc__bc", bins=25, min=0, max=1.25, latex=LatexObsNames.cs_base_h_so_n_0),
    MVAVarInfo(var="spKSFWVariables__bohso12__bc__bc", bins=25, min=-0.2, max=0.55, latex=LatexObsNames.cs_base_h_so_n_2),
    MVAVarInfo(var="spKSFWVariables__bohso14__bc__bc", bins=25, min=-0.2, max=0.3, latex=LatexObsNames.cs_base_h_so_n_4),
    MVAVarInfo(var="spKSFWVariables__bohso20__bc__bc", bins=25, min=0, max=0.7, latex=LatexObsNames.cs_base_h_so_m_0),
    MVAVarInfo(var="spKSFWVariables__bohso22__bc__bc", bins=25, min=-0.2, max=0.5, latex=LatexObsNames.cs_base_h_so_m_2),
    MVAVarInfo(var="spKSFWVariables__bohso24__bc__bc", bins=25, min=-0.2, max=0.3, latex=LatexObsNames.cs_base_h_so_m_4),
    MVAVarInfo(var="spKSFWVariables__bohoo0__bc__bc", bins=25, min=0.0, max=0.3, latex=LatexObsNames.cs_base_h_oo_0),
    MVAVarInfo(var="spKSFWVariables__bohoo1__bc__bc", bins=25, min=-0.04, max=0.04, latex=LatexObsNames.cs_base_h_oo_1),
    MVAVarInfo(var="spKSFWVariables__bohoo2__bc__bc", bins=25, min=-0.02, max=0.12, latex=LatexObsNames.cs_base_h_oo_2),
    MVAVarInfo(var="spKSFWVariables__bohoo3__bc__bc", bins=25, min=-0.04, max=0.04, latex=LatexObsNames.cs_base_h_oo_3),
    MVAVarInfo(var="spKSFWVariables__bohoo4__bc__bc", bins=25, min=-0.025, max=0.08, latex=LatexObsNames.cs_base_h_oo_4),
    MVAVarInfo(var="spCleoCone__bo1__bc__bc", bins=30, min=0.0, max=5.0, latex=LatexObsNames.cs_base_cleo_cone_1),
    MVAVarInfo(var="spCleoCone__bo2__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_2),
    MVAVarInfo(var="spCleoCone__bo3__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_3),
    MVAVarInfo(var="spCleoCone__bo4__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_4),
    MVAVarInfo(var="spCleoCone__bo5__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_5),
    MVAVarInfo(var="spCleoCone__bo6__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_6),
    MVAVarInfo(var="spCleoCone__bo7__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_7),
    MVAVarInfo(var="spCleoCone__bo8__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_8),
    MVAVarInfo(var="spCleoCone__bo9__bc__bc", bins=30, min=0.0, max=6.0, latex=LatexObsNames.cs_base_cleo_cone_9),
]  # type: List[MVAVarInfo]
# fmt: on


def get_cs_variable_dict(signal_side: bool = True) -> Dict[str, HistVariable]:
    if signal_side:
        b_id = "daughter__bo1__cm__"  # type: str
        b_name = LatexParticleNames.b_sig_str  # type: str
    else:
        b_id = "daughter__bo0__cm__"
        b_name = LatexParticleNames.b_tag_str
    separator = r"$\;$"  # type: str

    return {
        b_id
        + var_info.var: HistVariable(
            df_label=b_id + var_info.var,
            n_bins=var_info.bins,
            scope=(var_info.min, var_info.max),
            var_name=b_name + separator + var_info.latex,
            unit=var_info.unit,
            use_log_scale=var_info.log,
        )
        for var_info in base_cs_variable_infos
    }


def get_cs_variable_infos(signal_side: bool = True) -> List[MVAVarInfo]:
    if signal_side:
        b_id = "daughter__bo1__cm__"  # type: str
        b_name = LatexParticleNames.b_sig_str  # type: str
    else:
        b_id = "daughter__bo0__cm__"
        b_name = LatexParticleNames.b_tag_str
    separator = r"$\;$"  # type: str

    return [
        MVAVarInfo(
            var=b_id + info.var,
            latex=b_name + separator + info.latex,
            bins=info.bins,
            min=info.min,
            max=info.max,
        )
        for info in base_cs_variable_infos
    ]


# Variables required for validation and training
cs_validation_variables = [
    RDStarHistVars.pl_star,
    RDStarHistVars.missing_mass_q2_SidebandValidation,
    RDStarHistVars.e_miss_minus_p_miss,
]  # type: List[HistVariable]

_additional_vars = [
    "isNotContinuumEvent",  # Target variable for the mva training.
    "isSignalAcceptMissingNeutrino",  # Can be used to extract only correctly reconstructed signal events.
    "daughter__bo1__cm__spisSignalAcceptMissingNeutrino__bc",  # Can be used to extract corr. reconstructed sig. events.
    "daughter__bo0__cm__spisSignalAcceptMissingNeutrino__bc",  # Can be used to extract corr. reconstructed sig. events.
]  # type: List[str]

# List of additional variables needed for training process and validation but not as input for the cs mva.
cs_additional_variables = _additional_vars + [c.df_label for c in cs_validation_variables]  # type: List[str]


cs_vars_tag_side_name_space = SimpleNamespace(
    **{hv.df_label: hv for hv in get_cs_variable_dict(signal_side=False).values()}
)  # type: SimpleNamespace
cs_vars_sig_side_name_space = SimpleNamespace(
    **{hv.df_label: hv for hv in get_cs_variable_dict(signal_side=True).values()}
)  # type: SimpleNamespace


class ContinuumSuppression(MVABaseClass):
    def __init__(
        self,
        cs_variables: List[str],
        target_variable: str,
        weight: Optional[str] = None,
    ) -> None:
        super().__init__(
            variables=cs_variables,
            target_variable=target_variable,
            weight=weight,
        )


def run_cs_training(
    train_data_file_path: PathType,
    variable_file_path: PathType,
    weight_file_path: PathType,
    target_variable: str = "isNotContinuumEvent",
    fast_bdt_settings: Optional[Dict[str, Union[str, int]]] = None,
) -> None:
    """
    Trains a FastBDT for Continuum Suppression using a provided, prepared data set,
    and storing the resulting weight files as well as the list of used variables.

    The truth target used for the supervised training process is specified via the
    parameter target_variable.
    The FastBDT settings can be set via the optional parameter fast_bdt_settings,
    which expects a dictionary mapping the setting names to their respective values.

    The variables used as input for the MVA are provided as pickled list loaded from
    a pickle file specified by variable_file_path.

    The resulting classifier weights and the variable list required to apply the
    CS MVA are stored to a path defined by the argument weight_file_path.

    :param train_data_file_path: Path to the training data sample hdf5 file.
    :param variable_file_path: Path to the pickle file containing the variable to be used.
    :param weight_file_path: Path under which the trained weights and variables will be stored.
    :param target_variable: The variable used as truth target in the training.
    :param fast_bdt_settings: Settings used for the FastBDT MVA.
    :return: None
    """
    final_vars = ContinuumSuppression.load_pickled_variables(variable_file_path)
    cs_train_data = load_hdf5(hdf5_file_path=train_data_file_path, hdf5_key="train_sample")

    cs_train_data.loc[:, "cs_weights"] = np.ones(len(cs_train_data.index))

    cs = ContinuumSuppression(
        cs_variables=final_vars,
        target_variable=target_variable,
        weight="cs_weights",
    )

    if fast_bdt_settings is None:
        _fast_bdt_settings = {
            "nTrees": 200,
        }  # type: Dict[str, Union[str, int]]
    else:
        _fast_bdt_settings = fast_bdt_settings

    cs.add_classifier(classifier=FastBDT.Classifier(**_fast_bdt_settings))
    cs.train(cs_train_data)

    cs.save_classifier_weights(weight_file_path=weight_file_path, overwrite=True)
