"""
Providing latex strings for particle decay processes.

Felix Metzner 2021
"""

from typing import List, Dict, Tuple

from rdstar.latex_labels.latex_particle_names import LatexParticleNames as LPNs
from rdstar.latex_labels.latex_string_methods import LatexObsNameMethods as LoNM


__all__ = [
    "LatexDecayProcesses",
]


def get_all_reco_modes_str_for(particle: str, short: bool, all_prefix: bool = True) -> str:
    prefix: str = r"$\mathrm{All}\," if all_prefix else r"$"
    part_str: str = LoNM.get_bare_var_str(var_str=particle)
    if part_str:
        part_str += r"\,"
    if short:
        res: str = prefix + part_str + r"\mathrm{Reco.\,Modes}$"
    else:
        res = prefix + part_str + r"\mathrm{Reconstruction\,Modes}$"
    return res


class LatexDecayProcesses:
    # fmt: off

    # Some Tags
    _bkg_tag: str = r"$\mathrm{Bkg}$"
    _reco_tag: str = r"$\mathrm{Reco}$"
    _hadronic_tag: str = r"$\mathrm{Hadronic}$"
    _generic_tag: str = r"$\mathrm{Generic}$"
    _fake_tag: str = r"$\mathrm{Fake}$"
    _impossible_tag: str = r"$\mathrm{Impossible}$"

    _down_feed_suffix: str = r"$\mathrm{Down\ Feed}$"
    _up_feed_suffix: str = r"$\mathrm{Up\ Feed}$"
    _cross_feed_suffix: str = r"$\mathrm{Cross\ Feed}$"
    _daughter_mismatch_suffix: str = r"$\mathrm{Daughter\ Mismatch}$"

    neutral_d_st_to_d_pi: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.pi_zero_str])
    neutral_d_st_to_d_photon: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.gamma_str])
    anti_neutral_d_st_to_d_pi: str = LoNM.get_decay(from_part=LPNs.anti_d_st_zero_str, to_parts=[LPNs.anti_d_zero_str, LPNs.pi_zero_str])
    anti_neutral_d_st_to_d_photon: str = LoNM.get_decay(from_part=LPNs.anti_d_st_zero_str, to_parts=[LPNs.anti_d_zero_str, LPNs.gamma_str])
    _sub_decay_neutral_d_st_to_d_pi: str = LoNM.get_as_sub_decay(decay=neutral_d_st_to_d_pi, with_mother=True)
    _sub_decay_neutral_d_st_to_d_photon: str = LoNM.get_as_sub_decay(decay=neutral_d_st_to_d_photon, with_mother=True)
    _sub_decay_anti_neutral_d_st_to_d_pi: str = LoNM.get_as_sub_decay(decay=anti_neutral_d_st_to_d_pi, with_mother=True)
    _sub_decay_anti_neutral_d_st_to_d_photon: str = LoNM.get_as_sub_decay(decay=anti_neutral_d_st_to_d_photon, with_mother=True)

    #################################
    # Decay Strings without Charges #
    #################################

    b_to_d_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_w_wo_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.lepton_str, LPNs.nu_lep_str])

    bz_to_d_l_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_str, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_d_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_d_l_nu: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_str, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_d_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str, LPNs.nu_lep_str])

    bz_to_dm_l_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bz_to_dm_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bp_to_dz_l_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_l_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_pi, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_l_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_photon, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bm_to_dz_l_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_zero_str, LPNs.lepton_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.lepton_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_l_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_pi, LPNs.lepton_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_l_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_photon, LPNs.lepton_minus_str, LPNs.anti_nu_lep_str])

    bz_to_dm_e_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bz_to_dm_st_e_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bp_to_dz_e_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_e_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_e_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_pi, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_e_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_photon, LPNs.electron_plus_str, LPNs.nu_lep_str])
    bm_to_dz_e_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_zero_str, LPNs.electron_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_e_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.electron_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_e_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_pi, LPNs.electron_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_e_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_photon, LPNs.electron_minus_str, LPNs.anti_nu_lep_str])

    bz_to_dm_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bz_to_dm_st_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bp_to_dz_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_mu_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_pi, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bp_to_dz_st_mu_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_photon, LPNs.muon_plus_str, LPNs.nu_lep_str])
    bm_to_dz_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_zero_str, LPNs.muon_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_mu_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.muon_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_mu_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_pi, LPNs.muon_minus_str, LPNs.anti_nu_lep_str])
    bm_to_dz_st_mu_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_photon, LPNs.muon_minus_str, LPNs.anti_nu_lep_str])

    bp_to_dwwostz_l_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    bz_to_dwwostm_l_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])

    b_to_d_1_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_one_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d0st_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_sszero_st_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_dprime1_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_prime_one_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_dst_2_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_two_st_str, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_d_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.tau_str, LPNs.nu_tau_str])
    b_to_d_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.tau_str, LPNs.nu_tau_str])

    bz_to_d_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_str, LPNs.tau_str, LPNs.nu_tau_str])
    bz_to_d_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_str, LPNs.tau_str, LPNs.nu_tau_str])
    bpm_to_d_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_str, LPNs.tau_str, LPNs.nu_tau_str])
    bpm_to_d_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_st_str, LPNs.tau_str, LPNs.nu_tau_str])

    bz_to_dm_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bz_to_dm_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bp_to_dz_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bp_to_dz_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bp_to_dz_st_tau_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_pi, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bp_to_dz_st_tau_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[_sub_decay_neutral_d_st_to_d_photon, LPNs.tau_plus_str, LPNs.nu_tau_str])
    bm_to_dz_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_zero_str, LPNs.tau_minus_str, LPNs.anti_nu_tau_str])
    bm_to_dz_st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.tau_minus_str, LPNs.anti_nu_tau_str])
    bm_to_dz_st_tau_nu_dst_via_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_pi, LPNs.tau_minus_str, LPNs.anti_nu_tau_str])
    bm_to_dz_st_tau_nu_dst_via_ph: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[_sub_decay_anti_neutral_d_st_to_d_photon, LPNs.tau_minus_str, LPNs.anti_nu_tau_str])

    b_to_d_1_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_one_str, LPNs.tau_str, LPNs.nu_tau_str])
    b_to_d0st_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_sszero_st_str, LPNs.tau_str, LPNs.nu_tau_str])
    b_to_dprime1_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_prime_one_str, LPNs.tau_str, LPNs.nu_tau_str])
    b_to_dst_2_tau_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_two_st_str, LPNs.tau_str, LPNs.nu_tau_str])

    d_1_to_d_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_one_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_d_1_to_d_pi_pi: str = LoNM.get_as_sub_decay(decay=d_1_to_d_pi_pi, with_mother=True)
    b_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_1_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_d_eta_l_nu_gap: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.eta_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_st_eta_l_nu_gap: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.eta_str, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_d_pi_pi_l_nu_gap: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_st_pi_pi_l_nu_gap: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str, LPNs.lepton_str, LPNs.nu_lep_str])

    d_ss0_to_d_eta: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.eta_str])
    _sub_decay_d_ss0_to_d_eta: str = LoNM.get_as_sub_decay(decay=d_ss0_to_d_eta, with_mother=True)
    b_to_d_eta_l_nu_gap_via_d_st_0: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_eta, LPNs.lepton_str, LPNs.nu_lep_str])

    d_p1_to_d_st_eta: str = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.eta_str])
    _sub_decay_d_p1_to_d_st_eta: str = LoNM.get_as_sub_decay(decay=d_p1_to_d_st_eta, with_mother=True)
    b_to_d_st_eta_l_nu_gap_via_d_prime_1: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_st_eta, LPNs.lepton_str, LPNs.nu_lep_str])

    d_p1_to_d_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_d_p1_to_d_pi_pi: str = LoNM.get_as_sub_decay(decay=d_p1_to_d_pi_pi, with_mother=True)
    d_ss0_to_d_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_d_ss0_to_d_pi_pi: str = LoNM.get_as_sub_decay(decay=d_ss0_to_d_pi_pi, with_mother=True)
    b_to_d_pi_pi_l_nu_gap_via_d_prime_1: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_pi_pi_l_nu_gap_via_d_st_0: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])

    d_p1_to_d_st_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_d_p1_to_d_st_pi_pi: str = LoNM.get_as_sub_decay(decay=d_p1_to_d_st_pi_pi, with_mother=True)
    d_ss0_to_d_st_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_d_ss0_to_d_st_pi_pi: str = LoNM.get_as_sub_decay(decay=d_ss0_to_d_st_pi_pi, with_mother=True)
    b_to_d_st_pi_pi_l_nu_gap_via_d_prime_1: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_p1_to_d_st_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_st_pi_pi_l_nu_gap_via_d_st_0: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_d_ss0_to_d_st_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])

    dst_2_to_dst_pi_pi: str = LoNM.get_decay(from_part=LPNs.d_two_st_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.pi_str])
    _sub_decay_dst_2_to_dst_pi_pi: str = LoNM.get_as_sub_decay(decay=dst_2_to_dst_pi_pi, with_mother=True)
    b_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_NC: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dst_2_to_dst_pi_pi, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_d_s_kaon_l_nu_NC: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_strange_str, LPNs.k_str, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_d_st_s_kaon_l_nu_NC: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_strange_st_str, LPNs.k_str, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_dstst_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_st_str, LPNs.lepton_str, LPNs.nu_lep_str])

    b_to_xc_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.xc_str, LPNs.lepton_str, LPNs.nu_lep_str])

    dstst_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    dstst_to_d_wwost_pi_pm: str = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str])
    dstst_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])
    dstst_to_d_wwost_pi_minus: str = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_minus_str])
    _sub_decay_dstst_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_zero, with_mother=True)
    _sub_decay_dstst_to_d_wwost_pi_pm: str = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_pm, with_mother=True)
    _sub_decay_dstst_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_plus, with_mother=True)
    _sub_decay_dstst_to_d_wwost_pi_minus: str = LoNM.get_as_sub_decay(decay=dstst_to_d_wwost_pi_minus, with_mother=True)
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_plus, LPNs.lepton_str, LPNs.nu_lep_str])
    b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[_sub_decay_dstst_to_d_wwost_pi_minus, LPNs.lepton_str, LPNs.nu_lep_str])

    dzerostar_to_d_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.pi_zero_str])
    dzerostar_to_d_pi_pm: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_str, to_parts=[LPNs.d_str, LPNs.pi_pm_str])
    doneprime_to_dstar_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_zero_str])
    doneprime_to_dstar_pi_pm: str = LoNM.get_decay(from_part=LPNs.d_prime_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_pm_str])
    done_to_dstar_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_zero_str])
    done_to_dstar_pi_pm: str = LoNM.get_decay(from_part=LPNs.d_one_str, to_parts=[LPNs.d_st_str, LPNs.pi_pm_str])
    dtwostar_to_dwwostar_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_two_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    dtwostar_to_dwwostar_pi_pm: str = LoNM.get_decay(from_part=LPNs.d_two_st_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str])

    _sub_decay_dzerostar_to_d_pi_zero: str = LoNM.get_as_sub_decay(decay=dzerostar_to_d_pi_zero, with_mother=True)
    _sub_decay_dzerostar_to_d_pi_pm: str = LoNM.get_as_sub_decay(decay=dzerostar_to_d_pi_pm, with_mother=True)
    _sub_decay_doneprime_to_dstar_pi_zero: str = LoNM.get_as_sub_decay(decay=doneprime_to_dstar_pi_zero, with_mother=True)
    _sub_decay_doneprime_to_dstar_pi_pm: str = LoNM.get_as_sub_decay(decay=doneprime_to_dstar_pi_pm, with_mother=True)
    _sub_decay_done_to_dstar_pi_zero: str = LoNM.get_as_sub_decay(decay=done_to_dstar_pi_zero, with_mother=True)
    _sub_decay_done_to_dstar_pi_pm: str = LoNM.get_as_sub_decay(decay=done_to_dstar_pi_pm, with_mother=True)
    _sub_decay_dtwostar_to_dwwostar_pi_zero: str = LoNM.get_as_sub_decay(decay=dtwostar_to_dwwostar_pi_zero, with_mother=True)
    _sub_decay_dtwostar_to_dwwostar_pi_pm: str = LoNM.get_as_sub_decay(decay=dtwostar_to_dwwostar_pi_pm, with_mother=True)

    bpm_to_d0st_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_dzerostar_to_d_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_dprime1_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_doneprime_to_dstar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_d_1_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_done_to_dstar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_dst_2_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_dtwostar_to_dwwostar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_d0st_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_dzerostar_to_d_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_dprime1_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_doneprime_to_dstar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_d_1_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_done_to_dstar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bpm_to_dst_2_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[_sub_decay_dtwostar_to_dwwostar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])

    bz_to_d0st_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_dzerostar_to_d_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_dprime1_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_doneprime_to_dstar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_d_1_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_done_to_dstar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_dst_2_l_nu_with_pi_pm: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_dtwostar_to_dwwostar_pi_pm, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_d0st_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_dzerostar_to_d_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_dprime1_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_doneprime_to_dstar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_d_1_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_done_to_dstar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])
    bz_to_dst_2_l_nu_with_pi_zero: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[_sub_decay_dtwostar_to_dwwostar_pi_zero, LPNs.lepton_str, LPNs.nu_lep_str])

    # Pi L Nu
    b_to_pi_l_nu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.pi_str, LPNs.lepton_str, LPNs.nu_lep_str])

    #################################################################################
    # Old SigID and GenID Matched Labels (Only products, no charge, with Neutrinos) #
    #################################################################################
    to_d_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_l_nu)
    to_d_st_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_st_l_nu)
    to_d_st_l_nu_down_feed: str = LoNM.get_products_from_decay(process_label=b_to_d_st_l_nu)[:-1] + r"\;" + _down_feed_suffix[1:]

    to_dwwostz_l_nu: str = LoNM.get_products_from_decay(process_label=bp_to_dwwostz_l_nu)
    to_dwwostpm_l_nu: str = LoNM.get_products_from_decay(process_label=bz_to_dwwostm_l_nu)

    to_dz_l_nu: str = LoNM.get_products_from_decay(process_label=bp_to_dz_l_nu)
    to_dz_st_l_nu: str = LoNM.get_products_from_decay(process_label=bp_to_dz_st_l_nu)
    to_dz_st_l_nu_down_feed: str = LoNM.get_products_from_decay(process_label=bp_to_dz_st_l_nu)[:-1] + r"\;" + _down_feed_suffix[1:]
    to_dm_l_nu: str = LoNM.get_products_from_decay(process_label=bz_to_dm_l_nu)
    to_dm_st_l_nu: str = LoNM.get_products_from_decay(process_label=bz_to_dm_st_l_nu)
    to_dm_st_l_nu_down_feed: str = LoNM.get_products_from_decay(process_label=bz_to_dm_st_l_nu)[:-1] + r"\;" + _down_feed_suffix[1:]

    to_d_1_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_1_l_nu)
    to_d0st_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_d0st_l_nu)
    to_dprime1_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_dprime1_l_nu)
    to_dst_2_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_dst_2_l_nu)

    to_d_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_tau_nu)
    to_d_st_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_st_tau_nu)

    to_dz_tau_nu: str = LoNM.get_products_from_decay(process_label=bp_to_dz_tau_nu)
    to_dz_st_tau_nu: str = LoNM.get_products_from_decay(process_label=bp_to_dz_st_tau_nu)
    to_dm_tau_nu: str = LoNM.get_products_from_decay(process_label=bz_to_dm_tau_nu)
    to_dm_st_tau_nu: str = LoNM.get_products_from_decay(process_label=bz_to_dm_st_tau_nu)

    to_d_1_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_d_1_tau_nu)
    to_d0st_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_d0st_tau_nu)
    to_dprime1_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_dprime1_tau_nu)
    to_dst_2_tau_nu: str = LoNM.get_products_from_decay(process_label=b_to_dst_2_tau_nu)

    to_dstst_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu)

    to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero: str = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero)
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm: str = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm)
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus: str = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_plus)
    to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus: str = LoNM.get_products_from_decay(process_label=b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_minus)

    to_pi_l_nu: str = LoNM.get_products_from_decay(process_label=b_to_pi_l_nu)

    to_hadronic_bkg_txt: str = _hadronic_tag[:-1] + r"\;" + _bkg_tag[1:]
    to_b_bbar_bkg: str = LPNs.b_str[:-1] + LoNM.get_bare_var_str(LoNM.get_anti_particle(particle_str=LPNs.b_str)) + r"\;" + _bkg_tag[1:]
    to_continuum_txt: str = r"$\mathrm{Continuum}$"
    to_charm_continuum_txt: str = r"$\mathrm{Charm\,Continuum}$"
    to_uds_continuum_txt: str = r"$\mathrm{UDS\,Continuum}$"

    to_hadronic_tau_bkg_txt: str = _hadronic_tag[:-1] + r"\;" + LoNM.get_bare_var_str(LoNM.get_anti_particle(particle_str=LPNs.tau_str)) + r"\;" + _bkg_tag[1:]
    to_other_bkg_txt: str = r"$\mathrm{Other}$"
    to_inconclusive_txt: str = r"$\mathrm{Inconclusive}$"

    #########################################################
    # Precise decay Strings with Lepton Flavour and Charges #
    #########################################################

    # Tau Decay
    tau_plus_to_e_plus_nu_nu: str = LoNM.get_decay(from_part=LPNs.tau_plus_str, to_parts=[LPNs.electron_plus_str, LPNs.nu_e_str, LPNs.anti_nu_tau_str])
    tau_plus_to_mu_plus_nu_nu: str = LoNM.get_decay(from_part=LPNs.tau_plus_str, to_parts=[LPNs.muon_plus_str, LPNs.nu_mu_str, LPNs.anti_nu_tau_str])

    subdec_tau_plus_to_e_plus_nu_nu: str = LoNM.get_as_sub_decay(decay=tau_plus_to_e_plus_nu_nu, with_mother=True)
    subdec_tau_plus_to_mu_plus_nu_nu: str = LoNM.get_as_sub_decay(decay=tau_plus_to_mu_plus_nu_nu, with_mother=True)

    tau_to_hadrons_nu: str = LoNM.get_decay(from_part=LPNs.tau_str, to_parts=[LPNs.hadrons_str, LPNs.anti_nu_tau_str])
    tau_plus_to_hadrons_nu: str = LoNM.get_decay(from_part=LPNs.tau_plus_str, to_parts=[LPNs.hadrons_str, LPNs.anti_nu_tau_str])

    subdec_tau_to_hadrons_nu: str = LoNM.get_as_sub_decay(decay=tau_to_hadrons_nu, with_mother=True)
    subdec_tau_plus_to_hadrons_nu: str = LoNM.get_as_sub_decay(decay=tau_plus_to_hadrons_nu, with_mother=True)

    # D* Decay
    d_st_plus_to_d_pi: str = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_str, LPNs.pi_str])
    d_st_minus_to_d_pi: str = LoNM.get_decay(from_part=LPNs.d_st_minus_str, to_parts=[LPNs.d_str, LPNs.pi_str])
    d_st_zero_to_d_pi: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.pi_zero_str])
    anti_d_st_zero_to_d_pi: str = LoNM.get_decay(from_part=LPNs.anti_d_st_zero_str, to_parts=[LPNs.anti_d_zero_str, LPNs.pi_zero_str])

    d_st_zero_to_d_photon: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.gamma_str])
    anti_d_st_zero_to_d_photon: str = LoNM.get_decay(from_part=LPNs.anti_d_st_zero_str, to_parts=[LPNs.anti_d_zero_str, LPNs.gamma_str])

    subdec_d_st_plus_to_d_pi: str = LoNM.get_as_sub_decay(decay=d_st_plus_to_d_pi, with_mother=True)
    subdec_d_st_minus_to_d_pi: str = LoNM.get_as_sub_decay(decay=d_st_minus_to_d_pi, with_mother=True)
    subdec_d_st_zero_to_d_pi: str = LoNM.get_as_sub_decay(decay=d_st_zero_to_d_pi, with_mother=True)
    subdec_anti_d_st_zero_to_d_pi: str = LoNM.get_as_sub_decay(decay=anti_d_st_zero_to_d_pi, with_mother=True)

    subdec_d_st_zero_to_d_photon: str = LoNM.get_as_sub_decay(decay=d_st_zero_to_d_photon, with_mother=True)
    subdec_anti_d_st_zero_to_d_photon: str = LoNM.get_as_sub_decay(decay=anti_d_st_zero_to_d_photon, with_mother=True)

    impossible_d_st_plus_to_d_photon: str = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_minus_str, LPNs.gamma_str])
    impossible_d_st_minus_to_d_photon: str = LoNM.get_decay(from_part=LPNs.d_st_minus_str, to_parts=[LPNs.d_minus_str, LPNs.gamma_str])
    impossible_subdec_d_st_plus_to_d_photon: str = LoNM.get_as_sub_decay(decay=impossible_d_st_plus_to_d_photon, with_mother=True)
    impossible_subdec_d_st_minus_to_d_photon: str = LoNM.get_as_sub_decay(decay=impossible_d_st_minus_to_d_photon, with_mother=True)

    # D** Decay
    d_one_zero_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_one_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_one_minus_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_one_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_one_zero_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_one_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])
    d_one_minus_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_one_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])

    d_twostar_zero_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_two_st_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_twostar_minus_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_two_st_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_twostar_zero_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_two_st_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])
    d_twostar_minus_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_two_st_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])

    d_zerostar_zero_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_zerostar_minus_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_zerostar_zero_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])
    d_zerostar_minus_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_sszero_st_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])

    d_primeone_zero_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_prime_one_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_primeone_minus_to_d_wwost_pi_zero: str = LoNM.get_decay(from_part=LPNs.d_prime_one_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str])
    d_primeone_zero_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_prime_one_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])
    d_primeone_minus_to_d_wwost_pi_plus: str = LoNM.get_decay(from_part=LPNs.d_prime_one_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_plus_str])

    subdec_d_one_zero_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_one_zero_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_one_minus_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_one_minus_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_one_zero_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_one_zero_to_d_wwost_pi_plus, with_mother=True)
    subdec_d_one_minus_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_one_minus_to_d_wwost_pi_plus, with_mother=True)

    subdec_d_twostar_zero_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_twostar_zero_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_twostar_minus_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_twostar_minus_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_twostar_zero_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_twostar_zero_to_d_wwost_pi_plus, with_mother=True)
    subdec_d_twostar_minus_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_twostar_minus_to_d_wwost_pi_plus, with_mother=True)

    subdec_d_zerostar_zero_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_zerostar_zero_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_zerostar_minus_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_zerostar_minus_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_zerostar_zero_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_zerostar_zero_to_d_wwost_pi_plus, with_mother=True)
    subdec_d_zerostar_minus_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_zerostar_minus_to_d_wwost_pi_plus, with_mother=True)

    subdec_d_primeone_zero_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_primeone_zero_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_primeone_minus_to_d_wwost_pi_zero: str = LoNM.get_as_sub_decay(decay=d_primeone_minus_to_d_wwost_pi_zero, with_mother=True)
    subdec_d_primeone_zero_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_primeone_zero_to_d_wwost_pi_plus, with_mother=True)
    subdec_d_primeone_minus_to_d_wwost_pi_plus: str = LoNM.get_as_sub_decay(decay=d_primeone_minus_to_d_wwost_pi_plus, with_mother=True)

    # D** Decay
    d_stst_to_three_daugthers: str = LoNM.get_decay(from_part=LPNs.d_st_st_str, to_parts=[LPNs.small_x_str, LPNs.small_x_str, LPNs.small_x_str])
    subdec_d_stst_to_three_daugthers: str = LoNM.get_as_sub_decay(decay=d_stst_to_three_daugthers, with_mother=True)

    #
    # Signal

    b_plus_to_d_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_d_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_d_st_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_d_st_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])

    b_plus_to_d_st_zero_tau_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_pi, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_d_st_minus_tau_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_st_minus_to_d_pi, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_d_st_zero_tau_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_photon, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_d_st_minus_tau_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[impossible_subdec_d_st_minus_to_d_photon, LPNs.tau_plus_str, LPNs.nu_tau_str])[:-1] + r"\;" + _impossible_tag[1:]

    # With tau decay strings
    b_plus_to_d_zero_tau_plus_nu_via_e: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_plus_to_d_zero_tau_plus_nu_via_mu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_d_minus_tau_plus_nu_via_e: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_d_minus_tau_plus_nu_via_mu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])

    b_plus_to_dst_zero_tau_plus_nu_via_e: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_plus_to_dst_zero_tau_plus_nu_via_mu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_dst_minus_tau_plus_nu_via_e: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_dst_minus_tau_plus_nu_via_mu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])

    b_plus_to_dst_zero_tau_plus_nu_via_e_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_pi, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_plus_to_dst_zero_tau_plus_nu_via_mu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_pi, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_dst_minus_tau_plus_nu_via_e_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_st_minus_to_d_pi, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_dst_minus_tau_plus_nu_via_mu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_st_minus_to_d_pi, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])

    b_plus_to_dst_zero_tau_plus_nu_via_e_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_photon, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])
    b_plus_to_dst_zero_tau_plus_nu_via_mu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_photon, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])
    b_zero_to_dst_minus_tau_plus_nu_via_e_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[impossible_subdec_d_st_minus_to_d_photon, subdec_tau_plus_to_e_plus_nu_nu, LPNs.nu_tau_str])[:-1] + r"\;" + _impossible_tag[1:]
    b_zero_to_dst_minus_tau_plus_nu_via_mu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[impossible_subdec_d_st_minus_to_d_photon, subdec_tau_plus_to_mu_plus_nu_nu, LPNs.nu_tau_str])[:-1] + r"\;" + _impossible_tag[1:]

    #
    # Norm
    b_plus_to_d_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_d_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_d_st_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_d_st_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])

    b_plus_to_d_st_zero_e_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_pi, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_d_st_minus_e_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_st_minus_to_d_pi, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_d_st_zero_e_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_photon, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_d_st_minus_e_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[impossible_subdec_d_st_minus_to_d_photon, LPNs.electron_plus_str, LPNs.nu_e_str])[:-1] + r"\;" + _impossible_tag[1:]

    b_plus_to_d_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_d_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_d_st_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_d_st_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])

    b_plus_to_d_st_zero_mu_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_pi, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_d_st_minus_mu_plus_nu_with_Dst_to_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_st_minus_to_d_pi, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_d_st_zero_mu_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_st_zero_to_d_photon, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_d_st_minus_mu_plus_nu_with_Dst_to_photon: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[impossible_subdec_d_st_minus_to_d_photon, LPNs.muon_plus_str, LPNs.nu_mu_str])[:-1] + r"\;" + _impossible_tag[1:]

    #
    # B -> D** l nu
    # B -> Done l nu
    b_plus_to_dss_d_one_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_one_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_one_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_one_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_one_zero_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_one_minus_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_one_zero_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_one_minus_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])

    b_plus_to_dss_d_one_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_one_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_one_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_one_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_one_zero_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_one_minus_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_one_zero_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_one_minus_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])

    b_plus_to_dss_d_one_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_one_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_one_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_one_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_one_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_one_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_one_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_one_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])

    # B -> Dtwostar l nu
    b_plus_to_dss_d_twostar_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_two_st_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_twostar_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_two_st_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_twostar_zero_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_twostar_minus_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_twostar_zero_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_twostar_minus_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])

    b_plus_to_dss_d_twostar_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_two_st_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_twostar_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_two_st_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_twostar_zero_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_twostar_minus_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_twostar_zero_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_twostar_minus_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])

    b_plus_to_dss_d_twostar_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_two_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_twostar_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_two_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_twostar_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_twostar_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_twostar_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_twostar_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])

    # B -> Dzerostar l nu
    b_plus_to_dss_d_zerostar_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_sszero_st_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_zerostar_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_sszero_st_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_zerostar_zero_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_zerostar_minus_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_zerostar_zero_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_zerostar_minus_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])

    b_plus_to_dss_d_zerostar_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_sszero_st_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_zerostar_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_sszero_st_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_zerostar_zero_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_zerostar_minus_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_zerostar_zero_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_zerostar_minus_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])

    b_plus_to_dss_d_zerostar_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_sszero_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_zerostar_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_sszero_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_zerostar_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_zerostar_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_zerostar_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_zerostar_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])

    # B -> Dprimeone l nu
    b_plus_to_dss_d_primeone_zero_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_prime_one_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_primeone_minus_e_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_prime_one_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_primeone_zero_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_primeone_minus_e_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_zero, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_plus_to_dss_d_primeone_zero_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])
    b_zero_to_dss_d_primeone_minus_e_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_plus, LPNs.electron_plus_str, LPNs.nu_e_str])

    b_plus_to_dss_d_primeone_zero_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_prime_one_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_primeone_minus_mu_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_prime_one_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_primeone_zero_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_primeone_minus_mu_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_zero, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_plus_to_dss_d_primeone_zero_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])
    b_zero_to_dss_d_primeone_minus_mu_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_plus, LPNs.muon_plus_str, LPNs.nu_mu_str])

    b_plus_to_dss_d_primeone_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_prime_one_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_primeone_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_prime_one_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_primeone_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_primeone_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_zero, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_plus_to_dss_d_primeone_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_dss_d_primeone_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_plus, LPNs.tau_plus_str, LPNs.nu_tau_str])

    #
    # BKGs
    # Hadronic tau BKG
    # Hadronic B -> D tau_to_had nu
    hadronic_tau_b_plus_to_d_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_zero_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_d_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> D* tau_to_had nu
    hadronic_tau_b_plus_to_d_st_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_zero_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_d_st_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> D(*) tau_to_had nu
    hadronic_tau_b_plus_to_d_wwost_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_d_wwost_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> D(*) tau_to_had nu
    hadronic_tau_b_plus_to_d_wwostwwost_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_w_wo_st_zero_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_d_wwostwwost_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_w_wo_st_minus_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])

    # Hadronic B -> Done tau_to_had nu
    hadronic_tau_b_plus_to_dss_d_one_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_one_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_plus_to_dss_d_one_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_one_zero_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_one_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_one_minus_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> Dtwostar tau_to_had nu
    hadronic_tau_b_plus_to_dss_d_twostar_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_twostar_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_plus_to_dss_d_twostar_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_twostar_zero_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_twostar_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_twostar_minus_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> Dzerostar tau_to_had nu
    hadronic_tau_b_plus_to_dss_d_zerostar_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_zerostar_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_plus_to_dss_d_zerostar_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_zerostar_zero_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_zerostar_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_zerostar_minus_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    # Hadronic B -> Dprimeone tau_to_had nu
    hadronic_tau_b_plus_to_dss_d_primeone_zero_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_primeone_minus_tau_plus_nu_with_dss_via_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_zero, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_plus_to_dss_d_primeone_zero_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[subdec_d_primeone_zero_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    hadronic_tau_b_zero_to_dss_d_primeone_minus_tau_plus_nu_with_dss_via_pip: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[subdec_d_primeone_minus_to_d_wwost_pi_plus, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])

    #
    # Hadronic Background (HBKG) -> Reconstructed lepton is no lepton

    lepton_bkg: str = r"$\mathrm{Lepton}" + r"\;" + _bkg_tag[1:]
    long_lepton_fake_bkg: str = LPNs.lepton_str[:-1] + r"\," + _fake_tag[1:-1] + r"\," + _bkg_tag[1:]
    lepton_fake_bkg: str = LPNs.lepton_str[:-1] + r"\," + _fake_tag[1:]
    reco_lepton_fake_bkg: str = _reco_tag[:-1] + r"\," + LPNs.lepton_str[1:-1] + r"\," + _fake_tag[1:]
    # Generic Backgrounds
    b_to_dss_l_nu_with_dss_to_many_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[subdec_d_stst_to_three_daugthers, LPNs.lepton_str, LPNs.nu_str])
    lepton_without_relation_to_b_bkg: str = LPNs.b_str[:-1] + r"\,\not\to\," + LPNs.lepton_str[1:]
    d_fake_bkg: str = LPNs.d_w_wo_st_str[:-1] + r"\," + _fake_tag[1:]
    b_to_d_x_x_l_x_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.small_x_str, LPNs.small_x_str, LPNs.lepton_str, LPNs.small_x_str])
    lepton_from_other_b_decay_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.others_dots, LPNs.small_x_str, LPNs.others_dots])
    #
    # B -> pi l nu
    b_plus_to_pi_zero_l_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.pi_zero_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    b_plus_to_pi_zero_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.pi_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])
    b_zero_to_pi_minus_l_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.pi_minus_str, LPNs.lepton_plus_str, LPNs.nu_lep_str])
    b_zero_to_pi_minus_tau_plus_nu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.pi_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])

    #
    # Mismatch G BKGs
    _fake_e_from_reco_mu: str = _fake_tag[:-1] + r"\," + LPNs.electron_str[1:-1] + r"\,from\," + _reco_tag[1:-1] + r"\," + LPNs.mu_str[1:]
    _fake_mu_from_reco_e: str = _fake_tag[:-1] + r"\," + LPNs.mu_str[1:-1] + r"\,from\," + _reco_tag[1:-1] + r"\," + LPNs.electron_str[1:]

    tau_to_fake_e_from_reco_mu: str = LoNM.get_decay(from_part=LPNs.tau_str, to_parts=[_fake_e_from_reco_mu])
    tau_to_fake_mu_from_reco_e: str = LoNM.get_decay(from_part=LPNs.tau_str, to_parts=[_fake_mu_from_reco_e])
    subdec_tau_to_fake_e_from_reco_mu: str = LoNM.get_as_sub_decay(decay=tau_to_fake_e_from_reco_mu, with_mother=True)
    subdec_tau_to_fake_mu_from_reco_e: str = LoNM.get_as_sub_decay(decay=tau_to_fake_mu_from_reco_e, with_mother=True)

    lepton_cross_feed_tau_with_fake_mu_from_reco_e_bkg: str = tau_to_fake_mu_from_reco_e[:-1] + r"\ " + _cross_feed_suffix[1:]
    lepton_cross_feed_tau_with_fake_e_from_reco_mu_bkg: str = tau_to_fake_e_from_reco_mu[:-1] + r"\ " + _cross_feed_suffix[1:]
    lepton_cross_feed_fake_mu_from_reco_e_bkg: str = _fake_mu_from_reco_e[:-1] + r"\ " + _cross_feed_suffix[1:]
    lepton_cross_feed_fake_e_from_reco_mu_bkg: str = _fake_e_from_reco_mu[:-1] + r"\ " + _cross_feed_suffix[1:]

    #
    dss_daughter_mismatch_b_zero_to_dss_plus_e_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_zero_to_dss_plus_mu_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_zero_to_dss_plus_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_plus_e_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.electron_plus_str, LPNs.nu_e_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_plus_mu_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.muon_plus_str, LPNs.nu_mu_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_plus_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_minus_str, LPNs.tau_plus_str, LPNs.nu_tau_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_zero_to_dss_zero_e_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_zero_to_dss_zero_mu_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_zero_to_dss_zero_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_zero_e_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.electron_plus_str, LPNs.nu_e_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_zero_mu_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.muon_plus_str, LPNs.nu_mu_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]
    dss_daughter_mismatch_b_plus_to_dss_zero_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_st_zero_str, LPNs.tau_plus_str, LPNs.nu_tau_str])[:-1] + r"\,\mathrm{with}\," + LPNs.d_st_st_str[1:-1] + r"\," + _daughter_mismatch_suffix[1:]

    #
    # Generic B decays (all the events not covered by any other query)
    generic_bbbar_bkg: str = _generic_tag[:-1] + r"\ " + to_b_bbar_bkg[1:]
    #
    # Continuum BKG
    continuum_bkg: str = to_continuum_txt[:-1] + r"\ " + _bkg_tag[1:]
    charm_continuum_bkg: str = to_charm_continuum_txt[:-1] + r"\ " + _bkg_tag[1:]
    uds_continuum_bkg: str = to_uds_continuum_txt[:-1] + r"\ " + _bkg_tag[1:]

    # More labels for grouped background types:
    hadronic_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    b_to_dstst_hadronic_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_st_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    b_to_any_d_hadronic_tau_bkg: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_w_wo_st_str, subdec_tau_plus_to_hadrons_nu, LPNs.nu_tau_str])
    described_generic_bkg: str = r"$\mathrm{Known}" + r"\ " + _generic_tag[1:-1] + r"\ " + _bkg_tag[1:]
    lepton_cross_feed_bkg: str = LPNs.lepton_str[:-1] + r"\ " + _cross_feed_suffix[1:]
    b_to_dstst_with_dstst_daughter_mismatch_bkg: str = LPNs.d_st_st_str[:-1] + r"\ " + _daughter_mismatch_suffix[1:]

    #######################
    # General Reco Labels #
    #######################
    all_modes_label: str = get_all_reco_modes_str_for(particle="", short=False)
    all_modes_label_short: str = get_all_reco_modes_str_for(particle="", short=True)

    neutral_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.b_zero_str, short=False, all_prefix=False)
    neutral_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.b_zero_str, short=True, all_prefix=False)
    charged_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.b_plus_str, short=False, all_prefix=False)
    charged_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.b_plus_str, short=True, all_prefix=False)

    # B -> D(*) l
    reco_b0_to_dm_wwos_e: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.electron_plus_str])
    reco_bp_to_d0_wwos_e: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.electron_plus_str])
    reco_b0_to_dm_wwos_mu: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.muon_plus_str])
    reco_bp_to_d0_wwos_mu: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.muon_plus_str])

    reco_b0_to_dm_wwos_lepton: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.lepton_plus_str])
    reco_bp_to_d0_wwos_lepton: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.lepton_plus_str])

    reco_b_to_d_e: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.electron_plus_str])
    reco_b_to_d_mu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.muon_plus_str])
    reco_b_to_ds_e: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.electron_plus_str])
    reco_b_to_ds_mu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.muon_plus_str])

    reco_b_to_d_wwos_e: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.electron_plus_str])
    reco_b_to_d_wwos_mu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.muon_plus_str])

    reco_b_to_d: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.lepton_plus_str])
    reco_b_to_ds: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.lepton_plus_str])

    # B -> D** l
    reco_b0_to_dssm_e_with_dss_to_dm_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.pi_zero_str, LPNs.electron_plus_str])
    reco_bp_to_dss0_e_with_dss_to_d0_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.pi_zero_str, LPNs.electron_plus_str])
    reco_b0_to_dssm_mu_with_dss_to_dm_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.pi_zero_str, LPNs.muon_plus_str])
    reco_bp_to_dss0_mu_with_dss_to_d0_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.pi_zero_str, LPNs.muon_plus_str])

    reco_b0_to_dssm_e_with_dss_to_d0_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.pi_minus_str, LPNs.electron_plus_str])
    reco_bp_to_dss0_e_with_dss_to_dp_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_plus_str, LPNs.pi_minus_str, LPNs.electron_plus_str])
    reco_b0_to_dssm_mu_with_dss_to_d0_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.pi_minus_str, LPNs.muon_plus_str])
    reco_bp_to_dss0_mu_with_dss_to_dp_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_plus_str, LPNs.pi_minus_str, LPNs.muon_plus_str])

    reco_b0_to_dssm_e_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.electron_plus_str])
    reco_bp_to_dss0_e_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.electron_plus_str])
    reco_b0_to_dssm_mu_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.muon_plus_str])
    reco_bp_to_dss0_mu_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.muon_plus_str])

    reco_b_to_dss_e_with_dss_to_d_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.electron_plus_str])
    reco_b_to_dss_e_with_dss_to_d_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_minus_str, LPNs.electron_plus_str])
    reco_b_to_dss_mu_with_dss_to_d_wwos_pi0: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.muon_plus_str])
    reco_b_to_dss_mu_with_dss_to_d_wwos_pim: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_minus_str, LPNs.muon_plus_str])

    reco_b_to_dss_e_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.electron_plus_str])
    reco_b_to_dss_mu_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.muon_plus_str])

    reco_b_to_d_w_wo_st_w_wo_pi_e: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.wwo_pi_str, LPNs.electron_plus_str])
    reco_b_to_d_w_wo_st_w_wo_pi_mu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.wwo_pi_str, LPNs.muon_plus_str])

    reco_b0_to_dss_l_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_bp_to_dss_l_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b0_to_d_w_wo_st_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_bp_to_d_w_wo_st_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])

    reco_b_to_dss_l_with_dss_to_d_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_b_to_dss_l_with_dss_to_ds_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b_to_dss_l_with_dss_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b_to_d_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_b_to_ds_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])

    reco_b0_to_d_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_b0_to_ds_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_bp_to_d_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_bp_to_ds_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])

    reco_b_to_d0_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_zero_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_b_to_ds0_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_zero_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_b_to_dm_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_minus_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_b_to_dsm_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_minus_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b_to_d0_wwos_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_zero_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_b_to_dm_wwos_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_minus_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b_to_dm_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_minus_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_b_to_dsm_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_minus_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_b_to_d0_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_zero_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])
    reco_b_to_ds0_w_wo_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_zero_str, LPNs.wwo_pi_str, LPNs.lepton_plus_str])

    reco_b0_to_dm_wwos_wwos_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_w_wo_st_minus_str, LPNs.lepton_plus_str])
    reco_bp_to_d0_wwos_wwos_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_w_wo_st_zero_str, LPNs.lepton_plus_str])

    #
    reco_bz_to_d_wwos: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.lepton_str])
    reco_bpm_to_d_wwos: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.lepton_str])
    reco_bz_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_and_pi_str, LPNs.lepton_str])
    reco_bpm_to_d_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_w_wo_st_and_pi_str, LPNs.lepton_str])
    reco_bz_to_d_wwos_wwo_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_and_wwo_pi_str, LPNs.lepton_str])
    reco_bpm_to_d_wwos_wwo_pi: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_w_wo_st_and_wwo_pi_str, LPNs.lepton_str])

    reco_b_to_dpm_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.charged_d_w_wo_st_and_pi_str, LPNs.lepton_str])
    reco_b_to_dz_wwos_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.neutral_d_w_wo_st_and_pi_str, LPNs.lepton_str])
    reco_b_to_dpm_wwos_wwo_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.charged_d_w_wo_st_and_wwo_pi_str, LPNs.lepton_str])
    reco_b_to_dz_wwos_wwo_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.neutral_d_w_wo_st_and_wwo_pi_str, LPNs.lepton_str])

    # Reconstruction Labels (with Charge, no Neutrinos)
    reco_b0_to_dm_lp: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str])
    reco_b0_to_d_stm_lp: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str])
    reco_b0_to_dm_ep: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str])
    reco_b0_to_d_stm_ep: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str])
    reco_b0_to_dm_mup: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str])
    reco_b0_to_d_stm_mup: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str])
    reco_b0_to_dm_taup: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str])
    reco_b0_to_d_stm_taup: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str])

    reco_bm_to_d0_lm: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.lepton_minus_str])
    reco_bm_to_d_st0_lm: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.lepton_minus_str])
    reco_bm_to_d0_em: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_minus_str])
    reco_bm_to_d_st0_em: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_minus_str])
    reco_bm_to_d0_mum: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_minus_str])
    reco_bm_to_d_st0_mum: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_minus_str])
    reco_bm_to_d0_taum: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_minus_str])
    reco_bm_to_d_st0_taum: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_minus_str])

    reco_b0_to_dm_lp_pi0, reco_b0_to_dm_pi0_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_lp_pi0, reco_b0_to_d_stm_pi0_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_ep_pi0, reco_b0_to_dm_pi0_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.electron_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_ep_pi0, reco_b0_to_d_stm_pi0_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.electron_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_mup_pi0, reco_b0_to_dm_pi0_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.muon_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_mup_pi0, reco_b0_to_d_stm_pi0_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.muon_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_dm_taup_pi0, reco_b0_to_dm_pi0_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str, LPNs.pi_zero_str])  # type: str, str
    reco_b0_to_d_stm_taup_pi0, reco_b0_to_d_stm_pi0_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str, LPNs.pi_zero_str])  # type: str, str

    reco_bm_to_d0_lm_pi0, reco_bm_to_d0_pi0_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.lepton_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_lm_pi0, reco_bm_to_d_st0_pi0_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.lepton_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_em_pi0, reco_bm_to_d0_pi0_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.electron_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_em_pi0, reco_bm_to_d_st0_pi0_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.electron_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_mum_pi0, reco_bm_to_d0_pi0_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.muon_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_mum_pi0, reco_bm_to_d_st0_pi0_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.muon_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d0_taum_pi0, reco_bm_to_d0_pi0_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_zero_str, LPNs.tau_minus_str, LPNs.pi_zero_str])  # type: str, str
    reco_bm_to_d_st0_taum_pi0, reco_bm_to_d_st0_pi0_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_zero_str, LPNs.tau_minus_str, LPNs.pi_zero_str])  # type: str, str

    reco_b0_to_anti_d0_lp_pim, reco_b0_to_anti_d0_pim_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.lepton_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_lp_pim, reco_b0_to_anti_d_st0_pim_lp = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.lepton_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_ep_pim, reco_b0_to_anti_d0_pim_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.electron_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_ep_pim, reco_b0_to_anti_d_st0_pim_ep = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.electron_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_mup_pim, reco_b0_to_anti_d0_pim_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.muon_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_mup_pim, reco_b0_to_anti_d_st0_pim_mup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.muon_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d0_taup_pim, reco_b0_to_anti_d0_pim_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_zero_str), LPNs.tau_plus_str, LPNs.pi_minus_str])  # type: str, str
    reco_b0_to_anti_d_st0_taup_pim, reco_b0_to_anti_d_st0_pim_taup = LoNM.get_decay_and_last_switched(from_part=LPNs.b_zero_str, to_parts=[LoNM.get_anti_particle(LPNs.d_st_zero_str), LPNs.tau_plus_str, LPNs.pi_minus_str])  # type: str, str

    reco_bm_to_dp_lm_pim, reco_bm_to_dp_pim_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.lepton_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_lm_pim, reco_bm_to_d_stp_pim_lm = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.lepton_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_em_pim, reco_bm_to_dp_pim_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.electron_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_em_pim, reco_bm_to_d_stp_pim_em = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.electron_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_mum_pim, reco_bm_to_dp_pim_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.muon_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_mum_pim, reco_bm_to_d_stp_pim_mum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.muon_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_dp_taum_pim, reco_bm_to_dp_pim_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_plus_str, LPNs.tau_minus_str, LPNs.pi_minus_str])  # type: str, str
    reco_bm_to_d_stp_taum_pim, reco_bm_to_d_stp_pim_taum = LoNM.get_decay_and_last_switched(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_plus_str, LPNs.tau_minus_str, LPNs.pi_minus_str])  # type: str, str

    # B -> D(*) pi l  Reco Labels with B charge and D excitation
    reco_b0_to_d_pi_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_bp_to_d_pi_l: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.lepton_plus_str])

    reco_b0_to_ds_pi_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.lepton_plus_str])
    reco_bp_to_ds_pi_l: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.lepton_plus_str])

    # Reconstruction Labels (with Charge, no Neutrinos)
    reco_b0_to_pim_lp: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.pi_minus_str, LPNs.lepton_plus_str])
    reco_b0_to_pim_ep: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.pi_minus_str, LPNs.electron_plus_str])
    reco_b0_to_pim_mup: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.pi_minus_str, LPNs.muon_plus_str])

    reco_bp_to_pi0_lp: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.pi_zero_str, LPNs.lepton_plus_str])
    reco_bp_to_pi0_ep: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.pi_zero_str, LPNs.electron_plus_str])
    reco_bp_to_pi0_mup: str = LoNM.get_decay(from_part=LPNs.b_plus_str, to_parts=[LPNs.pi_zero_str, LPNs.muon_plus_str])

    reco_b_to_pi_e: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.pi_str, LPNs.electron_plus_str])
    reco_b_to_pi_mu: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.pi_str, LPNs.muon_plus_str])

    reco_b_to_pi_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.pi_str, LPNs.lepton_plus_str])

    #
    reco_b_to_dstst_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_st_str, LPNs.lepton_str])
    all_b_to_dss_l_modes_label: str = get_all_reco_modes_str_for(particle=reco_b_to_dstst_l, short=False)
    all_b_to_dss_l_modes_label_short: str = get_all_reco_modes_str_for(particle=reco_b_to_dstst_l, short=True)
    reco_b0_to_dstst_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_st_str, LPNs.lepton_str])
    neutral_b_to_dss_l_modes_label: str = get_all_reco_modes_str_for(particle=reco_b0_to_dstst_l, short=False)
    neutral_b_to_dss_l_modes_label_short: str = get_all_reco_modes_str_for(particle=reco_b0_to_dstst_l, short=True)
    reco_bpm_to_dstst_l: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_st_st_str, LPNs.lepton_str])
    charged_b_to_dss_l_modes_label: str = get_all_reco_modes_str_for(particle=reco_bpm_to_dstst_l, short=False)
    charged_b_to_dss_l_modes_label_short: str = get_all_reco_modes_str_for(particle=reco_bpm_to_dstst_l, short=True)

    reco_b_to_d_wwos_pi0_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    all_reco_b_to_d_wwos_pi0_l_label: str = get_all_reco_modes_str_for(particle=reco_b_to_d_wwos_pi0_l, short=False)
    all_reco_b_to_d_wwos_pi0_l_label_short: str = get_all_reco_modes_str_for(particle=reco_b_to_d_wwos_pi0_l, short=True)
    reco_b0_to_d_wwos_pi0_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    all_reco_b0_to_d_wwos_pi0_l_label: str = get_all_reco_modes_str_for(particle=reco_b0_to_d_wwos_pi0_l, short=False)
    all_reco_b0_to_d_wwos_pi0_l_label_short: str = get_all_reco_modes_str_for(particle=reco_b0_to_d_wwos_pi0_l, short=True)
    reco_bpm_to_d_wwos_pi0_l: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    all_reco_bpm_to_d_wwos_pi0_l_label: str = get_all_reco_modes_str_for(particle=reco_bpm_to_d_wwos_pi0_l, short=False)
    all_reco_bpm_to_d_wwos_pi0_l_label_short: str = get_all_reco_modes_str_for(particle=reco_bpm_to_d_wwos_pi0_l, short=True)

    reco_b_to_d_wwos_pip_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    all_reco_b_to_d_wwos_pip_l_label: str = get_all_reco_modes_str_for(particle=reco_b_to_d_wwos_pip_l, short=False)
    all_reco_b_to_d_wwos_pip_l_label_short: str = get_all_reco_modes_str_for(particle=reco_b_to_d_wwos_pip_l, short=True)
    reco_b0_to_d_wwos_pip_l: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    all_reco_b0_to_d_wwos_pip_l_label: str = get_all_reco_modes_str_for(particle=reco_b0_to_d_wwos_pip_l, short=False)
    all_reco_b0_to_d_wwos_pip_l_label_short: str = get_all_reco_modes_str_for(particle=reco_b0_to_d_wwos_pip_l, short=True)
    reco_bpm_to_d_wwos_pip_l: str = LoNM.get_decay(from_part=LPNs.b_pm_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    all_reco_bpm_to_d_wwos_pip_l_label: str = get_all_reco_modes_str_for(particle=reco_bpm_to_d_wwos_pip_l, short=False)
    all_reco_bpm_to_d_wwos_pip_l_label_short: str = get_all_reco_modes_str_for(particle=reco_bpm_to_d_wwos_pip_l, short=True)

    #
    # D** Validation Reco Mode Labels
    all_dss_val_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_st_st_str, short=False)
    all_dss_val_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.d_st_st_str, short=True)

    reco_dss_val_d_neutral_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.pi_zero_str, LPNs.lepton_str])
    reco_dss_val_ds_neutral_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    reco_dss_val_d_charged_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.pi_pm_str, LPNs.lepton_str])
    reco_dss_val_ds_charged_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    reco_dss_val_neutral_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    reco_dss_val_charged_pi: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    reco_dss_val_charged_b_neutral_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    reco_dss_val_charged_b_charged_pi: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    reco_dss_val_neutral_b_neutral_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_zero_str, LPNs.lepton_str])
    reco_dss_val_neutral_b_charged_pi: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_w_wo_st_str, LPNs.pi_pm_str, LPNs.lepton_str])
    reco_dss_val_neutral_b_d: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.lepton_str])
    reco_dss_val_charged_b_d: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.lepton_str])
    reco_dss_val_neutral_b_d_star: str = LoNM.get_decay(from_part=LPNs.b_zero_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.lepton_str])
    reco_dss_val_charged_b_d_star: str = LoNM.get_decay(from_part=LPNs.b_minus_str, to_parts=[LPNs.d_st_str, LPNs.pi_str, LPNs.lepton_str])

    #
    # Products Only Reconstruction Labels (with Charge, no Neutrinos)
    reco_to_dm_lp: str = LoNM.get_decay_products(to_parts=[LPNs.d_minus_str, LPNs.lepton_plus_str])
    reco_to_d_stm_lp: str = LoNM.get_decay_products(to_parts=[LPNs.d_st_minus_str, LPNs.lepton_plus_str])

    reco_to_dm_taup: str = LoNM.get_decay_products(to_parts=[LPNs.d_minus_str, LPNs.tau_plus_str])
    reco_to_d_stm_taup: str = LoNM.get_decay_products(to_parts=[LPNs.d_st_minus_str, LPNs.tau_plus_str])

    #
    # Reconstruction Labels (no Charge, no Neutrinos)
    reco_b_to_d_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.lepton_str])
    reco_b_to_d_st_l: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.lepton_str])

    reco_b_to_d_tau: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_str, LPNs.tau_str])
    reco_b_to_d_st_tau: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.tau_str])

    #
    # Products Only Reconstruction Labels (no Charge, no Neutrinos)
    reco_to_d_l: str = LoNM.get_decay_products(to_parts=[LPNs.d_str, LPNs.lepton_str])
    reco_to_d_st_l: str = LoNM.get_decay_products(to_parts=[LPNs.d_st_str, LPNs.lepton_str])

    reco_to_d_tau: str = LoNM.get_decay_products(to_parts=[LPNs.d_str, LPNs.tau_str])
    reco_to_d_st_tau: str = LoNM.get_decay_products(to_parts=[LPNs.d_st_str, LPNs.tau_str])

    #
    # Upsilon(4s) Reco Modes
    u4s_to_bp_tag_bm_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_plus_str, LPNs.b_sig_minus_str])
    u4s_to_b0_tag_anti_b0_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_anti_neutral_str])
    u4s_to_b0_tag_b0_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_neutral_str])
    u4s_to_b0_tag_bm_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_minus_str])
    u4s_to_b0_tag_bp_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_plus_str])
    u4s_to_bp_tag_b0_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_plus_str, LPNs.b_sig_neutral_str])
    u4s_to_bm_tag_b0_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_minus_str, LPNs.b_sig_neutral_str])

    u4s_to_neutral_bbbar: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_anti_neutral_str])[:-1] + r"\,\mathrm{with\,Mixing}$"

    u4s_to_b_tag_bm_sig_with_wrong_tag: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_pm_str])[:-1] + r"\,\mathrm{with\,wrong\,Tag}$"
    u4s_to_b_tag_0_sig_with_wrong_tag: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_neutral_str])[:-1] + r"\,\mathrm{with\,wrong\,Tag}$"

    u4s_to_all_correct_b_tag_b_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_str])
    u4s_to_all_wrong_b_tag_b_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_str])[:-1] + r"\,\mathrm{with\,wrong\,Tag}$"

    u4s_to_b_tag_bpm_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_pm_str])
    u4s_to_b_tag_b0_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_str, LPNs.b_sig_neutral_str])

    u4s_to_bpm_tag_b_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_pm_str, LPNs.b_sig_str])
    u4s_to_b0_tag_b_sig: str = LoNM.get_decay(from_part=LPNs.y4s_str, to_parts=[LPNs.b_tag_neutral_str, LPNs.b_sig_str])

    all_u4s_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.y4s_str, short=False)
    all_u4s_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.y4s_str, short=True)

    #
    # B_tag Reco Modes
    charged_tag_reco_modes: str = LPNs.b_tag_pm_str + " Decays"
    neutral_tag_reco_modes: str = LPNs.b_tag_neutral_str + " Decays"

    all_b_tag_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.b_tag_str, short=False)
    all_b_tag_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.b_tag_str, short=True)

    charged_b_to_d_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.anti_d_zero_str, LPNs.pi_plus_str, LPNs.others_dots])
    neutral_b_to_d_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_minus_str, LPNs.pi_plus_str, LPNs.others_dots])
    charged_b_to_d_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.anti_d_zero_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    neutral_b_to_d_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    charged_b_to_ocd_pipi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.d_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.others_dots])
    neutral_b_to_ocd_pipi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.anti_d_zero_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    charged_b_to_ds_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.pi_plus_str, LPNs.others_dots])
    neutral_b_to_ds_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_st_minus_str, LPNs.pi_plus_str, LPNs.others_dots])
    charged_b_to_ds_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.anti_d_st_zero_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    neutral_b_to_ds_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_st_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    charged_b_to_dd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.d_str, LPNs.d_str, LPNs.others_dots])
    neutral_b_to_dd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_str, LPNs.d_str, LPNs.others_dots])
    charged_b_to_dsd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.d_st_str, LPNs.d_str, LPNs.others_dots])
    neutral_b_to_dsd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.d_st_str, LPNs.d_str, LPNs.others_dots])
    charged_b_to_j_psi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[LPNs.j_psi_str, LPNs.others_dots])
    neutral_b_to_j_psi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[LPNs.j_psi_str, LPNs.others_dots])
    charged_b_to__other__tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_pm_str, to_parts=[r"$\mathrm{Others}$", LPNs.others_dots])
    neutral_b_to__other__tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_neutral_str, to_parts=[r"$\mathrm{Others}$", LPNs.others_dots])

    b_to_d_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_str, LPNs.pi_plus_str, LPNs.others_dots])
    b_to_d_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    b_to_ocd_pipi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_str, LPNs.pi_str, LPNs.pi_str, LPNs.others_dots])
    b_to_ds_1cpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_st_str, LPNs.pi_plus_str, LPNs.others_dots])
    b_to_ds_mcpi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_st_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.others_dots])
    b_to_dd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_str, LPNs.d_str, LPNs.others_dots])
    b_to_dsd_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.d_st_str, LPNs.d_str, LPNs.others_dots])
    b_to_j_psi_tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[LPNs.j_psi_str, LPNs.others_dots])
    b_to__other__tag_modes: str = LoNM.get_decay(from_part=LPNs.b_tag_str, to_parts=[r"$\mathrm{Others}$", LPNs.others_dots])

    #
    # D(*) Reco Modes
    # D*0 Channels
    dst0_to_d0_pi0: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.pi_zero_str])
    dst0_to_d0_gamma: str = LoNM.get_decay(from_part=LPNs.d_st_zero_str, to_parts=[LPNs.d_zero_str, LPNs.gamma_str])
    # D*+ Channels
    dstp_to_d0_pip: str = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_zero_str, LPNs.pi_plus_str])
    dstp_to_dp_pi0: str = LoNM.get_decay(from_part=LPNs.d_st_plus_str, to_parts=[LPNs.d_plus_str, LPNs.pi_zero_str])
    # D+ Channels
    dp_to_kaon_m_pip_pip: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str])
    dp_to_kaon_m_pip_pip_pi0: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_zero_str])
    dp_to_kaon_m_pip_pip_pip_pim: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])
    dp_to_kaon_s0_pip: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str])
    dp_to_kaon_s0_pip_pi0: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_zero_str])
    dp_to_kaon_s0_pip_pip_pim: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])
    dp_to_kaon_s0_kaon_p: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_short_str, LPNs.k_plus_str])
    dp_to_kaon_p_kaon_m_pip: str = LoNM.get_decay(from_part=LPNs.d_plus_str, to_parts=[LPNs.k_plus_str, LPNs.k_minus_str, LPNs.pi_plus_str])
    # D0 Channels
    d0_to_kaon_m_pip: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str])
    d0_to_kaon_m_pip_pi0: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_zero_str])
    d0_to_kaon_m_pip_pip_pim: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str])
    d0_to_kaon_m_pip_pip_pim_pi0: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.pi_plus_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.pi_zero_str])
    d0_to_kaon_s0_pi0: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_zero_str])
    d0_to_kaon_s0_pip_pim: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_minus_str])
    d0_to_kaon_s0_pip_pim_pi0: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.pi_plus_str, LPNs.pi_minus_str, LPNs.pi_zero_str])
    d0_to_kaon_m_kaon_p: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_minus_str, LPNs.k_plus_str])

    # D* Reco Mode Combinations
    dstpm_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_st_pm_str, short=True, all_prefix=False)
    dst0_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_st_zero_str, short=True, all_prefix=False)

    # D Reco Mode Combinations
    dpm_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_pm_str, short=True, all_prefix=False)
    d0_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_zero_str, short=True, all_prefix=False)

    # D+ Combinations
    dpm_to_kaon_mp_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[LPNs.k_mp_str, LPNs.others_dots])
    dpm_to_kaon_s0_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[LPNs.k_short_str, LPNs.others_dots])

    dpm_to_no_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[r"$\mathrm{no}\;$", LPNs.pi_str])
    dpm_to_1_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[r"$1\;$", LPNs.pi_str])
    dpm_to_2_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[r"$2\;$", LPNs.pi_str])
    dpm_to_multiple_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_pm_str, to_parts=[r"$\mathrm{multiple}\;$", LPNs.pi_str])

    # D0 Combinations
    d0_to_kaon_pm_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_mp_str, LPNs.others_dots])
    d0_to_kaon_s0_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[LPNs.k_short_str, LPNs.others_dots])

    d0_to_no_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[r"$\mathrm{no}\;$", LPNs.pi_str])
    d0_to_1_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[r"$1\;$", LPNs.pi_str])
    d0_to_2_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[r"$2\;$", LPNs.pi_str])
    d0_to_multiple_pi_reco_modes_label: str = LoNM.get_decay(from_part=LPNs.d_zero_str, to_parts=[r"$\mathrm{multiple}\;$", LPNs.pi_str])

    # D(*) Reco Mode Combinations
    all_d_meson_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_str, short=False)
    all_d_meson_reco_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.d_str, short=True)
    all_dst_meson_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_st_str, short=False)
    all_dst_meson_reco_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.d_st_str, short=True)
    all_d_wwost_meson_reco_modes_label: str = get_all_reco_modes_str_for(particle=LPNs.d_w_wo_st_str, short=False)
    all_d_wwost_meson_reco_modes_label_short: str = get_all_reco_modes_str_for(particle=LPNs.d_w_wo_st_str, short=True)

    #
    # The following strings are just for test purposes:
    __test_str_1: str = LoNM.get_decay(from_part=LPNs.b_str, to_parts=[LPNs.d_st_str, LPNs.electron_str, LPNs.anti_nu_e_str])
    # fmt: on

    @staticmethod
    def get_unfiltered_attributes_mapping() -> Dict[str, str]:
        _method_names: List[str] = [
            "get_unfiltered_attributes_mapping",
            "get_attributes_mapping",
        ]

        return {
            attr_name: val for attr_name, val in LatexDecayProcesses.__dict__.items() if not (attr_name in _method_names)
        }

    @staticmethod
    def get_attributes_mapping() -> Dict[str, str]:
        return {
            attr_name: val
            for attr_name, val in LatexDecayProcesses.get_unfiltered_attributes_mapping().items()
            if not (attr_name.endswith("__") or attr_name.startswith("_"))
        }


duplicate_ignore_dict: Dict[str, Tuple[str, ...]] = {
    "b_plus_to_d_zero_tau_plus_nu": ("bp_to_dz_tau_nu",),
    "b_zero_to_d_minus_tau_plus_nu": ("bz_to_dm_tau_nu",),
    "b_plus_to_d_st_zero_tau_plus_nu": ("bp_to_dz_st_tau_nu",),
    "b_zero_to_d_st_minus_tau_plus_nu": ("bz_to_dm_st_tau_nu",),
    "d_st_zero_to_d_pi": ("dst0_to_d0_pi0", "neutral_d_st_to_d_pi"),
    "d_st_zero_to_d_photon": ("dst0_to_d0_gamma", "neutral_d_st_to_d_photon"),
    "anti_neutral_d_st_to_d_pi": ("anti_d_st_zero_to_d_pi",),
    "anti_neutral_d_st_to_d_photon": ("anti_d_st_zero_to_d_photon",),
    "_sub_decay_neutral_d_st_to_d_pi": ("subdec_d_st_zero_to_d_pi",),
    "_sub_decay_neutral_d_st_to_d_photon": ("subdec_d_st_zero_to_d_photon",),
    "_sub_decay_anti_neutral_d_st_to_d_pi": ("subdec_anti_d_st_zero_to_d_pi",),
    "_sub_decay_anti_neutral_d_st_to_d_photon": ("subdec_anti_d_st_zero_to_d_photon",),
    "bp_to_dz_st_tau_nu_dst_via_pi": ("b_plus_to_d_st_zero_tau_plus_nu_with_Dst_to_pi",),
    "bp_to_dz_st_tau_nu_dst_via_ph": ("b_plus_to_d_st_zero_tau_plus_nu_with_Dst_to_photon",),
    "reco_b_to_d_wwos_pi0_l": ("reco_dss_val_neutral_pi",),
    "reco_b0_to_d_wwos_pi0_l": ("reco_dss_val_charged_pi",),
    "reco_b_to_d_wwos_pip_l": ("reco_dss_val_neutral_b_neutral_pi",),
    "reco_b0_to_d_wwos_pip_l": ("reco_dss_val_neutral_b_charged_pi",),
}


if __name__ == "__main__":
    from matplotlib import pyplot as plt

    print("Testing some stuff:")

    all_latex_process_keys = [
        process_name
        for process_name in LatexDecayProcesses.get_unfiltered_attributes_mapping().keys()
        if not (process_name.endswith("__"))
    ]
    all_process_labels = [LatexDecayProcesses.__dict__[process_name] for process_name in all_latex_process_keys]
    process_labels_dict = {pn: LatexDecayProcesses.__dict__[pn] for pn in all_latex_process_keys}

    accepted_duplicates: Tuple[str, ...] = tuple(
        list(duplicate_ignore_dict.keys()) + [vv for v in duplicate_ignore_dict.values() for vv in v]
    )
    n_accepted_duplicates: int = sum(len(v) for v in duplicate_ignore_dict.values())

    if not len(set(all_process_labels)) == len(all_process_labels) - n_accepted_duplicates:
        n_ud: int = len(all_process_labels) - len(set(all_process_labels)) - n_accepted_duplicates
        raise RuntimeError(
            f"Duplicates in process labels!\n\t"
            f"len(set(all_process_labels)) = {len(set(all_process_labels))}\n\t"
            f"len(all_process_labels) = {len(all_process_labels)}\n\t"
            f"len(all_process_labels) = {len(all_process_labels)}\n\t"
            f"n_accepted_duplicates = {n_accepted_duplicates}\n\t"
            f"number of unexpected duplicates = {n_ud}\n\t"
            f"Duplicates:\n\t\t - "
            + "\n\t\t - ".join(
                [
                    f"{k} ({v}): {[kk for kk, vv in process_labels_dict.items() if vv == v]}"
                    for k, v in process_labels_dict.items()
                    if all_process_labels.count(v) > 1 and k not in accepted_duplicates
                ]
            )
        )

    test_text = ""
    for process_name in all_latex_process_keys:
        process_label = LatexDecayProcesses.__dict__[process_name]
        print("\t- " + process_name + "  ->  " + process_label)
        assert process_label.count("{") == process_label.count("}")
        assert process_label.count("$") % 2 == 0, process_label.count("$")

        test_text += f"{process_label}\n"

        fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
        try:
            ax.annotate(text=process_label, xy=(0.5, 0.5), xytext=(0.6, 0.6))
            plt.show()
        except:  # noqa: E722  # Ignore this flake8 error for this test.
            plt.close(fig)
            fig.clf()
            raise ValueError("Error in " + process_name + "  ->  " + process_label)

        plt.close(fig)
        fig.clf()

    fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
    ax.annotate(text=test_text, xy=(0.5, 0.5), xytext=(0.6, 0.6))
    plt.show()

    print("Accepted Duplicates:")
    for k, v in duplicate_ignore_dict.items():
        print(f"\t{k}: {v}")

    # latex_process_label_attrs = LatexDecayProcesses.get_attributes_mapping()
    # from rdstar.latex_labels.latex_output_utilities import create_latex_command_file, \
    #     get_latex_cmd_to_attributes_mapping
    # create_latex_command_file(
    #     output_path="/home/metzner/Downloads/rdstar_latex_process_labels.tex",
    #     command_mapping_dict=get_latex_cmd_to_attributes_mapping(attributes_mapping=latex_process_label_attrs),
    #     command_prefix=None,
    #     overwrite=True
    # )
