#!/usr/bin/python

import os
import sys
import socket
import argparse
import subprocess


def check_host():
    """
    Checks if scripts are executed on the KEKCC. If not, the
    execution of the script is terminated.
    """
    if "cc.kek.jp" not in socket.gethostname():
        sys.exit("This script can only be executed on kekcc. Terminating!")
    if "BELLE_TOP_DIR" not in os.environ:
        sys.exit("You must source basf to run this script! Check the README if you don't know what to do...")
    if "BELLE2_LOCAL_DIR" in os.environ:
        sys.exit("If you want to run this script, basf2 must NOT be set up!")


def parse_command_line_args():
    parser = argparse.ArgumentParser(
        description="This script resubmits failed reconstruction jobs using input from makeGsimsummary."
    )
    parser.add_argument("-i", action="store", dest="summary_file", type=str, required=True)
    parsed_args = parser.parse_args()

    return parsed_args


def extract_failed_job_info(summary_file):

    with open(summary_file, "r") as file:
        failed_jobs_lines = list()

        for line in file:
            if line.startswith("WARNING"):
                failed_jobs_lines.append(line.rstrip("!\n").lstrip("'WARNING: NO output events number found in "))

        failed_job_info_dict_list = list()

        for line in failed_jobs_lines:
            source_dir = line.split("/gsim")[0]
            decay = line.split("log/")[1].split("/")[0]
            mdst_file = line.split("log/")[1].split("/")[1].split(".")[0] + ".gen"
            evtgen_mdst_path = os.path.join(source_dir, "evtgen", "gen", decay, mdst_file)
            exp_no = line.split("log/")[1].split("/")[1].split(".")[0].split("_")[2]
            formatted_exp_no = str(exp_no[1:] if exp_no.startswith("0") else exp_no)
            if formatted_exp_no.startswith("0"):
                raise ValueError(formatted_exp_no + " should not start with 0!")

            failed_job_info_dict_list.append(
                {"source_dir": source_dir,
                 "decay": decay,
                 "mdst_file": mdst_file,
                 "evtgen_mdst_path": evtgen_mdst_path,
                 "exp_no": formatted_exp_no}
            )

        return failed_job_info_dict_list


def resubmit_jobs(job_info_dict):
    file_name = job_info_dict["mdst_file"].split(".")[0]
    log_dir = os.path.join(job_info_dict["source_dir"], "gsim", "log", job_info_dict["decay"])
    error_log = os.path.join(log_dir, job_info_dict["mdst_file"]) + "error.log"
    log_log = os.path.join(log_dir, job_info_dict["mdst_file"]) + "log.log"
    log = os.path.join(log_dir, file_name) + ".log"

    mdst = os.path.join(job_info_dict["source_dir"], "gsim", "mdst", job_info_dict["decay"], file_name + ".mdst")
    hbook = os.path.join(job_info_dict["source_dir"], "gsim", "hbook", job_info_dict["decay"], file_name + ".hbook")

    for log_file_path in [error_log, log_log, log, mdst, hbook]:
        if os.path.isfile(log_file_path):

            if log_file_path == mdst:
                os.remove(log_file_path)
                print "Deleting old mdst file\n\t" + str(log_file_path)
                continue

            checking_new_path = True
            old_log_number = 1
            while checking_new_path:
                new_file_name = "old_log_" + str(old_log_number) + "_" + str(os.path.basename(log_file_path))
                new_path = os.path.join(os.path.dirname(log_file_path), new_file_name)

                if not os.path.isfile(new_path):
                    print "Renaming old log file\n\t" + str(log_file_path) + " to " + str(new_file_name)
                    os.rename(log_file_path, new_path)
                    checking_new_path = False

                old_log_number += 1

    command = ["bsub", "-q", "l", "-eo", error_log, "-oo", log_log,
               "./rec.csh", job_info_dict["evtgen_mdst_path"], job_info_dict["exp_no"], job_info_dict["decay"],
               job_info_dict["source_dir"]]

    print "Resubmitting Gsim job with the following commands: \n\t" + " ".join(command)

    subprocess.check_call(command)


if __name__=="__main__":
    check_host()

    args = parse_command_line_args()

    summary_file_path = args.summary_file

    for info in extract_failed_job_info(summary_file_path):
        resubmit_jobs(info)
