"""
B -> D* l nu FF Reweighting (CLN -> BGL) with eFFORT
"""

import numpy as np
import pandas as pd

from uncertainties import ufloat
from multiprocessing.pool import Pool
from typing import Optional, Dict, Tuple, List

from eFFORT.utility import PDG
from eFFORT.SLBToC.BToDstarLNu import BToDstarLNuBGL, BToDstarLNuCLNBelle, BToDstarLNu

from rdstar.utilities.logging import log_to_default

from rdstar.mc_reweighting.utility import FFReweightingBase

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCIDs
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name


__all__ = [
    "BtoDstarFFReweighting",
]


logger = log_to_default(name=__name__)


BMesonVarsTupleType = Tuple[Dict[str, str], Dict[str, str]]


class BtoDstarFFReweighting(FFReweightingBase):
    """
    Applies form factor reweighting for B->Dstlv decays for a given
    DataFrame.
    """

    weight_name: str = "weight_ff_btodst_"
    nom_weight_name: str = f"{weight_name}nom"
    base_var_weight_name: str = f"{weight_name}var_"

    daughter2_pdg_info_col: str = "daughter2_pdg"

    vcb_cln: float = 37.4e-3
    vcb_bgl: float = 41.6558e-3

    # old_params: np.ndarray = np.array(
    #     [
    #         ufloat(3.79139e-04, 2.49014e-04),  # a0
    #         ufloat(2.69537e-02, 9.31972e-03),  # a1
    #         ufloat(5.49846e-04, 2.34728e-05),  # b0
    #         ufloat(-2.04028e-03, 1.06439e-03),  # b1
    #         ufloat(-4.32818e-04, 2.63747e-04),  # c1
    #         ufloat(5.35029e-03, 4.60575e-03),  # c2
    #     ]
    # )
    # old_param_central_values: np.array = np.array([cv.nominal_value for cv in old_params])
    # old_param_uncertainties: np.array = np.array([cv.std_dev for cv in old_params])
    # old_param_correlation_matrix: np.ndarray = np.array(
    #     [
    #         [1.000, -0.952, -0.249, 0.417, 0.137, -0.054],  # a0
    #         [-0.952, 1.000, 0.383, -0.543, -0.268, 0.165],  # a1
    #         [-0.249, 0.383, 1.000, -0.793, -0.648, 0.461],  # b0
    #         [0.417, -0.543, -0.793, 1.000, 0.542, -0.333],  # b1
    #         [0.137, -0.268, -0.648, 0.542, 1.000, -0.953],  # c1
    #         [-0.054, 0.165, 0.461, -0.333, -0.953, 1.000],  # c2
    #     ]
    # )
    #
    # old_covariance_matrix: np.matrix = (
    #     np.outer(old_param_uncertainties, old_param_uncertainties) * old_param_correlation_matrix
    # )

    # The latest BGL fit result D*lnu form factors (https://arxiv.org/abs/2008.09341) untagged v3
    # from table 5, BGL(1,1,2) and the mapping {a_n, b_n, c_n} -> {a^g_n, a^f_n, a^(F1)_n}
    param_central_values: np.array = np.array(
        [
            1.00e-3,  # a0
            -2.35e-3,  # a1
            0.511e-3,  # b0
            0.67e-03,  # b1
            3.0e-04,  # c1
            -3.68e-03,  # c2
        ]
    )
    errors_stat: np.ndarray = np.array([0.02e-3, 0.61e-3, 0.004e-3, 0.17e-3, 0.06e-3, 1.26e-3])
    errors_sys: np.ndarray = np.array([0.02e-3, 0.66e-3, 0.013e-3, 0.30e-3, 0.08e-3, 1.2e-3])

    params: np.ndarray = np.array(
        [
            ufloat(c_val, np.sqrt(u_stat * u_stat + u_sys * u_sys))
            for c_val, u_stat, u_sys in zip(tuple(param_central_values), tuple(errors_stat), tuple(errors_sys))
        ]
    )

    correlation_matrix_stat: np.ndarray = np.array(
        [
            [1.000, -0.937, -0.128, 0.069, -0.081, 0.161],  # a0
            [-0.937, 1.000, 0.127, -0.222, 0.110, -0.192],  # a1
            [-0.218, 0.127, 1.000, -0.800, -0.751, 0.624],  # b0
            [0.069, -0.222, -0.800, 1.000, 0.443, -0.354],  # b1
            [-0.081, 0.110, -0.751, 0.443, 1.000, -0.978],  # c1
            [0.161, -0.192, 0.624, -0.354, -0.978, 1.000],  # c2
        ]
    )
    correlation_matrix_sys: np.ndarray = np.array(
        [
            [1.000, -0.940, -0.132, 0.085, -0.077, 0.158],  # a0
            [-0.940, 1.000, 0.129, -0.228, 0.107, -0.189],  # a1
            [-0.132, 0.129, 1.000, -0.806, -0.755, 0.629],  # b0
            [0.085, -0.228, -0.806, 1.000, 0.452, -0.362],  # b1
            [-0.077, 0.107, -0.755, 0.452, 1.000, -0.977],  # c1
            [0.158, -0.189, 0.629, -0.362, -0.977, 1.000],  # c2
        ]
    )
    covariance_matrix_stat: np.ndarray = np.outer(errors_stat, errors_stat) * correlation_matrix_stat
    covariance_matrix_sys: np.ndarray = np.outer(errors_sys, errors_sys) * correlation_matrix_sys
    covariance_matrix: np.ndarray = covariance_matrix_stat + covariance_matrix_sys

    def __init__(self, num_cpu: int = 1, cached_gammas: Optional[Dict[str, Dict[int, float]]] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        self._weight_name_str = BtoDstarFFReweighting.weight_name

        self.num_cpu: int = num_cpu
        self.cached_gammas: Optional[Dict[str, Dict[int, float]]] = cached_gammas

        self.b1_info_vars.update({BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(0): self.daughter2_pdg_info_col})
        self.b2_info_vars.update({BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(1): self.daughter2_pdg_info_col})

        self.B1_is_charged_btodstl: Optional[pd.Series] = None
        self.B2_is_charged_btodstl: Optional[pd.Series] = None
        self.B1_is_mixed_btodstl: Optional[pd.Series] = None
        self.B2_is_mixed_btodstl: Optional[pd.Series] = None

        self.B1_reweighting_vars: Dict[str, str] = {
            "w": self.b1_w,
            "cos_l": self.b1_costhetal,
            "cos_v": self.b1_costhetav,
            "chi": self.b1_chi,
        }

        self.B2_reweighting_vars: Dict[str, str] = {
            "w": self.b2_w,
            "cos_l": self.b2_costhetal,
            "cos_v": self.b2_costhetav,
            "chi": self.b2_chi,
        }

        self.varied_central_values = self.get_varied_central_values(
            central=self.param_central_values, covariance_matrix=self.covariance_matrix
        )

        self._btodst_charged_cln = BToDstarLNuCLNBelle(
            m_B=PDG.m_Bplus,
            m_Dstar=PDG.m_Dstarzero,
            V_cb=self.vcb_cln,
            cached_gammas=self.cached_gammas["cln_charged"] if self.cached_gammas is not None else None,
        )
        self._btodst_mixed_cln = BToDstarLNuCLNBelle(
            m_B=PDG.m_Bzero,
            m_Dstar=PDG.m_Dstarplus,
            V_cb=self.vcb_cln,
            cached_gammas=self.cached_gammas["cln_mixed"] if self.cached_gammas is not None else None,
        )
        assert self._btodst_charged_cln.R2_1 == 0.71, self._btodst_charged_cln.R2_1
        assert self._btodst_mixed_cln.R2_1 == 0.71, self._btodst_mixed_cln.R2_1
        logger.debug("Initialized CLN.")

        self._btodst_charged_bgl_nom = BToDstarLNuBGL(
            m_B=PDG.m_Bplus,
            m_Dstar=PDG.m_Dstarzero,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas["bgl_charged_nom"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(BtoDstarFFReweighting.param_central_values),
        )
        self._btodst_mixed_bgl_nom = BToDstarLNuBGL(
            m_B=PDG.m_Bzero,
            m_Dstar=PDG.m_Dstarplus,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas["bgl_mixed_nom"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(BtoDstarFFReweighting.param_central_values),
        )
        logger.debug("Initialized nominal BGL.")

        with Pool(processes=min(self.num_cpu, len(self.varied_central_values))) as pool:
            _btodst_bgl_variations = pool.map(self._init_variation_ff_classes, range(len(self.varied_central_values)))

        assert len(_btodst_bgl_variations) == len(self.varied_central_values)
        self._btodst_charged_bgl_vars, self._btodst_mixed_bgl_vars = zip(*_btodst_bgl_variations)

        assert all(isinstance(entry, tuple) for entry in self._btodst_charged_bgl_vars)
        assert all(isinstance(entry, tuple) for entry in self._btodst_mixed_bgl_vars)
        assert all(len(entry) == 2 for entry in self._btodst_charged_bgl_vars)
        assert all(len(entry) == 2 for entry in self._btodst_mixed_bgl_vars)
        assert all(isinstance(entry, BToDstarLNuBGL) for tup in self._btodst_charged_bgl_vars for entry in tup)
        assert all(isinstance(entry, BToDstarLNuBGL) for tup in self._btodst_mixed_bgl_vars for entry in tup)

    @staticmethod
    def get_d_daughter_mc_pdg_var(b_index: int) -> str:
        assert b_index in [0, 1]
        v = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmmcDaughter__bo1__cmPDG__bc__bc__bc__bc"
        return v

    def get_condition_tuples_and_vars(
        self,
    ) -> Tuple[Tuple[Tuple[pd.Series, pd.Series], Tuple[pd.Series, pd.Series]], BMesonVarsTupleType]:
        b_condition_tuples: Tuple[Tuple[pd.Series, pd.Series], Tuple[pd.Series, pd.Series]] = (
            (self.B1_is_charged_btodstl, self.B1_is_mixed_btodstl),
            (self.B2_is_charged_btodstl, self.B2_is_mixed_btodstl),
        )
        b_variables: BMesonVarsTupleType = (self.B1_reweighting_vars, self.B2_reweighting_vars)

        assert len(b_condition_tuples) == 2, len(b_condition_tuples)
        assert all(len(s.index) == len(b_condition_tuples[0][0].index) for t in b_condition_tuples for s in t)
        return b_condition_tuples, b_variables

    def _init_variation_ff_classes(
        self, variation: int
    ) -> Tuple[Tuple[BToDstarLNuBGL, BToDstarLNuBGL], Tuple[BToDstarLNuBGL, BToDstarLNuBGL]]:
        var = variation
        btodst_charged_bgl_var_up = BToDstarLNuBGL(
            m_B=PDG.m_Bplus,
            m_Dstar=PDG.m_Dstarzero,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_charged_var_up{var}"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(self.varied_central_values[variation][0]),
        )
        btodst_charged_bgl_var_down = BToDstarLNuBGL(
            m_B=PDG.m_Bplus,
            m_Dstar=PDG.m_Dstarzero,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_charged_var_down{var}"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(self.varied_central_values[variation][1]),
        )
        btodst_mixed_bgl_var_up = BToDstarLNuBGL(
            m_B=PDG.m_Bzero,
            m_Dstar=PDG.m_Dstarplus,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_mixed_var_up{var}"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(self.varied_central_values[variation][0]),
        )
        btodst_mixed_bgl_var_down = BToDstarLNuBGL(
            m_B=PDG.m_Bzero,
            m_Dstar=PDG.m_Dstarplus,
            V_cb=self.vcb_bgl,
            cached_gammas=self.cached_gammas[f"bgl_mixed_var_down{var}"] if self.cached_gammas is not None else None,
            exp_coeff=tuple(self.varied_central_values[variation][1]),
        )
        logger.debug(f"Initialized BGL variation {variation}.")

        return (btodst_charged_bgl_var_up, btodst_charged_bgl_var_down), (
            btodst_mixed_bgl_var_up,
            btodst_mixed_bgl_var_down,
        )

    def get_gamma_integrals(self) -> Dict[str, Dict[int, float]]:
        gamma_integrals = dict()
        gamma_integrals["cln_charged"] = self._btodst_charged_cln.get_gammas()
        gamma_integrals["cln_mixed"] = self._btodst_mixed_cln.get_gammas()

        gamma_integrals["bgl_charged_nom"] = self._btodst_charged_bgl_nom.get_gammas()
        gamma_integrals["bgl_mixed_nom"] = self._btodst_mixed_bgl_nom.get_gammas()

        for i, (up_charged_var, down_charged_var) in enumerate(self._btodst_charged_bgl_vars):
            gamma_integrals[f"bgl_charged_var_up{i}"] = up_charged_var.get_gammas()
            gamma_integrals[f"bgl_charged_var_down{i}"] = down_charged_var.get_gammas()
        for i, (up_mixed_var, down_mixed_var) in enumerate(self._btodst_mixed_bgl_vars):
            gamma_integrals[f"bgl_mixed_var_up{i}"] = up_mixed_var.get_gammas()
            gamma_integrals[f"bgl_mixed_var_down{i}"] = down_mixed_var.get_gammas()

        for val in gamma_integrals.values():
            BToDstarLNuBGL.check_precomputed_gammas_dict(val)

        return gamma_integrals

    def _add_nominal_weights(self, df: pd.DataFrame) -> None:
        self.nominal_weight_key = f"{self.weight_name_str}nom"
        df.loc[:, self.nominal_weight_key] = np.nan

        b_condition_tuples, variables_for_bs = self.get_condition_tuples_and_vars()

        for b_index, (b_condition_tuple, variables_for_b) in enumerate(zip(b_condition_tuples, variables_for_bs)):
            df.loc[b_condition_tuple[0], self.nominal_weight_key] = self._calculate_weights(
                old=self._btodst_charged_cln,
                new=self._btodst_charged_bgl_nom,
                w=df.loc[b_condition_tuple[0], variables_for_b["w"]].values,
                costhetal=df.loc[b_condition_tuple[0], variables_for_b["cos_l"]].values,
                costhetav=df.loc[b_condition_tuple[0], variables_for_b["cos_v"]].values,
                chiprime=df.loc[b_condition_tuple[0], variables_for_b["chi"]].values,
                pdg=df.loc[b_condition_tuple[0], self.get_d_daughter_mc_pdg_var(b_index=b_index)].values,
            )
            df.loc[b_condition_tuple[1], self.nominal_weight_key] = self._calculate_weights(
                old=self._btodst_mixed_cln,
                new=self._btodst_mixed_bgl_nom,
                w=df.loc[b_condition_tuple[1], variables_for_b["w"]].values,
                costhetal=df.loc[b_condition_tuple[1], variables_for_b["cos_l"]].values,
                costhetav=df.loc[b_condition_tuple[1], variables_for_b["cos_v"]].values,
                chiprime=df.loc[b_condition_tuple[1], variables_for_b["chi"]].values,
                pdg=df.loc[b_condition_tuple[1], self.get_d_daughter_mc_pdg_var(b_index=b_index)].values,
            )

        df.loc[df[self.nominal_weight_key].isnull(), self.nominal_weight_key] = 1.0
        logger.info("Nominal B->Dstlnu FF weights added.")

    def _add_varied_weights(self, df: pd.DataFrame) -> None:
        self.variation_weight_keys = list()

        b_condition_tuples, variables_for_bs = self.get_condition_tuples_and_vars()

        for i, (charged_var, mixed_var) in enumerate(zip(self._btodst_charged_bgl_vars, self._btodst_mixed_bgl_vars)):
            up_charged = charged_var[0]
            down_charged = charged_var[1]
            up_mixed = mixed_var[0]
            down_mixed = mixed_var[1]

            up_key = f"{self.weight_name_str}var_{i}_up"
            down_key = f"{self.weight_name_str}var_{i}_down"

            self.variation_weight_keys.append(up_key)
            self.variation_weight_keys.append(down_key)

            # for var_key, var_info in zip([up_key, down_key], [up, down]):
            for var_key, new_var in zip([up_key, down_key], [(up_charged, up_mixed), (down_charged, down_mixed)]):
                df.loc[:, var_key] = np.nan
                for b_index, (condition_tuple, variables_for_b) in enumerate(zip(b_condition_tuples, variables_for_bs)):
                    df.loc[condition_tuple[0], var_key] = self._calculate_weights(
                        old=self._btodst_charged_cln,
                        new=new_var[0],
                        w=df.loc[condition_tuple[0], variables_for_b["w"]].values,
                        costhetal=df.loc[condition_tuple[0], variables_for_b["cos_l"]].values,
                        costhetav=df.loc[condition_tuple[0], variables_for_b["cos_v"]].values,
                        chiprime=df.loc[condition_tuple[0], variables_for_b["chi"]].values,
                        pdg=df.loc[condition_tuple[0], self.get_d_daughter_mc_pdg_var(b_index=b_index)].values,
                    )
                    df.loc[condition_tuple[1], var_key] = self._calculate_weights(
                        old=self._btodst_mixed_cln,
                        new=new_var[1],
                        w=df.loc[condition_tuple[1], variables_for_b["w"]].values,
                        costhetal=df.loc[condition_tuple[1], variables_for_b["cos_l"]].values,
                        costhetav=df.loc[condition_tuple[1], variables_for_b["cos_v"]].values,
                        chiprime=df.loc[condition_tuple[1], variables_for_b["chi"]].values,
                        pdg=df.loc[condition_tuple[1], self.get_d_daughter_mc_pdg_var(b_index=b_index)].values,
                    )
                df.loc[df[var_key].isnull(), var_key] = 1.0

            logger.info(f"Variation {i} B->Dstlnu FF weights added.")

    @staticmethod
    def _calculate_weights(
        old: BToDstarLNu,
        new: BToDstarLNu,
        w: np.ndarray,
        costhetal: np.ndarray,
        costhetav: np.ndarray,
        chiprime: np.ndarray,
        pdg: np.ndarray,
    ) -> np.ndarray:
        old_d_gamma = old.dGamma_dw_dcosL_dcosV_dChi(w, costhetal, costhetav, chiprime, pdg)
        new_d_gamma = new.dGamma_dw_dcosL_dcosV_dChi(w, costhetal, costhetav, chiprime, pdg)
        return (old.Gamma(pdg) / new.Gamma(pdg)) * (new_d_gamma / old_d_gamma)

    def add_weights(self, df: pd.DataFrame) -> None:
        btodstlnu_gen_id: int = GenMCIDs.Dstl

        self.B1_is_charged_btodstl = (
            (df[gen_b1_id_col_name] == btodstlnu_gen_id)
            & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=0)]) == self.charged_b_meson_pdg_code)
        )
        self.B2_is_charged_btodstl = (
            (df[gen_b2_id_col_name] == btodstlnu_gen_id)
            & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=1)]) == self.charged_b_meson_pdg_code)
        )
        self.B1_is_mixed_btodstl = (
            (df[gen_b1_id_col_name] == btodstlnu_gen_id)
            & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=0)]) == self.mixed_b_meson_pdg_code)
        )
        self.B2_is_mixed_btodstl = (
            (df[gen_b2_id_col_name] == btodstlnu_gen_id)
            & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=1)]) == self.mixed_b_meson_pdg_code)
        )

        assert (self.B1_is_charged_btodstl & self.B1_is_mixed_btodstl).sum() == 0
        assert (self.B2_is_charged_btodstl & self.B2_is_mixed_btodstl).sum() == 0

        logger.info(
            f"Number of candidates to reweight:\n\t#Charged B meson = "
            f"{self.B1_is_charged_btodstl.sum() + self.B2_is_charged_btodstl.sum()}"
            f"\n\t#Mixed B meson = {self.B1_is_mixed_btodstl.sum() + self.B2_is_mixed_btodstl.sum()}"
        )

        self._add_nominal_weights(df)
        self._add_varied_weights(df)

        b1_cond = self.B1_is_charged_btodstl | self.B1_is_mixed_btodstl
        b2_cond = self.B2_is_charged_btodstl | self.B2_is_mixed_btodstl
        self._write_weight_info(df, conditions=(b1_cond, b2_cond))

    @staticmethod
    def finalize_weights(
        df: pd.DataFrame,
        weight_base_name: str = "weight_ff_btodst_",
        threshold: float = 5.0,
        set_to: float = 1.0,
        reset_special_to: float = 0.0,
        combo_name: Optional[str] = None,
    ) -> None:
        combo_info = ""
        if combo_name is not None:
            combo_info = f"events in {combo_name} combo "

        nom_weight_col = f"{weight_base_name}nom"
        assert nom_weight_col in df.columns

        wcs = [c for c in df.columns if c.startswith(weight_base_name)]

        btodstlnu_gen_ids: List[int] = [GenMCIDs.Dstl]

        BtoDstarFFReweighting.reset_problematic_special_mc(
            df=df, weight_base_name=weight_base_name, gen_ids=btodstlnu_gen_ids, reset_special_to=reset_special_to
        )

        update_mask = df[nom_weight_col] > threshold
        if update_mask.any():
            for col in [c for c in wcs if c.endswith("up") or c.endswith("down")]:
                if set_to == 0.0:
                    df.loc[update_mask, col] = set_to
                else:
                    df.loc[update_mask, col] = df.loc[update_mask, col] / df.loc[update_mask, nom_weight_col]
            df.loc[update_mask, nom_weight_col] = set_to

            print(
                f"Resetting the {weight_base_name} FF weights for {len(update_mask.index)} {combo_info}"
                f"to {set_to}, because their weight value was above {threshold}"
            )

        assert not pd.isnull(df[wcs]).any().any()
        assert (df.loc[:, nom_weight_col] >= 0.0).all()
        assert (df.loc[:, nom_weight_col] < threshold).all()


if __name__ == "__main__":
    bToDstar_BGL = BToDstarLNuBGL(m_B=PDG.m_Bplus, m_Dstar=PDG.m_Dstarzero, V_cb=41.6558e-3)
    btodstar_cln = BToDstarLNuCLNBelle(m_B=PDG.m_Bplus, m_Dstar=PDG.m_Dstarzero, V_cb=41.6558e-3)

    w_min: int = 1
    w_max: float = (bToDstar_BGL.m_B ** 2 + bToDstar_BGL.m_Dstar ** 2) / (2 * bToDstar_BGL.m_B * bToDstar_BGL.m_Dstar)
    n: int = 10
    w_range: np.array = np.linspace(w_min + 1e-7, w_max - 1e-7, n, endpoint=True)
    cosl_range: np.array = np.linspace(-1, 1, n, endpoint=True)
    cosnu_range: np.array = np.linspace(-1, 1, n, endpoint=True)
    chi_range: np.array = np.linspace(0, 2 * np.pi, n, endpoint=True)
    pdg_codes: np.array = np.random.choice([22, 111, 211], len(w_range))

    test_df: pd.DataFrame = pd.DataFrame(
        {
            gen_b1_id_col_name: np.full(n, 4),
            gen_b2_id_col_name: np.full(n, 5),
            default_sig_id_col_name: np.full(n, 4),
            "genUpsilon4S__bomcDaughter__bo0__cm__spwMC__bc__bc": w_range,
            "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetalMC__bc__bc": cosl_range,
            "genUpsilon4S__bomcDaughter__bo0__cm__spcosthetavMC__bc__bc": cosnu_range,
            "genUpsilon4S__bomcDaughter__bo0__cm__spchi_primeMC__bc__bc": chi_range,
            BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(0): pdg_codes,
            "genUpsilon4S__bomcDaughter__bo1__cm__spwMC__bc__bc": w_range,
            "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetalMC__bc__bc": cosl_range,
            "genUpsilon4S__bomcDaughter__bo1__cm__spcosthetavMC__bc__bc": cosnu_range,
            "genUpsilon4S__bomcDaughter__bo1__cm__spchi_primeMC__bc__bc": chi_range,
            BtoDstarFFReweighting.get_d_daughter_mc_pdg_var(1): pdg_codes,
        }
    )

    dst_rw = BtoDstarFFReweighting()
    dst_rw.add_weights(test_df)
    print(test_df)
