"""
Definition of MVA Selection Setups

Patrick Ecker und Felix Metzner 2020
"""

import copy
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Tuple, List, Dict, Any, Type

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.mva_methods.mva_methods import MVAVarInfo

from rdstar.continuum_suppression.continuum_suppression import get_cs_variable_infos

from rdstar.offline_analysis.rdstar_samples import mva_samples
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name


__all__ = [
    "MVASelectorSetupBase",
    "MVASelectorSetups",
]


class MVASelectorSetupBase(ABC):
    selection_mva_name: str = ""

    all_mva_variable_infos: Tuple[MVAVarInfo, ...] = ()

    _training_sample_size: int = 0
    _training_data_stream_key_dict: Dict[int, Tuple[str, ...]] = {}

    _required_variables: Tuple[str, ...] = ()
    _modified_variables: Dict[str, str] = {}  # Contains new_col_name: original_col_name

    _validation_variables: Tuple[HistVariable, ...] = (
        RDStarHistVars.missing_mass_q2_SidebandValidation,
        RDStarHistVars.e_miss_minus_p_miss,
        RDStarHistVars.pl_star,
    )

    _validation_columns: List[str] = [var.df_label for var in _validation_variables]

    _additional_validation_columns: List[str] = [
        "daughter__bo1__cm__spuseRestFrame__bodaughter__bo1__cm__spp__bc__bc__bc",
        "unique_event_id",
        "isSignalAcceptMissingNeutrino",
        "qSquaredReco",
        "daughter__bo0__cm__spMbc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        "daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
        "daughter__bo1__cm__spchiProb__bc",
        "ROE_neextra__boCleanROE__bc",
        "daughter__bo0__cm__spdeltaE__bc",
        "daughter__bo1__cm__spROE_neextra__boCleanROE__bc__bc",
        "ROE_deltae__boCleanROE__bc",
        "daughter__bo1__cm__spROE_deltae__boCleanROE__bc__bc",
        "daughter__bo1__cm__spROE_eextra__boCleanROE__bc__bc",
        "daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc",
    ]

    _validation_columns += _additional_validation_columns

    _fast_bdt_settings: Dict[str, Any] = {
        "nTrees": 200,
    }

    @classmethod
    def get_selection_variable_dict(cls) -> Dict[str, HistVariable]:
        assert isinstance(cls.all_mva_variable_infos, tuple), type(cls.all_mva_variable_infos)
        assert len(cls.all_mva_variable_infos) > 0, f"No variable infos defined for MVASelector class {cls.__name__}!"
        return {
            var_info.var: HistVariable(
                df_label=var_info.var,
                n_bins=var_info.bins,
                scope=(var_info.min, var_info.max),
                var_name=var_info.latex,
                unit=var_info.unit,
                use_log_scale=var_info.log,
            )
            for var_info in cls.all_mva_variable_infos
        }

    @classmethod
    def get_validation_variables(cls):
        return cls._validation_variables

    @classmethod
    def get_selection_mva_output_column_name(cls) -> str:
        assert isinstance(cls.selection_mva_name, str), type(cls.selection_mva_name)
        assert cls.selection_mva_name, f"selection_mva_name is not defined for MVASelector class {cls.__name__}!"
        return f"{cls.selection_mva_name}_output"

    @classmethod
    def get_target_column_name(cls) -> str:
        assert cls.selection_mva_name, f"selection_mva_name is not defined for MVASelector class {cls.__name__}!"
        return f"{cls.selection_mva_name}_truth_target"

    @classmethod
    def get_training_sample_size(cls) -> int:
        assert isinstance(cls._training_sample_size, int), type(cls._training_sample_size)
        assert cls._training_sample_size > 0, cls._training_sample_size
        return cls._training_sample_size

    @classmethod
    def get_training_data_info_tuples(cls) -> List[Tuple[str, int]]:
        assert isinstance(cls._training_data_stream_key_dict, dict), type(cls._training_data_stream_key_dict)
        assert (
            len(cls._training_data_stream_key_dict) > 0
        ), f"No Training data defined for MVASelector class {cls.__name__}!"
        return [
            (data_key, stream)
            for stream, data_keys in cls._training_data_stream_key_dict.items()
            for data_key in data_keys
        ]

    @classmethod
    def get_required_columns(cls) -> Tuple[str, ...]:
        train_vars = [
            mva_v.var if mva_v.var not in cls._modified_variables else cls._modified_variables[mva_v.var]
            for mva_v in cls.all_mva_variable_infos
        ]
        validation_vars = list(set(MVASelectorSetupBase._validation_columns + cls._validation_columns))
        return tuple(set(train_vars + list(cls._required_variables) + validation_vars))

    @classmethod
    def get_fast_bdt_settings(cls) -> Dict[str, Any]:
        return cls._fast_bdt_settings

    @classmethod
    @abstractmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _data_pre_processing(cls, df: pd.DataFrame) -> None:
        raise NotImplementedError()

    @classmethod
    def data_preparation(cls, df: pd.DataFrame, for_application_only: bool) -> None:
        if not for_application_only:
            cls._add_target_variable(df=df)
        cls._data_pre_processing(df=df)

        if for_application_only:
            assert all(c.var in df.columns for c in cls.all_mva_variable_infos), [
                c.var for c in cls.all_mva_variable_infos if c.var not in df.columns
            ]
        else:
            assert all(c in df.columns for c in cls.get_required_columns()), [
                c for c in cls.get_required_columns() if c not in df.columns
            ]

    @classmethod
    def data_preparation_for_validation(cls, df: pd.DataFrame) -> None:
        cls._add_target_variable(df=df)


class FirstSelectionMVASetup(MVASelectorSetupBase):
    selection_mva_name: str = "first_sel_mva"

    e_extra = MVAVarInfo(
        var=f"ROE_eextra__boCleanROE__bc_for_{selection_mva_name}",
        latex=LatexObsNames.roe_e_ecl_extra,
        bins=30,
        min=0.0,
        max=5.0,
        unit=LatexUnits.gev,
    )
    d_mass = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo0__cm__spM__bc__bc",
        latex=LatexObsNames.d_w_wo_star_meson_mass,
        bins=25,
        min=1.8,
        max=2.1,
        unit=LatexUnits.gev,
    )
    d_star_daughter_mass = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spM__bc__bc__bc",
        latex=LatexObsNames.d_w_wo_star_daughter_mass,
        bins=25,
        min=1.8,
        max=01.95,
        unit=LatexUnits.gev,
    )
    cs_mva = MVAVarInfo(
        var="cs_mva_output",
        latex=LatexObsNames.mva_cs_output,
        bins=25,
        min=0.0,
        max=1.0,
    )
    sig_b_decay_id = MVAVarInfo(
        var=Cols.decay_mode_column_name,
        latex=LatexObsNames.reco_bsig_reco_mode_id,
        bins=70,
        min=10.0,
        max=80.0,
    )
    upsilon_decay_id = MVAVarInfo(
        var="extraInfo__bodecayModeID__bc",
        latex=LatexObsNames.reco_y4s_reco_mode_id,
        bins=25,
        min=1000.0,
        max=1024.0,
    )

    all_mva_variable_infos: Tuple[MVAVarInfo, ...] = (
        e_extra,
        d_mass,
        d_star_daughter_mass,
        cs_mva,
        sig_b_decay_id,
        upsilon_decay_id,
    )

    _required_variables: Tuple[str, ...] = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
        gen_b1_id_col_name,
        gen_b2_id_col_name,
        default_sig_id_col_name,
    )

    # Contains information about modified variables in the form: {new_col_name: original_col_name}
    _modified_variables: Dict[str, str] = {e_extra.var: "ROE_eextra__boCleanROE__bc"}

    _training_sample_size: int = 50000
    _training_data_stream_key_dict: Dict[int, Tuple[str, ...]] = {
        -999: tuple(copy.copy(mva_samples)),
        0: ("Mixed", "Charged", "Charm", "UDS"),
    }

    sig_id_to_gen_id_mapping: Dict[int, Tuple[int, ...]] = {
        1: (1,),
        2: (2,),
        3: (3,),
        4: (4,),
        5: (51, 52, 53, 54, 55, 56, 57, 58),
        6: (51, 52, 53, 54, 55, 56, 57, 58),
    }

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        b_mdst_tau_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        b1_gen_id_col = gen_b1_id_col_name
        b2_gen_id_col = gen_b2_id_col_name
        sig_id_col = default_sig_id_col_name
        required_cols = [b_mdst_col, b_mdst_tau_col, b1_gen_id_col, b2_gen_id_col, sig_id_col]
        assert all(c in cls._required_variables for c in required_cols), [
            c for c in required_cols if c not in cls._required_variables
        ]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), [
            c for c in required_cols if c not in cls.get_required_columns()
        ]
        original_n_cols = len(df.columns)

        temp_sig_gen_id_col = "temp_mva_target_sig_gen_id"
        temp_sig_b_mdst_index_col = "temp_mva_target_sig_b_mdst_index"
        temp_columns = [temp_sig_gen_id_col, temp_sig_b_mdst_index_col]

        target_column = cls.get_target_column_name()
        new_columns = [target_column] + temp_columns
        assert all([c not in df.columns for c in new_columns]), [c for c in new_columns if c in df.columns]

        cond = df[b_mdst_col].isin([1, 2])
        cond_tau = df[b_mdst_tau_col].isin([1, 2])

        df.loc[:, temp_sig_b_mdst_index_col] = np.nan
        df.loc[cond, temp_sig_b_mdst_index_col] = df[b_mdst_col]
        df.loc[cond_tau, temp_sig_b_mdst_index_col] = df[b_mdst_tau_col]
        df.loc[(~cond) & (~cond_tau), temp_sig_b_mdst_index_col] = 0
        assert not pd.isnull(df[temp_sig_b_mdst_index_col]).any(), temp_sig_b_mdst_index_col

        cond_b1: pd.Series = df[temp_sig_b_mdst_index_col] == 1
        cond_b2: pd.Series = df[temp_sig_b_mdst_index_col] == 2

        df.loc[:, temp_sig_gen_id_col] = np.nan
        df.loc[cond_b1, temp_sig_gen_id_col] = df[b1_gen_id_col]
        df.loc[cond_b2, temp_sig_gen_id_col] = df[b2_gen_id_col]
        df.loc[(~cond_b1) & (~cond_b2), temp_sig_gen_id_col] = 0
        assert not pd.isnull(df[temp_sig_gen_id_col]).any(), temp_sig_gen_id_col

        target_condition = pd.Series(np.full(shape=len(df.index), fill_value=False, dtype=bool))
        for sig_id, gen_ids in cls.sig_id_to_gen_id_mapping.items():
            target_condition |= (df[sig_id_col] == sig_id) & (df[temp_sig_gen_id_col].isin(gen_ids))

        assert len(target_condition.index) == len(df.index), (len(target_condition.index), len(df.index))
        df.loc[:, target_column] = np.nan
        df.loc[target_condition.values, target_column] = 1
        df.loc[~target_condition.values, target_column] = 0

        for col in temp_columns:
            del df[col]

        assert not pd.isnull(df[target_column]).any(), target_column
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))

    @classmethod
    def _data_pre_processing(cls, df: pd.DataFrame) -> None:
        assert cls.e_extra.var not in df.columns, cls.e_extra.var
        assert cls.e_extra.var in cls._modified_variables.keys(), cls.e_extra.var

        original_e_extra_col = cls._modified_variables[cls.e_extra.var]
        assert original_e_extra_col in df.columns, original_e_extra_col

        df.loc[:, cls.e_extra.var] = df[original_e_extra_col]
        df.loc[df[original_e_extra_col] < 0.6, cls.e_extra.var] = 0.6


class DSSMVASetup(FirstSelectionMVASetup):
    selection_mva_name: str = "dss_first_sel_mva"

    _fast_bdt_settings: Dict[str, Any] = {
        "nTrees": 200,
        # "flatnessLoss": 2.0,
        # "numberOfFlatnessFeatures": 2,
    }

    _all_mva_variable_infos: List[MVAVarInfo] = list(FirstSelectionMVASetup.all_mva_variable_infos)

    # add the tag side cs vars
    _all_mva_variable_infos.extend(get_cs_variable_infos(signal_side=False))

    p_cms = MVAVarInfo(
        var="useCMSFrame__bop__bc", latex=LatexObsNames.y4s_cms_momentum, bins=30, min=0.0, max=3.0, unit=LatexUnits.gev
    )

    p_lep_cms = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo1__cm__spuseCMSFrame__bop__bc__bc__bc",
        latex=LatexObsNames.lepton_cms_p,
        bins=30,
        min=0.0,
        max=3.0,
        unit=LatexUnits.gev,
    )

    pt_lep_cms = MVAVarInfo(
        var="daughter__bo1__cm__spdaughter__bo1__cm__spuseCMSFrame__bopt__bc__bc__bc",
        latex=LatexObsNames.lepton_cms_p_t,
        bins=30,
        min=0.0,
        max=3.0,
        unit=LatexUnits.gev,
    )

    belle_ecl_energy = MVAVarInfo(
        var="belleECLEnergy",
        latex=LatexObsNames.ecl_belle_energy,
        bins=30,
        min=2.0,
        max=10.0,
        unit=LatexUnits.gev,
    )

    m_upsilon = MVAVarInfo(
        var="M",
        latex=LatexObsNames.y4s_mass,
        bins=30,
        min=7.0,
        max=11.0,
        unit=LatexUnits.gev,
    )

    q = MVAVarInfo(
        var="Q",
        latex=LatexObsNames.energy_q,
        bins=30,
        min=0.0,
        max=1.0,
        unit=LatexUnits.gev,
    )

    q_b_sig = MVAVarInfo(
        var="daughter__bo1__cm__spQ__bc",
        latex=LatexObsNames.b_sig_q,
        bins=30,
        min=0.0,
        max=4.0,
        unit=LatexUnits.gev,
    )

    cos_theta_l = MVAVarInfo(
        var="costhetalReco",
        latex=LatexObsNames.ff_cos_theta_l_reco,
        bins=30,
        min=-1.0,
        max=1.0,
        unit=LatexUnits.gev,
    )

    m2RecoilSignalSide_sig = MVAVarInfo(
        var="m2RecoilSignalSide_sig",
        latex=LatexObsNames.mva_mm2_sig,
        bins=30,
        min=-4.0,
        max=10.0,
        unit=LatexUnits.gev_sq,
    )

    m2RecoilSignalSide_bkg = MVAVarInfo(
        var="m2RecoilSignalSide_bkg",
        latex=LatexObsNames.mva_mm2_bkg,
        bins=30,
        min=-4.0,
        max=10.0,
        unit=LatexUnits.gev_sq,
    )

    additional_mva_infos: List[MVAVarInfo] = [
        p_cms,
        p_lep_cms,
        pt_lep_cms,
        belle_ecl_energy,
        m_upsilon,
        q,
        q_b_sig,
        cos_theta_l,
        # m2RecoilSignalSide_sig,
        # m2RecoilSignalSide_bkg,
    ]
    _all_mva_variable_infos.extend(additional_mva_infos)
    all_mva_variable_infos = tuple(_all_mva_variable_infos)

    # assert all_mva_variable_infos[-2] == m2RecoilSignalSide_sig, all_mva_variable_infos[-1]
    # assert all_mva_variable_infos[-1] == m2RecoilSignalSide_bkg, all_mva_variable_infos[-1]

    _required_variables: Tuple[str, ...] = (
        "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
        "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
    )

    _modified_variables: Dict[str, str] = {
        FirstSelectionMVASetup.e_extra.var: "ROE_eextra__boCleanROE__bc",
        # m2RecoilSignalSide_sig.var: "m2RecoilSignalSide",
        # m2RecoilSignalSide_bkg.var: "m2RecoilSignalSide"
    }

    dss_pdgs: List[int] = [10411, 10421, 10413, 10423, 20413, 20423, 415, 425]

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        b1_pdg_col = "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"
        b2_pdg_col = "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"
        b_mdst_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
        b_mdst_tau_col = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
        required_cols = [b_mdst_col, b_mdst_tau_col, b1_pdg_col, b2_pdg_col]
        assert all(c in cls._required_variables for c in required_cols), [
            c for c in required_cols if c not in cls._required_variables
        ]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), [
            c for c in required_cols if c not in cls.get_required_columns()
        ]
        original_n_cols = len(df.columns)

        temp_sig_b_mdst_index_col = "temp_mva_target_sig_b_mdst_index"
        temp_pdg_col = "temp_mva_target_sig_gen_id"
        temp_columns = [temp_sig_b_mdst_index_col, temp_pdg_col]

        target_column = cls.get_target_column_name()
        new_columns = [target_column] + temp_columns
        assert all([c not in df.columns for c in new_columns]), [c for c in new_columns if c in df.columns]

        cond = df[b_mdst_col].isin([1, 2])
        cond_tau = df[b_mdst_tau_col].isin([1, 2])

        df.loc[:, temp_sig_b_mdst_index_col] = np.nan
        df.loc[cond, temp_sig_b_mdst_index_col] = df[b_mdst_col]
        df.loc[cond_tau, temp_sig_b_mdst_index_col] = df[b_mdst_tau_col]
        df.loc[(~cond) & (~cond_tau), temp_sig_b_mdst_index_col] = 0
        assert not pd.isnull(df[temp_sig_b_mdst_index_col]).any(), temp_sig_b_mdst_index_col

        cond_b1 = df[temp_sig_b_mdst_index_col] == 1
        cond_b2 = df[temp_sig_b_mdst_index_col] == 2

        df.loc[:, temp_pdg_col] = np.nan
        df.loc[cond_b1, temp_pdg_col] = df[b1_pdg_col]
        df.loc[cond_b2, temp_pdg_col] = df[b2_pdg_col]
        df.loc[(~cond_b1) & (~cond_b2), temp_pdg_col] = 0
        assert not pd.isnull(df[temp_pdg_col]).any(), temp_pdg_col

        target_cond = abs(df[temp_pdg_col]).isin(cls.dss_pdgs)

        target_col = cls.get_target_column_name()
        df.loc[target_cond, target_col] = 1
        df.loc[(~target_cond), target_col] = 0

        for col in temp_columns:
            del df[col]

        assert not pd.isnull(df[target_column]).any(), target_column
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))

    # @classmethod
    # def _data_pre_processing(cls, df: pd.DataFrame) -> None:
    #     assert cls.e_extra.var not in df.columns, cls.e_extra.var
    #     assert cls.e_extra.var in cls._modified_variables.keys(), cls.e_extra.var
    #
    #     original_e_extra_col = cls._modified_variables[cls.e_extra.var]
    #     assert original_e_extra_col in df.columns, original_e_extra_col
    #
    #     df.loc[:, cls.e_extra.var] = df[original_e_extra_col]
    #     df.loc[df[original_e_extra_col] < 0.6, cls.e_extra.var] = 0.6
    #
    #     assert cls.m2RecoilSignalSide_sig.var not in df.columns, cls.m2RecoilSignalSide_sig.var
    #     assert cls.m2RecoilSignalSide_sig.var in cls._modified_variables.keys(), cls.m2RecoilSignalSide_sig.var
    #     assert cls.m2RecoilSignalSide_bkg.var not in df.columns, cls.m2RecoilSignalSide_bkg.var
    #     assert cls.m2RecoilSignalSide_bkg.var in cls._modified_variables.keys(), cls.m2RecoilSignalSide_bkg.var
    #
    #     m2RecoilColumn = cls._modified_variables[cls.m2RecoilSignalSide_sig.var]
    #     assert m2RecoilColumn in df.columns, m2RecoilColumn
    #     df.loc[:, cls.m2RecoilSignalSide_sig.var] = df[m2RecoilColumn]
    #     df.loc[:, cls.m2RecoilSignalSide_bkg.var] = df[m2RecoilColumn]
    #
    #     df.loc[df[cls.get_target_column_name()] == 0, cls.m2RecoilSignalSide_sig.var] = np.nan
    #     df.loc[df[cls.get_target_column_name()] == 1, cls.m2RecoilSignalSide_bkg.var] = np.nan


class SecondSelectionMVASetup(FirstSelectionMVASetup):
    selection_mva_name: str = "sec_sel_mva"
    sig_id_col_name: str = default_sig_id_col_name

    _required_variables: Tuple[str, ...] = (sig_id_col_name,)

    _target_sig_ids: List[int] = [1, 2, 3, 4]

    @classmethod
    def _add_target_variable(cls, df: pd.DataFrame) -> None:
        required_cols = [cls.sig_id_col_name]
        assert all(c in cls._required_variables for c in required_cols), [
            c for c in required_cols if c not in cls._required_variables
        ]
        assert all([c in df.columns for c in required_cols]), [c for c in required_cols if c not in df.columns]
        assert all([c in cls.get_required_columns() for c in required_cols]), [
            c for c in required_cols if c not in cls.get_required_columns()
        ]
        original_n_cols = len(df.columns)

        target_cond = df[cls.sig_id_col_name].isin(cls._target_sig_ids)

        target_col = cls.get_target_column_name()
        assert target_col not in df.columns, target_col
        df.loc[target_cond, target_col] = 1
        df.loc[(~target_cond), target_col] = 0

        assert not pd.isnull(df[target_col]).any(), target_col
        assert len(df.columns) == (1 + original_n_cols), (original_n_cols, len(df.columns))


class SecondSelectionDSSMVASetup(SecondSelectionMVASetup):
    selection_mva_name = "dss_sec_sel_mva"

    _target_sig_ids: List[int] = [5, 6]


class MVASelectorSetups:
    available_mva_setups: Dict[str, Type[MVASelectorSetupBase]] = {
        FirstSelectionMVASetup.selection_mva_name: FirstSelectionMVASetup,
        DSSMVASetup.selection_mva_name: DSSMVASetup,
        SecondSelectionMVASetup.selection_mva_name: SecondSelectionMVASetup,
        SecondSelectionDSSMVASetup.selection_mva_name: SecondSelectionDSSMVASetup,
    }

    @staticmethod
    def get_available_selection_mva_setup_names() -> Tuple[str, ...]:
        return tuple([sel_mva_name for sel_mva_name in MVASelectorSetups.available_mva_setups.keys()])

    @staticmethod
    def get_selection_mva_setup(setup_name: str):
        if setup_name not in MVASelectorSetups.available_mva_setups:
            raise KeyError(
                f"A Selection MVA with the name {setup_name} is not available."
                f"The options are:\n\t" + "\n\t-".join(MVASelectorSetups.get_available_selection_mva_setup_names())
            )
        return MVASelectorSetups.available_mva_setups[setup_name]


# Other possible mva variables (Check if binning is correct for all these variables):
# MVAVarInfo(var="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc", bins=25, min=0., max=1.,
#           latex=r"$\mathcal{P}_\mathrm{FEI}$")
# MVAVarInfo(var="daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc", bins=25, min=-0.2, max=0.8,
#           latex=r"$\Delta \mathrm{M}_{D^{(*)}}$")
# MVAVarInfo(var="daughter__bo1__cm__spchiProb__bc", bins=25, min=-1., max=1.,
#           latex=r"$\mathcal{P}_{\mathrm{B},\mathrm{sig},\mathrm{Vertex}}$")
# MVAVarInfo(var="ROE_neextra__boCleanROE__bc", bins=25, min=0., max=8.,
#           latex=r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra}$")
# MVAVarInfo(var="daughter__bo0__cm__spdeltaE__bc", bins=25, min=-0.15, max=0.1,
#           latex=r"$\Delta\mathrm{E}_\mathrm{tag}$")
# MVAVarInfo(var="daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spdM__bc__bc__bc", bins=25, min=-0.2,
#           max=0.2, latex=r"$\Delta \mathrm{M}_\mathrm{D}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_neextra__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\mathrm{nE}^\mathrm{ECL}_\mathrm{extra,sig}$")
# MVAVarInfo(var="ROE_deltae__boCleanROE__bc", bins=25, min=0., max=10.0,
#           latex=r"$\Delta \mathrm{E}_\mathrm{\Upsilon(4S),ROE}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_deltae__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\Delta \mathrm{E}_\mathrm{sig,ROE}$")
# MVAVarInfo(var="daughter__bo1__cm__spROE_eextra__boCleanROE__bc__bc", bins=25, min=0., max=10.0,
#           latex=r"$\mathrm{E}^\mathrm{ECL}_\mathrm{extra,sig}$")
