import os
import shutil
import numpy as np
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Union, Optional, Tuple, List, Dict, Generator

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.rdstar_samples import main_sample_keys
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name

from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils


__all__ = [
    "FittingPreparationTask",
    "CopyToSSDTask",
]


logger = log_to_default(name=__name__)


available_sidebands: Tuple[Union[str, None], ...] = (
    None,
    "mbc_sideband",
)


class FittingPreparationTask(utilities.BaseOfflineTask):
    selection: str = luigi.Parameter(positional=False)
    production_key: str = luigi.Parameter(positional=False)
    working_dir: str = luigi.Parameter(positional=False)

    sig_b_mdst_col: str = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
    sig_b_tau_mdst_col: str = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
    )
    advanced_sig_id_col: str = "advanced_sig_id"

    decay_mode_col: str = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"

    dss_pdgs: List[int] = [10411, 10421, 10413, 10423, 415, 425, 20413, 10423]
    dss_pdgs.extend(list(-np.array(dss_pdgs)))

    dss_sig_id_col: str = "dss_sig_id"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.data_key: str = "DataY4S"

        self.base_file_dict_path: str = f"/ceph/fmetzner/rdstar/{self.production_key}_stream0"
        self.base_input_path: str = (
            f"/ceph/fmetzner/rdstar/{self.production_key}_stream0/CombinedSamples/{self.selection}"
        )

        self.base_dir: PathType = os.path.join(self.working_dir, "samples")

    def get_data_path(
        self,
        data_type: str,
        sideband: Optional[str] = None,
    ) -> PathType:
        if not sideband:
            file_key = f"{data_type}_withFinalWeights_withMMSQResCorr_{self.selection}_with_CS_CS1"
        else:
            file_key = f"{data_type}_withFinalWeights_withMMSQResCorr_{self.selection}_with_CS_CS1_{sideband}"
        file_name = f"{file_key}.h5"
        return os.path.join(self.base_input_path, file_name)

    def get_eff_path(self, data_type: str) -> PathType:
        base_eff_path = f"/ceph/fmetzner/rdstar/{self.production_key}_stream0/{data_type}"
        file_name = f"{data_type}_Y4S_ntuple_{self.selection}_with_CS_CS1.h5"
        return os.path.join(base_eff_path, file_name)

    @staticmethod
    def apply_asimov_weights(df: pd.DataFrame) -> None:
        logger.debug("Applying Asimov Weights.")
        assert fitting_utils.final_weight_column_for_asimov not in df.columns
        base_weight_col: str = WeightManager.final_weight_with_bf_corrections_col
        assert base_weight_col in df.columns, base_weight_col
        df.loc[:, fitting_utils.final_weight_column_for_asimov] = df[base_weight_col].values

    @staticmethod
    def apply_qsquare_infos(df: pd.DataFrame) -> None:
        logger.debug("Applying q^2 Truth Information.")
        QSquaredResidualVariable.add_q_squared_residual_definitions(df=df)

    @staticmethod
    def get_cond_dict(df: pd.DataFrame):
        is_D_cond = df.sig_gen_D_pdg.isin([-411, -421, 411, 421])
        is_Ds_cond = df.sig_gen_D_pdg.isin([-413, -423, 413, 423])

        cond_dict = {
            101: (df.SIG_ID == 5) & is_D_cond,
            102: (df.SIG_ID == 5) & is_Ds_cond,
            103: (df.SIG_ID == 6) & is_D_cond,
            104: (df.SIG_ID == 6) & is_Ds_cond,
            105: (df.SIG_ID == 3),
            106: (df.SIG_ID == 4),
            107: (df.SIG_ID.isin([1, 2, 7, 8, 9])),
        }
        return cond_dict

    def create_advanced_sig_id(self, df: pd.DataFrame) -> None:
        logger.debug("Create advanced Sig ID.")
        test_cond = df[self.sig_b_mdst_col].isin([1, 2])

        df.loc[:, "b_mdst"] = np.nan
        df.loc[test_cond, "b_mdst"] = df[self.sig_b_mdst_col]

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        df.loc[:, "sig_gen_D_pdg"] = np.nan
        df.loc[b1_cond, "sig_gen_D_pdg"] = df.loc[
            b1_cond, "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"
        ]
        df.loc[b2_cond, "sig_gen_D_pdg"] = df.loc[
            b2_cond, "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"
        ]

        df.loc[:, self.advanced_sig_id_col] = np.nan
        for advanced_id, cond in self.get_cond_dict(df=df).items():
            df.loc[cond, self.advanced_sig_id_col] = advanced_id

    @staticmethod
    def get_dss_cond_dict(df: pd.DataFrame):
        pi_b1_col = "genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmmcDaughter__bo2__cmPDG__bc__bc__bc__bc"
        has_two_pi_b1 = df[pi_b1_col].isin([111, -111, 211, -211])
        pi_b2_col = "genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmmcDaughter__bo2__cmPDG__bc__bc__bc__bc"
        has_two_pi_b2 = df[pi_b2_col].isin([111, -111, 211, -211])

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        is_dss_b1 = df.genUpsilon4S__bomcDaughter__bo0__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc.isin(
            FittingPreparationTask.dss_pdgs
        )
        is_dss_b2 = df.genUpsilon4S__bomcDaughter__bo1__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc.isin(
            FittingPreparationTask.dss_pdgs
        )

        is_charged_B_cond = df.genUpsilon4S__bomcDaughter__bo0__cm__spPDG__bc__bc.isin([521, -521])

        cond_dict = {
            # B+
            201: (df.SIG_ID == 5) & df.sig_gen_id.isin([54]) & is_charged_B_cond,  # Dzerostar -> D(*) pi0
            202: (df.SIG_ID == 5) & df.sig_gen_id.isin([53]) & is_charged_B_cond,  # Doneprime -> D(*) pi0
            203: (df.SIG_ID == 5) & df.sig_gen_id.isin([51]) & is_charged_B_cond,  # Done -> D(*) pi0
            204: (df.SIG_ID == 5) & df.sig_gen_id.isin([52]) & is_charged_B_cond,  # Dtwostar -> D(*) pi0
            205: (df.SIG_ID == 6) & df.sig_gen_id.isin([54]) & is_charged_B_cond,  # Dzerostar -> D(*) pi+
            206: (df.SIG_ID == 6) & df.sig_gen_id.isin([53]) & is_charged_B_cond,  # Doneprime -> D(*) pi+
            207: (df.SIG_ID == 6) & df.sig_gen_id.isin([51]) & is_charged_B_cond,  # Done -> D(*) pi+
            208: (df.SIG_ID == 6) & df.sig_gen_id.isin([52]) & is_charged_B_cond,  # Dtwostar -> D(*) pi+
            209: (df.SIG_ID == 8) & b1_cond & has_two_pi_b1 & is_dss_b1 & is_charged_B_cond,  # Done -> D pi pi
            210: (df.SIG_ID == 8) & b2_cond & has_two_pi_b2 & is_dss_b2 & is_charged_B_cond,  # Done -> D pi pi
            211: (df.SIG_ID == 3) & is_charged_B_cond,  # D l nu
            212: (df.SIG_ID == 4) & is_charged_B_cond,  # D* l nu
            213: (df.SIG_ID == 8)
            & ~(b1_cond & has_two_pi_b1 & is_dss_b1)
            & ~(b2_cond & has_two_pi_b2 & is_dss_b2)
            & is_charged_B_cond,  # BBar BKG
            214: (df.SIG_ID == 1) & is_charged_B_cond,  # D tau nu
            215: (df.SIG_ID == 2) & is_charged_B_cond,  # D* tau nu,
            216: (df.SIG_ID == 7) & is_charged_B_cond,  # Hadronic BKG,
            # B0
            301: (df.SIG_ID == 5) & df.sig_gen_id.isin([54]) & ~is_charged_B_cond,  # Dzerostar -> D(*) pi0
            302: (df.SIG_ID == 5) & df.sig_gen_id.isin([53]) & ~is_charged_B_cond,  # Doneprime -> D(*) pi0
            303: (df.SIG_ID == 5) & df.sig_gen_id.isin([51]) & ~is_charged_B_cond,  # Done -> D(*) pi0
            304: (df.SIG_ID == 5) & df.sig_gen_id.isin([52]) & ~is_charged_B_cond,  # Dtwostar -> D(*) pi0
            305: (df.SIG_ID == 6) & df.sig_gen_id.isin([54]) & ~is_charged_B_cond,  # Dzerostar -> D(*) pi+
            306: (df.SIG_ID == 6) & df.sig_gen_id.isin([53]) & ~is_charged_B_cond,  # Doneprime -> D(*) pi+
            307: (df.SIG_ID == 6) & df.sig_gen_id.isin([51]) & ~is_charged_B_cond,  # Done -> D(*) pi+
            308: (df.SIG_ID == 6) & df.sig_gen_id.isin([52]) & ~is_charged_B_cond,  # Dtwostar -> D(*) pi+
            309: (df.SIG_ID == 8) & b1_cond & has_two_pi_b1 & is_dss_b1 & ~is_charged_B_cond,  # Done -> D pi pi
            310: (df.SIG_ID == 8) & b2_cond & has_two_pi_b2 & is_dss_b2 & ~is_charged_B_cond,  # Done -> D pi pi
            311: (df.SIG_ID == 3) & ~is_charged_B_cond,  # D l nu
            312: (df.SIG_ID == 4) & ~is_charged_B_cond,  # D* l nu
            313: (df.SIG_ID == 8)
            & ~(b1_cond & has_two_pi_b1 & is_dss_b1)
            & ~(b2_cond & has_two_pi_b2 & is_dss_b2)
            & ~is_charged_B_cond,  # BBar BKG
            314: (df.SIG_ID == 1) & ~is_charged_B_cond,  # D tau nu
            315: (df.SIG_ID == 2) & ~is_charged_B_cond,  # D* tau nu,
            316: (df.SIG_ID == 7) & ~is_charged_B_cond,  # Hadronic BKG,
            # Continuum
            401: (df.SIG_ID == 9),  # Continuum,
        }
        return cond_dict

    def create_dss_sig_id(self, df: pd.DataFrame) -> None:
        logger.debug("Create D** Sig ID.")
        test_cond = df[self.sig_b_mdst_col].isin([1, 2])
        test_cond_tau = df[self.sig_b_tau_mdst_col].isin([1, 2])

        df.loc[:, "b_mdst"] = np.nan
        df.loc[test_cond, "b_mdst"] = df[self.sig_b_mdst_col]
        df.loc[test_cond_tau, "b_mdst"] = df[self.sig_b_tau_mdst_col]

        b1_cond = df.b_mdst == 1
        b2_cond = df.b_mdst == 2

        df.loc[:, "sig_gen_id"] = np.nan
        df.loc[b1_cond, "sig_gen_id"] = df.loc[b1_cond, gen_b1_id_col_name]
        df.loc[b2_cond, "sig_gen_id"] = df.loc[b2_cond, gen_b2_id_col_name]

        df.loc[:, self.dss_sig_id_col] = np.nan
        for dss_id, sig_cond in self.get_dss_cond_dict(df=df).items():
            df.loc[sig_cond, self.dss_sig_id_col] = dss_id

    def run_fitting_preparation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()

        for sideband in available_sidebands:
            if not sideband:
                mc_output_key = f"combined_mc_prepared_{self.selection}"
                data_output_key = f"data_prepared_{self.selection}"
            else:
                mc_output_key = f"combined_mc_prepared_{self.selection}_{sideband}"
                data_output_key = f"data_prepared_{self.selection}_{sideband}"

            mc_output_path = os.path.join(self.base_dir, f"{mc_output_key}.h5")
            output_dict.update({mc_output_key: mc_output_path})

            data_output_path = os.path.join(self.base_dir, f"{data_output_key}.h5")
            output_dict.update({data_output_key: data_output_path})

            if not dry_run:
                mc_data_types = main_sample_keys
                data_types = mc_data_types + [self.data_key]
                data_paths = {
                    data_type: self.get_data_path(data_type=data_type, sideband=sideband) for data_type in data_types
                }

                # MC
                logger.debug("Load Dataframes.")
                mc_dfs = [pd.read_hdf(data_paths[data_type]) for data_type in mc_data_types]
                assert all(isinstance(df, pd.DataFrame) for df in mc_dfs)
                logger.debug("Merge Dataframes.")
                df_mc = pd.concat(mc_dfs, ignore_index=True, sort=False)

                # apply weights
                self.apply_asimov_weights(df=df_mc)
                # apply q^2 truth information
                self.apply_qsquare_infos(df=df_mc)
                # create dss sig id
                self.create_dss_sig_id(df=df_mc)
                # create advanced sig id
                self.create_advanced_sig_id(df=df_mc)

                tmp_mc_output_path = os.path.join(self.base_dir, f"tmp_{mc_output_key}.h5")
                df_mc.to_hdf(path_or_buf=tmp_mc_output_path, key="fitting")
                os.rename(src=tmp_mc_output_path, dst=mc_output_path)

                # Data
                df_data = pd.read_hdf(data_paths[self.data_key])
                assert isinstance(df_data, pd.DataFrame)
                tmp_data_output_path = os.path.join(self.base_dir, f"tmp_{data_output_key}.h5")
                df_data.to_hdf(path_or_buf=tmp_data_output_path, key="fitting")
                os.rename(src=tmp_data_output_path, dst=data_output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_fitting_preparation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_fitting_preparation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict


@inherits(FittingPreparationTask)
class CopyToSSDTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._ssd_dir: Optional[PathType] = None
        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def ssd_dir(self) -> PathType:
        if self._ssd_dir is None:
            user: str = utilities.get_user_name()
            host: str = utilities.get_host_name()
            if user == "pecker" and host == "deepthought":
                self._ssd_dir = "/local/scratch/ssd2/pecker/tmp"
            elif user == "fmetzner" and host in utilities.cache_dir_mapping.keys():
                self._ssd_dir = utilities.cache_dir_mapping[host]
            else:
                raise RuntimeError(f"Don't know what to do for user '{user}' on host '{host}'")

        return self._ssd_dir

    def run_copy_to_ssd(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()

        for sideband in available_sidebands:
            if not sideband:
                mc_output_key = f"combined_mc_prepared_{self.selection}"
                data_output_key = f"data_prepared_{self.selection}"
            else:
                mc_output_key = f"combined_mc_prepared_{self.selection}_{sideband}"
                data_output_key = f"data_prepared_{self.selection}_{sideband}"

            mc_output_path = os.path.join(self.ssd_dir, f"{mc_output_key}.h5")
            output_dict.update({mc_output_key: mc_output_path})

            data_output_path = os.path.join(self.ssd_dir, f"{data_output_key}.h5")
            output_dict.update({data_output_key: data_output_path})

            if not dry_run:
                mc_input_path = self.input_dict[mc_output_key]
                shutil.copyfile(src=mc_input_path, dst=mc_output_path)

                data_input_path = self.input_dict[data_output_key]
                shutil.copyfile(src=data_input_path, dst=data_output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_copy_to_ssd()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_copy_to_ssd(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(FittingPreparationTask)
