import json

import numpy as np
import pandas as pd
import uncertainties

from typing import Union, Optional, Tuple, List, Dict, NamedTuple, Any

from analysistools.statistics import limits as rdstar_limits

from rdstar.utilities import PathType
from rdstar.offline_analysis.mc_scaling_weights import MCSampleScaler
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.rdstar_samples import data_type_combinations
from rdstar.offline_analysis.rdstar_samples import get_number_of_produced_events
from rdstar.offline_analysis.weight_management.weight_information import WeightTypeCollection, WeightManager


__all__ = [
    "EfficiencyInfo",
    "DataTypeFraction",
    "EfficiencyNumbersContainer",
    "SimpleEfficiencyCalculator",
]


class EfficiencyInfo(NamedTuple):
    nominal_value: float
    std_dev_up: float
    std_dev_down: float


class DataTypeFraction(NamedTuple):
    data_types: Tuple[str, ...]
    fraction: float


class EventsByDecayModeEntry(NamedTuple):
    sig_id: int
    decay_mode_id: int
    n_events: float


_mixed_dstst_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["Dss"] if c.startswith("BZ")],
        *[c for c in data_type_combinations["DssTau"] if c.startswith("TBZ")],
    ]
)
_charged_dstst_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["Dss"] if c.startswith("BM")],
        *[c for c in data_type_combinations["DssTau"] if c.startswith("TBM")],
    ]
)


class EffNumbersEntry:
    def __init__(
        self,
        sig_id: int,
        produced_events: float,
        selected_events: float,
        selected_events_by_decay_mode: Tuple[EventsByDecayModeEntry, ...],
        selected_events_squared_weights: float,
        selected_events_squared_weights_by_decay_mode: Tuple[EventsByDecayModeEntry, ...],
    ) -> None:
        assert produced_events >= selected_events, (produced_events, selected_events)
        assert all(sbd.sig_id == sig_id for sbd in selected_events_by_decay_mode), (
            sig_id,
            [e.sig_id for e in selected_events_by_decay_mode],
        )

        _sum_of_events: float = sum([sbd.n_events for sbd in selected_events_by_decay_mode])
        assert selected_events == _sum_of_events, (selected_events, _sum_of_events)

        self.__decay_modes: Tuple[int, ...] = tuple([e.decay_mode_id for e in selected_events_by_decay_mode])
        assert len(self.__decay_modes) == len(set(self.__decay_modes)), (
            self.__decay_modes,
            len(self.__decay_modes),
            len(set(self.__decay_modes)),
        )

        self.__sig_id: int = sig_id
        self.__produced_events: float = produced_events
        self.__selected_events: float = selected_events
        self.__selected_events_by_decay_mode: Tuple[EventsByDecayModeEntry, ...] = selected_events_by_decay_mode
        self.__selected_events_squared_weights: float = selected_events_squared_weights
        self.__selected_events_squared_weights_by_decay_mode: Tuple[
            EventsByDecayModeEntry, ...
        ] = selected_events_squared_weights_by_decay_mode

    @property
    def sig_id(self) -> int:
        return self.__sig_id

    @sig_id.setter
    def sig_id(self, value: int) -> None:
        raise NotImplementedError("The property sig_id is not mutable")

    @property
    def produced_events(self) -> float:
        return self.__produced_events

    @produced_events.setter
    def produced_events(self, value: float) -> None:
        raise NotImplementedError("The property produced_events is not mutable")

    @property
    def selected_events(self) -> float:
        return self.__selected_events

    @selected_events.setter
    def selected_events(self, value: float) -> None:
        raise NotImplementedError("The property selected_events is not mutable")

    @property
    def selected_events_squared_weights(self) -> float:
        return self.__selected_events_squared_weights

    @selected_events_squared_weights.setter
    def selected_events_squared_weights(self, value: float) -> None:
        raise NotImplementedError("The property selected_events_squared_weights is not mutable")

    @property
    def selected_events_by_decay_mode(self) -> Tuple[EventsByDecayModeEntry, ...]:
        return self.__selected_events_by_decay_mode

    @selected_events_by_decay_mode.setter
    def selected_events_by_decay_mode(self, value: Tuple[EventsByDecayModeEntry, ...]) -> None:
        raise NotImplementedError("The property selected_events_by_decay_mode is not mutable")

    @property
    def selected_events_squared_weights_by_decay_mode(self) -> Tuple[EventsByDecayModeEntry, ...]:
        return self.__selected_events_by_decay_mode

    @selected_events_squared_weights_by_decay_mode.setter
    def selected_events_squared_weights_by_decay_mode(self, value: Tuple[EventsByDecayModeEntry, ...]) -> None:
        raise NotImplementedError("The property selected_events_squared_weights_by_decay_mode is not mutable")

    @property
    def decay_modes(self) -> Tuple[int, ...]:
        return self.__decay_modes

    def get_events_for_decay_mode(self, decay_mode: int) -> float:
        if decay_mode not in self.decay_modes:
            raise KeyError(f"No entry for decay_mode {decay_mode}!\nAvailable decay modes are {self.decay_modes}...")
        events_by_decay_mode_entries: List[EventsByDecayModeEntry] = [
            e for e in self.selected_events_by_decay_mode if e.decay_mode_id == decay_mode
        ]
        assert len(events_by_decay_mode_entries) == 1, (len(events_by_decay_mode_entries), events_by_decay_mode_entries)
        return events_by_decay_mode_entries[0].n_events

    def get_squared_sum_of_weights_for_decay_mode(self, decay_mode: int) -> float:
        if decay_mode not in self.decay_modes:
            raise KeyError(f"No entry for decay_mode {decay_mode}!\nAvailable decay modes are {self.decay_modes}...")
        sum_of_squared_weights_by_decay_mode_entries: List[EventsByDecayModeEntry] = [
            e for e in self.selected_events_squared_weights_by_decay_mode if e.decay_mode_id == decay_mode
        ]
        assert len(sum_of_squared_weights_by_decay_mode_entries) == 1, (
            len(sum_of_squared_weights_by_decay_mode_entries),
            sum_of_squared_weights_by_decay_mode_entries,
        )
        return sum_of_squared_weights_by_decay_mode_entries[0].n_events

    def get_info_for_decay_mode(self, decay_mode: int, info_type: str) -> float:
        if info_type == "event_count":
            return self.get_events_for_decay_mode(decay_mode=decay_mode)
        elif info_type == "sum_of_squared_weights":
            return self.get_squared_sum_of_weights_for_decay_mode(decay_mode=decay_mode)
        else:
            raise RuntimeError(
                f"Invalid info_type {info_type}, must be one of 'event_count' or 'sum_of_squared_weights'!"
            )

    def as_dict(self) -> Dict[str, Any]:
        dict_representation = {
            "sig_id": self.sig_id,
            "produced_events": self.produced_events,
            "selected_events": self.selected_events,
            "selected_events_by_decay_mode": [e._asdict() for e in self.selected_events_by_decay_mode],
        }
        return dict_representation

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "EffNumbersEntry":
        assert all(isinstance(e, dict) for e in input_dict["selected_events_by_decay_mode"])
        assert isinstance(input_dict["sig_id"], int) or input_dict["sig_id"].isdigit(), input_dict["sig_id"]
        events_by_dec_id: Tuple[EventsByDecayModeEntry, ...] = tuple(
            [EventsByDecayModeEntry(**e) for e in input_dict["selected_events_by_decay_mode"]]
        )
        sosqw_by_dec_id: Tuple[EventsByDecayModeEntry, ...] = tuple(
            [EventsByDecayModeEntry(**e) for e in input_dict["selected_events_squared_weights_by_decay_mode"]]
        )
        return EffNumbersEntry(
            sig_id=int(input_dict["sig_id"]),
            produced_events=input_dict["produced_events"],
            selected_events=input_dict["selected_events"],
            selected_events_by_decay_mode=events_by_dec_id,
            selected_events_squared_weights=input_dict["selected_events_squared_weights"],
            selected_events_squared_weights_by_decay_mode=sosqw_by_dec_id,
        )


class EfficiencyNumbersContainer:
    def __init__(self) -> None:
        self.__eff_numbers_dict: Dict[int, EffNumbersEntry] = dict()

    @property
    def is_empty(self) -> bool:
        return len(self.__eff_numbers_dict.keys()) == 0

    def add_entry(
        self,
        sig_id: int,
        produced_events: float,
        selected_events: float,
        selected_events_by_decay_mode: Tuple[EventsByDecayModeEntry, ...],
        selected_events_squared_weights: float,
        selected_events_squared_weights_by_decay_mode: Tuple[EventsByDecayModeEntry, ...],
    ) -> None:
        if sig_id in self.__eff_numbers_dict.keys():
            raise KeyError(f"Entry for sig id {sig_id} already exists!")

        new_entry = EffNumbersEntry(
            sig_id=sig_id,
            produced_events=produced_events,
            selected_events=selected_events,
            selected_events_by_decay_mode=selected_events_by_decay_mode,
            selected_events_squared_weights=selected_events_squared_weights,
            selected_events_squared_weights_by_decay_mode=selected_events_squared_weights_by_decay_mode,
        )
        self.__eff_numbers_dict.update({sig_id: new_entry})

    def get_entry(
        self,
        sig_id: int,
    ) -> EffNumbersEntry:
        assert sig_id in self.__eff_numbers_dict.keys(), (sig_id, self.__eff_numbers_dict.keys())
        return self.__eff_numbers_dict[sig_id]

    def get_produced_events(
        self,
        sig_ids: Union[int, Tuple[int, ...]],
    ) -> float:
        if isinstance(sig_ids, int):
            return self.get_entry(sig_id=sig_ids).produced_events
        elif isinstance(sig_ids, tuple):
            assert all(isinstance(sig_id, int) for sig_id in sig_ids), sig_ids
            return sum([self.get_entry(sig_id=sig_id).produced_events for sig_id in sig_ids])
        else:
            raise TypeError(f"Argument sig_ids must be int or tuple of int, but {type(sig_ids).__name__} was provided!")

    def get_info_for_selected(
        self,
        sig_ids: Union[int, Tuple[int, ...]],
        decay_mode_ids: Union[int, Tuple[int, ...]],
        info_type: str,
    ) -> float:
        if isinstance(sig_ids, int):
            eff_entries: List[EffNumbersEntry] = [self.get_entry(sig_id=sig_ids)]
        elif isinstance(sig_ids, tuple):
            assert all(isinstance(sig_id, int) for sig_id in sig_ids), sig_ids
            eff_entries = [self.get_entry(sig_id=sig_id) for sig_id in sig_ids]
        else:
            raise TypeError(f"Argument sig_ids must be int or tuple of int, but {type(sig_ids).__name__} was provided!")

        if info_type not in ("event_count", "sum_of_squared_weights"):
            raise RuntimeError(
                f"Invalid info_type {info_type}, must be one of 'event_count' or 'sum_of_squared_weights'!"
            )

        events_per_dec_id: List[float] = []
        if isinstance(decay_mode_ids, int):
            events_per_dec_id.extend(
                [
                    eff_entry.get_info_for_decay_mode(decay_mode=decay_mode_ids, info_type=info_type)
                    for eff_entry in eff_entries
                    if decay_mode_ids in eff_entry.decay_modes
                ]
            )
        elif isinstance(decay_mode_ids, tuple):
            assert all(isinstance(decay_mode_ids, int) for decay_mode_ids in decay_mode_ids), decay_mode_ids
            events_per_dec_id.extend(
                [
                    eff_entry.get_info_for_decay_mode(decay_mode=dec_id, info_type=info_type)
                    for dec_id in decay_mode_ids
                    for eff_entry in eff_entries
                    if dec_id in eff_entry.decay_modes
                ]
            )
        else:
            raise TypeError(f"Argument sig_ids must be int or tuple of int, but {type(sig_ids).__name__} was provided!")

        return sum(events_per_dec_id)

    def get_selected_events(
        self,
        sig_ids: Union[int, Tuple[int, ...]],
        decay_mode_ids: Union[int, Tuple[int, ...]],
    ) -> float:
        return self.get_info_for_selected(
            sig_ids=sig_ids,
            decay_mode_ids=decay_mode_ids,
            info_type="event_count",
        )

    def get_sum_of_squared_weights_for_selected_events(
        self,
        sig_ids: Union[int, Tuple[int, ...]],
        decay_mode_ids: Union[int, Tuple[int, ...]],
    ) -> float:
        return self.get_info_for_selected(
            sig_ids=sig_ids,
            decay_mode_ids=decay_mode_ids,
            info_type="sum_of_squared_weights",
        )

    def get_efficiency(
        self,
        sig_ids: Union[int, Tuple[int, ...]],
        decay_mode_ids: Union[int, Tuple[int, ...]],
    ) -> EfficiencyInfo:
        n_selected: float = self.get_selected_events(sig_ids=sig_ids, decay_mode_ids=decay_mode_ids)
        n_produced: float = self.get_produced_events(sig_ids=sig_ids)
        nominal_efficiency: float = n_selected / n_produced
        upper_value, lower_value = rdstar_limits.binomial_limit_approximation(
            n=n_produced, k=n_selected, sigma=1.0
        )  # type: float, float
        assert np.isclose(upper_value - nominal_efficiency, nominal_efficiency - lower_value), (
            lower_value,
            nominal_efficiency,
            upper_value,
        )

        assert lower_value <= nominal_efficiency <= upper_value, (lower_value, nominal_efficiency, upper_value)
        return EfficiencyInfo(
            nominal_value=nominal_efficiency,
            std_dev_up=upper_value - nominal_efficiency,
            std_dev_down=nominal_efficiency - lower_value,
        )

    def get_efficiencies_with_systematics(
        self,
        sig_ids_a: Union[int, Tuple[int, ...]],
        sig_ids_b: Union[int, Tuple[int, ...]],
        decay_mode_ids: Union[int, Tuple[int, ...]],
        systematics_cov_matrix: np.ndarray,
    ) -> Tuple[uncertainties.UFloat, uncertainties.UFloat]:
        selected_events: List[float] = []
        produced_events: List[float] = []

        for sig_ids in [sig_ids_a, sig_ids_b]:
            n_selected: float = self.get_selected_events(sig_ids=sig_ids, decay_mode_ids=decay_mode_ids)
            selected_events.append(n_selected)
            n_produced: float = self.get_produced_events(sig_ids=sig_ids)
            produced_events.append(n_produced)

        assert len(selected_events) == 2, selected_events
        assert len(produced_events) == 2, produced_events

        selected_events_with_correlated_sys_uncert: Tuple[
            uncertainties.UFloat, uncertainties.UFloat
        ] = uncertainties.correlated_values(nom_values=selected_events, covariance_mat=systematics_cov_matrix)
        assert len(selected_events_with_correlated_sys_uncert) == 2, selected_events_with_correlated_sys_uncert

        nominal_efficiency_values: Tuple[uncertainties.UFloat, ...] = tuple(
            n_sel / n_prod for n_sel, n_prod in zip(selected_events_with_correlated_sys_uncert, produced_events)
        )
        assert len(nominal_efficiency_values) == 2, nominal_efficiency_values

        (eff_a_with_sys, eff_b_with_sys) = nominal_efficiency_values  # type: uncertainties.UFloat, uncertainties.UFloat

        return eff_a_with_sys, eff_b_with_sys

    def as_dict(self) -> Dict[int, Any]:
        return {k: v.as_dict() for k, v in self.__eff_numbers_dict.items()}

    def fill_from_dict(self, dict_input: Dict[int, Any], convert_keys: bool = False) -> None:
        if not len(self.__eff_numbers_dict) == 0:
            raise RuntimeError("EfficiencyNumbersContainer is not empty! Cannot be filled from dictionary...")

        if convert_keys:
            _dict_input: Dict[int, Any] = {int(k): v for k, v in dict_input.items()}
        else:
            _dict_input = dict_input

        assert all([isinstance(k, int) for k in _dict_input.keys()]), list(_dict_input.keys())
        self.__eff_numbers_dict = {k: EffNumbersEntry.from_dict(v) for k, v in _dict_input.items()}

    def to_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path, "w") as json_file:
            json.dump(self.as_dict(), json_file, indent=2)

    def from_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path) as f:
            _input_dict = json.load(f)
        assert isinstance(_input_dict, dict), type(_input_dict).__name__
        assert all([isinstance(k, int) or k.isdigit() for k in _input_dict.keys()]), list(_input_dict.keys())
        input_dict = {int(k): v for k, v in _input_dict.items()}
        assert all([isinstance(k, int) for k in input_dict.keys()]), list(input_dict.keys())
        self.fill_from_dict(dict_input=input_dict)


class SimpleEfficiencyCalculator:
    data_type_column_name: str = Cols.data_type
    decay_mode_column_name: str = Cols.decay_mode_column_name

    own_eff_weight_column_name: str = "tmp_simple_efficiency_weight"

    def __init__(
        self,
        df: pd.DataFrame,
        sig_id_data_type_mapping: Dict[int, Tuple[DataTypeFraction, ...]],
        sig_id_column_name: str,
        file_dict_base_path: PathType,
        use_new_fei_calibration_weights: bool = False,
        additional_scaling_weight_columns: Optional[Tuple[str, ...]] = None,
    ) -> None:
        self._sig_id_column_name: str = sig_id_column_name
        self._sig_id_data_type_mapping: Dict[int, Tuple[DataTypeFraction, ...]] = sig_id_data_type_mapping
        self._file_dict_base_path: PathType = file_dict_base_path

        self._use_new_fei_calibration_weights: bool = use_new_fei_calibration_weights
        self._additional_scaling_weight_columns: Optional[Tuple[str, ...]] = additional_scaling_weight_columns

        self._scale_weight_map: Dict[str, float] = MCSampleScaler.generate_data_type_scaling_map(
            file_dict_dir_path=self._file_dict_base_path
        )

        self._df: pd.DataFrame = self.init_info_df(input_df=df)
        self._dt_scaling_map: Dict[int, Dict[str, float]] = self.get_data_type_scaling_map()
        self._eff_numbers_container: EfficiencyNumbersContainer = EfficiencyNumbersContainer()

    @property
    def eff_numbers_container(self) -> EfficiencyNumbersContainer:
        if self._eff_numbers_container.is_empty:
            self.generate_efficiency_numbers_dict()
        return self._eff_numbers_container

    def get_decay_mode_ids(self) -> Tuple[int, ...]:
        return tuple(self._df[self.decay_mode_column_name].unique())

    def init_info_df(self, input_df: pd.DataFrame) -> pd.DataFrame:
        used_cols: List[str] = [
            self._sig_id_column_name,
            self.data_type_column_name,
            self.decay_mode_column_name,
        ]
        eff_weight_cols: List[str] = list(WeightTypeCollection.efficiency_weight_columns)

        if self._use_new_fei_calibration_weights:
            eff_weight_cols.append(WeightManager.used_fei_calib_col)

        if self._additional_scaling_weight_columns is not None:
            assert all(c in input_df.columns for c in self._additional_scaling_weight_columns), [
                c for c in self._additional_scaling_weight_columns if c not in input_df.columns
            ]
            eff_weight_cols.extend(list(self._additional_scaling_weight_columns))

        required_cols: List[str] = used_cols + eff_weight_cols
        assert all(c in input_df.columns for c in required_cols), [c for c in required_cols if c not in input_df.columns]

        assert len(set(eff_weight_cols)) == len(eff_weight_cols), eff_weight_cols

        df: pd.DataFrame = input_df.loc[:, required_cols].copy(deep=True)
        df.loc[:, self.own_eff_weight_column_name] = df[eff_weight_cols].product(axis=1)

        return df

    def get_data_type_scaling_map(self) -> Dict[int, Dict[str, float]]:
        scaling_map: Dict[int, Dict[str, float]] = dict()

        for sig_id in self._sig_id_data_type_mapping.keys():
            data_types: Tuple[str, ...] = tuple(
                set([dt for dt_fracs in self._sig_id_data_type_mapping[sig_id] for dt in dt_fracs.data_types])
            )

            bare_scaling_map: Dict[str, float] = dict()

            for dt in data_types:
                assert dt not in bare_scaling_map.keys(), (dt, bare_scaling_map.keys())
                assert dt in self._scale_weight_map.keys(), (dt, self._scale_weight_map.keys())
                dt_scaling_factor: float = self._scale_weight_map[dt]

                dt_fraction_info: List[DataTypeFraction] = [
                    dtf for dtf in self._sig_id_data_type_mapping[sig_id] if dt in dtf.data_types
                ]
                assert len(dt_fraction_info) == 1, len(dt_fraction_info)
                # signal_fraction: float = dt_fraction_info[0].fraction
                signal_fraction: float = 1.0

                bare_scaling_map.update({dt: signal_fraction * dt_scaling_factor})

            if all(c in bare_scaling_map.keys() for c in ["Mixed", "Charged"]):
                base_scaling: float = max([bare_scaling_map[k] for k in ["Mixed", "Charged"]])
            elif all(c in bare_scaling_map.keys() for c in _mixed_dstst_dts):
                base_scaling = max([bare_scaling_map[k] for k in _mixed_dstst_dts])
            elif all(c in bare_scaling_map.keys() for c in _charged_dstst_dts):
                base_scaling = max([bare_scaling_map[k] for k in _charged_dstst_dts])
            elif any(c in bare_scaling_map.keys() for c in ["Mixed", "Charged"]):
                base_scaling = bare_scaling_map.get("Mixed", bare_scaling_map.get("Charged", -999.0))
                assert base_scaling is not -999.0
            else:
                base_scaling = max(bare_scaling_map.values())
            scale_map_entry: Dict[str, float] = {k: v / base_scaling for k, v in bare_scaling_map.items()}
            scaling_map.update({sig_id: scale_map_entry})
        return scaling_map

    def get_produced_events(
        self,
        data_type: str,
    ) -> int:
        return get_number_of_produced_events(data_type=data_type, file_dict_dir_path=self._file_dict_base_path)

    def get_produced_events_for_process(
        self,
        sig_id: int,
    ) -> float:
        assert sig_id in self._sig_id_data_type_mapping.keys()
        data_type_fractions: Tuple[DataTypeFraction, ...] = self._sig_id_data_type_mapping[sig_id]
        scaled_produced_events: float = 0.0

        for dt_fraction_info in data_type_fractions:
            produced_events: float = 0.0
            for data_type in dt_fraction_info.data_types:
                produced_events += self.get_produced_events(data_type=data_type) * self._dt_scaling_map[sig_id][data_type]

            scaled_produced_events += produced_events * dt_fraction_info.fraction

        return scaled_produced_events

    def get_selected_events_for_process(
        self,
        sig_id: int,
        decay_mode_id: int,
    ) -> Tuple[float, float]:
        relevant_data_types: Tuple[str, ...] = tuple(
            [dt for dt_info in self._sig_id_data_type_mapping[sig_id] for dt in dt_info.data_types]
        )

        decay_mode_cond: pd.Series = self._df[self.decay_mode_column_name] == decay_mode_id
        sig_id_cond: pd.Series = self._df[self._sig_id_column_name] == sig_id

        selected_events: float = 0.0
        sum_of_squared_weights: float = 0.0

        for data_type in relevant_data_types:
            data_type_cond: pd.Series = self._df[self.data_type_column_name] == data_type
            all_conditions: pd.Series = decay_mode_cond & sig_id_cond & data_type_cond

            scaling_weight: float = self._dt_scaling_map[sig_id][data_type]

            events: float = float(np.sum(self._df.loc[all_conditions, self.own_eff_weight_column_name].values))
            selected_events += events * scaling_weight

            squared_weights: float = float(
                np.sum(np.sqare(scaling_weight * self._df.loc[all_conditions, self.own_eff_weight_column_name].values))
            )
            sum_of_squared_weights += squared_weights

        return selected_events, sum_of_squared_weights

    def generate_efficiency_numbers_dict(self) -> None:
        assert self._eff_numbers_container.is_empty

        for sig_id in self._sig_id_data_type_mapping.keys():
            produced_events: float = self.get_produced_events_for_process(sig_id=sig_id)
            selected_events: float = 0.0
            selected_events_by_decay_ids: List[EventsByDecayModeEntry] = []

            selected_events_squared_weight: float = 0.0
            selected_events_squared_weight_by_decay_ids: List[EventsByDecayModeEntry] = []

            for decay_mode_id in self.get_decay_mode_ids():
                n_evts_for_decay_id, squared_weights_for_decay_id = self.get_selected_events_for_process(
                    sig_id=sig_id,
                    decay_mode_id=decay_mode_id,
                )  # type: float, float

                selected_events += n_evts_for_decay_id
                selected_events_squared_weight += squared_weights_for_decay_id

                selected_events_by_decay_ids.append(
                    EventsByDecayModeEntry(
                        sig_id=sig_id,
                        decay_mode_id=decay_mode_id,
                        n_events=n_evts_for_decay_id,
                    )
                )

                selected_events_squared_weight_by_decay_ids.append(
                    EventsByDecayModeEntry(
                        sig_id=sig_id,
                        decay_mode_id=decay_mode_id,
                        n_events=squared_weights_for_decay_id,
                    )
                )

            self._eff_numbers_container.add_entry(
                sig_id=sig_id,
                produced_events=produced_events,
                selected_events=selected_events,
                selected_events_by_decay_mode=tuple(selected_events_by_decay_ids),
                selected_events_squared_weights=selected_events_squared_weight,
                selected_events_squared_weights_by_decay_mode=tuple(selected_events_squared_weight_by_decay_ids),
            )

    def write_eff_dict_to_file(self, file_path: PathType) -> None:
        self.eff_numbers_container.to_file(file_path=file_path)

    def get_efficiency_info_container(self) -> EfficiencyNumbersContainer:
        return self.eff_numbers_container
