"""
Fitter class for extraction of B -> D** l nu BF and/or efficiency correction factors.

Felix Metzner 2022
"""

import numpy as np
import pandas as pd

from typing import Optional, Tuple

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.dss_cal_fitters.dss_cal_fitter_base import (
    DssBFCalibrationMappingType,
    DssBFCorrFitterBase,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.dss_cal_third_approach_fit_info import (
    dss_cal_fitter_setup_wide_dss_ids,
    dss_cal_fitter_setup_narrow_dss_ids,
)


__all__ = [
    "DssBFCorrFitter",
]


logger = log_to_default(name=__name__)


class DssBFCorrFitter(DssBFCorrFitterBase):

    this_fitter_setup: FitSetupTriplet = FitSetupCollection.dss_c_third_approach__fs

    ordered_parameter_names: Tuple[str, ...] = (
        # D** pi0 Components
        "B_wideDss_npi_l_yield",
        "B_narrowDss_npi_l_yield",
        # D** pi+/- Components
        "B_wideDss_cpi_l_yield",
        "B_narrowDss_cpi_l_yield",
        # D(*) Components
        "B_to_Dwwos_in_D_cpi_yield",
        "B_to_Dwwos_in_Ds_cpi_yield",
        "B_to_Dwwos_in_D_npi_yield",
        "B_to_Dwwos_in_Ds_npi_yield",
        # Bkg Components
        "bkg_in_d_cpi_yield",
        "bkg_in_ds_cpi_yield",
        "bkg_in_d_npi_yield",
        "bkg_in_ds_npi_yield",
    )

    B_wideDss_npi_l_yield_index: int = 0
    B_narrowDss_npi_l_yield_index: int = 1
    B_wideDss_cpi_l_yield_index: int = 2
    B_narrowDss_cpi_l_yield_index: int = 3

    B_to_Dwwos_in_D_cpi_yield_index: int = 4
    B_to_Dwwos_in_Ds_cpi_yield_index: int = 5
    B_to_Dwwos_in_D_npi_yield_index: int = 6
    B_to_Dwwos_in_Ds_npi_yield_index: int = 7

    bkg_in_d_cpi_yield_index: int = 8
    bkg_in_ds_cpi_yield_index: int = 9
    bkg_in_d_npi_yield_index: int = 10
    bkg_in_ds_npi_yield_index: int = 11

    end_yield_params_index: int = bkg_in_ds_npi_yield_index

    main_floating_params_slice = np.s_[B_wideDss_npi_l_yield_index : end_yield_params_index + 1 : 1]

    number_of_hardcoded_floating_params: int = end_yield_params_index + 1

    fixed_params_indices: Tuple[int, ...] = tuple([])
    manual_nuisances_params_indices: Tuple[int, ...] = tuple([])

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=None,
            **kwargs,
        )

        assert len(self.fit_setup_info.components) == self.end_yield_params_index + 1, (
            [(c.comp_index, c.name) for c in self.fit_setup_info.components],
            len(self.fit_setup_info.components),
            self.end_yield_params_index + 1,
        )

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    @property
    def _parameter_limits(self) -> Tuple[Tuple[int, Tuple[Optional[float], Optional[float]]], ...]:
        return tuple()

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        _chi2 = np.sum(
            (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
            / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
            axis=None,
        )
        assert isinstance(_chi2, float)
        return _chi2

    def gaussian_constrains(self, params: np.ndarray) -> float:
        # TODO!!!!!!
        return 0.0

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return params[self.main_floating_params_slice]

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = 0  # TODO!
        return self.number_of_hardcoded_floating_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        # Nuisances:
        automated_nuisances_initial_values: Tuple[float, ...] = tuple([])

        _initial_params = np.array(
            [
                _normalization_factors[self.B_wideDss_npi_l_yield_index],
                _normalization_factors[self.B_narrowDss_npi_l_yield_index],
                _normalization_factors[self.B_wideDss_cpi_l_yield_index],
                _normalization_factors[self.B_narrowDss_cpi_l_yield_index],
                #
                _normalization_factors[self.B_to_Dwwos_in_D_cpi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_Ds_cpi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_D_npi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_Ds_npi_yield_index],
                #
                _normalization_factors[self.bkg_in_d_cpi_yield_index],
                _normalization_factors[self.bkg_in_ds_cpi_yield_index],
                _normalization_factors[self.bkg_in_d_npi_yield_index],
                _normalization_factors[self.bkg_in_ds_npi_yield_index],
                #
                *automated_nuisances_initial_values,
            ]
        )

        assert len(_initial_params) == self.end_of_floating_params, (
            len(_initial_params),
            self.end_of_floating_params,
            _initial_params,
        )

        return _initial_params

    # endregion

    # region Correction application info
    @classmethod
    def get_correction_mapping(cls) -> DssBFCalibrationMappingType:
        dss_charged_pi_reco_ids: Tuple[int, ...] = (OldMCMatchingIDsQueryInfo.DSSlbkg.id,)
        dss_neutral_pi_reco_ids: Tuple[int, ...] = (OldMCMatchingIDsQueryInfo.DSSl.id,)

        wide_dss_ids: Tuple[int, ...] = dss_cal_fitter_setup_wide_dss_ids
        narrow_dss_ids: Tuple[int, ...] = dss_cal_fitter_setup_narrow_dss_ids

        _possible_result: DssBFCalibrationMappingType = {
            (dss_charged_pi_reco_ids, wide_dss_ids): cls.B_wideDss_npi_l_yield_index,
            (dss_charged_pi_reco_ids, narrow_dss_ids): cls.B_narrowDss_npi_l_yield_index,
            (dss_neutral_pi_reco_ids, wide_dss_ids): cls.B_wideDss_cpi_l_yield_index,
            (dss_neutral_pi_reco_ids, narrow_dss_ids): cls.B_narrowDss_cpi_l_yield_index,
        }

        return _possible_result

    # endregion
