"""
Fit setup info for D** calibration fit.

Felix Metzner 2023
"""

from typing import Tuple
from uncertainties import ufloat

from templatefitter.plotter.plot_style import KITColors, TangoColors

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollectionItem

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    dss_cal_fit_obs_setup,
    base_dss_cal_weight_info,
    dss_cal_bbbar_bkg_sig_ids,
    main_dss_cal_fit_component_id_column_name,
    add_main_dss_cal_fit_component_id_to,
    get_sig_ids_with_main_d_charge_info,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)


__all__ = [
    "dss_cal_fit_setup",
    "dss_cal_fit__component_info_collection",
    "dss_cal_fitter_setup_wide_dss_ids",
    "dss_cal_fitter_setup_narrow_dss_ids",
]


logger = log_to_default(name=__name__)


# Definition of components:

# (Main) D** Components

dss_neutral_pi_id: int = OldMCMatchingIDsQueryInfo.DSSl.id
dss_charged_pi_id: int = OldMCMatchingIDsQueryInfo.DSSlbkg.id

dwwos_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.Dtau.id,
    OldMCMatchingIDsQueryInfo.DStau.id,
    OldMCMatchingIDsQueryInfo.Dl.id,
    OldMCMatchingIDsQueryInfo.DSl.id,
)
bkg_ids: Tuple[int, ...] = dss_cal_bbbar_bkg_sig_ids + (OldMCMatchingIDsQueryInfo.CBKG.id,)


# D** pi0 Components

component_B_wideDss_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_neutral_pi_id,
    reco_type="all",
    main_d_type="all",
    true_dss_type="wide",
)

component_B_wideDss_npi_l: ComponentInfo = ComponentInfo(
    name="B_to_wideDss_npi_l",
    comp_index=10,
    latex_label=rf"Wide {LatexParticleNames.d_st_st_str} $+$ {LatexParticleNames.pi_zero_str}",
    sig_ids=component_B_wideDss_npi_ids,
    color=KITColors.kit_red,
)

component_B_narrowDss_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_neutral_pi_id,
    reco_type="all",
    main_d_type="all",
    true_dss_type="narrow",
)

component_B_narrowDss_npi_l: ComponentInfo = ComponentInfo(
    name="B_to_narrowDss_npi_l",
    comp_index=11,
    latex_label=rf"Nar. {LatexParticleNames.d_st_st_str} $+$ {LatexParticleNames.pi_zero_str}",
    sig_ids=component_B_narrowDss_npi_ids,
    color=TangoColors.scarletred2,
)


# D** pi+/- Components

component_B_wideDss_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_charged_pi_id,
    reco_type="all",
    main_d_type="all",
    true_dss_type="wide",
)

component_B_wideDss_cpi_l: ComponentInfo = ComponentInfo(
    name="B_to_wideDss_cpi_l",
    comp_index=20,
    latex_label=rf"Wide {LatexParticleNames.d_st_st_str} $+$ {LatexParticleNames.pi_pm_str}",
    sig_ids=component_B_wideDss_cpi_ids,
    color=TangoColors.orange3,
)

component_B_narrowDss_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_charged_pi_id,
    reco_type="all",
    main_d_type="all",
    true_dss_type="narrow",
)

component_B_narrowDss_cpi_l: ComponentInfo = ComponentInfo(
    name="B_to_narrowDss_cpi_l",
    comp_index=21,
    latex_label=rf"Nar. {LatexParticleNames.d_st_st_str} $+$ {LatexParticleNames.pi_pm_str}",
    sig_ids=component_B_narrowDss_cpi_ids,
    color=KITColors.kit_orange,
)


# D and D* Components

# _Dlnu_bkg_label = LatexDecayProcesses.b_to_d_w_wo_st_l_nu
_Dlnu_bkg_label = rf"{LatexParticleNames.d_w_wo_st_str} {LatexParticleNames.lepton_str} {LatexParticleNames.nu_lep_str}"

component_B_to_Dwwos_in_D_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dwwos_ids,
    reco_type="d_cpi",
    main_d_type="all",
)
component_B_to_Dwwos_in_D_cpi: ComponentInfo = ComponentInfo(
    name="BDwwos_inDCPi",
    comp_index=40,
    latex_label=rf"{_Dlnu_bkg_label} in {LatexDecayProcesses.reco_dss_val_d_charged_pi}",
    sig_ids=component_B_to_Dwwos_in_D_cpi_ids,
    color=KITColors.kit_blue,
)

component_B_to_Dwwos_in_Ds_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dwwos_ids,
    reco_type="dst_cpi",
    main_d_type="all",
)
component_B_to_Dwwos_in_Ds_cpi: ComponentInfo = ComponentInfo(
    name="BDwwos_inDsCPi",
    comp_index=41,
    latex_label=rf"{_Dlnu_bkg_label} in {LatexDecayProcesses.reco_dss_val_ds_charged_pi}",
    sig_ids=component_B_to_Dwwos_in_Ds_cpi_ids,
    color=TangoColors.skyblue1,
)


component_B_to_Dwwos_in_D_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dwwos_ids,
    reco_type="d_npi",
    main_d_type="all",
)
component_B_to_Dwwos_in_D_npi: ComponentInfo = ComponentInfo(
    name="BDwwos_inDNPi",
    comp_index=42,
    latex_label=rf"{_Dlnu_bkg_label} in {LatexDecayProcesses.reco_dss_val_d_neutral_pi}",
    sig_ids=component_B_to_Dwwos_in_D_npi_ids,
    color=KITColors.kit_cyan,
)

component_B_to_Dwwos_in_Ds_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dwwos_ids,
    reco_type="dst_npi",
    main_d_type="all",
)
component_B_to_Dwwos_in_Ds_npi: ComponentInfo = ComponentInfo(
    name="BDwwos_inDsNPi",
    comp_index=43,
    latex_label=rf"{_Dlnu_bkg_label} in {LatexDecayProcesses.reco_dss_val_ds_neutral_pi}",
    sig_ids=component_B_to_Dwwos_in_Ds_npi_ids,
    color="cyan",
)


# Background Components

dss_cal_bkg_in_d_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=bkg_ids,
    reco_type="d_cpi",
    main_d_type=None,
)

component_bkg_in_d_cpi: ComponentInfo = ComponentInfo(
    name="BKG_in_D_cpi",
    comp_index=60,
    latex_label=rf"BKG in {LatexDecayProcesses.reco_dss_val_d_charged_pi}",
    sig_ids=dss_cal_bkg_in_d_cpi_ids,
    color=TangoColors.aluminium2,
)

dss_cal_bkg_in_ds_cpi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=bkg_ids,
    reco_type="dst_cpi",
    main_d_type=None,
)

component_bkg_in_ds_cpi: ComponentInfo = ComponentInfo(
    name="BKG_in_Ds_cpi",
    comp_index=61,
    latex_label=rf"BKG in {LatexDecayProcesses.reco_dss_val_ds_charged_pi}",
    sig_ids=dss_cal_bkg_in_ds_cpi_ids,
    color=TangoColors.aluminium3,
)

dss_cal_bkg_in_d_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=bkg_ids,
    reco_type="d_npi",
    main_d_type=None,
)

component_bkg_in_d_npi: ComponentInfo = ComponentInfo(
    name="BKG_in_D_npi",
    comp_index=62,
    latex_label=rf"BKG in {LatexDecayProcesses.reco_dss_val_d_neutral_pi}",
    sig_ids=dss_cal_bkg_in_d_npi_ids,
    color=TangoColors.aluminium4,
)

dss_cal_bkg_in_ds_npi_ids: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=bkg_ids,
    reco_type="dst_npi",
    main_d_type=None,
)

component_bkg_in_ds_npi: ComponentInfo = ComponentInfo(
    name="BKG_in_Ds_npi",
    comp_index=63,
    latex_label=rf"BKG in {LatexDecayProcesses.reco_dss_val_ds_neutral_pi}",
    sig_ids=dss_cal_bkg_in_ds_npi_ids,
    color=TangoColors.aluminium5,
)


dss_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    # D** pi0 Components
    component_B_wideDss_npi_l,
    component_B_narrowDss_npi_l,
    # D** pi+/- Components
    component_B_wideDss_cpi_l,
    component_B_narrowDss_cpi_l,
    #
    # D(*) Components
    component_B_to_Dwwos_in_D_cpi,
    component_B_to_Dwwos_in_Ds_cpi,
    component_B_to_Dwwos_in_D_npi,
    component_B_to_Dwwos_in_Ds_npi,
    # Bkg Components
    component_bkg_in_d_cpi,
    component_bkg_in_ds_cpi,
    component_bkg_in_d_npi,
    component_bkg_in_ds_npi,
)


__sig_id__dss_cal_fit__collection_definition: Tuple[MCMatchingCollectionItem, ...] = tuple(
    [
        MCMatchingCollectionItem(
            name=comp.name,
            sig_ids=comp.sig_ids,
            new_id=comp.comp_index,
            latex_label=comp.latex_label,
            color=comp.color,
        )
        for comp in dss_cal_fit_comps
    ]
)
dss_cal_fit__component_info_collection: MCMatchingCollection = MCMatchingCollection(
    collection_name="original_sig_id__dss_cal_fit",
    sig_id_col_name="__tmp__" + main_dss_cal_fit_component_id_column_name,
    original_sig_id_col_name=main_dss_cal_fit_component_id_column_name,
    collection_definition=__sig_id__dss_cal_fit__collection_definition,
    base_mc_matching_id_query_infos=OldMCMatchingIDsQueryInfo,
    is_reduced_subset_of_original=False,
)


dss_cal_fit_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="B_to_DLNeutralPi",
            ch_index=0,
            reco_mode_info=RecoModeInfos.dss_d_l_neutral_pion_modes,
            components=dss_cal_fit_comps,
            observable_infos=dss_cal_fit_obs_setup.dss_obs_info_B_to_DLNeutralPi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_DStarLNeutralPi",
            ch_index=1,
            reco_mode_info=RecoModeInfos.dss_ds_l_neutral_pion_modes,
            components=dss_cal_fit_comps,
            observable_infos=dss_cal_fit_obs_setup.dss_obs_info_B_to_DStarLNeutralPi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_DLChargedPi",
            ch_index=2,
            reco_mode_info=RecoModeInfos.dss_d_l_charged_pion_modes,
            components=dss_cal_fit_comps,
            observable_infos=dss_cal_fit_obs_setup.dss_obs_info_B_to_DLChargedPi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_DStarLChargedPi",
            ch_index=3,
            reco_mode_info=RecoModeInfos.dss_ds_l_charged_pion_modes,
            components=dss_cal_fit_comps,
            observable_infos=dss_cal_fit_obs_setup.dss_obs_info_B_to_DStarLChargedPi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
    ]
)


dss_cal_fit_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[],
    key_order=(),
)


_dss_cal_asimov_fit_setup = FitSetupInfoContainer(
    components=dss_cal_fit_comps,
    reco_channels=dss_cal_fit_chs,
    observable_infos=dss_cal_fit_obs_setup.dss_cal_obs_infos,
    weight_info=base_dss_cal_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=dss_cal_fit_tag_eff_corr_info,
    reco_mode_column_name=Cols.decay_mode_column_name,
    component_id_column_name=main_dss_cal_fit_component_id_column_name,
    component_id_function=add_main_dss_cal_fit_component_id_to,
)


dss_cal_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_dss_cal_asimov_fit_setup)

component_B_wideDss_npi_ids_for_other_reco_modes: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_neutral_pi_id,
    reco_type="other",
    main_d_type="all",
    true_dss_type="wide",
)

component_B_wideDss_cpi_ids_for_other_reco_modes: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_charged_pi_id,
    reco_type="other",
    main_d_type="all",
    true_dss_type="wide",
)

component_B_narrowDss_npi_ids_for_other_reco_modes: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_neutral_pi_id,
    reco_type="other",
    main_d_type="all",
    true_dss_type="narrow",
)

component_B_narrowDss_cpi_ids_for_other_reco_modes: Tuple[int, ...] = get_sig_ids_with_main_d_charge_info(
    sig_ids=dss_charged_pi_id,
    reco_type="other",
    main_d_type="all",
    true_dss_type="narrow",
)

dss_cal_fitter_setup_wide_dss_ids: Tuple[int, ...] = (
    component_B_wideDss_npi_ids
    + component_B_wideDss_cpi_ids
    + component_B_wideDss_npi_ids_for_other_reco_modes
    + component_B_wideDss_cpi_ids_for_other_reco_modes
)
dss_cal_fitter_setup_narrow_dss_ids: Tuple[int, ...] = (
    component_B_narrowDss_npi_ids
    + component_B_narrowDss_cpi_ids
    + component_B_narrowDss_npi_ids_for_other_reco_modes
    + component_B_narrowDss_cpi_ids_for_other_reco_modes
)
