"""
Dedicated R(D) and R(D*) fitter class.

See project document for more details: https://www.overleaf.com/project/635692514ef3600c05c4c004

Felix Metzner 2022
"""

import copy
import numpy as np
import pandas as pd

from iminuit import Minuit
from typing import Union, Optional, Tuple, Dict, List, ClassVar, Any

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.branching_fractions import constants_and_measurements as bf_calc_consts

from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCIDs
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name

from rdstar.offline_analysis.fitting.dedicated_fit_approach import rdstar_systematics as rdstar_sys

from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import immutable_cached_property
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_base import FitterBase, MinuitSettings
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer

from rdstar.offline_analysis.fitting.dedicated_fit_approach.rdstar_systematics import BKGHandling
from rdstar.offline_analysis.fitting.dedicated_fit_approach.systematics_manager.shape_sys_evaluater import (
    ReducedSystematicsDetails,
    UpDownAndDiffSystematicsDetails,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.dedicated_sys_handler import (
    RDStarSystematicsContainer,
    BKGHandlingInfoContainer,
    RDStarSystematicsInfos,
)


__all__ = [
    "RDStarFitter",
]


logger = log_to_default(name=__name__)


rdstar_minuit_settings = MinuitSettings(
    error_factor=0.05,
    error_def=Minuit.LEAST_SQUARES,
    print_level=1,
    n_call=200_000,
    n_iterate=3,
)


class RDStarFitter(FitterBase):
    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        file_dict_dir_path: Optional[PathType] = None,
        **kwargs,
    ) -> None:

        _use_old_ratio_values: bool = RDStarFitter.pop_kwarg_with_default(
            kw_args=kwargs,
            key="use_old_ratio_values",
            default_value=False,
        )

        __check_additive_effects_shape: bool = RDStarFitter.pop_kwarg_with_default(
            kw_args=kwargs,
            key="check_additive_effects_shape",
            default_value=True,
        )

        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=file_dict_dir_path,
            **kwargs,
        )

        self.minuit_settings: MinuitSettings = rdstar_minuit_settings

        self._systematics: RDStarSystematicsContainer = copy.deepcopy(RDStarFitter.full_systematics_container)
        self._bkg_handler: BKGHandlingInfoContainer = copy.deepcopy(RDStarFitter.basic_bkg_handler)

        self._use_special_eff_calculation: bool = False
        self._use_old_ratio_values: bool = _use_old_ratio_values

        assert self.n_full_yield_entries == self.n_signal_and_norm_components + self.n_bkg_components, (
            self.n_full_yield_entries,
            self.n_signal_and_norm_components,
            self.n_bkg_components,
        )

        self._check_additive_effects_shape: bool = __check_additive_effects_shape

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    # region Yields parameter constants
    tau_zp_nom: ClassVar[float] = 1.0 / bf_calc_consts.tau_p0.nominal_value
    mc_tau_pz: ClassVar[float] = bf_calc_consts.tau_p0.nominal_value  # We don't use uncerts. on this MC truth!
    mc_f_pz: ClassVar[float] = bf_calc_consts.ratio_fpm_over_f00.nominal_value  # We don't use uncerts on this MC truth!
    mcRD: ClassVar[float] = bf_calc_consts.sm_r_d.nominal_value  # We don't use uncerts. on this MC truth!
    mcRDs: ClassVar[float] = bf_calc_consts.sm_r_d_st.nominal_value  # We don't use uncerts. on this MC truth!

    huschle_RD: ClassVar[float] = bf_calc_consts.huschle_r_d.nominal_value  # We don't use uncerts. on this value
    huschle_RDs: ClassVar[float] = bf_calc_consts.huschle_r_d_st.nominal_value  # We don't use uncerts. on this value

    # Definition of index ranges in params vector:
    n_floating_signal_components: ClassVar[int] = 4  # RD, RD*, y_D0l, y_D*0l
    n_calculated_signal_components: ClassVar[int] = 4
    n_signal_and_norm_components: ClassVar[int] = 8
    n_bkg_components: ClassVar[int] = 6  # Assuming 3 * 2 Background templates atm!

    n_floating_yield_params: ClassVar[int] = n_floating_signal_components + n_bkg_components
    n_full_yield_entries: ClassVar[int] = n_floating_yield_params + n_calculated_signal_components

    # Parameter slices and indices
    main_floating_params_slice = np.s_[0:n_floating_yield_params:1]

    # endregion

    # region Information derived from yields parameter constants
    efficiency_factors_padding_width: ClassVar[int] = n_full_yield_entries - n_signal_and_norm_components
    # endregion

    # region Nuisances Infos
    full_systematics_container: ClassVar[RDStarSystematicsContainer] = RDStarSystematicsContainer(
        name=RDStarSystematicsContainer.full_systematics_container_setup_name,
        first_sys_nuisance_param_index=n_floating_yield_params,
        inactive_sys_names=None,
    )

    empty_systematics_container: ClassVar[
        RDStarSystematicsContainer
    ] = full_systematics_container.get_inactive_systematics_setup()
    # endregion

    # region Nuisances Infos
    basic_bkg_handler: ClassVar[BKGHandlingInfoContainer] = BKGHandlingInfoContainer(
        dss_bkg=BKGHandling.BKG_CONSTRAINT,
        bbbar_bkg=BKGHandling.BKG_CONSTRAINT,
        continuum_bkg=BKGHandling.BKG_CONSTRAINT,
        bkg_constraint_param_id_start=RDStarSystematicsInfos.total_number_of_nuisance_parameters
        + n_floating_yield_params,
    )
    # endregion

    # region Constants for yield relations and efficiency relations
    yield_map_matrix: ClassVar[np.ndarray] = np.array(
        [
            [1, 0, 1, 0, 0, 0, 0, 0, 0, 0],  # B+ -> D0 tau+ nu
            [1, 0, 1, 0, 0, 0, 0, 0, 0, 0],  # B0 -> D- tau+ nu
            [0, 1, 0, 1, 0, 0, 0, 0, 0, 0],  # B+ -> D*0 tau+ nu
            [0, 1, 0, 1, 0, 0, 0, 0, 0, 0],  # B0 -> D*- tau+ nu
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],  # B+ -> D0 l+ nu
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],  # B0 -> D- l+ nu
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],  # B+ -> D*0 l+ nu
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],  # B0 -> D*- l+ nu
            # Dstst Bkg Components
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],  # D** BKG in B+ reco channels
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],  # D** BKG in B0 reco channels
            # BBbar Bkg Components
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],  # BBbar BKG in B+ reco channels
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],  # BBbar BKG in B0 reco channels
            # Continuum Bkg Components
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],  # Continuum BKG in B+ reco channels
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1],  # Continuum BKG in B0 reco channels
        ]
    )

    # -------------------------------------------------------- Sig       | Norm      | Bkgs
    additional_params_matrix: ClassVar[np.ndarray] = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0]) * tau_zp_nom
    anti_addit_params_matrix: ClassVar[np.ndarray] = np.array([1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1])

    eff_p1_map_matrix: ClassVar[np.ndarray] = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            # Part for normalization:
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
        ]
    )
    eff_p2_map_matrix: ClassVar[np.ndarray] = np.array(
        [
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 1],
            # Part for normalization:
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )

    @property
    def systematics(self) -> RDStarSystematicsContainer:
        return self._systematics

    @property
    def bkg_handler(self) -> BKGHandlingInfoContainer:
        return self._bkg_handler

    def set_systematics_states(self, states_infos: Dict[str, bool]) -> None:
        self.systematics.set_states(states_infos=states_infos)

    @immutable_cached_property
    def used_rd_value(self) -> float:
        """ R(D) value used for the scaling of the MC samples """
        if self._use_old_ratio_values:
            return self.huschle_RD
        else:
            return self.mcRD

    @immutable_cached_property
    def used_rd_star_value(self) -> float:
        """ R(D*) value used for the scaling of the MC samples """
        if self._use_old_ratio_values:
            return self.huschle_RDs
        else:
            return self.mcRDs

    @immutable_cached_property
    def eff_where(self) -> np.ndarray:
        assert not np.any(self.eff_p1_map_matrix * self.eff_p2_map_matrix)
        return (self.eff_p1_map_matrix + self.eff_p2_map_matrix) != 0

    @immutable_cached_property
    def yield_constants(self) -> np.ndarray:
        """
        Definition of constant factors in yield calculation
        Assuming 3 Background templates atm
        """

        # c_to_n_const: float = self.mc_tau_pz * self.mc_f_pz
        c_to_n_const: float = self.mc_tau_pz
        return np.array(
            [
                1.0 / self.used_rd_value,
                c_to_n_const / self.used_rd_value,  # tau_zp is multiplied in the get_yields method!
                1.0 / self.used_rd_star_value,
                c_to_n_const / self.used_rd_star_value,  # tau_zp is multiplied in the get_yields method!
                1.0,
                c_to_n_const,  # tau_zp is multiplied in the get_yields method!
                1.0,
                c_to_n_const,  # tau_zp is multiplied in the get_yields method!
                # D** BKG
                1.0,
                1.0,
                # BBbar BKG
                1.0,
                1.0,
                # Continuum BKG
                1.0,
                1.0,
            ]
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return (
            np.prod(
                np.multiply(self.yield_map_matrix, params[self.main_floating_params_slice]),
                axis=1,
                where=self.yield_map_matrix != 0,
            )
            * self.yield_constants
            * self.bkg_yield_constraints_effect(params=params)
            * self.get_tau_zp_factors(params=params)
            * self.get_efficiency_factors(params=params)
            * self.get_multiplicative_nuisance_effect(params=params)
        )

    # endregion

    @immutable_cached_property
    def constraint_effect_matrix(self) -> np.ndarray:
        _n_bkg_const_params: int = self.bkg_handler.number_of_bkg_constraints
        assert self.n_bkg_components == _n_bkg_const_params, (self.n_bkg_components, _n_bkg_const_params)

        _bkg_const_matrix_base: np.ndarray = np.diag(
            np.array(
                [
                    rdstar_sys.BKGHandlingDss.relative_uncert_cb,
                    rdstar_sys.BKGHandlingDss.relative_uncert_nb,
                    rdstar_sys.BKGHandlingBBBar.relative_uncert_cb,
                    rdstar_sys.BKGHandlingBBBar.relative_uncert_nb,
                    rdstar_sys.BKGHandlingContinuum.relative_uncert_cb,
                    rdstar_sys.BKGHandlingContinuum.relative_uncert_nb,
                ]
            )
        )

        _bkg_const_matrix: np.ndarray = np.pad(
            _bkg_const_matrix_base,
            pad_width=((0, 0), (self.n_full_yield_entries - self.n_bkg_components, 0)),
            mode="constant",
            constant_values=0.0,
        )

        _m_shape: Tuple[int, ...] = _bkg_const_matrix.shape

        assert _m_shape[0] == _n_bkg_const_params, (_m_shape, _n_bkg_const_params, self.n_full_yield_entries)
        assert _m_shape[1] == self.n_full_yield_entries, (_m_shape, _n_bkg_const_params, self.n_full_yield_entries)

        return _bkg_const_matrix

    constraint_effect_anti_vector: ClassVar[np.ndarray] = np.concatenate(
        (
            np.ones(shape=(n_signal_and_norm_components,)),
            np.zeros(shape=(n_bkg_components,)),
        ),
        axis=0,
    )

    @immutable_cached_property
    def bkg_constraint_params_ids(self) -> np.ndarray:
        return np.array(self.bkg_handler.bkg_constraint_nuisance_param_ids)

    # region Background constraints effects
    def bkg_yield_constraints_effect(self, params: np.ndarray) -> np.ndarray:
        return (
            1.0 + (params[self.bkg_constraint_params_ids] @ self.constraint_effect_matrix)
        ) + self.constraint_effect_anti_vector

    # endregion

    # region pau_zp factors
    def get_tau_zp_factors(self, params: np.ndarray) -> np.ndarray:
        return (
            self.additional_params_matrix
            * (
                1.0
                + rdstar_sys.MultSysTauZeroPlusRatio.relative_uncertainty
                * params[self.systematics.mult_sys__tau_zp_ratio.nuisance_param_id]
            )
            * (
                1.0
                + rdstar_sys.MultSysQEDModel.relative_uncertainty
                * params[self.systematics.mult_sys__qed_model.nuisance_param_id]
            )
            + self.anti_addit_params_matrix
        )

    # endregion

    # region Efficiency factors
    def _prepare_eff_information(
        self,
        eff_info_dict: Dict[int, float],
    ) -> np.ndarray:
        _comp_indices: Tuple[int, ...] = tuple(c.comp_index for c in self.fit_setup_info.components)
        assert list(_comp_indices) == sorted(_comp_indices), _comp_indices
        assert list(eff_info_dict.keys()) == list(_comp_indices), (list(eff_info_dict.keys()), _comp_indices)

        return np.array(
            [
                eff_info_dict[_comp_indices[0]],
                eff_info_dict[_comp_indices[1]],
                eff_info_dict[_comp_indices[2]],
                eff_info_dict[_comp_indices[3]],
                #
                eff_info_dict[_comp_indices[5]],
                eff_info_dict[_comp_indices[7]],
                #
                eff_info_dict[_comp_indices[4]],
                #
                eff_info_dict[_comp_indices[6]],
            ]
        )

    @immutable_cached_property
    def special_evt_count_value_dict(self) -> Dict[int, float]:
        _eff_info: List[Tuple[int, float]] = []

        _reco_ids: Tuple[int, ...] = tuple([i for ch in self.fit_setup_info.reco_channels for i in ch.reco_channel_ids])
        for c_id, comp in enumerate(self.fit_setup_info.components):
            if c_id >= self.n_signal_and_norm_components:
                break

            evt_count: float = self.global_eff_info.eff_numbers_container.get_selected_events(
                sig_ids=comp.sig_ids,
                decay_mode_ids=_reco_ids,
            )

            _eff_info.append((comp.comp_index, evt_count))

        return dict(_eff_info)

    @immutable_cached_property
    def special_evt_count_error_dict(self) -> Dict[int, float]:
        _eff_error_info: List[Tuple[int, float]] = []

        _reco_ids: Tuple[int, ...] = tuple([i for ch in self.fit_setup_info.reco_channels for i in ch.reco_channel_ids])
        for c_id, comp in enumerate(self.fit_setup_info.components):
            if c_id >= self.n_signal_and_norm_components:
                break

            sum_of_squared_weights: float = (
                self.global_eff_info.eff_numbers_container.get_sum_of_squared_weights_for_selected_events(
                    sig_ids=comp.sig_ids,
                    decay_mode_ids=_reco_ids,
                )
            )

            _eff_error_info.append((comp.comp_index, np.sqrt(sum_of_squared_weights)))

        return dict(_eff_error_info)

    @immutable_cached_property
    def eff_base_values(self) -> np.ndarray:
        _initial_effs: Dict[int, float] = dict(self._local_efficiency_infos.get_initial_sum_of_weights_per_component())

        if self._use_special_eff_calculation:
            assert all(k in _initial_effs for k in self.special_evt_count_value_dict)
            _initial_effs.update(self.special_evt_count_value_dict)

        result: np.ndarray = self._prepare_eff_information(eff_info_dict=_initial_effs)
        assert len(result) == self.systematics.mult_sys__mc_statistics.number_of_nuisances, (
            len(result),
            self.systematics.mult_sys__mc_statistics.number_of_nuisances,
        )
        return result

    @immutable_cached_property
    def eff_base_errors(self) -> np.ndarray:
        _eff_errors: Dict[int, float] = dict(self._local_efficiency_infos.get_error_on_sum_of_weights_per_component())

        if self._use_special_eff_calculation:
            assert all(k in _eff_errors for k in self.special_evt_count_error_dict)
            _eff_errors.update(self.special_evt_count_error_dict)

        return self._prepare_eff_information(eff_info_dict=_eff_errors)

    @immutable_cached_property
    def eff_mult_sys_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.mult_sys__mc_statistics.nuisance_param_ids)

    @immutable_cached_property
    def relative_eff_base_errors(self) -> np.ndarray:
        return self.eff_base_errors / self.eff_base_values

    def get_efficiency_factors(self, params: np.ndarray) -> np.ndarray:
        # Here the multiplicative MC stats error is considered:
        eff_params: np.ndarray = params[self.eff_mult_sys_nuisance_param_ids]
        eff_vec: np.ndarray = self.eff_base_values * (1.0 + self.relative_eff_base_errors * eff_params)

        res = np.multiply(self.eff_p1_map_matrix, eff_vec) + np.multiply(self.eff_p2_map_matrix, 1.0 / eff_vec)
        eff_factors: np.ndarray = np.pad(
            np.prod(res, axis=1, where=self.eff_where, initial=1.0),
            pad_width=(0, self.efficiency_factors_padding_width),
            mode="constant",
            constant_values=1.0,
        )
        return eff_factors

    # endregion

    # region Multiplicative uncertainty factors
    @immutable_cached_property
    def fei_corr_sys_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.mult_sys__fei_eff.nuisance_param_ids)

    @immutable_cached_property
    def fei_corr_sys_relative_uncertainties(self) -> np.ndarray:
        _ordered_uncertainties: Tuple[float, ...] = (
            rdstar_sys.MultSysFEIEff.fei_corr_relative_uncert_neutral_b,
            rdstar_sys.MultSysFEIEff.fei_corr_relative_uncert_charged_b,
        )
        return np.array(_ordered_uncertainties)

    # FEI Correction relations based on full yield definition given the current template setup:
    # B+ -> D0 tau+ nu          => None
    # B0 -> D- tau+ nu          => neutral B fei corr / charged B fei corr
    # B+ -> D*0 tau+ nu         => None
    # B0 -> D*- tau+ nu         => neutral B fei corr / charged B fei corr
    # B+ -> D0 l+ nu            => None
    # B0 -> D- l+ nu            => neutral B fei corr / charged B fei corr
    # B+ -> D*0 l+ nu           => None
    # B0 -> D*- l+ nu           => neutral B fei corr / charged B fei corr
    # D** BKG in B+ reco        => charged B fei corr
    # D** BKG in B0 reco        => neutral B fei corr
    # BBbar BKG in B+ reco      => charged B fei corr
    # BBbar BKG in B0 reco      => neutral B fei corr
    # Continuum BKG in B+ reco  => FEI uncertainties are handled via Continuum constraint!
    # Continuum BKG in B0 reco  => FEI uncertainties are handled via Continuum constraint!

    fei_factor_matrix: ClassVar[np.ndarray] = np.array(
        [
            # Sig      | Norm      | Bkgs
            [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0],  # neutral B factors
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0],  # charged B factors
        ]
    )
    fei_inverse_factor_matrix: ClassVar[np.ndarray] = np.array(
        [
            # Sig      | Norm      | Bkgs
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # neutral B factors
            [0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0],  # charged B factors
        ]
    )
    anti_fei_factor_w_bgks_matrix: ClassVar[np.ndarray] = np.array([1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1])
    anti_fei_factor_wo_bkg_matrix: ClassVar[np.ndarray] = np.array([1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1])

    def get_multiplicative_nuisance_effect(self, params: np.ndarray) -> np.ndarray:
        # TODO: Maybe apply these effects already in get_efficiency_factors to eff_vec instead of in get_yields?!
        #  and add part that affects backgrounds (or maybe more) in get_yields via a different method and to get_yields?

        fei_params: np.ndarray = params[self.fei_corr_sys_param_ids]

        base_factors: np.ndarray = 1.0 + self.fei_corr_sys_relative_uncertainties * fei_params
        fei_nuisance_effect: np.ndarray = (
            (base_factors @ self.fei_factor_matrix) + self.anti_fei_factor_w_bgks_matrix
        ) * ((1.0 / base_factors) @ self.fei_inverse_factor_matrix + self.anti_fei_factor_wo_bkg_matrix)

        return fei_nuisance_effect

    # endregion

    # region Additive uncertainty factors

    def _calculated_rel_shape_sys_uncert_matrix_from_vars_for(
        self,
        shape_sys_info: ReducedSystematicsDetails,
        shape_sys_name: str,
        param_ids: np.ndarray,
    ) -> np.ndarray:
        rearranged_eigenvecs: np.ndarray = shape_sys_info.get_rearranged_eigenvectors(fit_setup=self.fit_setup_info)

        eigen_vals_times_vecs = rearranged_eigenvecs * np.sqrt(shape_sys_info.reduced_eigenvalues)

        pure_counts: np.ndarray = self.pure_mc_templates

        if self.fit_setup_info.remove_empty_bins:
            try:
                eigen_vals_times_vecs = np.delete(eigen_vals_times_vecs, self.empty_bin_mask, axis=0)
            except ValueError as ve:
                logger.info(
                    f"Current shapes for {shape_sys_name}:"
                    f"\n\teigen_vals_times_vecs: {eigen_vals_times_vecs.shape}"
                    f"\n\tempty_bin_mask: {self.empty_bin_mask.shape}"
                    f"\n\tpure_counts: {pure_counts.shape}"
                )
                raise ve

        assert len(pure_counts.shape) == 2, pure_counts.shape
        assert len(eigen_vals_times_vecs.shape) == 3, eigen_vals_times_vecs.shape

        assert eigen_vals_times_vecs.shape[0] == pure_counts.shape[0], (eigen_vals_times_vecs.shape, pure_counts.shape)
        assert eigen_vals_times_vecs.shape[1] == pure_counts.shape[1], (eigen_vals_times_vecs.shape, pure_counts.shape)

        _normalization: np.ndarray = copy.deepcopy(self.pure_mc_templates)
        _normalization[_normalization == 0.0] = 1.0

        normalized_sys_vector: np.ndarray = eigen_vals_times_vecs / _normalization[:, :, np.newaxis]
        normalized_sys_vector[self.pure_mc_templates <= 0.0] = 0.0

        if not np.all(np.abs(normalized_sys_vector) < 1.0):
            _cond: np.ndarray = np.abs(normalized_sys_vector) >= 1.0
            _param_cond: np.ndarray = np.any(np.abs(normalized_sys_vector) < 1.0, axis=(0, 1))
            _new_val: np.ndarray = 0.25 * np.sign(eigen_vals_times_vecs)

            _norm_for_comp: np.ndarray = (np.ones_like(eigen_vals_times_vecs) * _normalization[:, :, np.newaxis])[_cond]
            assert _param_cond.shape == param_ids.shape, (_param_cond.shape, param_ids.shape)

            logger.warning(
                f"Resetting shape sys values in {np.sum(_cond)} cases "
                f"for bins with max(n_events in bin) = {np.max(_norm_for_comp)} "
                f"for {shape_sys_name}!"
                f"\n\t the affected nuisance parameters are {tuple(param_ids[_param_cond])}"
                f"\n\t top ten largest affected bin counts {sorted(list(_norm_for_comp), reverse=True)[:10]}"
            )

            assert np.all(np.abs(eigen_vals_times_vecs[_cond]) > _norm_for_comp), (
                shape_sys_name,
                f"\n\t{eigen_vals_times_vecs[_cond]}\n\t{_norm_for_comp}\n\t{_new_val[_cond]}",
            )

            normalized_sys_vector[_cond] = _new_val[_cond]

        assert np.all(np.abs(normalized_sys_vector) < 1.0), (
            shape_sys_name,
            np.sum(np.abs(normalized_sys_vector) > 1.0),
            normalized_sys_vector[np.abs(normalized_sys_vector) > 1.0].flatten(),
            eigen_vals_times_vecs[np.abs(normalized_sys_vector) > 1.0].flatten(),
            _normalization[np.any(np.abs(normalized_sys_vector) > 1.0, axis=2)].flatten(),
        )

        assert eigen_vals_times_vecs.shape[2] == shape_sys_info.number_of_retained_vals, (
            shape_sys_name,
            eigen_vals_times_vecs.shape,
            shape_sys_info.number_of_retained_vals,
        )

        return normalized_sys_vector

    def _calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
        self,
        shape_sys_info: UpDownAndDiffSystematicsDetails,
        shape_sys_name: str,
        param_ids: np.ndarray,
    ) -> np.ndarray:
        rearranged_rel_shape_diff: np.ndarray = shape_sys_info.get_rearranged_relative_shape_effect(
            fit_setup=self.fit_setup_info,
        )

        if self.fit_setup_info.remove_empty_bins:
            try:
                rearranged_rel_shape_diff = np.delete(rearranged_rel_shape_diff, self.empty_bin_mask, axis=0)
            except ValueError as ve:
                logger.info(
                    f"Current shapes for {shape_sys_name}:"
                    f"\n\trearranged_rel_shape_diff: {rearranged_rel_shape_diff.shape}"
                    f"\n\tempty_bin_mask: {self.empty_bin_mask.shape}"
                )
                raise ve

        assert len(rearranged_rel_shape_diff.shape) == 3, rearranged_rel_shape_diff.shape
        _pc: np.ndarray = self.pure_mc_templates

        assert rearranged_rel_shape_diff.shape[0] == _pc.shape[0], (rearranged_rel_shape_diff.shape, _pc.shape)
        assert rearranged_rel_shape_diff.shape[1] == _pc.shape[1], (rearranged_rel_shape_diff.shape, _pc.shape)

        if not np.all(np.abs(rearranged_rel_shape_diff) < 1.0):
            _cond: np.ndarray = np.abs(rearranged_rel_shape_diff) >= 1.0
            _param_cond: np.ndarray = np.any(np.abs(rearranged_rel_shape_diff) < 1.0, axis=(0, 1))
            _new_val: np.ndarray = np.zeros_like(rearranged_rel_shape_diff)

            assert _param_cond.shape == param_ids.shape, (_param_cond.shape, param_ids.shape)

            logger.warning(
                f"Resetting shape sys values in {np.sum(_cond)} cases "
                f"for {shape_sys_name}!"
                f"\n\t the affected nuisance parameters are {tuple(param_ids[_param_cond])}"
            )

            rearranged_rel_shape_diff[_cond] = _new_val[_cond]

        assert np.all(np.abs(rearranged_rel_shape_diff) < 1.0), (
            shape_sys_name,
            np.sum(np.abs(rearranged_rel_shape_diff) > 1.0),
            rearranged_rel_shape_diff[np.abs(rearranged_rel_shape_diff) > 1.0].flatten(),
        )

        assert rearranged_rel_shape_diff.shape[2] == shape_sys_info.number_of_variations, (
            shape_sys_name,
            rearranged_rel_shape_diff.shape,
            shape_sys_info.number_of_variations,
        )

        return rearranged_rel_shape_diff

    @immutable_cached_property
    def lepton_id_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        lid_shape_sys_info: ReducedSystematicsDetails = rdstar_sys.AddSysPIDLepton.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_vars_for(
            shape_sys_info=lid_shape_sys_info,
            shape_sys_name="lid_shape_sys",
            param_ids=self.lid_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def lepton_fake_id_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        fake_lid_shape_sys_info: ReducedSystematicsDetails = rdstar_sys.AddSysPIDLeptonFake.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_vars_for(
            shape_sys_info=fake_lid_shape_sys_info,
            shape_sys_name="fake_lid_shape_sys",
            param_ids=self.fake_lid_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def hadron_id_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        hid_shape_sys_infos: Tuple[ReducedSystematicsDetails, ...] = rdstar_sys.AddSysPIDHadron.get_shape_sys_info()

        _param_ranges: List[slice] = []
        _start_param_id: int = 0
        for hid_sys_info in hid_shape_sys_infos:
            _next_param_id: int = _start_param_id + hid_sys_info.number_of_retained_vals
            _param_ranges.append(slice(_start_param_id, _next_param_id, 1))
            _start_param_id = _next_param_id

        hid_ucert_matrices: Tuple[np.ndarray, ...] = tuple(
            self._calculated_rel_shape_sys_uncert_matrix_from_vars_for(
                shape_sys_info=_hid_shape_sys_info,
                shape_sys_name=f"hadron_id_shape_sys_{i}",
                param_ids=self.h_id_shape_nuisance_param_ids[_param_ranges[i]],
            )
            for i, _hid_shape_sys_info in enumerate(hid_shape_sys_infos)
        )

        assert all(len(hid_ucert_matrices[0].shape) == len(m.shape) for m in hid_ucert_matrices), [
            m.shape for m in hid_ucert_matrices
        ]
        assert all(hid_ucert_matrices[0].shape[0] == m.shape[0] for m in hid_ucert_matrices), [
            m.shape for m in hid_ucert_matrices
        ]
        assert all(hid_ucert_matrices[0].shape[1] == m.shape[1] for m in hid_ucert_matrices), [
            m.shape for m in hid_ucert_matrices
        ]

        return np.concatenate(hid_ucert_matrices, axis=2)

    @immutable_cached_property
    def slow_pi_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        slow_pi_shape_sys_info: ReducedSystematicsDetails = rdstar_sys.AddSysSlowPion.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_vars_for(
            shape_sys_info=slow_pi_shape_sys_info,
            shape_sys_name="slow_pi_shape_sys",
            param_ids=self.slow_pi_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def k_short_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        k_short_shape_sys_info: ReducedSystematicsDetails = rdstar_sys.AddSysKShortEff.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_vars_for(
            shape_sys_info=k_short_shape_sys_info,
            shape_sys_name="k_short_shape_sys",
            param_ids=self.k_short_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def ff_sig_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        ff_sig_shape_sys_info: UpDownAndDiffSystematicsDetails = rdstar_sys.AddSysFFSignal.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
            shape_sys_info=ff_sig_shape_sys_info,
            shape_sys_name="ff_sig_shape_sys",
            param_ids=self.ff_sig_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def ff_norm_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        ff_norm_shape_sys_info: UpDownAndDiffSystematicsDetails = rdstar_sys.AddSysFFNorm.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
            shape_sys_info=ff_norm_shape_sys_info,
            shape_sys_name="ff_norm_shape_sys",
            param_ids=self.ff_norm_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def wide_dss_ff_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        w_dss_ff_ssis: Tuple[UpDownAndDiffSystematicsDetails, ...] = rdstar_sys.AddSysDSSFFWide.get_shape_sys_info()
        assert len(w_dss_ff_ssis) == 2, len(w_dss_ff_ssis)

        w_dss_ff_ucert_matrices: Tuple[np.ndarray, ...] = tuple(
            self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
                shape_sys_info=_w_dss_ff_shape_sys_info,
                shape_sys_name=f"w_dss_ff_shape_sys_{i}",
                param_ids=self.wide_dss_ff_shape_nuisance_param_ids,
            )
            for i, _w_dss_ff_shape_sys_info in enumerate(w_dss_ff_ssis)
        )

        assert all(m.shape == w_dss_ff_ucert_matrices[0].shape for m in w_dss_ff_ucert_matrices), [
            m.shape for m in w_dss_ff_ucert_matrices
        ]
        assert len(w_dss_ff_ucert_matrices[0].shape) == 3, w_dss_ff_ucert_matrices[0].shape
        assert len(w_dss_ff_ucert_matrices) == 2, len(w_dss_ff_ucert_matrices)

        return np.add(w_dss_ff_ucert_matrices[0], w_dss_ff_ucert_matrices[1])

    @immutable_cached_property
    def narrow_dss_ff_rel_shape_sys_uncert_matrix(self) -> np.ndarray:
        n_dss_ff_ssis: Tuple[UpDownAndDiffSystematicsDetails, ...] = rdstar_sys.AddSysDSSFFNarrow.get_shape_sys_info()
        assert len(n_dss_ff_ssis) == 2, len(n_dss_ff_ssis)

        n_dss_ff_ucert_matrices: Tuple[np.ndarray, ...] = tuple(
            self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
                shape_sys_info=_n_dss_ff_shape_sys_info,
                shape_sys_name=f"n_dss_ff_shape_sys_{i}",
                param_ids=self.narrow_dss_ff_shape_nuisance_param_ids,
            )
            for i, _n_dss_ff_shape_sys_info in enumerate(n_dss_ff_ssis)
        )

        assert all(m.shape == n_dss_ff_ucert_matrices[0].shape for m in n_dss_ff_ucert_matrices), [
            m.shape for m in n_dss_ff_ucert_matrices
        ]
        assert len(n_dss_ff_ucert_matrices[0].shape) == 3, n_dss_ff_ucert_matrices[0].shape
        assert len(n_dss_ff_ucert_matrices) == 2, len(n_dss_ff_ucert_matrices)

        return np.add(n_dss_ff_ucert_matrices[0], n_dss_ff_ucert_matrices[1])

    @immutable_cached_property
    def qed_rel_shape_sys_effect(self) -> np.ndarray:
        qed_model_shape_sys_info: UpDownAndDiffSystematicsDetails = rdstar_sys.AddSysQEDModel.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
            shape_sys_info=qed_model_shape_sys_info,
            shape_sys_name="qed_model_shape_sys",
            param_ids=self.qed_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def tracking_rel_shape_sys_effect(self) -> np.ndarray:
        tracking_shape_sys_info: UpDownAndDiffSystematicsDetails = rdstar_sys.AddSysTracking.get_shape_sys_info()
        return self._calculated_rel_shape_sys_uncert_matrix_from_diffs_for(
            shape_sys_info=tracking_shape_sys_info,
            shape_sys_name="tracking_shape_sys",
            param_ids=self.tracking_shape_nuisance_param_ids,
        )

    @immutable_cached_property
    def lid_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__pid_lepton.nuisance_param_ids)

    @immutable_cached_property
    def fake_lid_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__pid_lepton_fake.nuisance_param_ids)

    @immutable_cached_property
    def h_id_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__pid_hadron.nuisance_param_ids)

    @immutable_cached_property
    def slow_pi_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__slow_pion.nuisance_param_ids)

    @immutable_cached_property
    def k_short_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__k_short_eff.nuisance_param_ids)

    @immutable_cached_property
    def ff_sig_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__ff_signal.nuisance_param_ids)

    @immutable_cached_property
    def ff_norm_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__ff_norm.nuisance_param_ids)

    @immutable_cached_property
    def wide_dss_ff_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__dss_wide_ff.nuisance_param_ids)

    @immutable_cached_property
    def narrow_dss_ff_shape_nuisance_param_ids(self) -> np.ndarray:
        return np.array(self.systematics.add_sys__dss_narrow_ff.nuisance_param_ids)

    @immutable_cached_property
    def qed_shape_nuisance_param_ids(self) -> np.ndarray:
        _param_ids: Tuple[int, ...] = self.systematics.add_sys__qed_model.nuisance_param_ids
        assert len(_param_ids) == 1, (len(_param_ids), _param_ids)
        return np.array(_param_ids)

    @immutable_cached_property
    def tracking_shape_nuisance_param_ids(self) -> np.ndarray:
        _param_ids: Tuple[int, ...] = self.systematics.add_sys__tracking.nuisance_param_ids
        assert len(_param_ids) == 1, (len(_param_ids), _param_ids)
        return np.array(_param_ids)

    def get_additive_nuisance_effect(self, params: np.ndarray) -> np.ndarray:
        # TODO: Add effect of systematics via eigendirections here!
        # TODO: Ignore systematic, if they are not active!!!

        # Reco Sys
        l_id_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.lepton_id_rel_shape_sys_uncert_matrix * params[self.lid_shape_nuisance_param_ids],
            axis=2,
        )
        l_fake_id_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.lepton_fake_id_rel_shape_sys_uncert_matrix * params[self.fake_lid_shape_nuisance_param_ids],
            axis=2,
        )
        hadron_id_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.hadron_id_rel_shape_sys_uncert_matrix * params[self.h_id_shape_nuisance_param_ids],
            axis=2,
        )
        slow_pi_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.slow_pi_rel_shape_sys_uncert_matrix * params[self.slow_pi_shape_nuisance_param_ids],
            axis=2,
        )
        k_short_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.k_short_rel_shape_sys_uncert_matrix * params[self.k_short_shape_nuisance_param_ids],
            axis=2,
        )

        # HAMMER FF Sys
        ff_sig_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.ff_sig_rel_shape_sys_uncert_matrix * params[self.ff_sig_shape_nuisance_param_ids],
            axis=2,
        )
        ff_norm_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.ff_norm_rel_shape_sys_uncert_matrix * params[self.ff_norm_shape_nuisance_param_ids],
            axis=2,
        )

        # D** FF Sys
        wide_dss_ff_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.wide_dss_ff_rel_shape_sys_uncert_matrix * params[self.wide_dss_ff_shape_nuisance_param_ids],
            axis=2,
        )
        narrow_dss_ff_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.narrow_dss_ff_rel_shape_sys_uncert_matrix * params[self.narrow_dss_ff_shape_nuisance_param_ids],
            axis=2,
        )

        # Other Sys
        qed_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.qed_rel_shape_sys_effect * params[self.qed_shape_nuisance_param_ids],
            axis=2,
        )
        tracking_sys_shape_effect: np.ndarray = np.prod(
            1.0 + self.tracking_rel_shape_sys_effect * params[self.tracking_shape_nuisance_param_ids],
            axis=2,
        )

        if self._check_additive_effects_shape:
            sys_effects: Tuple[Tuple[str, int, np.ndarray, Any, Any], ...] = (
                (
                    "l_id",
                    int(np.sum(l_id_sys_shape_effect < 0)),
                    l_id_sys_shape_effect[l_id_sys_shape_effect < 0].flatten(),
                    tuple(self.lid_shape_nuisance_param_ids),
                    params[self.lid_shape_nuisance_param_ids],
                ),
                (
                    "l_fake_id",
                    int(np.sum(l_fake_id_sys_shape_effect < 0)),
                    l_fake_id_sys_shape_effect[l_fake_id_sys_shape_effect < 0].flatten(),
                    tuple(self.fake_lid_shape_nuisance_param_ids),
                    params[self.fake_lid_shape_nuisance_param_ids],
                ),
                (
                    "hadron_id",
                    int(np.sum(hadron_id_sys_shape_effect < 0)),
                    hadron_id_sys_shape_effect[hadron_id_sys_shape_effect < 0].flatten(),
                    tuple(self.h_id_shape_nuisance_param_ids),
                    params[self.h_id_shape_nuisance_param_ids],
                ),
                (
                    "slow_pi",
                    int(np.sum(slow_pi_sys_shape_effect < 0)),
                    slow_pi_sys_shape_effect[slow_pi_sys_shape_effect < 0].flatten(),
                    tuple(self.slow_pi_shape_nuisance_param_ids),
                    params[self.slow_pi_shape_nuisance_param_ids],
                ),
                (
                    "k_short",
                    int(np.sum(k_short_sys_shape_effect < 0)),
                    k_short_sys_shape_effect[k_short_sys_shape_effect < 0].flatten(),
                    tuple(self.k_short_shape_nuisance_param_ids),
                    params[self.k_short_shape_nuisance_param_ids],
                ),
                (
                    "qed_sys_shape",
                    int(np.sum(qed_sys_shape_effect < 0)),
                    qed_sys_shape_effect[qed_sys_shape_effect < 0].flatten(),
                    tuple(self.qed_shape_nuisance_param_ids),
                    params[self.qed_shape_nuisance_param_ids],
                ),
            )

            if any([v2 > 0 for _, v2, _, _, _ in sys_effects]):
                raise RuntimeError(
                    "Problem in sys effects:\n\t"
                    + "\n\t".join([f"{v1}: {v2}\n\t\t{v3}\n\t\t{v4}\n\t\t{v5}" for v1, v2, v3, v4, v5 in sys_effects])
                )

        sys_shape_effect: np.ndarray = (
            l_id_sys_shape_effect
            * l_fake_id_sys_shape_effect
            * hadron_id_sys_shape_effect
            * slow_pi_sys_shape_effect
            * k_short_sys_shape_effect
            * ff_sig_sys_shape_effect
            * ff_norm_sys_shape_effect
            * wide_dss_ff_sys_shape_effect
            * narrow_dss_ff_sys_shape_effect
            * qed_sys_shape_effect
            * tracking_sys_shape_effect
        )

        if self._check_additive_effects_shape:
            assert self.template_shapes.shape == sys_shape_effect.shape, (
                self.template_shapes.shape,
                sys_shape_effect.shape,
            )
            self._check_additive_effects_shape = False

        return sys_shape_effect

    # endregion

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        if self.systematics.add_sys__mc_statistics.is_active:
            _chi2 = np.sum(
                (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
                / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
                axis=None,
            )
        else:
            _chi2 = np.sum(
                (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
                / self.data_bin_stat_errors_squared,
                axis=None,
            )

        assert isinstance(_chi2, float)
        return _chi2

    # endregion

    # region Gaussian constraints
    @immutable_cached_property
    def constraint_nuisance_uncertainty_matrix(self) -> np.ndarray:
        # Unitary matrix for constraints on nuisance parameters from systematics and background yield constraints.
        # Constraints are handled as gaussian nuisances with std of 1.0 around mean of 0.0:
        _n_nuisances: int = self.systematics.number_of_nuisance_parameters + self.bkg_handler.number_of_bkg_constraints
        return np.identity(_n_nuisances)

    @immutable_cached_property
    def full_nuisance_params_ids_info(self) -> Union[np.ndarray, slice]:
        nuisance_params_ids: Tuple[int, ...] = self.systematics.all_nuisance_parameter_ids
        constraint_params_ids: Tuple[int, ...] = self.bkg_handler.bkg_constraint_nuisance_param_ids

        _param_ids: Tuple[int, ...] = nuisance_params_ids + constraint_params_ids

        assert len(_param_ids) == len(set(_param_ids)), [i for i in _param_ids if _param_ids.count(i) > 1]

        if tuple(np.arange(len(_param_ids)) + _param_ids[0]) == _param_ids:
            return np.s_[_param_ids[0] : _param_ids[-1] + 1 : 1]
        else:
            return np.array(_param_ids)

    def gaussian_constrains(self, params: np.ndarray) -> float:
        """
        Nuisances and other constraint params are handled as gaussian nuisances with std of 1.0 around mean of 0.0.

        This is done with nuisances considered as (1 + relative_uncertainty * theta), with
        theta being a standard-normal distributed nuisance varying with standard deviation of 1.0 around a mean of 0.0.
        This is implemented as:
            nuisance_vector @ inverse_correlation_matrix @ nuisance_vector
        where inverse_correlation_matrix is a unity matrix with shape (n_nuisances, n_nuisances),
        while nuisance_vector is a vector of length n_nuisances
        that contains all nuisances for systematics and constraints.

        An alternative approach would be for instance:
            np.sum(((constraint_values - constraint_pars) / self.constraint_sigmas) ** 2)
        if non standard-normal distributed nuisances are required.
        """
        nu_params: np.ndarray = params[self.full_nuisance_params_ids_info]

        return nu_params @ self.constraint_nuisance_uncertainty_matrix @ nu_params

    # endregion

    # region Normalized template shapes

    @immutable_cached_property
    def reco_ch_bin_slices(self) -> Tuple[slice, ...]:
        bin_ranges: Dict[int, Tuple[int, int]] = self.get_bin_indices_range_per_reco_channel()
        return tuple(np.s_[bin_range[0] : bin_range[0] : 1] for bin_range in bin_ranges.values())

    def normalized_template_shapes(self, params: np.ndarray) -> np.ndarray:
        if not self.systematics.has_additive_systematics:
            return self.template_shapes
        else:
            _new_shapes: np.ndarray = self.template_shapes * self.get_additive_nuisance_effect(params=params)

            for reco_ch_bin_slice in self.reco_ch_bin_slices:
                ch_norm_denominator: np.ndarray = _new_shapes[reco_ch_bin_slice].sum(axis=0)
                ch_norm_denominator[ch_norm_denominator == 0] = 1.0
                _new_shapes /= ch_norm_denominator[np.newaxis, :]

            return _new_shapes

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            (
                self.get_yields(params=params)
                * self.local_efficiencies  # TODO: Make local efficiencies floating?
                * self.normalized_template_shapes(params=params)
            ),
            axis=1,
        )

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = self.systematics.number_of_nuisance_parameters
        return self.n_floating_yield_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        y_dz_l_nu: float = _normalization_factors[4]
        y_dzstar_l_nu: float = _normalization_factors[6]
        y_cb_dss_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 0]
        y_nb_dss_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 1]
        y_cb_bbbar_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 2]
        y_nb_bbbar_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 3]
        y_cb_continuum_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 4]
        y_nb_continuum_bkg: float = _normalization_factors[self.n_signal_and_norm_components + 5]

        _initial_yield_params: np.ndarray = np.array(
            [
                self.used_rd_value,
                self.used_rd_star_value,
                y_dz_l_nu,
                y_dzstar_l_nu,
                y_cb_dss_bkg,
                y_nb_dss_bkg,
                y_cb_bbbar_bkg,
                y_nb_bbbar_bkg,
                y_cb_continuum_bkg,
                y_nb_continuum_bkg,
            ]
        )

        assert len(_initial_yield_params) == self.n_floating_yield_params, (
            len(_initial_yield_params),
            self.n_floating_yield_params,
            _initial_yield_params,
        )

        _initial_nuisance_params: np.ndarray = np.zeros(self.systematics.number_of_nuisance_parameters)
        assert len(_initial_nuisance_params) == self.systematics.number_of_nuisance_parameters, (
            len(_initial_nuisance_params),
            self.systematics.number_of_nuisance_parameters,
        )

        _initial_constraint_params: np.ndarray = np.zeros(self.bkg_handler.number_of_bkg_constraints)
        assert len(_initial_constraint_params) == self.bkg_handler.number_of_bkg_constraints, (
            len(_initial_constraint_params),
            self.systematics.number_of_nuisance_parameters,
        )

        return np.concatenate((_initial_yield_params, _initial_nuisance_params, _initial_constraint_params), axis=0)

    # endregion

    # region MC df rescaling to old Belle R(D) and R(D*) Values
    @staticmethod
    def get_mc_df_reweighted_to_old_belle_values(
        df: pd.DataFrame,
        fit_weight_col_name: str,
    ) -> pd.DataFrame:
        """
        BF(B -> D tau nu) = R(D) * BF(B -> D l nu)
        BF'(B -> D tau nu) = BF(B -> D tau nu) * R'(D) / R(D)
        weight = R'(D) / R(D)
        """
        assert fit_weight_col_name in df.columns, fit_weight_col_name
        assert gen_b1_id_col_name in df.columns, gen_b1_id_col_name
        assert gen_b2_id_col_name in df.columns, gen_b2_id_col_name

        d_factor: float = RDStarFitter.huschle_RD / RDStarFitter.mcRD
        dst_factor: float = RDStarFitter.huschle_RDs / RDStarFitter.mcRDs

        is_b1_to_d_tau: np.ndarray = df[gen_b1_id_col_name] == GenMCIDs.Dtau
        is_b2_to_d_tau: np.ndarray = df[gen_b2_id_col_name] == GenMCIDs.Dtau
        is_b1_to_dst_tau: np.ndarray = df[gen_b1_id_col_name] == GenMCIDs.Dsttau
        is_b2_to_dst_tau: np.ndarray = df[gen_b2_id_col_name] == GenMCIDs.Dsttau

        weight_vals: np.ndarray = copy.deepcopy(df[fit_weight_col_name].values)

        weight_vals[is_b1_to_d_tau] *= d_factor
        weight_vals[is_b2_to_d_tau] *= d_factor
        weight_vals[is_b1_to_dst_tau] *= dst_factor
        weight_vals[is_b2_to_dst_tau] *= dst_factor

        _df = df.copy(deep=True)
        assert not all(_df[fit_weight_col_name].values == weight_vals)
        _df.loc[:, fit_weight_col_name] = weight_vals
        return _df

    # endregion
