"""
Systematics for R(D) + R(D*) fit.

Felix Metzner 2023
"""

import os
import numpy as np

from abc import ABC
from enum import Enum
from uncertainties import ufloat, UFloat
from typing import ClassVar, Tuple, List, Optional

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses
from rdstar.branching_fractions import constants_and_measurements as bf_calc_consts

from rdstar.offline_analysis.corrections.qed_correction import new_coulomb_correction_col_name

from rdstar.offline_analysis.fitting.dedicated_fit_approach.systematics_manager.shape_sys_evaluater import (
    sys_info_cache_dir_path,
    ReducedSystematicsDetails,
    UpDownAndDiffSystematicsDetails,
)


# TODO: Add systematics for difference between nominal BLPRXP and old BGL FF weights for normalization

# TODO: Consider pi_zero Reco Efficiency Correction as Systematic
#       -> https://kds.kek.jp/event/6029/contributions/138400/attachments/107673/127649/MC20101115-BGM.pdf


__all__ = [
    "BKGHandling",
    #
    "AbstractSysInfo",
    #
    "AbstractMultSysInfo",
    "AbstractAddSysInfo",
    "AbstractBKGHandlingInfo",
    # Background Handling
    "BKGHandlingDss",
    "BKGHandlingBBBar",
    "BKGHandlingContinuum",
    #
    # Multiplicative Systematics:
    "MultSysMCStatistics",
    # "MultSysBFTauToLNuNu",  # TODO
    # "MultSysDwwosMesonBF",  # TODO: D(*) Meson Branching Fractions
    # "MultSysFFSignal",  # TODO
    # "MultSysFFNorm",  # TODO
    "MultSysQEDModel",
    "MultSysTauZeroPlusRatio",
    # "MultSysPIDLepton",  # TODO
    # "MultSysPIDHadron",  # TODO
    # "MultSysSlowPion",  # TODO
    # "MultSysTracking",  # TODO
    "MultSysFEIEff",
    # "MultSysFEISlowPion",  # TODO: Might not be needed if closure in FEI Cal
    # "MultSysKShortEff",  # TODO: K_Short Reco Efficiency Correction
    # "MultSysPiZeroEff",  # TODO: pi0 Reco Efficiency Correction
    #
    # Additive Systematics:
    "AddSysMCStatistics",
    # "AddSysDwwosMesonBF",  # TODO: D(*) Meson Branching Fractions
    # "AddSysBFBtoDwwos",  # TODO: BF B -> D(*) (Added by Florian to white board)
    "AddSysFFSignal",
    "AddSysFFNorm",
    "AddSysQEDModel",
    # "AddSysDSSBF",  # TODO
    "AddSysDSSFFWide",
    "AddSysDSSFFNarrow",
    # "AddSysGapBKGModelling",  # TODO: Resonant vs. non-resonant gap description
    # "AddSysGapBKGBFComposition",  # TODO: Gap BFs
    # "AddSysBBbarBKGBFComposition",  # TODO: BBbar BKG BFs
    # "AddSysContinuumBKGComposition",  # TODO
    "AddSysPIDLepton",
    "AddSysPIDLeptonFake",
    "AddSysPIDHadron",
    "AddSysSlowPion",
    "AddSysTracking",
    # "AddSysFEISlowPion",  # TODO: Might not be needed if closure in FEI Cal
    "AddSysKShortEff",
    # "AddSysPiZeroEff",  # TODO: pi0 Reco Efficiency Correction
]


logger = log_to_default(name=__name__)


class BKGHandling(Enum):
    BKG_FIXED: ClassVar[str] = "BkgFix"
    BKG_FLOATING: ClassVar[str] = "BkgFlo"
    BKG_CONSTRAINT: ClassVar[str] = "BkgCon"

    @staticmethod
    def get_valid_members_info() -> List[str]:
        return [f"{m.name}: {m.value}" for m in BKGHandling]

    @staticmethod
    def get_valid_names() -> List[str]:
        return [m.name for m in BKGHandling]

    @staticmethod
    def get_valid_values() -> List[str]:
        return [m.value for m in BKGHandling]

    @staticmethod
    def bkg_handling_to_int_code(item: "BKGHandling") -> int:
        if item == BKGHandling.BKG_FIXED:
            return 0
        elif item == BKGHandling.BKG_FLOATING:
            return 1
        elif item == BKGHandling.BKG_CONSTRAINT:
            return 2
        else:
            raise ValueError(f"Unknown BKGHandling {item.name}: {item.value}")


class AbstractSysInfo(ABC):
    _systematics_key: ClassVar[str] = ""
    _latex_str: ClassVar[str] = ""
    _number_of_nuisances: ClassVar[int] = -1

    _sys_type: ClassVar[str] = ""
    _is_standard_normal: ClassVar[bool] = True

    _always_fixed_parameter_indices: ClassVar[Optional[Tuple[int, ...]]] = None

    @classmethod
    def get_systematics_key(cls) -> str:
        assert cls._systematics_key, (cls._systematics_key, cls.__name__)
        return cls._systematics_key

    @classmethod
    def get_latex_str(cls) -> str:
        assert cls._latex_str, (cls._latex_str, cls.__name__)
        return cls._latex_str

    @classmethod
    def number_of_nuisances(cls) -> int:
        assert cls._number_of_nuisances >= 0, (cls._number_of_nuisances, cls.get_systematics_key(), cls.__name__)
        return cls._number_of_nuisances

    @classmethod
    def get_systematics_type(cls) -> str:
        assert cls._sys_type, (cls._sys_type, cls.__name__)
        return cls._sys_type

    @classmethod
    def get_is_standard_normal_nuisance(cls) -> bool:
        return cls._is_standard_normal

    @classmethod
    def get_always_fixed_parameter_indices(cls) -> Optional[Tuple[int, ...]]:
        return cls._always_fixed_parameter_indices


class AbstractMultSysInfo(AbstractSysInfo):
    _sys_type: ClassVar[str] = "multiplicative"


class AbstractAddSysInfo(AbstractSysInfo):
    _sys_type: ClassVar[str] = "additive"


class AbstractBKGHandlingInfo(AbstractSysInfo):
    _sys_type: ClassVar[str] = "bkg_handling"

    _bkg_id_char: ClassVar[str] = ""
    _bkg_yield_indices: ClassVar[Tuple[int, ...]] = tuple()

    @classmethod
    def get_bkg_yield_indices(cls) -> Tuple[int, ...]:
        assert cls._bkg_yield_indices, (cls._systematics_key, cls._bkg_yield_indices)
        return cls._bkg_yield_indices

    @classmethod
    def get_bkg_id_char(cls) -> str:
        assert cls._bkg_id_char, (cls._systematics_key, cls._bkg_id_char, cls._bkg_yield_indices)
        return cls._bkg_id_char


###############
# Constraints #
###############


class BKGHandlingDss(AbstractBKGHandlingInfo):
    _systematics_key: ClassVar[str] = "bkg_handling_dss_yields"
    _latex_str: ClassVar[str] = LatexParticleNames.d_st_str[:-1] + r"\, \mathrm{BKG}$"
    _number_of_nuisances: ClassVar[int] = 2

    _bkg_id_char: ClassVar[str] = "D"
    _bkg_yield_indices: ClassVar[Tuple[int, ...]] = (4, 5)

    # TODO: Which values to use here? Choose from options:
    #           1.) Use 4 instead of 2 Templates? -> would require different structure for yield calculation!
    #           2.) Calculated from templated reweighted with values for narrow and wide cases below?
    #           3.) Use as shape uncertainty, as well? In particular in combination with option 2.)

    corr_narrow_dss_to_pi0: ClassVar[UFloat] = ufloat(1.02, 0.32)  # TODO: Use?
    corr_narrow_dss_to_pip: ClassVar[UFloat] = ufloat(1.27, 0.05)  # TODO: Use?
    corr_wide_dss_to_pi0: ClassVar[UFloat] = ufloat(1.24, 0.44)  # TODO: Use?
    corr_wide_dss_to_pip: ClassVar[UFloat] = ufloat(0.77, 0.07)  # TODO: Use?

    corr_in_cb_reco_ch: ClassVar[UFloat] = ufloat(1.0, 0.0)  # TODO: Calculated from values above?
    corr_in_nb_reco_ch: ClassVar[UFloat] = ufloat(1.0, 0.0)  # TODO: Calculated from values above?

    relative_uncert_cb: ClassVar[float] = corr_in_cb_reco_ch.std_dev / corr_in_cb_reco_ch.nominal_value
    relative_uncert_nb: ClassVar[float] = corr_in_nb_reco_ch.std_dev / corr_in_nb_reco_ch.nominal_value


class BKGHandlingBBBar(AbstractBKGHandlingInfo):
    _systematics_key: ClassVar[str] = "bkg_handling_bb_bar_yields"
    _latex_str: ClassVar[str] = LatexParticleNames.b_b_bar_str[:-1] + r"\, \mathrm{BKG}$"
    _number_of_nuisances: ClassVar[int] = 2

    _bkg_id_char: ClassVar[str] = "B"
    _bkg_yield_indices: ClassVar[Tuple[int, ...]] = (6, 7)

    # TODO: Which values should we use here? -> Let this BKG float completely or only use FEI sys (MultSysFEIEff)?
    corr_in_cb_reco_ch: ClassVar[UFloat] = ufloat(1.0, 0.1)  # TODO: This is probably not final!
    corr_in_nb_reco_ch: ClassVar[UFloat] = ufloat(1.0, 0.1)  # TODO: This is probably not final!

    relative_uncert_cb: ClassVar[float] = corr_in_cb_reco_ch.std_dev / corr_in_cb_reco_ch.nominal_value
    relative_uncert_nb: ClassVar[float] = corr_in_nb_reco_ch.std_dev / corr_in_nb_reco_ch.nominal_value


class BKGHandlingContinuum(AbstractBKGHandlingInfo):
    _systematics_key: ClassVar[str] = "bkg_handling_continuum_yields"
    _latex_str: ClassVar[str] = r"Continuum\, \mathrm{BKG}$"
    _number_of_nuisances: ClassVar[int] = 2

    _bkg_id_char: ClassVar[str] = "C"
    _bkg_yield_indices: ClassVar[Tuple[int, ...]] = (8, 9)

    # TODO: This is the actual FEI Uncertainty on the continuum setup!
    corr_in_cb_reco_ch: ClassVar[UFloat] = ufloat(0.87, 0.05)
    corr_in_nb_reco_ch: ClassVar[UFloat] = ufloat(1.08, 0.04)

    relative_uncert_cb: ClassVar[float] = corr_in_cb_reco_ch.std_dev / corr_in_cb_reco_ch.nominal_value
    relative_uncert_nb: ClassVar[float] = corr_in_nb_reco_ch.std_dev / corr_in_nb_reco_ch.nominal_value


##############################
# Multiplicative Systematics #
##############################

# MC Statistics
class MultSysMCStatistics(AbstractMultSysInfo):
    """
    Uncertainty on efficiency values (=number of candidates in MC = sum of weights in MC for a given event category)
    N_MCEvents -> N_MCEvents * (1 + relative_uncertainty * nuisance)
        with relative_uncertainty = sqrt(sum_of_squared_wegihts) / sum_of_weights
        and a standard normal distributed nuisance
        per event category -> 8 nuisances, 2 of which are fixed!
    """

    _systematics_key: ClassVar[str] = "multiplicative_mc_stat_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \, MC\, Stat.}$"
    _number_of_nuisances: ClassVar[int] = 8

    _number_of_always_fixed_nuisances: ClassVar[int] = 2
    _always_fixed_parameter_indices: ClassVar[Optional[Tuple[int, ...]]] = (  # Start counting at 0!
        4,  # = N^{MC Reco}{D- lep+}        (5th efficiency nuisance parameter)
        6,  # = N^{MC Reco}{D*- lep+}       (7th efficiency nuisance parameter)
    )


# BFs
class MultSysBFTauToLNuNu(AbstractMultSysInfo):  # TODO: BF tau -> l nu nu
    _systematics_key: ClassVar[str] = "multiplicative_bf_tau_to_l_nu_nu_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \,}" + LatexParticleNames.tau_str[1:-1] + r"\, \mathrm{Decay\, BF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class MultSysDwwosMesonBF(AbstractMultSysInfo):  # TODO: D(*) Meson Branching Fractions
    _systematics_key: ClassVar[str] = "multiplicative_dwwos_meson_bf_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\,}" + LatexParticleNames.d_w_wo_st_str[1:-1] + r"\,\mathrm{Decay\,BF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# Form Factors


class MultSysFFSignal(AbstractMultSysInfo):  # TODO: Form Factors for Signal
    _systematics_key: ClassVar[str] = "multiplicative_ff_signal_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\, Signal\, FF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class MultSysFFNorm(AbstractMultSysInfo):  # TODO: Form Factors for Normalization
    _systematics_key: ClassVar[str] = "multiplicative_ff_norm_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\, Norm\, FF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# QED Modelling / Iso-spin Symmetry
class MultSysQEDModel(AbstractMultSysInfo):
    """
    This is the tau_0p -> tau_0p * (1 + alpha pi * theta_nuisance) replacement in the yield calculation
    that covers the uncertainty on the isospin-symmetry assumption via the factor alpha pi,
    which is basically the relative uncertainty to tau_0p in this case.
    """

    _systematics_key: ClassVar[str] = "multiplicative_qed_model_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\, QED\, Model}$"
    _number_of_nuisances: ClassVar[int] = 1

    __one_over_alpha: ClassVar[float] = 137.035999
    relative_uncertainty: ClassVar[float] = np.pi / __one_over_alpha  # -> already the rel. uncertainty in this case!


class MultSysTauZeroPlusRatio(AbstractMultSysInfo):
    """
    Uncertainty on lifetime ratio tau_zp = tau_B0 / tau_B+ = 1 / tau_pz
    """

    _systematics_key: ClassVar[str] = "multiplicative_tau_zp_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\,} \tau{}_{0+}\, \mathrm{Uncert.}$"
    _number_of_nuisances: ClassVar[int] = 1

    __tau_zp: ClassVar[UFloat] = 1.0 / bf_calc_consts.tau_p0
    tau_zp_nominal_value: ClassVar[UFloat] = __tau_zp.nominal_value
    relative_uncertainty: ClassVar[float] = __tau_zp.std_dev / __tau_zp.nominal_value


# Particle IDs
class MultSysPIDLepton(AbstractMultSysInfo):  # TODO: Lepton particle IDs
    _systematics_key: ClassVar[str] = "multiplicative_pid_lepton_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\,}" + LatexParticleNames.lepton_str[1:-1] + r"\, \mathrm{ID}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class MultSysPIDHadron(AbstractMultSysInfo):  # TODO: Hadron particle IDs
    _systematics_key: ClassVar[str] = "multiplicative_pid_hadron_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \, Hadron \, ID}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# Slow Pion Efficiency
class MultSysSlowPion(AbstractMultSysInfo):  # TODO: Slow Pion Efficiency
    _systematics_key: ClassVar[str] = "multiplicative_slow_pion_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \, Slow}\, " + LatexParticleNames.pi_str[1:-1] + r"\, \mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# Tracking Efficiency
class MultSysTracking(AbstractMultSysInfo):  # TODO: Tracking Efficiency
    _systematics_key: ClassVar[str] = "multiplicative_tracking_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \, Tracking \, Uncert.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# FEI Efficiency
class MultSysFEIEff(AbstractMultSysInfo):
    _systematics_key: ClassVar[str] = "multiplicative_fei_eff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult. \, FEI \, Corr.}$"
    _number_of_nuisances: ClassVar[int] = 2

    # This defines the order of the systematics parameters!
    neutral_b_fei_corr_sys_index: ClassVar[int] = 0
    charged_b_fei_corr_sys_index: ClassVar[int] = 1

    # TODO: Consider correlations between fit results!? In combinations with MultSysFEISlowPion!?
    # TODO: Correlation between factors based on correlations of nominal B decay BFs (if using isospin averaged bfs)

    fei_corr_nb: ClassVar[UFloat] = ufloat(0.755954546482212, 0.032800461365742176)
    fei_corr_cb: ClassVar[UFloat] = ufloat(0.8103050602849804, 0.03363438836346298)

    fei_corr_relative_uncert_neutral_b: ClassVar[float] = fei_corr_nb.std_dev / fei_corr_nb.nominal_value
    fei_corr_relative_uncert_charged_b: ClassVar[float] = fei_corr_cb.std_dev / fei_corr_cb.nominal_value


class MultSysFEISlowPion(AbstractMultSysInfo):  # TODO: FEI Slow Pion Inefficiency?
    _systematics_key: ClassVar[str] = "multiplicative_fei_slow_pion_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.\, FEI\, Slow}\," + LatexParticleNames.pi_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = 2

    # TODO: Consider correlations between fit results!? In combinations with MultSysFEIEff!?

    fei_cf_corr_nb: ClassVar[UFloat] = ufloat(0.6438078214016749, 0.041064884370931275)  # TODO: Use!
    fei_cf_corr_cb: ClassVar[UFloat] = ufloat(0.8092324319886378, 0.03392612316517229)  # TODO: Use!

    fei_cf_corr_relative_uncert_neutral_b: ClassVar[float] = fei_cf_corr_nb.std_dev / fei_cf_corr_nb.nominal_value
    fei_cf_corr_relative_uncert_charged_b: ClassVar[float] = fei_cf_corr_cb.std_dev / fei_cf_corr_cb.nominal_value


# Other Reco Efficiencies
class MultSysKShortEff(AbstractMultSysInfo):  # TODO: K_Short Reco Efficiency Correction
    _systematics_key: ClassVar[str] = "multiplicative_k_short_eff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.}\, " + LatexParticleNames.k_short_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class MultSysPiZeroEff(AbstractMultSysInfo):  # TODO: pi0 Reco Efficiency Correction
    _systematics_key: ClassVar[str] = "multiplicative_pi_zero_eff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{mult.}\, " + LatexParticleNames.pi_zero_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


########################
# Additive Systematics #
########################


class AddSysMCStatistics(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_mc_stat_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, MC\, Stat.}$"
    _number_of_nuisances: ClassVar[int] = 0
    _is_standard_normal: ClassVar[bool] = False


# BFs
class AddSysDwwosMesonBF(AbstractAddSysInfo):  # TODO: D(*) Meson Branching Fractions
    _systematics_key: ClassVar[str] = "additive_dwwos_meson_bf_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \,}" + LatexParticleNames.d_w_wo_st_str[1:-1] + r"\, \mathrm{Decay\, BF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class AddSysBFBtoDwwos(AbstractAddSysInfo):  # TODO: BF B -> D(*) (Added by Florian to white board)
    _systematics_key: ClassVar[str] = "additive_b_to_dwwos_bf_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \,}" + LatexDecayProcesses.b_to_xc_l_nu[1:-1] + r"\, \mathrm{BF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# Form Factors
class AddSysFFSignal(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_ff_signal_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, Signal\, FF}$"
    _number_of_nuisances: ClassVar[int] = 9

    _shape_info_file_name: str = "hammer_sig_ff_shape_sys_info.pkl"

    @staticmethod
    def get_shape_sys_info() -> UpDownAndDiffSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysFFSignal._shape_info_file_name)

        _shape_sys_info: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        return _shape_sys_info


class AddSysFFNorm(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_ff_norm_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, Norm\, FF}$"
    _number_of_nuisances: ClassVar[int] = 9

    _shape_info_file_name: str = "hammer_norm_ff_shape_sys_info.pkl"

    @staticmethod
    def get_shape_sys_info() -> UpDownAndDiffSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysFFNorm._shape_info_file_name)

        _shape_sys_info: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        return _shape_sys_info


# QED Modelling
class AddSysQEDModel(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_qed_model_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, QED\, Model}$"
    _number_of_nuisances: ClassVar[int] = 1

    qed_correction_col_name: ClassVar[str] = new_coulomb_correction_col_name

    _shape_info_file_name: str = "qed_corr_shape_sys_info.pkl"

    @staticmethod
    def double_fermi_dirac_distribution(x: np.ndarray, a: float, b: float, w: float = 50) -> np.ndarray:  # TODO: Use!?
        # TODO: Should type of x and return value really be a np.ndarray?
        return (2 * (1 + np.exp(w * (x - b))) * (1 + np.exp(-w * (x - a)))) ** -1

    @staticmethod
    def add_constraint_with_double_fd_dist(x: np.ndarray, a: float, b: float) -> np.ndarray:  # TODO: Use!?
        return -2 * np.log(AddSysQEDModel.double_fermi_dirac_distribution(x, a, b)) + 2 * np.log(  # <- could be cached?!
            AddSysQEDModel.double_fermi_dirac_distribution(np.mean([a, b]), a, b)
        )

    @staticmethod
    def get_shape_sys_info() -> UpDownAndDiffSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysQEDModel._shape_info_file_name)

        _shape_sys_info: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        return _shape_sys_info


class AddSysDSSBF(AbstractAddSysInfo):  # TODO: D** BF (Difference between Composition and BF?)
    _systematics_key: ClassVar[str] = "additive_dss_bf_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \,}" + LatexParticleNames.d_st_st_str[1:-1] + r"\, \mathrm{BF}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class AddSysDSSFFWide(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_dss_wide_ff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, wide\,}" + LatexParticleNames.d_st_st_str[1:-1] + r"\, \mathrm{FF}$"
    _number_of_nuisances: ClassVar[int] = 3

    d_primeone_ff_shape_info_file_name: str = "d_primeone_d_st_st_ff_shape_sys_info.pkl"
    d_zerostar_ff_shape_info_file_name: str = "d_zerostar_d_st_st_ff_shape_sys_info.pkl"

    @staticmethod
    def get_shape_sys_info() -> Tuple[UpDownAndDiffSystematicsDetails, ...]:
        d_po_ff_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysDSSFFWide.d_primeone_ff_shape_info_file_name)
        d_zs_ff_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysDSSFFWide.d_zerostar_ff_shape_info_file_name)

        d_po_ssi: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(file_path=d_po_ff_fp)
        d_zs_ssi: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(file_path=d_zs_ff_fp)

        _all_var_vals: Tuple[int, ...] = (
            d_po_ssi.number_of_variations,
            d_zs_ssi.number_of_variations,
        )

        assert all(nv == AddSysDSSFFWide._number_of_nuisances for nv in _all_var_vals), (
            _all_var_vals,
            AddSysDSSFFWide._number_of_nuisances,
        )

        return d_po_ssi, d_zs_ssi


class AddSysDSSFFNarrow(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_dss_narrow_ff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, narrow\,}" + LatexParticleNames.d_st_st_str[1:-1] + r"\, \mathrm{FF}$"
    _number_of_nuisances: ClassVar[int] = 4

    d_one_ff_shape_info_file_name: str = "d_one_d_st_st_ff_shape_sys_info.pkl"
    d_twostar_ff_shape_info_file_name: str = "d_twostar_d_st_st_ff_shape_sys_info.pkl"

    @staticmethod
    def get_shape_sys_info() -> Tuple[UpDownAndDiffSystematicsDetails, ...]:
        d_o_ff_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysDSSFFNarrow.d_one_ff_shape_info_file_name)
        d_ts_ff_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysDSSFFNarrow.d_twostar_ff_shape_info_file_name)

        d_o_ssi: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(file_path=d_o_ff_fp)
        d_ts_ssi: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(file_path=d_ts_ff_fp)

        _all_var_vals: Tuple[int, ...] = (
            d_o_ssi.number_of_variations,
            d_ts_ssi.number_of_variations,
        )

        assert all(nv == AddSysDSSFFNarrow._number_of_nuisances for nv in _all_var_vals), (
            _all_var_vals,
            AddSysDSSFFNarrow._number_of_nuisances,
        )

        return d_o_ssi, d_ts_ssi


# Other Backgrounds


class AddSysGapBKGModelling(AbstractAddSysInfo):  # TODO: Resonant vs. non-resonant gap description
    _systematics_key: ClassVar[str] = "additive_gap_modelling_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, Gap\, Model}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class AddSysGapBKGBFComposition(AbstractAddSysInfo):  # TODO: Gap Background Composition
    _systematics_key: ClassVar[str] = "additive_gap_bf_composition_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, Gap\, BF\, Comp.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class AddSysBBbarBKGBFComposition(AbstractAddSysInfo):  # TODO: BBbar Background Composition
    _systematics_key: ClassVar[str] = "additive_bbbar_bf_composition_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, }" + LatexParticleNames.b_b_bar_str[1:-1] + r"\, \mathrm{BKG\, Comp.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


class AddSysContinuumBKGComposition(AbstractAddSysInfo):  # TODO: Continuum Background Composition
    _systematics_key: ClassVar[str] = "additive_continuum_composition_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\, Continuum\, BKG\, Comp.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO


# Particle IDs
class AddSysPIDLepton(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_pid_lepton_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\,}" + LatexParticleNames.lepton_str[1:-1] + r"\, \mathrm{ID}$"
    _number_of_nuisances: ClassVar[int] = 10

    _shape_info_file_name: str = "new_lepton_pid_shape_sys_infos_evd_with_10_evs.pkl"

    @staticmethod
    def get_shape_sys_info() -> ReducedSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDLepton._shape_info_file_name)

        _shape_sys_info: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        assert _shape_sys_info.number_of_retained_vals == AddSysPIDLepton._number_of_nuisances, (
            _shape_sys_info.number_of_retained_vals,
            AddSysPIDLepton._number_of_nuisances,
        )
        assert _shape_sys_info.is_eigen_value_decomposition

        return _shape_sys_info


class AddSysPIDLeptonFake(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_pid_lepton_fake_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.\,}" + LatexParticleNames.lepton_str[1:-1] + r"\, \mathrm{Fake \, ID}$"
    _number_of_nuisances: ClassVar[int] = 6

    _shape_info_file_name: str = "new_lepton_fake_pid_shape_sys_infos_evd_with_6_evs.pkl"

    @staticmethod
    def get_shape_sys_info() -> ReducedSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDLeptonFake._shape_info_file_name)

        _shape_sys_info: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        assert _shape_sys_info.number_of_retained_vals == AddSysPIDLeptonFake._number_of_nuisances, (
            _shape_sys_info.number_of_retained_vals,
            AddSysPIDLeptonFake._number_of_nuisances,
        )
        assert _shape_sys_info.is_eigen_value_decomposition

        return _shape_sys_info


class AddSysPIDHadron(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_pid_hadron_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, Hadron \, ID}$"
    _number_of_nuisances: ClassVar[int] = 6 + 6 + 10 + 10

    _pion_id_shape_info_file_name: str = "pion_id_shape_sys_infos_evd_with_6_evs.pkl"
    _kaon_id_shape_info_file_name: str = "kaon_id_shape_sys_infos_evd_with_6_evs.pkl"
    _pion_fake_id_shape_info_file_name: str = "pion_fake_id_shape_sys_infos_evd_with_10_evs.pkl"
    _kaon_fake_id_shape_info_file_name: str = "kaon_fake_id_shape_sys_infos_evd_with_10_evs.pkl"

    @staticmethod
    def get_shape_sys_info() -> Tuple[ReducedSystematicsDetails, ...]:
        _pi_si_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDHadron._pion_id_shape_info_file_name)
        _k_si_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDHadron._kaon_id_shape_info_file_name)
        _pi_f_si_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDHadron._pion_fake_id_shape_info_file_name)
        _K_f_si_fp: PathType = os.path.join(sys_info_cache_dir_path, AddSysPIDHadron._kaon_fake_id_shape_info_file_name)

        _pi_ssi: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(file_path=_pi_si_fp)
        _k_ssi: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(file_path=_k_si_fp)
        _pi_f_ssi: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(file_path=_pi_f_si_fp)
        _K_f_ssi: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(file_path=_K_f_si_fp)

        _all_retained_vals: Tuple[int, ...] = (
            _pi_ssi.number_of_retained_vals,
            _k_ssi.number_of_retained_vals,
            _pi_f_ssi.number_of_retained_vals,
            _K_f_ssi.number_of_retained_vals,
        )
        sum_of_retained_vals: int = sum(_all_retained_vals)

        assert sum_of_retained_vals == AddSysPIDHadron._number_of_nuisances, (
            sum_of_retained_vals,
            _all_retained_vals,
            AddSysPIDHadron._number_of_nuisances,
        )

        assert _pi_ssi.is_eigen_value_decomposition
        assert _k_ssi.is_eigen_value_decomposition
        assert _pi_f_ssi.is_eigen_value_decomposition
        assert _K_f_ssi.is_eigen_value_decomposition

        return (
            _pi_ssi,
            _k_ssi,
            _pi_f_ssi,
            _K_f_ssi,
        )


# Slow Pion Efficiency
class AddSysSlowPion(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_slow_pion_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, Slow}\, " + LatexParticleNames.pi_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = 6

    _shape_info_file_name: str = "slow_pion_shape_sys_infos_evd_with_6_evs.pkl"

    @staticmethod
    def get_shape_sys_info() -> ReducedSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysSlowPion._shape_info_file_name)

        _shape_sys_info: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        assert _shape_sys_info.number_of_retained_vals == AddSysSlowPion._number_of_nuisances, (
            _shape_sys_info.number_of_retained_vals,
            AddSysSlowPion._number_of_nuisances,
        )
        assert _shape_sys_info.is_eigen_value_decomposition

        return _shape_sys_info


# Tracking Efficiency
class AddSysTracking(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_tracking_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, Tracking \, Uncert.}$"
    _number_of_nuisances: ClassVar[int] = 1

    _shape_info_file_name: str = "tracking_shape_sys_info.pkl"

    @staticmethod
    def get_shape_sys_info() -> UpDownAndDiffSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysTracking._shape_info_file_name)

        _shape_sys_info: UpDownAndDiffSystematicsDetails = UpDownAndDiffSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        return _shape_sys_info


# FEI Efficiency


class AddSysFEISlowPion(AbstractAddSysInfo):  # TODO: FEI Slow Pion Inefficiency?
    _systematics_key: ClassVar[str] = "additive_fei_slow_pion_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add. \, FEI \, Slow}\," + LatexParticleNames.pi_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = 2

    __fei_cf_corr_nb: ClassVar[UFloat] = MultSysFEISlowPion.fei_cf_corr_nb
    __fei_cf_corr_cb: ClassVar[UFloat] = MultSysFEISlowPion.fei_cf_corr_cb

    fei_cf_corr_relative_uncert_neutral_b: ClassVar[float] = MultSysFEISlowPion.fei_cf_corr_relative_uncert_neutral_b
    fei_cf_corr_relative_uncert_charged_b: ClassVar[float] = MultSysFEISlowPion.fei_cf_corr_relative_uncert_charged_b


# Other Reco Efficiencies
class AddSysKShortEff(AbstractAddSysInfo):
    _systematics_key: ClassVar[str] = "additive_k_short_eff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.}\, " + LatexParticleNames.k_short_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = 6

    _shape_info_file_name: str = "k_short_shape_sys_infos_evd_with_6_evs.pkl"

    @staticmethod
    def get_shape_sys_info() -> ReducedSystematicsDetails:
        _shape_info_file_path: PathType = os.path.join(sys_info_cache_dir_path, AddSysKShortEff._shape_info_file_name)

        _shape_sys_info: ReducedSystematicsDetails = ReducedSystematicsDetails.load_from(
            file_path=_shape_info_file_path,
        )

        assert _shape_sys_info.number_of_retained_vals == AddSysKShortEff._number_of_nuisances, (
            _shape_sys_info.number_of_retained_vals,
            AddSysKShortEff._number_of_nuisances,
        )
        assert _shape_sys_info.is_eigen_value_decomposition

        return _shape_sys_info


class AddSysPiZeroEff(AbstractAddSysInfo):  # TODO: pi0 Reco Efficiency Correction
    _systematics_key: ClassVar[str] = "additive_pi_zero_eff_sys"
    _latex_str: ClassVar[str] = r"$\mathrm{add.}\, " + LatexParticleNames.pi_zero_str[1:-1] + r"\,\mathrm{Corr.}$"
    _number_of_nuisances: ClassVar[int] = -999  # TODO
