#!/usr/bin/env python
# coding: utf-8

import os
import re
import copy
import pickle
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from dataclasses import dataclass
from pandas.core.groupby.generic import DataFrameGroupBy
from typing import Tuple, List, Dict, Union, Optional, Iterable, ClassVar


from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector as TrkCorr
from rdstar.offline_analysis.systematics import (
    SysColsInfo,
    new_full_systematics_dict,
    new_shape_sys_info_names,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer


__all__ = [
    "sys_info_cache_dir_path",
    "BinningMaskManager",
    "SysColVarManager",
    "MatrixEigenDecoInfo",
    "MatrixSingularValueDecoInfo",
    "ReducedSystematicsDetails",
    "UpDownAndDiffSystematicsDetails",
    "CovMatEvaluator",
    "FitSetupBinningPerRecoChInfo",
    "CovEvaluator",
    "SysCovEvaluationManager",
]


logger = log_to_default(name=__name__)


sys_info_cache_dir_path: PathType = "/ceph/fmetzner/rdstar/data_set_prod_24thMarch2022_stream0/ShapeSysInfoDataCache"


# region Binning Info Definitions


@dataclass(frozen=True)
class BinningMaskManager:
    all_mask: pd.Series
    binning_df: pd.DataFrame
    ordered_binning_cols: Tuple[str, ...]

    @property
    def index_groupby(self) -> DataFrameGroupBy:
        return self.binning_df[self.all_mask].groupby(list(self.ordered_binning_cols)).indices

    def get_groupby_for(
        self,
        df: pd.DataFrame,
        cols: Union[None, str, Iterable[str]],
    ) -> DataFrameGroupBy:
        if cols is None:
            return self.index_groupby

        if isinstance(cols, str):
            groupby_cols: Union[str, Tuple[str, ...]] = cols
        else:
            try:
                groupby_cols = tuple(v for v in cols)
            except TypeError as t_e:
                raise TypeError(
                    f"Argument cols must be string or Iterable of strings, but it is of type {type(cols)}!\n{t_e}"
                )

        assert df.index.equals(self.binning_df.index), (len(df.index), len(self.binning_df.index))
        if isinstance(groupby_cols, str):
            assert groupby_cols in df.columns, groupby_cols
        else:
            assert all(c in df.columns for c in groupby_cols), [c for c in groupby_cols if c not in df.columns]
            assert len(groupby_cols), groupby_cols

        required_cols = (
            [
                groupby_cols,
            ]
            if isinstance(groupby_cols, str)
            else list(groupby_cols)
        )
        __df = df[required_cols].copy(deep=True)

        for bin_col in self.ordered_binning_cols:
            assert bin_col not in __df.columns, bin_col
            __df.loc[:, bin_col] = self.binning_df[bin_col].values

        grouped_res: DataFrameGroupBy = __df[self.all_mask].groupby(list(self.ordered_binning_cols))[groupby_cols]

        for bin_col in self.ordered_binning_cols:
            del __df[bin_col]

        return grouped_res

    def delete_binning_cols_from_df(self, df: pd.DataFrame) -> None:
        for bin_col in self.ordered_binning_cols:
            if bin_col in df.columns:
                del df[bin_col]


@dataclass(frozen=True)
class UpDownVarInfos:
    vars_group_by: DataFrameGroupBy
    #
    nominal_col: str
    up_var_cols: Tuple[str, ...]
    down_var_cols: Tuple[str, ...]
    #
    number_of_vars: int

    def __post_init__(self) -> None:
        assert len(self.up_var_cols) == len(self.down_var_cols), (len(self.up_var_cols), len(self.down_var_cols))
        assert len(self.up_var_cols) == self.number_of_vars, (len(self.up_var_cols), self.number_of_vars)

    def get_up_and_down_col_name_pairs(self) -> Tuple[Tuple[str, str], ...]:
        col_pairs: List[Tuple[str, str]] = []
        for up_col, dw_col in zip(self.up_var_cols, self.down_var_cols):
            SysColVarManager.compare_up_and_down_col_names(up_col=up_col, down_col=dw_col)
            col_pairs.append((up_col, dw_col))
        return tuple(col_pairs)


@dataclass(frozen=True)
class DiffVarInfos:
    vars_group_by: DataFrameGroupBy
    #
    nominal_col: str
    var_cols: Tuple[str, ...]
    #
    number_of_vars: int

    def __post_init__(self) -> None:
        assert len(self.var_cols) == self.number_of_vars, (len(self.var_cols), self.number_of_vars, self.var_cols)


# Systematic Columns
class SysColVarManager:

    component_bin_index_col: ClassVar[str] = "COMP_BIN"
    reco_ch_bin_index_col: ClassVar[str] = "RECO_CH_BIN"

    def __init__(
        self,
        base_df: pd.DataFrame,
        fit_setup: FitSetupInfoContainer,
    ) -> None:
        self._base_df: pd.DataFrame = base_df
        self._fit_setup: FitSetupInfoContainer = fit_setup

        self._mask_manager: BinningMaskManager = self.get_bin_and_comp_mask_manager()

    @staticmethod
    def get_binning_bin_col_name_for(dim: int) -> str:
        assert isinstance(dim, int), (dim, type(dim))
        assert 0 <= dim < 10, dim
        return f"BINNING_DIM{dim}_BIN"

    @property
    def fit_setup(self) -> FitSetupInfoContainer:
        return self._fit_setup

    @property
    def main_weight_col(self) -> str:
        return self.fit_setup.weight_info.col_name

    def get_bin_and_comp_mask_manager(self) -> "BinningMaskManager":
        fit_obs_cols: Tuple[str, ...] = tuple(
            set([o.col_name for reco_ch in self.fit_setup.reco_channels for o in reco_ch.observable_infos])
        )
        other_obs: Tuple[str, ...] = (
            self.fit_setup.component_id_column_name,
            self.fit_setup.reco_mode_col,
            self.main_weight_col,
        )
        relevant_obs: Tuple[str, ...] = fit_obs_cols + other_obs

        this_df: pd.DataFrame = self._base_df[list(relevant_obs)].copy(deep=True)

        this_df.loc[:, self.component_bin_index_col] = -999
        this_df.loc[:, self.reco_ch_bin_index_col] = -999

        for fit_obs_index, fit_obs_info in enumerate(self.fit_setup.observable_infos):
            assert fit_obs_index < 2, (fit_obs_index, fit_obs_info.col_name)
            this_df.loc[:, self.get_binning_bin_col_name_for(dim=fit_obs_index)] = -999

        # Components
        for comp_index, fit_comp in enumerate(self.fit_setup.components):
            comp_condition = this_df[self.fit_setup.component_id_column_name].isin(fit_comp.sig_ids)
            this_df.loc[comp_condition, self.component_bin_index_col] = comp_index + 1

        # Binning Infos
        for reco_ch_id, fit_reco_ch in enumerate(self.fit_setup.reco_channels):
            reco_mode_condition = fit_reco_ch.get_mask(df=this_df, reco_mode_column_str=self.fit_setup.reco_mode_col)
            this_df.loc[reco_mode_condition, self.reco_ch_bin_index_col] = reco_ch_id + 1

            for obs_index, obs_info in enumerate(fit_reco_ch.observable_infos):
                assert obs_index < 2, (obs_index, obs_info.col_name)

                obs_values = this_df.loc[reco_mode_condition, obs_info.col_name].values
                this_df.loc[reco_mode_condition, self.get_binning_bin_col_name_for(dim=obs_index)] = np.digitize(
                    obs_values,
                    obs_info.bin_edges,
                    right=False,
                )

                assert not np.any(
                    this_df.loc[reco_mode_condition, self.get_binning_bin_col_name_for(dim=obs_index)] <= 0
                ), "Values below lowest bond!"
                assert not np.any(
                    this_df.loc[reco_mode_condition, self.get_binning_bin_col_name_for(dim=obs_index)]
                    >= len(obs_info.bin_edges)
                ), "Values above highest bond!"

        all_mask = this_df[self.component_bin_index_col] > -990
        all_mask &= this_df[self.reco_ch_bin_index_col] > -990
        for obs_index in range(len(self.fit_setup.observable_infos)):
            all_mask &= this_df[self.get_binning_bin_col_name_for(dim=obs_index)] > -990

        _obs_bin_cols: Tuple[str, ...] = tuple(
            self.get_binning_bin_col_name_for(dim=i) for i in range(len(self.fit_setup.observable_infos))
        )
        _ordered_binning_cols: Tuple[str, ...] = (
            self.component_bin_index_col,
            self.reco_ch_bin_index_col,
        ) + _obs_bin_cols

        return BinningMaskManager(
            all_mask=all_mask,
            binning_df=this_df,
            ordered_binning_cols=_ordered_binning_cols,
        )

    def prepare_varied_weights_for(
        self,
        df: pd.DataFrame,
        nominal_weight_col: str,
        varied_weight_cols: Tuple[str, ...],
    ) -> Tuple[np.ndarray, ...]:
        assert self.main_weight_col in df.columns, self.main_weight_col
        assert nominal_weight_col in df.columns, nominal_weight_col
        assert all(c in df.columns for c in varied_weight_cols), [c for c in varied_weight_cols if c not in df.columns]

        main_weight = df[self.main_weight_col].values
        bare_weight = main_weight / df[nominal_weight_col].values
        varied_weights = [bare_weight * df[var_c].values for var_c in varied_weight_cols]
        for var_w in varied_weights:
            is_nan_mask = np.isnan(var_w)
            var_w[is_nan_mask] = main_weight[is_nan_mask]

        return tuple(varied_weights)

    def add_varied_main_weights_to_df_for(
        self,
        df: pd.DataFrame,
        sys_info_key: str,
    ) -> Tuple[str, ...]:
        assert sys_info_key in new_full_systematics_dict, sys_info_key
        sys_info: SysColsInfo = new_full_systematics_dict[sys_info_key]

        _tmp_nominal_sys_weight_col_name: str = "__tmp_nominal_weight_col_for_diff_sys_case__"
        assert _tmp_nominal_sys_weight_col_name not in df.columns, _tmp_nominal_sys_weight_col_name

        if sys_info.from_df:
            _nominal_sys_weight_col: Optional[str] = sys_info.ratio
            assert _nominal_sys_weight_col is not None, sys_info_key
            nominal_sys_weight_col: str = _nominal_sys_weight_col

            if sys_info.up_down_from_df is None:
                sys_var_cols: Tuple[str, ...] = tuple([c for c in df.columns if c.startswith(sys_info.var)])
            elif isinstance(sys_info.up_down_from_df, tuple):
                _nom_col, _sys_var_cols_list = sys_info.get_systematics_input_for_df(df=df)
                _sys_var_cols: Tuple[str, ...] = tuple(_sys_var_cols_list)
                assert isinstance(_sys_var_cols, tuple) and all(isinstance(c, str) for c in _sys_var_cols), (
                    _sys_var_cols,
                    type(_sys_var_cols),
                    [type(c) for c in _sys_var_cols if not isinstance(c, str)],
                )
                sys_var_cols = _sys_var_cols
            else:
                raise RuntimeError("Unknown case encountered!")
        else:
            if sys_info.ratio is None:
                assert sys_info.var in df.columns, sys_info.var
                nominal_sys_weight_col = _tmp_nominal_sys_weight_col_name
                assert nominal_sys_weight_col not in df.columns, nominal_sys_weight_col
                df.loc[:, nominal_sys_weight_col] = 1.0

                if sys_info.var == TrkCorr.tracking_systematic_name:
                    applied_track_sys_weight_col: str = "__applied_tracking_systematic__"
                    assert applied_track_sys_weight_col not in df.columns, applied_track_sys_weight_col
                    _applied_track_sys_weights = 1.0 + df[sys_info.var].values
                    df.loc[:, applied_track_sys_weight_col] = _applied_track_sys_weights
                    assert not np.any(np.isnan(df[applied_track_sys_weight_col].values))

                    sys_var_cols = (applied_track_sys_weight_col,)
                else:
                    sys_var_cols = (sys_info.var, )
            else:
                raise RuntimeError("Unknown case encountered!")

        varied_weights: Tuple[np.ndarray, ...] = self.prepare_varied_weights_for(
            df=df,
            nominal_weight_col=nominal_sys_weight_col,
            varied_weight_cols=sys_var_cols,
        )

        if nominal_sys_weight_col == _tmp_nominal_sys_weight_col_name:
            df.drop(columns=nominal_sys_weight_col, inplace=True)
            assert nominal_sys_weight_col not in df.columns, nominal_sys_weight_col

        new_col_names: List[str] = []
        for sys_i, w_var in enumerate(varied_weights):
            sys_i_mod: int = 0
            if sys_info.up_down_from_df is None:
                up_down_tag: str = ""
            else:
                _up_tag, _down_tag = sys_info.up_down_from_df
                _this_var_col_name: str = sys_var_cols[sys_i]
                if _up_tag in _this_var_col_name:
                    up_down_tag = "_up"
                elif _down_tag in _this_var_col_name:
                    assert len(sys_var_cols) % 2 == 0, (len(sys_var_cols), len(sys_var_cols) % 2, sys_var_cols)
                    sys_i_mod = - int(len(sys_var_cols) / 2)
                    up_down_tag = "_down"
                else:
                    raise RuntimeError(f"No match for {_up_tag} or {_down_tag} in {_this_var_col_name}!")

            new_col_name = f"__sys_varied_weight_for_{sys_info.var}{up_down_tag}_{sys_i + sys_i_mod}__"
            assert new_col_name not in df.columns, new_col_name
            new_col_names.append(new_col_name)
            df.loc[:, new_col_name] = w_var

        return tuple(new_col_names)

    def get_var_info_for(
        self,
        df: pd.DataFrame,
        sys_info_key: str,
        normalize_to_nominal: bool = False,
    ) -> DataFrameGroupBy:
        varied_cols: Tuple[str, ...] = self.add_varied_main_weights_to_df_for(df=df, sys_info_key=sys_info_key)
        relevant_cols: Tuple[str, ...] = (self.main_weight_col,) + varied_cols

        if normalize_to_nominal:
            _df = df.copy(deep=True)
            nominal_sow = np.sum(_df[self.main_weight_col].values)
            assert nominal_sow > 0.0, nominal_sow

            for w_col in relevant_cols:
                _df[w_col] /= nominal_sow

        else:
            _df = df

        res: DataFrameGroupBy = self._mask_manager.get_groupby_for(df=_df, cols=relevant_cols)
        for var_col in varied_cols:
            if var_col in df.columns:
                del df[var_col]

        return res

    @staticmethod
    def _get_up_down_int_from(col_name: str) -> int:
        _all_ints: List[int] = list(map(int, re.findall(r'\d+', col_name)))
        return _all_ints[-1]

    @staticmethod
    def compare_up_and_down_col_names(up_col: str, down_col: str) -> None:
        assert "up" in up_col, up_col
        assert "down" in down_col, down_col
        _up_int: int = SysColVarManager._get_up_down_int_from(col_name=up_col)
        _down_int: int = SysColVarManager._get_up_down_int_from(col_name=down_col)
        assert _up_int == _down_int, (_up_int, _down_int, up_col, down_col)

    def get_up_down_info_for(
        self,
        df: pd.DataFrame,
        sys_info_key: str,
    ) -> UpDownVarInfos:
        up_down_cols: Tuple[str, ...] = self.add_varied_main_weights_to_df_for(df=df, sys_info_key=sys_info_key)

        assert len(up_down_cols) % 2 == 0, (len(up_down_cols), len(up_down_cols) % 2)
        _number_of_up_vars: int = int(len(up_down_cols) / 2)

        _up_var_cols: Tuple[str, ...] = tuple(c for c in up_down_cols[:_number_of_up_vars])
        _down_var_cols: Tuple[str, ...] = tuple(c for c in up_down_cols[_number_of_up_vars:])

        for u_col, d_col in zip(_up_var_cols, _down_var_cols):
            self.compare_up_and_down_col_names(up_col=u_col, down_col=d_col)

        relevant_cols: Tuple[str, ...] = (self.main_weight_col,) + _up_var_cols + _down_var_cols
        _tmp_cols: Tuple[str, ...] = _up_var_cols + _down_var_cols

        group_by_res: DataFrameGroupBy = self._mask_manager.get_groupby_for(df=df, cols=relevant_cols)

        for _tmp_col in _tmp_cols:
            if _tmp_col in df.columns:
                del df[_tmp_col]

        return UpDownVarInfos(
            vars_group_by=group_by_res,
            number_of_vars=_number_of_up_vars,
            nominal_col=self.main_weight_col,
            up_var_cols=_up_var_cols,
            down_var_cols=_down_var_cols,
        )

    def get_diff_info_for(
        self,
        df: pd.DataFrame,
        sys_info_key: str,
    ) -> DiffVarInfos:
        _var_cols: Tuple[str, ...] = self.add_varied_main_weights_to_df_for(df=df, sys_info_key=sys_info_key)
        assert len(_var_cols) == 1, (len(_var_cols), _var_cols)
        var_col: str = _var_cols[0]

        relevant_cols: Tuple[str, ...] = (self.main_weight_col, var_col)

        group_by_res: DataFrameGroupBy = self._mask_manager.get_groupby_for(df=df, cols=relevant_cols)

        if var_col in df.columns:
            del df[var_col]

        return DiffVarInfos(
            vars_group_by=group_by_res,
            number_of_vars=1,
            nominal_col=self.main_weight_col,
            var_cols=_var_cols,
        )


# endregion


# Covariance Calculations
# region Covariance Calculations


@dataclass(frozen=True)
class MatrixEigenDecoInfo:
    """
    Eigenvalue decomposition
    with
        eigenvalues, eigenvectors = np.linalg.eigh(a)
        => np.dot(a, eigenvectors[:, i]) = eigenvalues[i] * eigenvectors[:, i].
    """

    eigen_values: np.ndarray
    eigen_vectors: np.ndarray
    inverted_eigen_vectors: np.ndarray

    @property
    def matrix_of_eigen_vectors_as_rows(self) -> np.ndarray:
        return self.eigen_vectors.T

    def get_indices_for_k_largest_values(self, k: int) -> np.ndarray:
        return np.flip(np.argsort(self.eigen_values, axis=0)[-k:])

    def get_eigen_values_array_with_only_x_largest_as_none_zero(self, x: int) -> np.ndarray:
        modified_eigen_values: np.ndarray = copy.deepcopy(self.eigen_values)

        x_largest_eigenvalue: float = np.sort(self.eigen_values)[-x]
        smaller_than_xth_mask: np.ndarray = np.less(self.eigen_values, x_largest_eigenvalue)

        modified_eigen_values[smaller_than_xth_mask] = 0.0
        return modified_eigen_values

    @staticmethod
    def from_evd(matrix: np.ndarray) -> "MatrixEigenDecoInfo":
        e_vals, e_vecs_as_column = np.linalg.eigh(matrix)

        assert np.linalg.cond(e_vecs_as_column) < 1.0 / np.finfo(e_vecs_as_column.dtype).eps, (
            e_vals,
            e_vecs_as_column,
            np.linalg.cond(e_vecs_as_column) < 1.0 / np.finfo(e_vecs_as_column.dtype).eps,
        )

        for i in range(e_vals.shape[0]):
            assert np.allclose(np.dot(matrix, e_vecs_as_column[:, i]) - e_vals[i] * e_vecs_as_column[:, i], 0.0), (
                i,
                e_vals[i],
                e_vecs_as_column[:, i],
            )

        assert np.all(np.isreal(e_vals)), e_vals
        assert np.all(np.isreal(e_vecs_as_column)), e_vecs_as_column

        v_mat = copy.deepcopy(e_vecs_as_column)
        e_vecs_as_columns_inv = np.linalg.inv(v_mat)
        assert np.all(np.isreal(e_vecs_as_columns_inv)), e_vecs_as_columns_inv

        return MatrixEigenDecoInfo(
            eigen_values=e_vals,
            eigen_vectors=e_vecs_as_column,
            inverted_eigen_vectors=e_vecs_as_columns_inv,
        )


@dataclass(frozen=True)
class MatrixSingularValueDecoInfo:
    """
    Singular value decomposition
    with
        u, s, vh = np.linalg.svd(a)
        => a = u @ np.diag(s) @ vh = (u * s) @ vh
    """

    s_vals: np.ndarray
    u_mat: np.ndarray
    vh_mat: np.ndarray

    def get_indices_for_k_largest_values(self, k: int) -> np.ndarray:
        return np.flip(np.argsort(self.s_vals, axis=0)[-k:])

    def get_s_values_array_with_only_x_largest_as_none_zero(self, x: int) -> np.ndarray:
        modified_s_vals: np.ndarray = copy.deepcopy(self.s_vals)

        x_largest_s_val: float = np.sort(self.s_vals)[-x]
        smaller_than_xth_mask: np.ndarray = np.less(self.s_vals, x_largest_s_val)

        modified_s_vals[smaller_than_xth_mask] = 0.0
        return modified_s_vals

    @staticmethod
    def from_svd(matrix: np.ndarray) -> "MatrixSingularValueDecoInfo":
        # with u, s, vh = np.linalg.svd(a) => a = u @ np.diag(s) @ vh = (u * s) @ vh
        u_mat, s_vals, vh_mat = np.linalg.svd(
            matrix,
            full_matrices=False,
            compute_uv=True,
            hermitian=True,
        )

        assert np.all(np.isreal(s_vals)), s_vals
        assert np.all(np.isreal(u_mat)), u_mat
        assert np.all(np.isreal(vh_mat)), vh_mat

        np.allclose(matrix, np.dot(u_mat, np.dot(np.diag(s_vals), vh_mat))), (
            matrix,
            u_mat,
            s_vals,
            vh_mat,
        )

        return MatrixSingularValueDecoInfo(
            s_vals=s_vals,
            u_mat=u_mat,
            vh_mat=vh_mat,
        )


class CovMatEvaluator:
    def __init__(
        self,
        cov_matrix: np.ndarray,
    ) -> None:
        self._check_cov_mat(cov_mat=cov_matrix)

        self._cov_mat: np.ndarray = cov_matrix

        self._evd_info: Optional[MatrixEigenDecoInfo] = None
        self._svd_info: Optional[MatrixSingularValueDecoInfo] = None

    @property
    def cov_mat(self) -> np.ndarray:
        return self._cov_mat

    @staticmethod
    def _check_cov_mat(cov_mat: np.ndarray) -> None:
        assert np.array_equal(cov_mat, cov_mat.conj().T), (cov_mat.shape, cov_mat)
        assert np.all(np.isreal(cov_mat)), (cov_mat.shape, np.sum(~np.isreal(cov_mat)))
        assert not np.any(np.isnan(cov_mat)), (cov_mat.shape, np.sum(np.isnan(cov_mat)))

    def _check_ev_argument(self, ev_arg: int, caller_name: Optional[str] = None) -> None:
        assert isinstance(ev_arg, int), (caller_name, ev_arg, type(ev_arg))
        assert ev_arg > 0, (caller_name, ev_arg)
        assert ev_arg < self.cov_mat.shape[0], (caller_name, ev_arg, self.cov_mat.shape)

    def get_eigen_info(self) -> MatrixEigenDecoInfo:
        if self._evd_info is None:
            self._evd_info = MatrixEigenDecoInfo.from_evd(matrix=self.cov_mat)
        return self._evd_info

    def get_singular_info(self) -> MatrixSingularValueDecoInfo:
        if self._svd_info is None:
            self._svd_info = MatrixSingularValueDecoInfo.from_svd(matrix=self.cov_mat)
        return self._svd_info

    def evaluate_cov_mat_with_evd(
        self,
        number_of_eigenvalues_to_keep: int,
    ) -> np.ndarray:
        """
        Evaluation with eigenvalue decomposition
        """
        self._check_ev_argument(ev_arg=number_of_eigenvalues_to_keep, caller_name="evaluate_cov_mat_with_evd")

        eigen_deco_info: MatrixEigenDecoInfo = self.get_eigen_info()
        modified_eigen_values: np.ndarray = eigen_deco_info.get_eigen_values_array_with_only_x_largest_as_none_zero(
            x=number_of_eigenvalues_to_keep,
        )

        v_mat: np.ndarray = eigen_deco_info.eigen_vectors
        v_mat_inv: np.ndarray = eigen_deco_info.inverted_eigen_vectors
        new_cov: np.ndarray = v_mat @ np.diag(modified_eigen_values) @ v_mat_inv

        assert new_cov.shape == self.cov_mat.shape, (number_of_eigenvalues_to_keep, new_cov.shape, self.cov_mat.shape)
        assert np.all(np.isreal(new_cov)), (number_of_eigenvalues_to_keep, new_cov)
        return new_cov

    def evaluate_cov_mat_with_svd(
        self,
        number_of_singular_values_to_keep: int,
    ) -> np.ndarray:
        """
        Evaluation with singular value decomposition
        """

        self._check_ev_argument(ev_arg=number_of_singular_values_to_keep, caller_name="evaluate_cov_mat_with_svd")

        sv_deco_info: MatrixSingularValueDecoInfo = self.get_singular_info()
        modified_s_vals: np.ndarray = sv_deco_info.get_s_values_array_with_only_x_largest_as_none_zero(
            x=number_of_singular_values_to_keep,
        )

        new_cov: np.ndarray = sv_deco_info.u_mat @ np.diag(modified_s_vals) @ sv_deco_info.vh_mat

        assert new_cov.shape == self.cov_mat.shape, (
            number_of_singular_values_to_keep,
            new_cov.shape,
            self.cov_mat.shape,
        )
        assert np.all(np.isreal(new_cov)), (number_of_singular_values_to_keep, new_cov)
        return new_cov


# endregion


# region Fit Setup Binning Info


@dataclass(frozen=True)
class FitSetupBinningPerRecoChInfo:
    padding_infos: Dict[int, int]
    slice_infos: Dict[int, Tuple[int, int]]
    n_bins_per_component: int
    n_reco_channels: int
    n_components: int

    @property
    def reco_ch_split_ids(self) -> Tuple[int, ...]:
        return tuple(self.slice_infos[rch_id][1] for rch_id in range(self.n_reco_channels - 1))

    @staticmethod
    def init_from(fit_setup: FitSetupInfoContainer) -> "FitSetupBinningPerRecoChInfo":
        padding_infos: Dict[int, int] = {}
        slice_infos: Dict[int, Tuple[int, int]] = {}
        current_first_bin_id: int = 0

        for reco_ch_id, reco_ch_info in enumerate(fit_setup.reco_channels):
            padding_infos.update({reco_ch_id: fit_setup.max_n_flattened_bins_per_ch - reco_ch_info.n_bins_total})
            slice_infos.update({reco_ch_id: (current_first_bin_id, current_first_bin_id + reco_ch_info.n_bins_total)})
            current_first_bin_id += reco_ch_info.n_bins_total

        assert len(padding_infos) == len(slice_infos) == fit_setup.n_reco_channels, (
            len(padding_infos),
            len(slice_infos),
            fit_setup.n_reco_channels,
        )

        return FitSetupBinningPerRecoChInfo(
            padding_infos=padding_infos,
            slice_infos=slice_infos,
            n_bins_per_component=fit_setup.n_bins_total,
            n_reco_channels=fit_setup.n_reco_channels,
            n_components=fit_setup.n_components,
        )

# endregion


# region Systematics Info Container helper functions

def get_rearranged_shape_sys_array_for(
    input_array: np.ndarray,
    shape_sys_dimension: int,
    fit_setup: FitSetupInfoContainer,
) -> np.ndarray:
    binning_info: FitSetupBinningPerRecoChInfo = FitSetupBinningPerRecoChInfo.init_from(fit_setup=fit_setup)

    _comp_split_array: np.ndarray = np.stack(np.split(input_array, binning_info.n_bins_per_component, axis=0))

    _reco_ch_splits = np.split(_comp_split_array, binning_info.reco_ch_split_ids, axis=0)
    _padded_reco_ch_splits = [
        np.pad(
            v,
            pad_width=((0, binning_info.padding_infos[rch_id]), (0, 0), (0, 0)),
            mode="constant",
            constant_values=0.0,
        )
        for rch_id, v in enumerate(_reco_ch_splits)
    ]

    padded_array_per_component: np.ndarray = np.concatenate(_padded_reco_ch_splits, axis=0)

    assert padded_array_per_component.shape[0] == fit_setup.max_n_bins_total, (
        padded_array_per_component.shape,
        f"{fit_setup.max_n_flattened_bins_per_ch} * {fit_setup.n_reco_channels} = {fit_setup.max_n_bins_total}",
    )
    assert padded_array_per_component.shape[1] == fit_setup.n_components, (
        padded_array_per_component.shape,
        fit_setup.n_components,
    )
    assert padded_array_per_component.shape[2] == shape_sys_dimension, (
        padded_array_per_component.shape,
        shape_sys_dimension,
    )

    return padded_array_per_component


def write_sys_shape_info_object_to(
    sys_shape_info_object,
    filename: str,
    target_dir_path: PathType,
    overwrite: bool = False,
) -> None:
    os.makedirs(target_dir_path, exist_ok=True)

    file_path: PathType = os.path.join(target_dir_path, filename)
    tmp_file_path: PathType = os.path.join(target_dir_path, f"tmp_{filename}")
    assert os.path.splitext(file_path)[1] == ".pkl", (
        filename,
        file_path,
        os.path.splitext(file_path),
    )

    if os.path.exists(file_path):
        if overwrite:
            assert os.path.isfile(file_path), file_path
            os.remove(file_path)
        else:
            raise FileExistsError(file_path)

    assert not os.path.exists(tmp_file_path), tmp_file_path

    logger.info(f"Writing systematic shape info to: {file_path}")

    with open(tmp_file_path, "wb") as output_file:
        pickle.dump(sys_shape_info_object, output_file, pickle.HIGHEST_PROTOCOL)
    os.rename(src=tmp_file_path, dst=file_path)

# endregion


# region Reduced Systematics Info Container


@dataclass(frozen=True)
class ReducedSystematicsDetails:
    # Fit Setup Info
    expected_number_of_bins: int
    number_of_retained_vals: int
    #
    # with v, w = np.linalg.eig(a)  =>  a = w v w^(-1)
    reduced_eigenvalues: Optional[np.ndarray] = None
    reduced_eigenvectors: Optional[np.ndarray] = None
    #
    # with u, s, vh = np.linalg.svd(a)  =>  a = u @ np.diag(s) @ vh = (u * s) @ vh
    reduced_s_vals: Optional[np.ndarray] = None
    reduced_u_mat: Optional[np.ndarray] = None
    reduced_vh_mat: Optional[np.ndarray] = None

    def __post_init__(self) -> None:
        if self.reduced_eigenvalues is None:
            assert self.reduced_eigenvectors is None
            assert self.reduced_s_vals is not None
            assert self.reduced_u_mat is not None
            assert self.reduced_vh_mat is not None

            assert self.reduced_s_vals.shape[0] == self.number_of_retained_vals, (
                self.reduced_s_vals.shape,
                self.number_of_retained_vals,
                self.reduced_u_mat.shape,
                self.reduced_vh_mat.shape,
                self.expected_number_of_bins,
            )

            assert self.reduced_s_vals.shape[0] == self.reduced_u_mat.shape[1], (
                self.reduced_s_vals.shape,
                self.reduced_u_mat.shape,
                self.reduced_vh_mat.shape,
                self.expected_number_of_bins,
            )
            assert self.reduced_s_vals.shape[0] == self.reduced_vh_mat.shape[0], (
                self.reduced_s_vals.shape,
                self.reduced_u_mat.shape,
                self.reduced_vh_mat.shape,
                self.expected_number_of_bins,
            )
            assert self.reduced_u_mat.shape[0] == self.expected_number_of_bins, (
                self.expected_number_of_bins,
                self.reduced_u_mat.shape,
                self.reduced_s_vals.shape,
                self.reduced_vh_mat.shape,
            )
            assert self.reduced_u_mat.shape == self.reduced_vh_mat.T.shape, (
                self.reduced_s_vals.shape,
                self.reduced_u_mat.shape,
                self.reduced_vh_mat.shape,
                self.expected_number_of_bins,
            )
        else:
            assert self.reduced_eigenvectors is not None
            assert self.reduced_s_vals is None
            assert self.reduced_u_mat is None
            assert self.reduced_vh_mat is None

            assert self.reduced_eigenvalues.shape[0] == self.number_of_retained_vals, (
                self.reduced_eigenvalues.shape,
                self.number_of_retained_vals,
                self.reduced_eigenvectors.shape,
                self.expected_number_of_bins,
            )

            assert self.reduced_eigenvalues.shape[0] == self.reduced_eigenvectors.shape[1], (
                self.reduced_eigenvalues.shape,
                self.reduced_eigenvectors.shape,
                self.expected_number_of_bins,
            )
            assert self.reduced_eigenvectors.shape[0] == self.expected_number_of_bins, (
                self.reduced_eigenvectors.shape,
                self.expected_number_of_bins,
                self.reduced_eigenvalues.shape,
            )

    @property
    def is_eigen_value_decomposition(self) -> bool:
        return self.reduced_eigenvalues is not None

    @property
    def is_singular_value_decomposition(self) -> bool:
        return not self.is_eigen_value_decomposition

    def get_rearranged_array_for(
        self,
        input_array: np.ndarray,
        fit_setup: FitSetupInfoContainer,
    ) -> np.ndarray:
        return get_rearranged_shape_sys_array_for(
            input_array=input_array,
            shape_sys_dimension=self.number_of_retained_vals,
            fit_setup=fit_setup,
        )

    def get_rearranged_eigenvectors(self, fit_setup: FitSetupInfoContainer) -> np.ndarray:
        assert self.reduced_eigenvectors is not None
        return self.get_rearranged_array_for(input_array=self.reduced_eigenvectors, fit_setup=fit_setup)

    def get_rearranged_u_mat(self, fit_setup: FitSetupInfoContainer) -> np.ndarray:
        assert self.reduced_u_mat is not None
        return self.get_rearranged_array_for(input_array=self.reduced_u_mat, fit_setup=fit_setup)

    def get_rearranged_vh_mat(self, fit_setup: FitSetupInfoContainer) -> np.ndarray:
        assert self.reduced_vh_mat is not None
        _vh_t_mat: np.ndarray = self.get_rearranged_array_for(input_array=self.reduced_vh_mat.T, fit_setup=fit_setup)
        _new_vh_mat: np.ndarray = _vh_t_mat.transpose((2, 0, 1))
        assert _new_vh_mat.shape[0] == self.reduced_vh_mat.shape[0], (_new_vh_mat.shape, self.reduced_vh_mat.shape)
        return _new_vh_mat

    def write_to(
        self,
        filename: str,
        target_dir_path: PathType,
        overwrite: bool = False,
    ) -> None:
        write_sys_shape_info_object_to(
            sys_shape_info_object=self,
            filename=filename,
            target_dir_path=target_dir_path,
            overwrite=overwrite,
        )

    @staticmethod
    def load_from(file_path: PathType) -> "ReducedSystematicsDetails":
        if not os.path.exists(file_path):
            raise FileNotFoundError(file_path)

        logger.info(f"Loading systematic shape info from: {os.path.basename(file_path)}")

        with open(file_path, "rb") as input_file:
            reduced_shape_sys_details: ReducedSystematicsDetails = pickle.load(input_file)

        return reduced_shape_sys_details


# endregion


# region Up-Down Systematics Info Container

@dataclass(frozen=True)
class UpDownAndDiffSystematicsDetails:
    # Fit Setup Info
    expected_number_of_bins: int
    #
    number_of_variations: int
    #
    shape_effects: np.ndarray
    relative_shape_effects: np.ndarray
    #
    sys_type_str: str

    def __post_init__(self) -> None:
        assert self.shape_effects.shape == self.relative_shape_effects.shape, (
            self.shape_effects.shape,
            self.relative_shape_effects.shape,
        )

        assert self.shape_effects.shape[0] == self.expected_number_of_bins, (
            self.shape_effects.shape,
            self.expected_number_of_bins,
        )

        assert self.shape_effects.shape[1] == self.number_of_variations, (
            self.shape_effects.shape,
            self.number_of_variations,
        )

    def get_rearranged_shape_effect(self, fit_setup: FitSetupInfoContainer) -> np.ndarray:
        return get_rearranged_shape_sys_array_for(
            input_array=self.shape_effects,
            shape_sys_dimension=self.number_of_variations,
            fit_setup=fit_setup,
        )

    def get_rearranged_relative_shape_effect(self, fit_setup: FitSetupInfoContainer) -> np.ndarray:
        return get_rearranged_shape_sys_array_for(
            input_array=self.relative_shape_effects,
            shape_sys_dimension=self.number_of_variations,
            fit_setup=fit_setup,
        )

    def write_to(
        self,
        filename: str,
        target_dir_path: PathType,
        overwrite: bool = False,
    ) -> None:
        write_sys_shape_info_object_to(
            sys_shape_info_object=self,
            filename=filename,
            target_dir_path=target_dir_path,
            overwrite=overwrite,
        )

    @staticmethod
    def load_from(file_path: PathType) -> "UpDownAndDiffSystematicsDetails":
        if not os.path.exists(file_path):
            raise FileNotFoundError(file_path)

        logger.info(f"Loading up-down systematic shape info from: {os.path.basename(file_path)}")

        with open(file_path, "rb") as input_file:
            up_down_shape_sys_details: UpDownAndDiffSystematicsDetails = pickle.load(input_file)

        return up_down_shape_sys_details


# endregion


# Evaluation

# region Covariance matrix evaluation


class CovEvaluator:
    def __init__(
        self,
        sys_key: str,
        sys_var_info: DataFrameGroupBy,
        fit_setup: FitSetupInfoContainer,
        keep_values: Tuple[int, ...],
    ) -> None:
        assert len(set(keep_values)) == len(keep_values), keep_values

        self._sys_key: str = sys_key

        self._sys_info: DataFrameGroupBy = sys_var_info

        self._fit_setup: FitSetupInfoContainer = fit_setup
        self._main_weight_col: str = self._fit_setup.weight_info.col_name

        self._keep_values: Tuple[int, ...] = keep_values

        self._base_cov_df: pd.DataFrame = self.reindex_matrix_df_to_expected_full_binning_indices(
            df=self._sys_info.sum().T.reset_index().cov(),
        )
        self._base_corr_df: pd.DataFrame = self.reindex_matrix_df_to_expected_full_binning_indices(
            self._sys_info.sum().T.reset_index().corr(),
            check_for_nans=False,
        )

        self._base_cov: np.ndarray = self._base_cov_df.values

        self._cov_mat_evaluator: CovMatEvaluator = CovMatEvaluator(cov_matrix=self.base_cov)

        self._evd_keepers: Dict[int, np.ndarray] = {
            v: self._cov_mat_evaluator.evaluate_cov_mat_with_evd(number_of_eigenvalues_to_keep=v)
            for v in self._keep_values
        }
        self._svd_keepers: Dict[int, np.ndarray] = {
            -v: self._cov_mat_evaluator.evaluate_cov_mat_with_svd(number_of_singular_values_to_keep=v)
            for v in self._keep_values
        }
        assert len(set(list(self._evd_keepers.keys()) + list(self._svd_keepers.keys()))) == len(self._evd_keepers) + len(
            self._svd_keepers
        ), (
            list(self._evd_keepers.keys()),
            list(self._svd_keepers.keys()),
        )

        self._keepers: Dict[int, np.ndarray] = {k: v for k, v in self._evd_keepers.items()}
        self._keepers.update({k: v for k, v in self._svd_keepers.items()})

    @property
    def sys_key(self) -> str:
        return self._sys_key

    @property
    def sys_info(self) -> DataFrameGroupBy:
        return self._sys_info

    @property
    def base_cov_df(self) -> pd.DataFrame:
        return self._base_cov_df

    @property
    def base_corr_df(self) -> pd.DataFrame:
        if not np.any(np.isnan(self._base_corr_df.values)):
            return self._base_corr_df
        else:
            corr_values = self.get_own_corr_from_cov(cov=self.base_cov)
            _corr_df = self.base_cov_df.copy(deep=True)
            for c_id, c in enumerate(_corr_df.columns):
                _corr_df[c] = corr_values[:, c_id]
            self._base_corr_df = _corr_df
            return self._base_corr_df

    @property
    def base_cov(self) -> np.ndarray:
        return self._base_cov

    @property
    def binning_indices(self) -> Tuple[Tuple[int, ...], ...]:
        return tuple(self.base_cov_df.index)

    def get_keep(self, k: int) -> np.ndarray:
        assert k in self._keepers, (k, list(self._keepers.keys()))
        return self._keepers[k]

    @property
    def keep_values(self) -> Tuple[int, ...]:
        return self._keep_values

    def update_keepers(
        self,
        new_keep_values: Tuple[int, ...],
    ) -> None:
        _new_vs: Tuple[int, ...] = tuple([v for v in new_keep_values if v not in self.keep_values])
        assert all(v > 0 for v in _new_vs), (_new_vs,)
        assert all(v < self._base_cov.shape[0] for v in _new_vs), (_new_vs, self._base_cov.shape[0])

        assert all(v not in self._evd_keepers for v in _new_vs), (v for v in _new_vs if v in self._evd_keepers)
        assert all(-v not in self._svd_keepers for v in _new_vs), (v for v in _new_vs if -v in self._svd_keepers)
        assert all(v not in self._keepers for v in _new_vs), (v for v in _new_vs if v in self._keepers)
        assert all(-v not in self._keepers for v in _new_vs), (v for v in _new_vs if -v in self._keepers)

        new_evd_keepers: Dict[int, np.ndarray] = {
            v: self._cov_mat_evaluator.evaluate_cov_mat_with_evd(number_of_eigenvalues_to_keep=v) for v in _new_vs
        }
        new_svd_keepers: Dict[int, np.ndarray] = {
            -v: self._cov_mat_evaluator.evaluate_cov_mat_with_svd(number_of_singular_values_to_keep=v) for v in _new_vs
        }

        self._keep_values = tuple(sorted(list(self.keep_values + _new_vs)))

        self._evd_keepers.update(new_evd_keepers)
        self._svd_keepers.update(new_svd_keepers)
        self._keepers.update({k: v for k, v in new_evd_keepers.items()})
        self._keepers.update({k: v for k, v in new_svd_keepers.items()})

    @staticmethod
    def get_expected_full_binning_indices(fit_setup: FitSetupInfoContainer) -> Tuple[Tuple[int, int, int, int], ...]:
        res: List[Tuple[int, int, int, int]] = []
        for comp_id in range(len(fit_setup.components)):
            for reco_ch_id, fit_reco_ch in enumerate(fit_setup.reco_channels):
                obs_info_1 = fit_reco_ch.observable_infos[0]
                obs_info_2 = fit_reco_ch.observable_infos[1]
                for obs1_bin_id in range(obs_info_1.n_bins):
                    for obs2_bin_id in range(obs_info_2.n_bins):
                        res.append((comp_id + 1, reco_ch_id + 1, obs1_bin_id + 1, obs2_bin_id + 1))

        return tuple(res)

    def reindex_matrix_df_to_expected_full_binning_indices(
        self,
        df: pd.DataFrame,
        fill_value: float = 0.0,
        check_for_nans: bool = True,
    ) -> pd.DataFrame:
        assert df.shape == df.T.shape, f"Input data frame is not symmetric! It has the shape {df.shape}"
        if check_for_nans:
            assert not np.any(np.isnan(df.values)), "Input data frame contains NaN values!"

        _new_df: pd.DataFrame = copy.deepcopy(df)

        _expected_multiindex: pd.MultiIndex = pd.MultiIndex.from_tuples(
            self.get_expected_full_binning_indices(fit_setup=self._fit_setup)
        )

        _new_df = _new_df.reindex(_expected_multiindex).T.reindex(_expected_multiindex).T.fillna(value=fill_value)

        if check_for_nans:
            assert not np.any(np.isnan(_new_df.values)), "New, re-indexed data frame contains NaN values!"

        return _new_df

    def reindex_column_df_to_expected_full_binning_indices(
        self,
        df: pd.DataFrame,
        fill_value: float = 0.0,
        check_for_nans: bool = True,
    ) -> pd.DataFrame:
        assert len(df.shape) == 1, f"Input data frame is not a column! It has the shape {df.shape}"
        if check_for_nans:
            assert not np.any(np.isnan(df.values)), "Input data frame contains NaN values!"

        _new_df: pd.DataFrame = copy.deepcopy(df)
        _expected_multiindex: pd.MultiIndex = pd.MultiIndex.from_tuples(
            self.get_expected_full_binning_indices(fit_setup=self._fit_setup)
        )
        _new_df = _new_df.reindex(_expected_multiindex).fillna(value=fill_value)

        if check_for_nans:
            assert not np.any(np.isnan(_new_df.values)), "New, re-indexed data frame contains NaN values!"

        return _new_df

    def get_boiled_down_systematics_details(
        self,
        keep: int,
    ) -> ReducedSystematicsDetails:
        if keep > 0:
            _evd_info: MatrixEigenDecoInfo = self._cov_mat_evaluator.get_eigen_info()
            _reduced_eigen_value_indices: np.ndarray = _evd_info.get_indices_for_k_largest_values(k=keep)

            _reduced_eigenvalues: np.ndarray = copy.deepcopy(_evd_info.eigen_values[_reduced_eigen_value_indices])
            _reduced_eigenvectors: np.ndarray = copy.deepcopy(_evd_info.eigen_vectors[:, _reduced_eigen_value_indices])

            # Testing the condition np.dot(a, eigenvectors[:, i]) = eigenvalues[i] * eigenvectors[:, i]
            for value_id, eigenvalue in enumerate(_reduced_eigenvalues):
                _left: np.ndarray = np.dot(self._base_cov, _reduced_eigenvectors[:, value_id])
                _right: np.ndarray = eigenvalue * _reduced_eigenvectors[:, value_id]
                assert np.allclose(_left, _right, atol=1e-5, rtol=1e-5), (
                    _left,
                    _right,
                    value_id,
                    eigenvalue,
                    _reduced_eigenvectors[value_id],
                )

            return ReducedSystematicsDetails(
                expected_number_of_bins=self._fit_setup.n_bins_total * self._fit_setup.n_components,
                number_of_retained_vals=keep,
                reduced_eigenvalues=_reduced_eigenvalues,
                reduced_eigenvectors=_reduced_eigenvectors,
                reduced_s_vals=None,
                reduced_u_mat=None,
                reduced_vh_mat=None,
            )
        elif keep < 0:
            _svd_info: MatrixSingularValueDecoInfo = self._cov_mat_evaluator.get_singular_info()
            _reduced_singular_value_indices: np.ndarray = _svd_info.get_indices_for_k_largest_values(k=-keep)

            _reduced_singular_values: np.ndarray = copy.deepcopy(_svd_info.s_vals[_reduced_singular_value_indices])
            _reduced_u_mat: np.ndarray = copy.deepcopy(_svd_info.u_mat[:, _reduced_singular_value_indices])
            _reduced_vh_mat: np.ndarray = copy.deepcopy(_svd_info.vh_mat[_reduced_singular_value_indices, :])

            # Testing the condition np.dot(a, vh.H[:, i]) = np.dot(u[:, i] * s[i])
            for value_id, singular_value in enumerate(_reduced_singular_values):
                _left = np.asarray(np.dot(self._base_cov, np.matrix(_reduced_vh_mat).H[:, value_id]))[:, 0]
                _right = np.dot(_reduced_u_mat[:, value_id], singular_value).T
                assert np.allclose(_left, _right, atol=1e-5, rtol=1e-5), (
                    _left,
                    _right,
                    value_id,
                    singular_value,
                    _reduced_u_mat[value_id],
                    _reduced_vh_mat[value_id],
                )

            return ReducedSystematicsDetails(
                expected_number_of_bins=self._fit_setup.n_bins_total * self._fit_setup.n_components,
                number_of_retained_vals=-keep,
                reduced_eigenvalues=None,
                reduced_eigenvectors=None,
                reduced_s_vals=_reduced_singular_values,
                reduced_u_mat=_reduced_u_mat,
                reduced_vh_mat=_reduced_vh_mat,
            )
        else:
            raise ValueError(f"Invalid keep value: {keep}!")

    @property
    def mean_bin_uncertainty(self) -> float:
        _norm_df: pd.DataFrame = self._sys_info.sum()[self._main_weight_col]
        _norm: np.ndarray = self.reindex_column_df_to_expected_full_binning_indices(df=_norm_df).values
        _norm[_norm == 0.0] = 1.0

        res = np.mean(np.sqrt(np.diag(self.base_cov_df.values)) / _norm)

        assert isinstance(res, float) or (isinstance(res, np.ndarray) and res.shape == (1,)), (res, type(res))
        return float(res)

    def mean_bin_uncertainty_for(
        self,
        k: int,
        set_small_negative_to_zero: bool = False,
        bound: float = -1e-20,
    ) -> float:
        this_cov: np.ndarray = self.get_keep(k=k)

        cov_diagonal: np.ndarray = np.diag(this_cov)
        if set_small_negative_to_zero:
            cov_diagonal = self.set_small_negative_values_to(a=cov_diagonal, bound=bound, new_value=0.0)

        _norm_df: pd.DataFrame = self._sys_info.sum()[self._main_weight_col]
        _norm: np.ndarray = self.reindex_column_df_to_expected_full_binning_indices(df=_norm_df).values
        _norm[_norm == 0.0] = 1.0

        res = np.mean(np.sqrt(cov_diagonal) / _norm)
        assert isinstance(res, float) or (isinstance(res, np.ndarray) and res.shape == (1,)), (res, type(res))
        return float(res)

    def evaluate_mean_bin_uncertainty(
        self,
        set_small_negative_to_zero: bool = False,
        bound: float = -1e-20,
    ):
        res: Tuple[Tuple[Union[int, str], float], ...] = tuple(
            (k, self.mean_bin_uncertainty_for(k, set_small_negative_to_zero=set_small_negative_to_zero, bound=bound))
            for k in self._keepers
        )
        return (("all", self.mean_bin_uncertainty),) + res

    def evaluate_diff_of_mean_bin_uncertainty(self, set_small_negative_to_zero=False, bound=-1e-20):
        _res_dict = dict(
            self.evaluate_mean_bin_uncertainty(set_small_negative_to_zero=set_small_negative_to_zero, bound=bound)
        )
        base_value = _res_dict.pop("all")
        return tuple((k, v - base_value) for k, v in _res_dict.items())

    def evaluate_min_max(self) -> Tuple[Tuple[Union[int, str], float, float], ...]:
        _covs = (("all", self.base_cov),) + tuple((k, v) for k, v in self._keepers.items())
        return tuple((k, float(np.min(v)), float(np.max(v))) for k, v in _covs)

    def evaluate_abs_min_max(self) -> Tuple[Tuple[Union[int, str], float, float, float, float], ...]:
        _covs = (("all", self.base_cov),) + tuple((k, v) for k, v in self._keepers.items())
        return tuple(
            (k, float(np.min(np.abs(v))), float(np.min(np.abs(v[v != 0.0]))), float(np.min(v)), float(np.max(v)))
            for k, v in _covs
        )

    def evaluate_max_diff(self) -> Tuple[Tuple[int, float], ...]:
        return tuple((k, np.max(self.base_cov - v, axis=None)) for k, v in self._keepers.items())

    def evaluate_max_diff_relative_to_max_of_all(self) -> Tuple[Tuple[int, float], ...]:
        return tuple(
            (k, np.max((self.base_cov - v) / np.max(self.base_cov), axis=None)) for k, v in self._keepers.items()
        )

    def evaluate_max_diff_relative_to_max_of_row(self) -> Tuple[Tuple[int, float], ...]:
        base_cov_norm = np.max(self.base_cov, axis=0)
        base_cov_norm[base_cov_norm == 0.0] = 1.0
        return tuple((k, np.max((self.base_cov - v) / base_cov_norm, axis=None)) for k, v in self._keepers.items())

    def evaluate_max_diff_relative_to_element(self) -> Tuple[Tuple[int, float], ...]:
        base_cov_norm = copy.deepcopy(self.base_cov)
        base_cov_norm[base_cov_norm == 0.0] = 1.0
        return tuple((k, np.max((self.base_cov - v) / base_cov_norm, axis=None)) for k, v in self._keepers.items())

    def evaluate_max_diff_relative_to_diagonal(self) -> Tuple[Tuple[int, float], ...]:
        base_cov_norm = copy.deepcopy(np.diag(self.base_cov))
        base_cov_norm[base_cov_norm == 0.0] = 1.0
        return tuple((k, np.max((self.base_cov - v) / base_cov_norm, axis=None)) for k, v in self._keepers.items())

    def check_if_max_relative_diff_is_below(
        self,
        bound: float,
        relative_to: Optional[str] = None,
    ) -> Tuple[Tuple[int, bool], ...]:
        _valid_values: Tuple[Optional[str], ...] = (None, "max_of_all", "max_of_row", "diagonal", "element")
        if relative_to not in _valid_values:
            raise ValueError(
                f"Invalid value for parameter 'relative_to':\n\t"
                f"Must be one of {_valid_values}, but {relative_to} was provided!"
            )
        if relative_to == "max_of_all" or relative_to is None:
            _res: Tuple[Tuple[int, float], ...] = self.evaluate_max_diff_relative_to_max_of_all()
        elif relative_to == "max_of_row":
            _res = self.evaluate_max_diff_relative_to_max_of_row()
        elif relative_to == "diagonal":
            _res = self.evaluate_max_diff_relative_to_diagonal()
        elif relative_to == "element":
            _res = self.evaluate_max_diff_relative_to_element()
        else:
            raise ValueError("Invalid value for parameter 'relative_to'...")

        return tuple((v[0], v[1] < bound) for v in _res)

    def evaluate_diff_is_below(
        self,
        max_diff: float,
    ) -> Tuple[Tuple[int, bool], ...]:
        return tuple((k, np.all(np.abs(self.base_cov - v) < max_diff)) for k, v in self._keepers.items())

    @staticmethod
    def get_own_corr_from_cov(cov: np.ndarray) -> np.ndarray:
        d_inv = np.nan_to_num(np.diag(1.0 / np.sqrt(np.diag(cov))))
        return np.matmul(d_inv, np.matmul(cov, d_inv))

    @staticmethod
    def set_small_negative_values_to(
        a: np.ndarray,
        bound: float,
        new_value: float,
    ) -> np.ndarray:
        assert bound < 0, bound
        _a: np.ndarray = copy.deepcopy(a)
        _a[(_a < 0.0) & (_a > bound)] = new_value
        return _a


# endregion


# region Shape systematics from variations evaluation manager


class SysCovEvaluationManager:
    possible_sys_col_base_names: ClassVar[Tuple[str, ...]] = new_shape_sys_info_names

    _default_eval_settings: ClassVar[Dict[str, Tuple[int, ...]]] = {
        "new_lepton_id": (2, 3, 4, 6, 10, 20),
        "new_lepton_fake_id": (2, 3, 4, 6, 10, 20),
        "lepton_id": (2, 3, 4, 6, 10, 20),
        #
        "slow_pion": (2, 3, 4, 6, 10),
        #
        "pion_id": (2, 3, 4, 6, 10),
        "pion_fake_id": (2, 3, 4, 6, 10),
        "kaon_id": (2, 3, 4, 6, 10),
        "kaon_fake_id": (2, 3, 4, 6, 10),
        #
        "k_short": (2, 3, 4, 6, 10),
    }

    def __init__(
        self,
        df: pd.DataFrame,
        fit_setup: FitSetupInfoContainer,
    ) -> None:

        self._fit_setup: FitSetupInfoContainer = fit_setup

        self._sys_col_var_manager: SysColVarManager = SysColVarManager(
            base_df=df,
            fit_setup=self._fit_setup,
        )

        assert all(v in self.possible_sys_col_base_names for v in self._default_eval_settings.keys()), (
            [v for v in self._default_eval_settings.keys() if v not in self.possible_sys_col_base_names]
        )

        self._evaluated_systematics: Dict[str, CovEvaluator] = dict()

    @property
    def sys_col_var_manager(self) -> SysColVarManager:
        return self._sys_col_var_manager

    @property
    def fit_setup(self) -> FitSetupInfoContainer:
        return self._fit_setup

    def evaluate(
        self,
        df: pd.DataFrame,
        sys_key: str,
        keep_values: Tuple[int, ...],
    ) -> CovEvaluator:

        if sys_key not in self.possible_sys_col_base_names:
            raise ValueError(f"Invalid sys_key '{sys_key}'! Must be one of {list(self.possible_sys_col_base_names)}")
        if sys_key not in self._default_eval_settings:
            raise ValueError(f"Invalid sys_key '{sys_key}'! Must be one of {list(self._default_eval_settings.keys())}")

        if sys_key in self._evaluated_systematics:
            cov_eval_info: CovEvaluator = self._evaluated_systematics[sys_key]
        else:
            sys_var_info: DataFrameGroupBy = self.sys_col_var_manager.get_var_info_for(
                df=df,
                sys_info_key=sys_key,
                normalize_to_nominal=False,
            )
            cov_eval_info = CovEvaluator(
                sys_key=sys_key,
                sys_var_info=sys_var_info,
                fit_setup=self.fit_setup,
                keep_values=keep_values,
            )

            self._evaluated_systematics[sys_key] = cov_eval_info

        if not all(kv in cov_eval_info.keep_values for kv in keep_values):
            cov_eval_info.update_keepers(new_keep_values=keep_values)

        return cov_eval_info

    def get_systematics_evaluation_infos_for(self, sys_key: str) -> CovEvaluator:
        if sys_key not in self._evaluated_systematics:
            raise KeyError(f"Systematics key {sys_key} was not evaluated, yet...")
        return self._evaluated_systematics[sys_key]


# endregion


# region Shape systematics from weight difference evaluation manager


class SysDiffEvaluationManager:
    possible_sys_col_base_names: ClassVar[Tuple[str, ...]] = new_shape_sys_info_names
    sys_diff_infos: ClassVar[Dict[str, str]] = {
        "hammer_norm_ff_sys": "UpDown",
        "hammer_sig_ff_sys": "UpDown",
        #
        "d_ff_sys": "UpDown",
        "d_st_ff_sys": "UpDown",
        #
        "d_one_d_st_st_ff_sys": "UpDown",
        "d_twostar_d_st_st_ff_sys": "UpDown",
        "d_primeone_d_st_st_ff_sys": "UpDown",
        "d_zerostar_d_st_st_ff_sys": "UpDown",
        #
        "tracking_sys": "SimpleDiff",
        #
        "qed_shape_correction_sys": "SimpleDiff",
    }

    def __init__(
        self,
        df: pd.DataFrame,
        fit_setup: FitSetupInfoContainer,
    ) -> None:

        self._fit_setup: FitSetupInfoContainer = fit_setup

        self._sys_shape_diff_manager: SysColVarManager = SysColVarManager(
            base_df=df,
            fit_setup=self._fit_setup,
        )

        assert all(v in self.possible_sys_col_base_names for v in self.sys_diff_infos.keys()), (
            [v for v in self.sys_diff_infos.keys() if v not in self.possible_sys_col_base_names]
        )

        self._evaluated_systematics: Dict[str, UpDownAndDiffSystematicsDetails] = dict()

    @property
    def sys_shape_diff_manager(self) -> SysColVarManager:
        return self._sys_shape_diff_manager

    @property
    def fit_setup(self) -> FitSetupInfoContainer:
        return self._fit_setup

    @property
    def main_weight_col(self) -> str:
        return self.fit_setup.weight_info.col_name

    def _calculate_shape_effect_from_up_and_down_vars(
        self,
        up_down_vars_info: UpDownVarInfos,
    ) -> Tuple[np.ndarray, np.ndarray]:
        _vars_group_by: DataFrameGroupBy = up_down_vars_info.vars_group_by

        vars_df: pd.DataFrame = self.reindex_shape_effect_to_expected_full_binning_indices(
            df=_vars_group_by.sum(),
            fill_value=0.0,
            check_for_nans=True,
        )

        _nom_vals: np.ndarray = vars_df[up_down_vars_info.nominal_col].values

        _res_arrays: List[np.ndarray] = []
        _rel_res_arrays: List[np.ndarray] = []
        for up_col, down_col in up_down_vars_info.get_up_and_down_col_name_pairs():
            assert up_col in vars_df.columns, (up_col, vars_df.columns)
            assert down_col in vars_df.columns, (down_col, vars_df.columns)

            _up_vals: np.ndarray = vars_df[up_col].values
            _down_vals: np.ndarray = vars_df[down_col].values

            _up_diff: np.ndarray = _up_vals - _nom_vals
            _dw_diff: np.ndarray = _down_vals - _nom_vals

            assert not np.any(np.isnan(_up_diff)), (up_col, down_col, np.sum(np.isnan(_up_diff)))
            assert not np.any(np.isnan(_dw_diff)), (up_col, down_col, np.sum(np.isnan(_dw_diff)))

            _res_vals: np.ndarray = np.maximum(np.abs(_up_diff), np.abs(_dw_diff)) * np.sign(_up_diff)
            assert not np.any(np.isnan(_res_vals)), (up_col, down_col, np.sum(np.isnan(_res_vals)))

            not_zero_mask: np.ndarray = np.not_equal(_nom_vals, 0.0)
            _rel_res_vals: np.ndarray = np.zeros_like(_res_vals)
            np.place(_rel_res_vals, not_zero_mask, _res_vals[not_zero_mask] / _nom_vals[not_zero_mask])
            assert not np.any(np.isnan(_rel_res_vals)), (up_col, down_col, np.sum(np.isnan(_rel_res_vals)))

            _res_arrays.append(_res_vals)
            _rel_res_arrays.append(_rel_res_vals)

        shape_effect: np.ndarray = np.stack(_res_arrays, axis=1)
        relative_shape_effect: np.ndarray = np.stack(_rel_res_arrays, axis=1)

        return shape_effect, relative_shape_effect

    def _calculate_shape_effect_from_diff_vars(
        self,
        diff_vars_info: DiffVarInfos,
    ) -> Tuple[np.ndarray, np.ndarray]:
        _vars_group_by: DataFrameGroupBy = diff_vars_info.vars_group_by
        assert len(diff_vars_info.var_cols) == 1, (len(diff_vars_info.var_cols), diff_vars_info.var_cols)
        var_col: str = diff_vars_info.var_cols[0]

        vars_df: pd.DataFrame = self.reindex_shape_effect_to_expected_full_binning_indices(
            df=_vars_group_by.sum(),
            fill_value=0.0,
            check_for_nans=True,
        )

        _nom_vals: np.ndarray = vars_df[diff_vars_info.nominal_col].values
        _diff_vals: np.ndarray = vars_df[var_col].values

        shape_effect: np.ndarray = _diff_vals - _nom_vals
        assert not np.any(np.isnan(shape_effect)), (var_col, np.sum(np.isnan(shape_effect)))

        not_zero_mask: np.ndarray = np.not_equal(_nom_vals, 0.0)
        rel_shape_effect: np.ndarray = np.zeros_like(shape_effect)
        np.place(rel_shape_effect, not_zero_mask, shape_effect[not_zero_mask] / _nom_vals[not_zero_mask])
        assert not np.any(np.isnan(rel_shape_effect)), (var_col, np.sum(np.isnan(shape_effect)))

        return shape_effect[:, np.newaxis], rel_shape_effect[:, np.newaxis]

    def reindex_shape_effect_to_expected_full_binning_indices(
        self,
        df: pd.DataFrame,
        fill_value: float = 0.0,
        check_for_nans: bool = True,
    ) -> pd.DataFrame:
        assert isinstance(df, pd.DataFrame), type(df)

        if check_for_nans:
            assert not np.any(np.isnan(df.values)), "Input data frame contains NaN values!"

        _new_df: pd.DataFrame = copy.deepcopy(df)
        _expected_multiindex: pd.MultiIndex = pd.MultiIndex.from_tuples(
            CovEvaluator.get_expected_full_binning_indices(fit_setup=self._fit_setup)
        )
        _new_df = _new_df.reindex(_expected_multiindex).fillna(value=fill_value)

        if check_for_nans:
            assert not np.any(np.isnan(_new_df.values)), "New, re-indexed data frame contains NaN values!"

        return _new_df

    def evaluate(
        self,
        df: pd.DataFrame,
        sys_key: str,
    ) -> UpDownAndDiffSystematicsDetails:

        if sys_key not in self.possible_sys_col_base_names:
            raise ValueError(f"Invalid sys_key '{sys_key}'! Must be one of {list(self.possible_sys_col_base_names)}")
        if sys_key not in self.sys_diff_infos:
            raise ValueError(f"Invalid sys_key '{sys_key}'! Must be one of {list(self.sys_diff_infos.keys())}")

        if sys_key in self._evaluated_systematics:
            _eval_info: UpDownAndDiffSystematicsDetails = self._evaluated_systematics[sys_key]
        else:

            if self.sys_diff_infos[sys_key] == "UpDown":
                u_d_info: UpDownVarInfos = self.sys_shape_diff_manager.get_up_down_info_for(
                    df=df,
                    sys_info_key=sys_key,
                )

                _shape_effect, _rel_shape_effect = self._calculate_shape_effect_from_up_and_down_vars(
                    up_down_vars_info=u_d_info,
                )

                _eval_info = UpDownAndDiffSystematicsDetails(
                    expected_number_of_bins=self._fit_setup.n_bins_total * self._fit_setup.n_components,
                    number_of_variations=u_d_info.number_of_vars,
                    shape_effects=_shape_effect,
                    relative_shape_effects=_rel_shape_effect,
                    sys_type_str=self.sys_diff_infos[sys_key],
                )

            elif self.sys_diff_infos[sys_key] == "SimpleDiff":
                diff_info: DiffVarInfos = self.sys_shape_diff_manager.get_diff_info_for(
                    df=df,
                    sys_info_key=sys_key,
                )

                _shape_effect, _rel_shape_effect = self._calculate_shape_effect_from_diff_vars(
                    diff_vars_info=diff_info,
                )

                _eval_info = UpDownAndDiffSystematicsDetails(
                    expected_number_of_bins=self._fit_setup.n_bins_total * self._fit_setup.n_components,
                    number_of_variations=diff_info.number_of_vars,
                    shape_effects=_shape_effect,
                    relative_shape_effects=_rel_shape_effect,
                    sys_type_str=self.sys_diff_infos[sys_key],
                )
            else:
                raise ValueError(f"Unknown sys type: '{self.sys_diff_infos[sys_key]}' for sys_key = '{sys_key}'!")

            self._evaluated_systematics[sys_key] = _eval_info

        return _eval_info

    def get_systematics_evaluation_infos_for(self, sys_key: str) -> UpDownAndDiffSystematicsDetails:
        if sys_key not in self._evaluated_systematics:
            raise KeyError(f"Systematics key {sys_key} was not evaluated, yet...")
        return self._evaluated_systematics[sys_key]


# endregion


# region Plotting Methods


def replace_df_matrix_values_from_np_array_if_not_none(
    matrix_df: pd.DataFrame,
    array: Optional[np.ndarray],
) -> pd.DataFrame:
    if array is not None:
        assert isinstance(array, np.ndarray)
        assert array.shape == matrix_df.values.shape
        _df = matrix_df.copy(deep=True)
        for c_id, c in enumerate(_df.columns):
            _df[c] = array[:, c_id]
    else:
        _df = matrix_df
    return _df


def plot_heatmap(
    df: pd.DataFrame,
    v_min: Optional[float] = None,
    v_max: Optional[float] = None,
    replace_values_with: Optional[np.ndarray] = None,
) -> None:
    # Seaborn Settings
    sns.set_theme()
    sns.set(rc={"figure.dpi": 300, "savefig.dpi": 300})

    _df: pd.DataFrame = replace_df_matrix_values_from_np_array_if_not_none(
        matrix_df=df,
        array=replace_values_with,
    )

    fig, ax = plt.subplots(figsize=(20, 20), dpi=200)

    sns.heatmap(
        _df,
        annot=False,
        cmap="crest",
        vmin=v_min,
        vmax=v_max,
        ax=ax,
    )


def plot_clustermap(
    df: pd.DataFrame,
    fit_setup: FitSetupInfoContainer,
    replace_values_with: Optional[np.ndarray] = None,
    **kwargs,
) -> None:
    # Seaborn Settings
    sns.set_theme()
    sns.set(rc={"figure.dpi": 300, "savefig.dpi": 300})

    comp_values = list(df.index.get_level_values(0).unique())

    # comp_pal = sns.husl_palette(len(comp_values), s=.45)
    # comp_pal = sns.color_palette("rainbow", len(comp_values))
    comp_pal = sns.color_palette([c.color for c in fit_setup.components])

    comp_lut = dict(zip(comp_values, comp_pal))

    # Convert the palette to vectors that will be drawn on the side of the matrix
    components = df.index.get_level_values(0)
    comp_colors = pd.Series(components, index=df.index).map(comp_lut)

    _df: pd.DataFrame = replace_df_matrix_values_from_np_array_if_not_none(
        matrix_df=df,
        array=replace_values_with,
    )

    g = sns.clustermap(
        _df,
        center=0,
        cmap="vlag",
        method=kwargs.get("method", "ward"),
        metric=kwargs.get("metric", "euclidean"),
        row_colors=comp_colors,
        col_colors=comp_colors,
        dendrogram_ratio=(0.1, 0.2),
        cbar_pos=(0.02, 0.32, 0.03, 0.2),
        figsize=(25, 25),
    )

    g.ax_row_dendrogram.remove()


# endregion
