"""
Tools for running and evaluating the tag cal fits

Felix Metzner 2023
"""

import os
import copy
import json
import numpy as np
import pandas as pd

from uncertainties import unumpy as unp

from iminuit import Minuit
from uncertainties import ufloat
from iminuit.minuit import mutil
from dataclasses import dataclass
from typing import Tuple, List, Dict, Type, Any, Optional, Callable

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.branching_fractions import constants_and_measurements as bfs_and_consts

from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.comparison_tools.data_mc_comparer import DataMCComparer

from rdstar.offline_analysis.systematics import full_systematics_cols
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.rdstar_samples import alternative_sample_keys
from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo

from rdstar.offline_analysis.observables.sig_id_observables import SigIDObservables
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeGroupInfo
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BTagRecoModes, BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos, U4SRecoModeInfos, RecoModeInfos
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import (
    MCMatchingCollections,
    MCMatchingCollection,
    MCMatchingCollectionItem,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import get_file_path_for

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters.tag_cal_fitter_base import TagCorrFitterBase
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    RecoChannelInfo,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import (
    tc_fitter_dict,
    keys_to_evaluate,
    right_sig_tag_combination_selection,
    # wrong_sig_tag_combination_selection,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import tag_cal_main_fit_info as tag_fit_info
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters.fitter_tag_cal_main import (
    TagCorrectionFitter as MainTagCalFitter,
)


__all__ = [
    "TagCalEvalResult",
    "TagCalFitEvaluator",
]


logger = log_to_default(name=__name__)


# region sideband eval infos
sideband_data_component_label: str = "Sideband Data"
q2_sideband_title: str = rf"{LatexObsNames.fitting_q2} Sideband"
q2_sideband_text: str = LatexObsNames.fitting_q2[:-1] + r" < 4.0" + LatexUnits.gev_sq[1:]

post_fit_plot_reco_mode_info_collection = BSigRecoModes.main_modes
# post_fit_plot_mc_matching_collection = MCMatchingCollections.extended_sig_id__basic__collection
post_fit_plot_mc_matching_collection = tag_fit_info.tag_cal_fit__component_info_collection

_eval_sb_plot_infos: Tuple[SidebandInfo, ...] = (
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_for_main_fei_cal.get_modified_histogram_variable(n_bins=10),
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.66),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.pl_star_signal_fit,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.neextra_for_fit,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.58),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
)


sb_key = "q2_sb_eval"
sb_eval_title = f"{RDStarHistVars.q2_reco_in_sideband_region.variable_name}-Sideband"
sb_eval_info_text = q2_sideband_text

plot_reco_mode_info_collection = BSigRecoModes.detailed_main_modes
plot_mc_matching_collection = MCMatchingCollections.extended_sig_id__charge__collection

_more_eval_sb_plot_infos: Tuple[SidebandInfo, ...] = (
    SidebandInfo(
        key=sb_key,
        hist_var=RDStarHistVars.missing_mass_for_main_fei_cal.get_modified_histogram_variable(
            df_label=Cols.alt_res_corr_missing_mass
        ),
        title=sb_eval_title,
        info_text=sb_eval_info_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.67),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
    ),
    SidebandInfo(
        key=sb_key,
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=sb_eval_title,
        info_text=sb_eval_info_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.67),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
    ),
    SidebandInfo(
        key=sb_key,
        hist_var=RDStarHistVars.neextra_fine,
        title=sb_eval_title,
        info_text=sb_eval_info_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.67),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
    ),
    SidebandInfo(
        key=sb_key,
        hist_var=RDStarHistVars.cs_mva_output,
        title=sb_eval_title,
        info_text=sb_eval_info_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.66),
        text_v_pos=(0.97, 0.63),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 1.6),
        systematics=None,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
    ),
    SidebandInfo(
        key=sb_key,
        hist_var=RDStarHistVars.q2_reco_in_sideband_region,
        title=sb_eval_title,
        info_text=sb_eval_info_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.67),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
    ),
)


post_fit_plot_infos: Tuple[SidebandInfo, ...] = _eval_sb_plot_infos
eval_plot_infos: Tuple[SidebandInfo, ...] = _more_eval_sb_plot_infos
eval_sb_plot_infos: Tuple[SidebandInfo, ...] = _eval_sb_plot_infos + _more_eval_sb_plot_infos

# endregion


# region FEI Tag calibration info to component and reco channel mapping


@dataclass(frozen=True)
class TagComponentToRecoChAndSigIDMapInfo:
    component_name: str
    reco_channel_info_to_sig_ids: Dict[RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]]

    @property
    def all_reco_channel_ids(self) -> Tuple[int, ...]:
        return tuple(ch_id for k in self.reco_channel_info_to_sig_ids for ch_id in k.ids)

    @property
    def all_reco_channel_group_names(self) -> Tuple[str, ...]:
        return tuple(k.name for k in self.reco_channel_info_to_sig_ids)

    @property
    def all_tag_cal_reco_channel_group_names(self) -> Tuple[str, ...]:
        _map: Dict[str, str] = self.get_reco_ch_name_to_tag_cal_reco_ch_name_map()
        return tuple(_map[k.name] for k in self.reco_channel_info_to_sig_ids)

    @property
    def all_mc_matching_ids(self) -> Tuple[int, ...]:
        return tuple(
            set(
                sig_id
                for mc_colls in self.reco_channel_info_to_sig_ids.values()
                for mc_coll in mc_colls
                for sig_id in mc_coll.sig_ids
            )
        )

    @property
    def all_mc_matching_group_names(self) -> Tuple[str, ...]:
        return tuple(set(coll.name for mc_colls in self.reco_channel_info_to_sig_ids.values() for coll in mc_colls))

    @staticmethod
    def get_tag_cal_reco_ch_names_to_reco_ch_name_map() -> Dict[str, str]:
        return {
            tag_cal_reco_ch_info.name: tag_cal_reco_ch_info.reco_mode_info.name
            for tag_cal_reco_ch_info in tag_fit_info.tag_cal_fit_chs
        }

    @staticmethod
    def get_reco_ch_name_to_tag_cal_reco_ch_name_map() -> Dict[str, str]:
        return {
            v: k for k, v in TagComponentToRecoChAndSigIDMapInfo.get_tag_cal_reco_ch_names_to_reco_ch_name_map().items()
        }


_tag_mc_matching_item_cb_to_d: Tuple[MCMatchingCollectionItem, ...] = (
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_Dtaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_Dlnu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_Dtaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_Dlnu"),
)
_tag_mc_matching_item_nb_to_d: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d

_tag_mc_matching_item_cb_to_ds: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d + (
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_DStaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_DSlnu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_DStaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_DSlnu"),
)
_tag_mc_matching_item_nb_to_ds: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_ds

_tag_mc_matching_item_cb_to_d_bkg: Tuple[MCMatchingCollectionItem, ...] = (
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="DSSpi0lnu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="DSSpiplnu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="HadronicTau_HBKG"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Lepton_GBKG"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="GBKG"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="CBKG"),
)
_tag_mc_matching_item_nb_to_d_bkg: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d_bkg
_tag_mc_matching_item_cb_to_ds_bkg: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d_bkg
_tag_mc_matching_item_nb_to_ds_bkg: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d_bkg

_tag_mc_matching_item_cb_to_d_ds_df: Tuple[MCMatchingCollectionItem, ...] = (
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_DStaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Charged_B_To_DSlnu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_DStaunu"),
    MCMatchingCollections.extended_sig_id__charge__collection.get_group_definition(name="Neutral_B_To_DSlnu"),
)
_tag_mc_matching_item_nb_to_d_ds_df: Tuple[MCMatchingCollectionItem, ...] = _tag_mc_matching_item_cb_to_d_ds_df


tag_fit_component_Bp_Dwwos_l_ch_and_mc_id_mapping: Dict[RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]] = {
    RecoModeInfos.charged_b_to_d_modes: _tag_mc_matching_item_cb_to_d,
    RecoModeInfos.charged_b_to_ds_modes: _tag_mc_matching_item_cb_to_ds,
}
tag_fit_component_Bz_Dwwos_l_ch_and_mc_id_mapping: Dict[RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]] = {
    RecoModeInfos.neutral_b_to_d_modes: _tag_mc_matching_item_nb_to_d,
    RecoModeInfos.neutral_b_to_ds_modes: _tag_mc_matching_item_nb_to_ds,
}
tag_fit_component_charged_b_to_d_bkg_ch_and_mc_id_mapping: Dict[
    RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]
] = {
    RecoModeInfos.charged_b_to_d_modes: _tag_mc_matching_item_cb_to_d_bkg,
}
tag_fit_component_neutral_b_to_d_bkg_ch_and_mc_id_mapping: Dict[
    RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]
] = {
    RecoModeInfos.neutral_b_to_d_modes: _tag_mc_matching_item_nb_to_d_bkg,
}
tag_fit_component_charged_b_to_ds_bkg_ch_and_mc_id_mapping: Dict[
    RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]
] = {
    RecoModeInfos.charged_b_to_ds_modes: _tag_mc_matching_item_cb_to_ds_bkg,
}
tag_fit_component_neutral_b_to_ds_bkg_ch_and_mc_id_mapping: Dict[
    RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]
] = {
    RecoModeInfos.neutral_b_to_ds_modes: _tag_mc_matching_item_nb_to_ds_bkg,
}
tag_fit_component_Bp_DzS_l_df_ch_and_mc_id_mapping: Dict[RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]] = {
    RecoModeInfos.charged_b_to_d_modes: _tag_mc_matching_item_cb_to_d_ds_df,
}
tag_fit_component_Bz_DmS_l_df_ch_and_mc_id_mapping: Dict[RecoModeGroupInfo, Tuple[MCMatchingCollectionItem, ...]] = {
    RecoModeInfos.neutral_b_to_d_modes: _tag_mc_matching_item_nb_to_d_ds_df,
}


tag_cal_fit_component_to_reco_ch_and_sig_id_tuple: Tuple[TagComponentToRecoChAndSigIDMapInfo, ...] = (
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_Bp_Dwwos_l.name,
        reco_channel_info_to_sig_ids=tag_fit_component_Bp_Dwwos_l_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_Bz_Dwwos_l.name,
        reco_channel_info_to_sig_ids=tag_fit_component_Bz_Dwwos_l_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_charged_b_to_d_bkg.name,
        reco_channel_info_to_sig_ids=tag_fit_component_charged_b_to_d_bkg_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_neutral_b_to_d_bkg.name,
        reco_channel_info_to_sig_ids=tag_fit_component_neutral_b_to_d_bkg_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_charged_b_to_ds_bkg.name,
        reco_channel_info_to_sig_ids=tag_fit_component_charged_b_to_ds_bkg_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_neutral_b_to_ds_bkg.name,
        reco_channel_info_to_sig_ids=tag_fit_component_neutral_b_to_ds_bkg_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_Bp_DzS_l_df.name,
        reco_channel_info_to_sig_ids=tag_fit_component_Bp_DzS_l_df_ch_and_mc_id_mapping,
    ),
    TagComponentToRecoChAndSigIDMapInfo(
        component_name=tag_fit_info.component_Bz_DmS_l_df.name,
        reco_channel_info_to_sig_ids=tag_fit_component_Bz_DmS_l_df_ch_and_mc_id_mapping,
    ),
)

tag_cal_fit_component_to_reco_ch_and_sig_id_mapping: Dict[str, TagComponentToRecoChAndSigIDMapInfo] = {
    v.component_name: v for v in tag_cal_fit_component_to_reco_ch_and_sig_id_tuple
}

bf_infos_for_uncerts: Dict[str, ufloat] = {
    "Neutral_B_To_Dlnu": bfs_and_consts.bf_b0_to_d_l_nu_pdg_eval.ufloat,
    "Charged_B_To_Dlnu": bfs_and_consts.bf_bp_to_d_l_nu_pdg_eval.ufloat,
    "Neutral_B_To_DSlnu": bfs_and_consts.bf_b0_to_d_st_l_nu_pdg_eval.ufloat,
    "Charged_B_To_DSlnu": bfs_and_consts.bf_bp_to_d_st_l_nu_pdg_eval.ufloat,
}

bf_relative_uncerts: Dict[str, float] = {k: v.std_dev / v.nominal_value for k, v in bf_infos_for_uncerts.items()}

# endregion


# region TagCalEvalResult
class NumpyNDArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class TaggedEventsPerChannelAndComponentCalculator:
    def __init__(
        self,
        df: pd.DataFrame,
        weight_column_name: str,
        reco_mode_column_name: str,
        reco_channels_info: Tuple[RecoChannelInfo, ...],
        components_mc_matching_collection: MCMatchingCollection,
    ) -> None:
        self._data_frame: pd.DataFrame = df

        self.weight_column_name: str = weight_column_name
        self.reco_mode_column_name: str = reco_mode_column_name

        self._reco_ch_info: Tuple[RecoChannelInfo, ...] = reco_channels_info
        self._components_info: MCMatchingCollection = components_mc_matching_collection

        self._base_dict: Dict[str, Dict[str, Tuple[float, float]]] = self.calculate_sow_dict()
        self.sum_of_weights_per_channel_per_component: Dict[str, Dict[str, float]] = {
            k: {kk: vv[0] for kk, vv in v.items()} for k, v in self._base_dict.items()
        }
        self.sum_of_squared_weights_per_channel_per_component: Dict[str, Dict[str, float]] = {
            k: {kk: vv[1] for kk, vv in v.items()} for k, v in self._base_dict.items()
        }

    def calculate_sow_dict(self) -> Dict[str, Dict[str, Tuple[float, float]]]:
        sig_id_col: str = self._components_info.sig_id_col_name

        per_ch_per_comp_dict: Dict[str, Dict[str, Tuple[float, float]]] = {}

        for ch_info in self._reco_ch_info:
            assert ch_info.name not in per_ch_per_comp_dict, (ch_info.name, list(per_ch_per_comp_dict.keys()))

            ch_mask: np.ndarray = ch_info.get_mask(
                df=self._data_frame,
                reco_mode_column_str=self.reco_mode_column_name,
            )

            per_comp_dict: Dict[str, Tuple[float, float]] = {}
            for comp_info in self._components_info.collection_definition:
                assert comp_info.name not in per_comp_dict, (comp_info.name, list(per_comp_dict.keys()))

                comp_mask: np.ndarray = self._data_frame[sig_id_col].values == comp_info.new_id

                _sum_of_weights = np.sum(self._data_frame[ch_mask & comp_mask][self.weight_column_name])
                assert isinstance(_sum_of_weights, float), (_sum_of_weights, type(_sum_of_weights))
                sum_of_weights: float = _sum_of_weights

                _sum_of_sq_weights = np.sum(np.square(self._data_frame[ch_mask & comp_mask][self.weight_column_name]))
                assert isinstance(_sum_of_sq_weights, float), (_sum_of_sq_weights, type(_sum_of_sq_weights))
                sum_of_squared_weights: float = _sum_of_sq_weights

                per_comp_dict[comp_info.name] = (sum_of_weights, sum_of_squared_weights)

            per_ch_per_comp_dict[ch_info.name] = per_comp_dict

        return per_ch_per_comp_dict


class TagCalEvalResult:
    def __init__(
        self,
        selection_name: str,
        fit_scenario_name: str,
        fit_obs_setup_tag: str,
        initial_parameters: np.ndarray,
        cov_mat: np.ndarray,
        corr_mat: np.ndarray,
        res_values: np.ndarray,
        res_errors: np.ndarray,
        param_names_mapping: Dict[int, str],
        component_mapping: Dict[int, str],
        sum_of_weights_per_comp_per_channel: Dict[str, Dict[str, float]],
        sum_of_squared_weights_per_comp_per_channel: Dict[str, Dict[str, float]],
        sow_mc_matching_collection_name: str,
    ) -> None:
        self.selection_name: str = selection_name
        self.fit_scenario_name: str = fit_scenario_name
        self.fit_obs_setup_tag: str = fit_obs_setup_tag

        self.initial_parameters: np.ndarray = initial_parameters

        self.cov_mat: np.ndarray = cov_mat
        self.corr_mat: np.ndarray = corr_mat
        self.res_values: np.ndarray = res_values
        self.res_errors: np.ndarray = res_errors
        self.param_names_mapping: Dict[int, str] = param_names_mapping
        self.component_mapping: Dict[int, str] = component_mapping

        self.sum_of_weights_per_comp_per_channel: Dict[str, Dict[str, float]] = sum_of_weights_per_comp_per_channel
        self.sum_of_squared_weights_per_comp_per_channel: Dict[
            str, Dict[str, float]
        ] = sum_of_squared_weights_per_comp_per_channel
        self.sow_mc_matching_collection_name: str = sow_mc_matching_collection_name

        self.inverted_param_names_mapping: Dict[str, int] = {v: i for i, v in self.param_names_mapping.items()}
        self.inverted_component_mapping: Dict[str, int] = {v: i for i, v in self.component_mapping.items()}

    @property
    def data_over_mc_ratio_array(self) -> unp.uarray:
        fit_res_array = unp.uarray(nominal_values=self.res_values, std_devs=self.res_errors)
        return fit_res_array / self.initial_parameters

    def get_latex_table_col_str_array(self, precision: int) -> np.ndarray:
        return np.array([f"${_v:.{precision}L}$" for _v in self.data_over_mc_ratio_array])

    def get_latex_table_col_str(self, precision: int) -> str:
        return "\n".join([f"${_v:.{precision}L}$" for _v in self.data_over_mc_ratio_array])

    @property
    def params_names_col_str(self) -> str:
        return "\n".join(self.param_names_mapping.values())

    @property
    def params_names_latex_labels_array(self) -> Tuple[str, ...]:
        tag_fit_setup = self.fit_scenario.this_fitter_setup
        assert tag_fit_setup is not None
        return tuple(tag_fit_setup.asimov.components[i].latex_label for i in self.param_names_mapping.keys())

    def as_dict(self) -> Dict[str, Any]:
        return {
            "selection_name": self.selection_name,
            "fit_scenario_name": self.fit_scenario_name,
            "fit_obs_setup_tag": self.fit_obs_setup_tag,
            "initial_parameters": self.initial_parameters,
            "cov_mat": self.cov_mat,
            "corr_mat": self.corr_mat,
            "res_values": self.res_values,
            "res_errors": self.res_errors,
            "inverted_param_names_mapping": self.inverted_param_names_mapping,
            "inverted_component_mapping": self.inverted_component_mapping,
            "sum_of_weights_per_comp_per_channel": self.sum_of_weights_per_comp_per_channel,
            "sum_of_squared_weights_per_comp_per_channel": self.sum_of_squared_weights_per_comp_per_channel,
            "sow_mc_matching_collection_name": self.sow_mc_matching_collection_name,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "TagCalEvalResult":
        _inverted_param_names_mapping = input_dict["inverted_param_names_mapping"]
        assert isinstance(_inverted_param_names_mapping, dict), type(_inverted_param_names_mapping).__name__
        _inverted_component_mapping = input_dict["inverted_component_mapping"]
        assert isinstance(_inverted_component_mapping, dict), type(_inverted_component_mapping).__name__
        _param_names_mapping: Dict[int, str] = {int(v): i for i, v in _inverted_param_names_mapping.items()}
        _component_mapping: Dict[int, str] = {int(v): i for i, v in _inverted_component_mapping.items()}

        sow_per_comp_per_ch: Dict[str, Dict[str, float]] = input_dict.get("sum_of_weights_per_comp_per_channel", {})
        sosqw_per_comp_per_ch: Dict[str, Dict[str, float]] = input_dict.get(
            "sum_of_squared_weights_per_comp_per_channel", {}
        )
        sow_mc_matching_collection_name: str = input_dict.get("sow_mc_matching_collection_name", "NONE")

        return TagCalEvalResult(
            selection_name=input_dict["selection_name"],
            fit_scenario_name=input_dict["fit_scenario_name"],
            fit_obs_setup_tag=input_dict["fit_obs_setup_tag"],
            initial_parameters=np.asarray(input_dict["initial_parameters"]),
            cov_mat=np.asarray(input_dict["cov_mat"]),
            corr_mat=np.asarray(input_dict["corr_mat"]),
            res_values=np.asarray(input_dict["res_values"]),
            res_errors=np.asarray(input_dict["res_errors"]),
            param_names_mapping=_param_names_mapping,
            component_mapping=_component_mapping,
            sum_of_weights_per_comp_per_channel=sow_per_comp_per_ch,
            sum_of_squared_weights_per_comp_per_channel=sosqw_per_comp_per_ch,
            sow_mc_matching_collection_name=sow_mc_matching_collection_name,
        )

    def save_to_file(self, file_path: PathType) -> None:
        with open(file_path, "w") as fp:
            json.dump(obj=self.as_dict(), fp=fp, indent=2, cls=NumpyNDArrayEncoder)

    @staticmethod
    def load_from_file(file_path: PathType) -> "TagCalEvalResult":
        with open(file_path) as f:
            _loaded_dict = json.load(f)
        assert isinstance(_loaded_dict, dict), type(_loaded_dict).__name__
        return TagCalEvalResult.from_dict(input_dict=_loaded_dict)

    @property
    def fit_scenario(self) -> Type[TagCorrFitterBase]:
        return tc_fitter_dict[self.fit_scenario_name]

    @staticmethod
    def compare_results(
        dir_path: PathType,
        first_eval_name: str,
        first_scenario_name: str,
        first_selection_name: str,
        first_obs_setup_tag: str,
        second_eval_name: str,
        second_scenario_name: str,
        second_selection_name: str,
        second_obs_setup_tag: str,
        precision: int = 2,
    ) -> pd.DataFrame:
        first_file_name: str = f"{first_scenario_name}__{first_selection_name}__{first_obs_setup_tag}.json"
        second_file_name: str = f"{second_scenario_name}__{second_selection_name}__{second_obs_setup_tag}.json"

        first_file_path: PathType = os.path.join(dir_path, first_eval_name, first_file_name)
        second_file_path: PathType = os.path.join(dir_path, second_eval_name, second_file_name)

        assert os.path.isfile(first_file_path), first_file_path
        assert os.path.isfile(second_file_path), second_file_path

        first_result_info = TagCalEvalResult.load_from_file(file_path=first_file_path)
        second_result_info = TagCalEvalResult.load_from_file(file_path=second_file_path)

        first_df = pd.DataFrame(
            {
                first_scenario_name: first_result_info.params_names_latex_labels_array,
                first_selection_name: first_result_info.get_latex_table_col_str_array(precision=precision),
            }
        )
        second_df = pd.DataFrame(
            {
                second_scenario_name: second_result_info.params_names_latex_labels_array,
                second_selection_name: second_result_info.get_latex_table_col_str_array(precision=precision),
            }
        )
        return pd.concat([first_df, second_df], axis=1)

    def get_mc_sample_with_correction_for(
        self,
        correction_info_mapping: Dict[Tuple[Tuple[int, ...], Tuple[int, ...]], int],
        df: pd.DataFrame,
        apply_for_continuum: bool = True,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        _sample = copy.deepcopy(df)

        assert self.fit_scenario.this_fitter_setup is not None
        reco_id_col: str = self.fit_scenario.this_fitter_setup.asimov.reco_mode_col
        component_id_col: str = self.fit_scenario.this_fitter_setup.asimov.component_id_column_name
        weight_col: str = self.fit_scenario.this_fitter_setup.asimov.weight_info.col_name

        if component_id_col not in _sample.columns:
            self.fit_scenario.this_fitter_setup.asimov.prepare_mc_df(df=_sample)

        if reco_id_col == TagRecoModeInfos.fei_reco_mode_col_name and reco_id_col not in _sample.columns:
            TagRecoModeInfos.add_new_fei_reco_mode_col(df=df, check_if_new=True)

        if matching_tag_selection is None:
            tag_match_mask: np.ndarray = np.full(shape=len(_sample), fill_value=True)
        elif matching_tag_selection is True:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(U4SRecoModeInfos.u4s_from_all_correct_modes.ids)
        elif matching_tag_selection is False:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(
                U4SRecoModeInfos.u4s_from_all_wrong_tag_modes.ids
            )
        else:
            raise RuntimeError(f"Invalid option 'matching_tag_selection' = {matching_tag_selection}!")

        test_mask: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        for (reco_ids, comp_sig_ids), ratio_index in correction_info_mapping.items():
            reco_mask: np.ndarray = _sample[reco_id_col].isin(reco_ids)
            comp_mask: np.ndarray = _sample[component_id_col].isin(comp_sig_ids)

            not_continuum: np.ndarray = np.equal(_sample[Cols.is_continuum_evt].values, np.zeros(shape=comp_mask.shape))

            if apply_for_continuum:
                mask: np.ndarray = reco_mask & comp_mask & tag_match_mask
            else:
                mask = reco_mask & comp_mask & tag_match_mask & not_continuum

            scale_factor: float = self.data_over_mc_ratio_array[ratio_index].nominal_value
            _sample.loc[mask, weight_col] *= scale_factor

            assert not np.any(test_mask & mask), np.sum(test_mask & mask)
            test_mask |= mask

        return _sample

    def get_mc_sample_with_correction(
        self,
        df: pd.DataFrame,
        apply_for_continuum: bool = True,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        return self.get_mc_sample_with_correction_for(
            correction_info_mapping=self.fit_scenario.get_correction_mapping(),
            df=df,
            apply_for_continuum=apply_for_continuum,
            matching_tag_selection=matching_tag_selection,
        )

    def get_mc_sample_with_full_correction(
        self,
        df: pd.DataFrame,
        apply_for_continuum: bool = True,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        return self.get_mc_sample_with_correction_for(
            correction_info_mapping=self.fit_scenario.get_full_correction_mapping(),
            df=df,
            apply_for_continuum=apply_for_continuum,
            matching_tag_selection=matching_tag_selection,
        )

    def get_mc_sample_with_simple_correction_for_bkg_only(
        self,
        df: pd.DataFrame,
        apply_for_continuum: bool = True,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        return self.get_mc_sample_with_correction_for(
            correction_info_mapping=self.fit_scenario.get_correction_mapping_for_bkg(),
            df=df,
            apply_for_continuum=apply_for_continuum,
            matching_tag_selection=matching_tag_selection,
        )

    def get_rdstar_fei_correction_uncertainties_dict(self) -> Dict[str, ufloat]:
        allowed_fit_scenario_names: Tuple[str, ...] = ("tc_main",)
        assert self.fit_scenario_name in allowed_fit_scenario_names, (self.fit_scenario_name, allowed_fit_scenario_names)

        assert self.sow_mc_matching_collection_name != "NONE"
        assert self.sum_of_weights_per_comp_per_channel

        tag_cal_comp_to_param_index: Dict[str, int] = {
            tag_fit_info.component_Bp_Dwwos_l.name: MainTagCalFitter.bp_to_dwwos_l_yield_index,
            tag_fit_info.component_Bz_Dwwos_l.name: MainTagCalFitter.bz_to_dwwos_l_yield_index,
            tag_fit_info.component_charged_b_to_d_bkg.name: MainTagCalFitter.charged_b_to_d_bkg_yield_index,
            tag_fit_info.component_neutral_b_to_d_bkg.name: MainTagCalFitter.neutral_b_to_d_bkg_yield_index,
            tag_fit_info.component_charged_b_to_ds_bkg.name: MainTagCalFitter.charged_b_to_ds_bkg_yield_index,
            tag_fit_info.component_neutral_b_to_ds_bkg.name: MainTagCalFitter.neutral_b_to_ds_bkg_yield_index,
            tag_fit_info.component_Bp_DzS_l_df.name: MainTagCalFitter.bp_to_dsl_df_yield_index,
            tag_fit_info.component_Bz_DmS_l_df.name: MainTagCalFitter.bz_to_dsl_df_yield_index,
        }

        tag_cal_ch_name_map: Dict[
            str, str
        ] = TagComponentToRecoChAndSigIDMapInfo.get_reco_ch_name_to_tag_cal_reco_ch_name_map()

        res_dict: Dict[str, ufloat] = {}

        fit_res_array = unp.uarray(nominal_values=self.res_values, std_devs=self.res_errors)
        initial_params_array = self.initial_parameters

        for tag_cal_component_name, cal_info in tag_cal_fit_component_to_reco_ch_and_sig_id_mapping.items():
            result_index: int = tag_cal_comp_to_param_index[tag_cal_component_name]

            fit_result: ufloat = fit_res_array[result_index]
            initial_value: float = initial_params_array[result_index]

            sum_of_weights: float = 0.0

            for expected_channel, expected_mc_matching in cal_info.reco_channel_info_to_sig_ids.items():
                expected_tag_cal_fit_ch_name: str = tag_cal_ch_name_map[expected_channel.name]
                expected_mc_group_names: Tuple[str, ...] = tuple(v.name for v in expected_mc_matching)

                sum_of_weights += sum(
                    [
                        sow
                        for ch_name, v in self.sum_of_weights_per_comp_per_channel.items()
                        for mc_matching_group_name, sow in v.items()
                        if mc_matching_group_name in expected_mc_group_names and ch_name == expected_tag_cal_fit_ch_name
                    ]
                )

            logger.warning(
                f"{np.isclose(initial_value, sum_of_weights, rtol=0.01, atol=5)}, "
                f"{(tag_cal_component_name, initial_value, sum_of_weights)}"
            )
            assert np.isclose(initial_value, sum_of_weights, rtol=0.01, atol=5), (
                tag_cal_component_name,
                initial_value,
                sum_of_weights,
            )

            sum_of_weights_var_up = 0.0
            sum_of_weights_var_down = 0.0

            for reco_info, comp_infos in cal_info.reco_channel_info_to_sig_ids.items():
                reco_ch_name_requirement: str = tag_cal_ch_name_map[reco_info.name]
                comp_requirement: Tuple[str, ...] = tuple(c.name for c in comp_infos)

                _sow_up_list: List[float] = [
                    sow * (1.0 + bf_relative_uncerts[comp_name]) if comp_name in bf_relative_uncerts else sow
                    for comp_name, sow in self.sum_of_weights_per_comp_per_channel[reco_ch_name_requirement].items()
                    if comp_name in comp_requirement
                ]

                _sow_down_list: List[float] = [
                    sow * (1.0 - bf_relative_uncerts[comp_name]) if comp_name in bf_relative_uncerts else sow
                    for comp_name, sow in self.sum_of_weights_per_comp_per_channel[reco_ch_name_requirement].items()
                    if comp_name in comp_requirement
                ]

                sum_of_weights_var_up += np.sum(_sow_up_list)
                sum_of_weights_var_down += np.sum(_sow_down_list)

            uncert_from_sow: float = 0.5 * (sum_of_weights_var_up - sum_of_weights_var_down)

            initial_value_with_uncert: ufloat = ufloat(initial_value, uncert_from_sow)

            nominal_ratio: ufloat = fit_result / initial_value_with_uncert

            res_dict[tag_cal_component_name] = nominal_ratio

        return res_dict

    def get_mc_sample_with_simple_correction(
        self,
        df: pd.DataFrame,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        _sample = copy.deepcopy(df)

        assert self.fit_scenario.this_fitter_setup is not None
        reco_id_col: str = self.fit_scenario.this_fitter_setup.asimov.reco_mode_col
        weight_col: str = self.fit_scenario.this_fitter_setup.asimov.weight_info.col_name

        if reco_id_col == TagRecoModeInfos.fei_reco_mode_col_name and reco_id_col not in _sample.columns:
            TagRecoModeInfos.add_new_fei_reco_mode_col(df=df, check_if_new=True)

        if matching_tag_selection is None:
            tag_match_mask: np.ndarray = np.full(shape=len(_sample), fill_value=True)
        elif matching_tag_selection is True:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(U4SRecoModeInfos.u4s_from_all_correct_modes.ids)
        elif matching_tag_selection is False:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(
                U4SRecoModeInfos.u4s_from_all_wrong_tag_modes.ids
            )
        else:
            raise RuntimeError(f"Invalid option 'matching_tag_selection' = {matching_tag_selection}!")

        test_mask: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        for reco_ids, ratio_index in self.fit_scenario.get_simple_correction_mapping().items():
            reco_mask: np.ndarray = _sample[reco_id_col].isin(reco_ids)

            mask: np.ndarray = reco_mask & tag_match_mask

            scale_factor: float = self.data_over_mc_ratio_array[ratio_index].nominal_value
            _sample.loc[mask, weight_col] *= scale_factor

            assert not np.any(test_mask & mask), np.sum(test_mask & mask)
            test_mask |= mask

        return _sample


# endregion


# region TagCalFitEvaluator
class TagCalFitEvaluator:

    selections: Tuple[str, ...] = (
        "main_with_mva_selection_sec_sel_mva",
        # "main_selection",
        # "optimized_mva_selection_sec_sel_mva",
        # "loose_mva_selection_sec_sel_mva",
        # "first_selection",
    )

    u4s_combinations_selection: Dict[str, Optional[SimpleCutInfo]] = {
        # "all_Y4S_combinations": None,
        # "with_wrong_sig_tag_comb": wrong_sig_tag_combination_selection,
        "with_right_sig_tag_comb": right_sig_tag_combination_selection,
    }

    post_fit_plot_infos: Tuple[SidebandInfo, ...] = post_fit_plot_infos
    eval_plot_infos: Tuple[SidebandInfo, ...] = eval_plot_infos

    def __init__(
        self,
        evaluation_name: str,
        base_input_path: PathType,
        base_output_path: PathType,
        fit_obs_setup_tag: str,
        produce_post_fit_plots: bool = True,
        produce_evaluation_plots: bool = True,
        produce_plots_with_tag_cal_reco_modes: bool = False,
        fix_backgrounds: bool = False,
        error_definition_dict: Optional[Dict[str, ufloat]] = None,
        mc_sample_preparation_functions: Tuple[Callable, ...] = (),
    ) -> None:

        assert os.path.isdir(base_input_path), base_input_path
        self.evaluation_name: str = evaluation_name
        self.base_input_path: PathType = base_input_path
        self.base_output_path: PathType = base_output_path
        self.fit_obs_setup_tag: str = fit_obs_setup_tag

        self.produce_post_fit_plots: bool = produce_post_fit_plots
        self.produce_evaluation_plots: bool = produce_evaluation_plots
        self.produce_plots_with_tag_cal_reco_modes: bool = produce_plots_with_tag_cal_reco_modes

        self.fix_backgrounds: bool = fix_backgrounds
        self.error_definition_dict: Optional[Dict[str, ufloat]] = error_definition_dict
        self.mc_sample_preparation_functions: Tuple[Callable, ...] = mc_sample_preparation_functions

        self._mc_sample: Optional[pd.DataFrame] = None

        self.output_path: PathType = os.path.join(self.base_output_path, self.evaluation_name)

        self.fit_scenario_dict: Dict[str, Type[TagCorrFitterBase]] = copy.deepcopy(tc_fitter_dict)

    @property
    def mc_sample(self) -> pd.DataFrame:
        assert self._mc_sample is not None
        return self._mc_sample

    @mc_sample.setter
    def mc_sample(self, df: pd.DataFrame) -> None:
        assert self._mc_sample is None
        self._mc_sample = df

    def reset_mc_sample(self, df: pd.DataFrame) -> None:
        self._mc_sample = df

    def run_evaluation(self) -> None:
        assert os.path.isdir(self.base_output_path), self.base_output_path
        os.makedirs(self.output_path, exist_ok=True)

        logger.info("Starting evaluation!")
        for selection_name in self.selections:
            logger.info(f"\tEvaluating selection {selection_name}")
            self.evaluate_selection(selection_name=selection_name)

    def evaluate_selection(
        self,
        selection_name: str,
    ) -> None:
        logger.info("\t\tloading mc data...")
        self.reset_mc_sample(df=self.load_mc(selection_name=selection_name))
        logger.info("\t\tloading recorded data...")
        data_sample: pd.DataFrame = self.load_data(selection_name=selection_name)

        TagRecoModeInfos.add_new_fei_reco_mode_col(df=data_sample, check_if_new=True)

        for fit_scenario_name in self.fit_scenario_dict:
            if keys_to_evaluate and fit_scenario_name not in keys_to_evaluate:
                continue

            logger.info(f"\t\t- fit_scenario {fit_scenario_name}")

            for u4s_sel_name, u4s_sel in self.u4s_combinations_selection.items():
                logger.info(f"\t\t- fit_scenario {fit_scenario_name} with {u4s_sel_name} Y(4S) Selection")

                _u4s_mc_sample: pd.DataFrame = copy.deepcopy(self.mc_sample)
                _u4s_data_sample: pd.DataFrame = copy.deepcopy(data_sample)

                if u4s_sel is not None:
                    u4s_sel.apply_cut(df=_u4s_mc_sample)
                    u4s_sel.apply_cut(df=_u4s_data_sample)

                    _u4s_mc_sample = _u4s_mc_sample.reset_index(drop=True)
                    _u4s_data_sample = _u4s_data_sample.reset_index(drop=True)

                self.evaluate_scenario(
                    mc_sample_copy=_u4s_mc_sample,
                    data_sample=_u4s_data_sample,
                    selection_name=f"{selection_name}_{u4s_sel_name}",
                    fit_scenario_name=fit_scenario_name,
                )

    def evaluate_scenario(
        self,
        mc_sample_copy: pd.DataFrame,
        data_sample: pd.DataFrame,
        selection_name: str,
        fit_scenario_name: str,
    ) -> None:
        tag_corr_fitter: Type[TagCorrFitterBase] = self.fit_scenario_dict[fit_scenario_name]
        assert tag_corr_fitter.this_fitter_setup is not None
        fit_scenario_triplet: FitSetupTriplet = tag_corr_fitter.this_fitter_setup
        data_fit_setup: FitSetupInfoContainer = fit_scenario_triplet.data

        logger.info("\t\t\tpreparing mc sample...")
        data_fit_setup.prepare_mc_df(
            df=mc_sample_copy,
            fail_if_already_prepared=False,
        )

        logger.info("\t\t\tsetting up fit...")
        fitter: TagCorrFitterBase = tag_corr_fitter(
            fit_setup_info=data_fit_setup,
            mc_df=mc_sample_copy,
            data_df=data_sample,
        )

        initial_parameters: np.ndarray = np.array(list(fitter.initial_parameters))

        logger.info("\t\t\tfitting...")
        m = Minuit(
            fitter.cost_fcn,
            initial_parameters,
            name=tuple([str(i) for i in range(len(list(fitter.initial_parameters)))]),
        )

        fitter.set_iminuit_properties(minuit_instance=m)

        if self.fix_backgrounds:
            for bkg_index in fitter.get_background_indices():
                m.fixed[bkg_index] = True

                if self.error_definition_dict is not None:
                    try:
                        bkg_comp_name = fitter.fit_setup_info.components[bkg_index].name
                        uncert_values: ufloat = self.error_definition_dict[bkg_comp_name]
                        relative_error = uncert_values.std_dev / uncert_values.nominal_value
                        m.errors[bkg_index] = fitter.initial_parameters[bkg_index] * relative_error
                    except KeyError:
                        pass

        data_fmin: mutil.FMin = (
            m.simplex()
            .migrad(ncall=fitter.minuit_settings.n_call, iterate=fitter.minuit_settings.n_iterate)
            .minos()
            .hesse()
            .fmin
        )

        assert data_fmin.is_valid and data_fmin.has_valid_parameters and data_fmin.has_covariance
        logger.info("\t\t\tDone! Saving now...")

        eval_result: TagCalEvalResult = self.generate_result_output(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            fitter=fitter,
            minimizer=m,
            initial_parameters=initial_parameters,
        )

        self.save_results(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            result=eval_result,
        )

        if self.produce_post_fit_plots:
            logger.info("\t\t\tProducing post fit plots...")
            self.produce_plots_for(
                plot_infos=self.post_fit_plot_infos,
                selection_name=selection_name,
                fit_scenario_name=fit_scenario_name,
                mc_sample=mc_sample_copy,
                data_sample=data_sample,
                fitter=fitter,
                minimizer=m,
                produce_plots_with_tag_cal_reco_modes=True,
            )
        else:
            logger.info("\t\t\tSkipping post fit plots!")
        if self.produce_evaluation_plots:
            logger.info("\t\t\tProducing evaluation plots...")
            self.produce_plots_for(
                plot_infos=self.eval_plot_infos,
                selection_name=selection_name,
                fit_scenario_name=fit_scenario_name,
                mc_sample=mc_sample_copy,
                data_sample=data_sample,
                fitter=fitter,
                minimizer=m,
                produce_plots_with_tag_cal_reco_modes=False,
            )
        else:
            logger.info("\t\t\tSkipping evaluation plots!")

    def generate_result_output(
        self,
        selection_name: str,
        fit_scenario_name: str,
        fitter: TagCorrFitterBase,
        minimizer: Minuit,
        initial_parameters: np.ndarray,
    ) -> TagCalEvalResult:

        assert np.all(initial_parameters == fitter.initial_parameters), (initial_parameters, fitter.initial_parameters)

        sow_calculator = TaggedEventsPerChannelAndComponentCalculator(
            df=fitter.internal_mc_sample_df,
            weight_column_name=fitter.mc_weight_col_name,
            reco_mode_column_name=fitter.fit_setup_info.reco_mode_col,
            reco_channels_info=fitter.fit_setup_info.reco_channels,
            components_mc_matching_collection=MCMatchingCollections.extended_sig_id__charge__collection,
        )

        return TagCalEvalResult(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            fit_obs_setup_tag=self.fit_obs_setup_tag,
            initial_parameters=initial_parameters,
            cov_mat=np.array(minimizer.covariance),
            corr_mat=np.array(minimizer.covariance.correlation()),
            res_values=np.array(minimizer.values),
            res_errors=np.array(minimizer.errors),
            param_names_mapping=fitter.parameter_name_mapping(),
            component_mapping=fitter.fit_setup_info.components_name_to_id_mapping,
            sum_of_weights_per_comp_per_channel=sow_calculator.sum_of_weights_per_channel_per_component,
            sum_of_squared_weights_per_comp_per_channel=sow_calculator.sum_of_squared_weights_per_channel_per_component,
            sow_mc_matching_collection_name=MCMatchingCollections.extended_sig_id__charge__collection.collection_name,
        )

    def produce_plots_for(
        self,
        plot_infos: Tuple[SidebandInfo, ...],
        selection_name: str,
        fit_scenario_name: str,
        mc_sample: pd.DataFrame,
        data_sample: pd.DataFrame,
        fitter: TagCorrFitterBase,
        minimizer: Minuit,
        produce_plots_with_tag_cal_reco_modes: bool,
    ) -> None:
        output_name: str = f"{fit_scenario_name}__{selection_name}__{self.fit_obs_setup_tag}"
        output_path: PathType = os.path.join(self.output_path, output_name)

        os.makedirs(output_path, exist_ok=True)

        use_tag_cal_reco_mode_options: Tuple[bool, ...] = (False,)
        if produce_plots_with_tag_cal_reco_modes and self.produce_plots_with_tag_cal_reco_modes:
            use_tag_cal_reco_mode_options = (False, True)

        rescaled_mc_sample = fitter.get_rescaled_mc_sample_with(
            sample=mc_sample,
            weight_col_name=fitter.mc_weight_col_name,
            parameters=np.array(minimizer.values),
            add_required_sig_id_col=False,
        )

        for sb_info in plot_infos:
            if sb_info.file_name_appendix is None:
                sb_plot_dir_name: str = sb_info.hist_var.df_label
            else:
                sb_plot_dir_name = f"{sb_info.hist_var.df_label}_{sb_info.file_name_appendix}"

            sb_plot_output_path: PathType = os.path.join(output_path, sb_plot_dir_name)
            if os.path.isdir(sb_plot_output_path):
                return

            os.makedirs(sb_plot_output_path, exist_ok=True)

            for use_sys_for_plot in (True, False):
                if use_sys_for_plot:
                    sb_info_with_sys = SidebandInfo.create_comparer_infos_with_systematics(
                        comparer_info_list=[sb_info],
                        systematics_collection=full_systematics_cols,
                        sys_name_tag="with_sys",
                    )[0]
                else:
                    sb_info_with_sys = sb_info

                for use_tag_reco_mode in use_tag_cal_reco_mode_options:
                    if use_tag_reco_mode:
                        if sb_info_with_sys.file_name_appendix is None:
                            continue

                        _sb_info_with_sys = SidebandInfo.create_modified_comparer_info(
                            comparer_info=sb_info_with_sys,
                            reco_mode_info_collection=BTagRecoModes.b_tag_by_groups_and_by_b_charge,
                            file_name_appendix=f"{sb_info_with_sys.file_name_appendix}_inFEITagModes",
                        )
                        data_mc_comparer = DataMCComparer(comparer_info=_sb_info_with_sys)
                    else:
                        data_mc_comparer = DataMCComparer(comparer_info=sb_info_with_sys)

                    data_mc_comparer.plot_data_mc_comparison(
                        mc_df=rescaled_mc_sample,
                        data_df=data_sample,
                        dry_run=False,
                        export_path=sb_plot_output_path,
                        weight_col=fitter.mc_weight_col_name,
                        reco_mode_col=TagRecoModeInfos.fei_reco_mode_col_name if use_tag_reco_mode else None,
                        style="stacked",
                        normalize_mc_to_data=False,
                        ratio_type="vs_uncert",
                        gof_check_method="toys",
                        adaptive_binning=False,
                        show_plots=False,
                    )

    def save_results(
        self,
        selection_name: str,
        fit_scenario_name: str,
        result: TagCalEvalResult,
    ) -> None:
        _file_name: str = f"{fit_scenario_name}__{selection_name}__{self.fit_obs_setup_tag}.json"
        output_tmp_path: PathType = os.path.join(self.output_path, f"tmp_{_file_name}")
        output_path: PathType = os.path.join(self.output_path, _file_name)

        result.save_to_file(file_path=output_tmp_path)
        os.rename(src=output_tmp_path, dst=output_path)

    def get_file_path_from_key(
        self,
        key: str,
        selection_name: str,
    ) -> PathType:
        return get_file_path_for(
            base_path=self.base_input_path,
            selection_name=selection_name,
            sample_key=key,
            sb_tag="_q2_sideband",
            use_hammered_samples=True,
        )

    def load_mc(self, selection_name: str) -> pd.DataFrame:
        mc_paths = [self.get_file_path_from_key(key=k, selection_name=selection_name) for k in alternative_sample_keys]
        _df = pd.concat([pd.read_hdf(mcp) for mcp in mc_paths], ignore_index=True, sort=False).reset_index(drop=True)

        TagRecoModeInfos.add_new_fei_reco_mode_col(df=_df, check_if_new=True)
        SigIDObservables.add_sig_id_observables_definitions(df=_df)

        for preparation_function in self.mc_sample_preparation_functions:
            _df = preparation_function(df=_df)

        return _df

    def load_data(self, selection_name: str) -> pd.DataFrame:
        data_path: PathType = self.get_file_path_from_key(key="DataY4S", selection_name=selection_name)
        return pd.read_hdf(data_path).reset_index(drop=True)


# endregion
