"""
Dedicated R(D) and R(D*) fitter class.

Felix Metzner 2022
"""

import os
import pandas as pd

from typing import Callable, Type, Optional

from rdstar.utilities import PathType

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos
from rdstar.offline_analysis.observables.offline_observable_definition_base import OfflineObservableBase


__all__ = [
    "immutable_cached_property",
    "get_file_base_name_for",
    "get_file_path_for",
    "generate_sig_id_application_function",
]


class immutable_cached_property:
    def __init__(self, function: Callable) -> None:
        self._function: Callable = function

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        value = self._function(obj)
        setattr(obj, self._function.__name__, value)
        return value


def get_file_base_name_for(
    selection_name: str,
    sb_tag: Optional[str],
    use_hammered_samples: bool,
) -> str:
    if sb_tag is None:
        _sb_tag: str = ""
    else:
        _sb_tag = "sb_tag" if sb_tag.startswith("_") else f"_{sb_tag}"

    file_base_name: str = "wHW_wEBGLfC_wMMSQRC" if use_hammered_samples else "withFinalWeights_withMMSQResCorr"

    return f"{file_base_name}_{selection_name}_with_CS_CS1{_sb_tag}"


def get_file_path_for(
    base_path: PathType,
    selection_name: str,
    sample_key: str,
    sb_tag: Optional[str] = None,
    use_hammered_samples: bool = True,
) -> PathType:
    if sb_tag is None:
        _sb_tag: str = ""
    else:
        _sb_tag = "sb_tag" if sb_tag.startswith("_") else f"_{sb_tag}"

    file_base_name: str = get_file_base_name_for(
        selection_name=selection_name,
        sb_tag=sb_tag,
        use_hammered_samples=use_hammered_samples,
    )

    file_name: str = f"{sample_key}_{file_base_name}.h5"

    file_path = os.path.join(base_path, file_name)
    assert os.path.exists(file_path), (sample_key, selection_name, sb_tag, _sb_tag, use_hammered_samples, file_path)
    return file_path


def generate_sig_id_application_function(
    sig_col_obs_definition_class: Type[OfflineObservableBase],
    new_col_name: str,
    add_modified_fei_reco_mode_col: bool = False,
    add_combined_tag_and_sig_reco_mode_col: bool = False,
) -> Callable:
    assert new_col_name == sig_col_obs_definition_class.get_new_column_name()

    def application_function(df: pd.DataFrame) -> None:
        assert new_col_name not in df.columns, new_col_name

        if add_modified_fei_reco_mode_col:
            if TagRecoModeInfos.fei_reco_mode_col_name not in df.columns:
                TagRecoModeInfos.add_new_fei_reco_mode_col(df=df, check_if_new=False)

        if add_combined_tag_and_sig_reco_mode_col:
            if TagRecoModeInfos.fei_plus_sig_d_reco_mode_col_name not in df.columns:
                TagRecoModeInfos.add_sig_d_plus_fei_reco_mode_col(df=df, check_if_new=False)

        new_values = sig_col_obs_definition_class.calculate_new_column_values(df=df)
        assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
        df.loc[:, sig_col_obs_definition_class.new_column_name] = new_values

        assert not pd.isnull(df[new_col_name]).any(), new_col_name

    return application_function
