"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""

import copy

from inspect import signature
from typing import Union, Optional, Tuple, List, Dict, Sequence, Any, NamedTuple

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.binned_distributions.binning import BinsInputType
from templatefitter.minimizer import available_template_fitter_minimizer

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.systematics import SysColsInfo

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType

from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import DataTypeFraction


__all__ = [
    "TemplateSetupType",
    "TemplateSetup",
    "FitVariable",
    "FitVarDictType",
    "FitCase",
]


TemplateSetupType = Union[Dict[str, List[str]], Dict[str, Tuple[str, ...]], List[str], Tuple[str, ...]]


class FitVariable(NamedTuple):
    variable: HistVariable
    alternative_binning: Optional[BinsInputType]
    alternative_scope: Optional[Tuple[float, float]]

    @property
    def binning(self) -> BinsInputType:
        if self.alternative_binning is not None:
            return self.alternative_binning
        else:
            return self.variable.n_bins

    @property
    def scope(self) -> Tuple[float, float]:
        if self.alternative_scope is not None:
            return self.alternative_scope
        else:
            return self.variable.scope


class TemplateSetup:
    def __init__(
        self,
        template_setup: TemplateSetupType,
    ) -> None:
        self.__template_setup = self.init_template_setup(template_setup_arg=template_setup)

    @staticmethod
    def init_template_setup(
        template_setup_arg: TemplateSetupType,
    ) -> TemplateSetupType:
        if isinstance(template_setup_arg, dict):
            __all_values = [v for vs in template_setup_arg.values() for v in vs]
            assert len(set(__all_values)) == len(__all_values), template_setup_arg
            return copy.copy(template_setup_arg)
        elif isinstance(template_setup_arg, (list, tuple)):
            assert len(template_setup_arg) == len(set(template_setup_arg)), template_setup_arg
            return tuple([k for k in template_setup_arg])
        else:
            raise TypeError(
                f"Argument template_setup of TemplateSetup must be of type dict, list or tuple, "
                f"but is of type {type(template_setup_arg).__name__}:\n{template_setup_arg}"
            )

    @property
    def template_setup(self) -> TemplateSetupType:
        return self.__template_setup


FitVarDictType = Dict[str, Tuple[FitVariable, ...]]


class FitCase:
    def __init__(
        self,
        *,
        name: str,
        base_dir_name: str,
        sample_key: str,
        output_tag: str,
        plot_info_text: str,
        fix_nuisance_parameters: bool,
        asimov_data: bool,
        round_asimov_data: bool,
        fit_variables_by_mode_dict: FitVarDictType,
        weight_column_name: str,
        systematics_cols: Tuple[SysColsInfo, ...],
        reco_mode_infos: Sequence[Tuple[str, Sequence[int]]],
        template_setup: Union[TemplateSetup, Dict[str, TemplateSetup]],
        sig_id_to_data_type_mapping: Dict[int, Tuple[DataTypeFraction, ...]],
        plot_colors: Dict[str, str],
        plot_labels: Dict[str, str],
        fix_component_yields: Optional[Sequence[str]] = None,
        sig_id_column_name: Optional[str] = None,
        sig_queries: Optional[MCSigIDNameToQueryMappingType] = None,
        reco_mode_labels_dict: Optional[Dict[str, str]] = None,
        alternative_template_colors: Optional[Dict[str, str]] = None,
        mc_sample_keys: Optional[Sequence[str]] = None,
        minimizer_id: Optional[str] = None,
        run_significance_evaluation: bool = True,
        create_plots: bool = True,
        create_projection_plots: bool = True,
        create_template_plots: bool = True,
        create_2d_template_plots: bool = True,
        add_new_fei_calibration_factors: bool = False,
        add_reco_mode_efficiency_correction: bool = False,
        additional_info: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.__name: str = name
        self.__base_dir_name: str = base_dir_name
        self.__sample_key: str = sample_key
        self.__output_tag: str = output_tag
        self.__plot_info_text: str = plot_info_text
        self.__fix_nuisance_parameters: bool = fix_nuisance_parameters
        self.__asimov_data: bool = asimov_data
        self.__round_asimov_data: bool = round_asimov_data

        self.__fit_variables_by_mode_dict: FitVarDictType = fit_variables_by_mode_dict

        self.__weight_column_name: str = weight_column_name
        self.__systematics_cols: Tuple[SysColsInfo, ...] = systematics_cols

        self.__reco_mode_infos: Tuple[Tuple[str, Tuple[int, ...]], ...] = tuple(
            [(name, tuple(ids)) for name, ids in reco_mode_infos]
        )

        assert set(self.__fit_variables_by_mode_dict.keys()) == set([k for k, _ in self.__reco_mode_infos]), (
            list(self.__fit_variables_by_mode_dict.keys()),
            [k for k, _ in self.__reco_mode_infos],
        )

        self.__template_setup = template_setup
        if isinstance(template_setup, dict):
            self.__template_setup_dict = template_setup
        else:
            self.__template_setup_dict = {rm: template_setup for rm, _ in reco_mode_infos}

        self.__sig_id_to_data_type_mapping: Dict[int, Tuple[DataTypeFraction, ...]] = sig_id_to_data_type_mapping

        self.__plot_colors: Dict[str, str] = copy.copy(plot_colors)
        assert all(comp in self.plot_colors for comp in self.all_components)

        self.__plot_labels: Dict[str, str] = copy.copy(plot_labels)
        assert all(comp in self.plot_labels for comp in self.all_components)

        self.__fix_component_yields: Optional[Tuple[str, ...]] = None
        if fix_component_yields is not None:
            assert all(c in self.all_components for c in fix_component_yields), (
                fix_component_yields,
                self.all_components,
            )
            self.__fix_component_yields = tuple(fix_component_yields)

        if sig_id_column_name is not None:
            self.__sig_id_column_name: str = sig_id_column_name
        else:
            self.__sig_id_column_name = default_sig_id_col_name

        if sig_queries is not None:
            self.__sig_queries: MCSigIDNameToQueryMappingType = sig_queries
        else:
            self.__sig_queries = OldMCMatchingIDsQueryInfo.get_sig_id_mapping()

        _reco_mode_keys: Tuple[str, ...] = tuple([k for k, _ in self.reco_mode_infos])
        if reco_mode_labels_dict is not None:
            assert all(rmk in reco_mode_labels_dict for rmk in _reco_mode_keys), [
                k for k in _reco_mode_keys if k not in reco_mode_labels_dict
            ]
            self.__reco_mode_labels_dict: Dict[str, str] = reco_mode_labels_dict
        else:
            self.__reco_mode_labels_dict = {k: k for k in _reco_mode_keys}

        self.__alternative_template_colors: Optional[Dict[str, str]] = alternative_template_colors

        if mc_sample_keys is not None:
            assert all(k in rdstar_samples.data_type_combinations.keys() for k in mc_sample_keys), [
                k for k in mc_sample_keys if k not in rdstar_samples.data_type_combinations.keys()
            ]
            self.__mc_sample_keys: Tuple[str, ...] = tuple(mc_sample_keys)
        else:
            self.__mc_sample_keys = tuple(rdstar_samples.main_sample_keys)
        assert len(self.__mc_sample_keys) == len(set(self.__mc_sample_keys)), self.__mc_sample_keys

        if minimizer_id is not None:
            self.__minimizer_id: str = minimizer_id
        else:
            self.__minimizer_id = "iminuit"
        assert self.__minimizer_id in available_template_fitter_minimizer.keys(), (
            self.__minimizer_id,
            available_template_fitter_minimizer.keys(),
        )

        self.__run_significance_evaluation: bool = run_significance_evaluation

        self.__create_plots: bool = create_plots
        self.__create_projection_plots: bool = create_projection_plots
        self.__create_template_plots: bool = create_template_plots
        self.__create_2d_template_plots: bool = create_2d_template_plots

        self.__add_new_fei_calibration_factors: bool = add_new_fei_calibration_factors
        self.__add_reco_mode_efficiency_correction: bool = add_reco_mode_efficiency_correction

        self.__additional_info: Optional[Dict[str, Any]] = additional_info

        self.__sig_id_dict: Dict[str, Tuple[int, ...]] = self.init_sig_id_dict()

        self._check_sig_queries()

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def base_dir_name(self) -> str:
        return self.__base_dir_name

    @base_dir_name.setter
    def base_dir_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sample_key(self) -> str:
        return self.__sample_key

    @sample_key.setter
    def sample_key(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def output_tag(self) -> str:
        return self.__output_tag

    @output_tag.setter
    def output_tag(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def plot_info_text(self) -> str:
        return self.__plot_info_text

    @plot_info_text.setter
    def plot_info_text(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def fix_nuisance_parameters(self) -> bool:
        return self.__fix_nuisance_parameters

    @fix_nuisance_parameters.setter
    def fix_nuisance_parameters(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def asimov_data(self) -> bool:
        return self.__asimov_data

    @asimov_data.setter
    def asimov_data(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def round_asimov_data(self) -> bool:
        return self.__round_asimov_data

    @round_asimov_data.setter
    def round_asimov_data(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def fit_variables_by_mode_dict(self) -> FitVarDictType:
        return self.__fit_variables_by_mode_dict

    @fit_variables_by_mode_dict.setter
    def fit_variables_by_mode_dict(self, value: FitVarDictType) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def weight_column_name(self) -> str:
        return self.__weight_column_name

    @weight_column_name.setter
    def weight_column_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def systematics_cols(self) -> Tuple[SysColsInfo, ...]:
        return self.__systematics_cols

    @systematics_cols.setter
    def systematics_cols(self, value: Sequence[SysColsInfo]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def reco_mode_infos(self) -> Tuple[Tuple[str, Tuple[int, ...]], ...]:
        return self.__reco_mode_infos

    @reco_mode_infos.setter
    def reco_mode_infos(self, value: Tuple[Tuple[str, Tuple[int, ...]], ...]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def template_setup(self) -> Union[TemplateSetup, Dict[str, TemplateSetup]]:
        return self.__template_setup

    @template_setup.setter
    def template_setup(
        self,
        value: Union[TemplateSetup, Dict[str, TemplateSetup]],
    ) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    def template_setup_keys_for_mode(
        self,
        mode: str,
    ) -> Tuple[str, ...]:
        template_setup = self.__template_setup_dict[mode].template_setup

        if isinstance(template_setup, tuple):
            _unpacked_template_setup_keys: Tuple[str, ...] = template_setup
        else:
            assert isinstance(template_setup, dict), (template_setup, type(template_setup).__name__)
            _unpacked_template_setup_keys = tuple([str(key) for key in template_setup.keys()])
        return _unpacked_template_setup_keys

    @property
    def sig_id_to_data_type_mapping(self) -> Dict[int, Tuple[DataTypeFraction, ...]]:
        return self.__sig_id_to_data_type_mapping

    @sig_id_to_data_type_mapping.setter
    def sig_id_to_data_type_mapping(self, value: Dict[int, Tuple[DataTypeFraction, ...]]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    def plot_order_for_mode(
        self,
        mode: str,
    ) -> Tuple[str, ...]:
        return self.template_setup_keys_for_mode(mode=mode)

    @property
    def plot_colors(self) -> Dict[str, str]:
        return self.__plot_colors

    @plot_colors.setter
    def plot_colors(self, value: Dict[str, str]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def plot_labels(self) -> Dict[str, str]:
        return self.__plot_labels

    @plot_labels.setter
    def plot_labels(self, value: Dict[str, str]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def fix_component_yields(self) -> Tuple[str, ...]:
        if self.__fix_component_yields is None:
            return tuple([])
        else:
            return self.__fix_component_yields

    @fix_component_yields.setter
    def fix_component_yields(self, value: Optional[Sequence[str]]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sig_id_column_name(self) -> str:
        return self.__sig_id_column_name

    @sig_id_column_name.setter
    def sig_id_column_name(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def sig_queries(self) -> MCSigIDNameToQueryMappingType:
        return self.__sig_queries

    @sig_queries.setter
    def sig_queries(self, value: MCSigIDNameToQueryMappingType) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def reco_mode_labels_dict(self) -> Dict[str, str]:
        return self.__reco_mode_labels_dict

    @reco_mode_labels_dict.setter
    def reco_mode_labels_dict(self, value: Dict[str, str]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def alternative_template_colors(self) -> Optional[Dict[str, str]]:
        return self.__alternative_template_colors

    @alternative_template_colors.setter
    def alternative_template_colors(self, value: Optional[Dict[str, str]]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def mc_sample_keys(self) -> Tuple[str, ...]:
        return self.__mc_sample_keys

    @mc_sample_keys.setter
    def mc_sample_keys(self, value: Tuple[str, ...]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def minimizer_id(self) -> str:
        return self.__minimizer_id

    @minimizer_id.setter
    def minimizer_id(self, value: str) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def run_significance_evaluation(self) -> bool:
        return self.__run_significance_evaluation

    @run_significance_evaluation.setter
    def run_significance_evaluation(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def create_plots(self) -> bool:
        return self.__create_plots

    @create_plots.setter
    def create_plots(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def create_projection_plots(self) -> bool:
        return self.__create_projection_plots

    @create_projection_plots.setter
    def create_projection_plots(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def create_template_plots(self) -> bool:
        return self.__create_template_plots

    @create_template_plots.setter
    def create_template_plots(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def create_2d_template_plots(self) -> bool:
        return self.__create_2d_template_plots

    @create_2d_template_plots.setter
    def create_2d_template_plots(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def add_new_fei_calibration_factors(self) -> bool:
        return self.__add_new_fei_calibration_factors

    @add_new_fei_calibration_factors.setter
    def add_new_fei_calibration_factors(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def add_reco_mode_efficiency_correction(self) -> bool:
        return self.__add_reco_mode_efficiency_correction

    @add_reco_mode_efficiency_correction.setter
    def add_reco_mode_efficiency_correction(self, value: bool) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def additional_info(self) -> Dict[str, Any]:
        return dict() if self.__additional_info is None else self.__additional_info

    @additional_info.setter
    def additional_info(self, value: Optional[Dict[str, Any]]) -> None:
        raise TypeError("The FitCase object does not support item assignment.")

    @property
    def all_components(self) -> Tuple[str, ...]:
        return tuple(set([cn for m_name, _ in self.reco_mode_infos for cn in self.plot_order_for_mode(mode=m_name)]))

    def init_sig_id_dict(self) -> Dict[str, Tuple[int, ...]]:
        sig_id_dict: Dict[str, Tuple[int, ...]] = dict()
        used_sig_ids: List[int] = []
        for mode_name, _ in self.reco_mode_infos:
            template_setup = self.__template_setup_dict[mode_name].template_setup
            for component in self.plot_order_for_mode(mode=mode_name):
                if isinstance(template_setup, dict):
                    sig_ids = tuple([self.sig_queries[comp].id for comp in template_setup[component]])
                else:
                    sig_ids = (self.sig_queries[component].id,)

                if component in sig_id_dict.keys():
                    assert set(sig_ids) == set(sig_id_dict[component]), (set(sig_ids), set(sig_id_dict[component]))
                else:
                    sig_id_dict[component] = sig_ids
                    used_sig_ids.extend(sig_ids)

        assert len(used_sig_ids) == len(set(used_sig_ids)), used_sig_ids
        return sig_id_dict

    def get_sig_id_for_component(self, component_name: str) -> Tuple[int, ...]:
        return self.__sig_id_dict[component_name]

    def get_additional_setting(self, setting_key: str) -> Any:
        if setting_key not in self.additional_info.keys():
            raise KeyError(
                f"No additional setting for key {setting_key} available in this fit case '{self.name}'!\n"
                f"Available keys are {list(self.additional_info.keys())}"
            )
        return self.additional_info[setting_key]

    def get_additional_setting_with_default(self, setting_key: str, default_setting: Any) -> Any:
        return self.additional_info.get(setting_key, default_setting)

    def has_additional_setting(self, setting_key: str) -> bool:
        return setting_key in self.additional_info.keys()

    def get_reco_mode_id_map(self) -> Dict[str, Tuple[int, ...]]:
        return {rmn: ids for rmn, ids in self.reco_mode_infos}

    def get_cut_query(self, reco_mode_column_str: str) -> str:
        reco_mode_id_map: Dict[str, Tuple[int, ...]] = self.get_reco_mode_id_map()

        fit_var_queries_by_mode: Dict[str, str] = {
            r_mode_key: " & ".join([f"({fv.scope[0]} <= {fv.variable.df_label} <= {fv.scope[1]})" for fv in fit_vars])
            for r_mode_key, fit_vars in self.fit_variables_by_mode_dict.items()
        }

        string_queries: List[str] = [
            f"({fit_var_query} & ({reco_mode_column_str} in {[rm_id for rm_id in reco_mode_id_map[r_mode_key]]}))"
            for r_mode_key, fit_var_query in fit_var_queries_by_mode.items()
        ]

        return "(" + " | ".join(string_queries) + ")"

    @property
    def max_dimensions(self) -> int:
        return max(len(h_vars) for h_vars in self.fit_variables_by_mode_dict.values())

    @property
    def has_channels_with_multiple_dimensions(self) -> bool:
        return self.max_dimensions > 1

    @staticmethod
    def create_modified_fit_case(fit_case: "FitCase", **kwargs) -> "FitCase":
        fit_case_signature = signature(FitCase)
        fc_param_names: List[str] = list(fit_case_signature.parameters.keys())
        assert all(k in fc_param_names for k in kwargs.keys()), [k for k in kwargs.keys() if k not in fc_param_names]

        return FitCase(**{kw: kwargs.get(kw, getattr(fit_case, kw)) for kw in fc_param_names})

    def _check_sig_queries(self) -> None:
        if isinstance(self.template_setup, dict):
            for ts in self.template_setup.values():
                template_setup = ts.template_setup
                if isinstance(template_setup, dict):
                    assert all(
                        comp in self.sig_queries for template_comp in template_setup.values() for comp in template_comp
                    )
                else:
                    assert isinstance(template_setup, list) or isinstance(template_setup, tuple)
                    assert all(comp in self.sig_queries for comp in template_setup)
        else:
            template_setup = self.template_setup.template_setup
            if isinstance(template_setup, dict):
                assert all(
                    comp in self.sig_queries for template_comp in template_setup.values() for comp in template_comp
                )
            else:
                assert isinstance(template_setup, list) or isinstance(template_setup, tuple)
                assert all(comp in self.sig_queries for comp in template_setup)
