"""
Tools for the optimization of fit setups.

Felix Metzner 2021
"""

import os
import json

from matplotlib import pyplot as plt
from typing import Tuple, Dict, List, Optional, NamedTuple, Any

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors

from rdstar.utilities import PathType

from rdstar.latex_labels.latex_commons import LatexCommonStrings
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.result_calculation.ratio_calculator import UpDownUValue
from rdstar.offline_analysis.fitting.mva_cut_info import MVACutInfo, CutInfoTupleType
from rdstar.offline_analysis.fitting.fit_routine import FitRoutine, FitResultContainer


__all__ = [
    "FitOptimizationValidator",
]


MVACutKeyType = str
OptimizationInputDictType = Dict[MVACutKeyType, Dict[str, PathType]]


class PlotInfoMapping(NamedTuple):
    key: str
    latex_label: str


class FitOptimizationValidator:
    r_d_plot_key = "RD_MVA_OptimizationPlot"  # type: str
    r_dst_plot_key = "RDStar_MVA_OptimizationPlot"  # type: str
    optimization_result_key = "MVA_OptimizationResult"  # type: str

    rd_result_key = "best_rd_value"  # type: str
    rd_st_result_key = "best_rd_st_value"  # type: str
    rd_all_key = "all_rd_values"  # type: str
    rd_st_all_key = "all_rd_st_values"  # type: str

    rd_result_type = "rd"  # type: str
    rd_star_result_type = "rd_star"  # type: str
    result_types = (rd_result_type, rd_star_result_type)  # type: Tuple[str, ...]

    plot_info_mapping = {
        rd_result_type: PlotInfoMapping(key=r_d_plot_key, latex_label=LatexCommonStrings.rd_str),
        rd_star_result_type: PlotInfoMapping(key=r_dst_plot_key, latex_label=LatexCommonStrings.rd_star_str),
    }  # type: Dict[str, PlotInfoMapping]

    def __init__(
        self,
        fit_case: FitCase,
        base_mva_cut_info: MVACutInfo,
        base_dir_path: PathType,
    ) -> None:
        self.fit_case = fit_case  # type: FitCase
        self.base_mva_cut_info = base_mva_cut_info  # type: MVACutInfo
        self.base_dir_path = base_dir_path  # type: PathType

        self.expected_keys = tuple(self.base_mva_cut_info.get_probe_combinations_dict().keys())  # type: Tuple[str, ...]

    def validate(self, input_dict: OptimizationInputDictType) -> None:
        fit_result_container_key = FitRoutine.get_task_output_key(
            fit_case=self.fit_case,
            output_key=FitRoutine.result_info_key,
        )  # type: str

        assert all(fit_result_container_key in v.keys() for v in input_dict.values())

        fit_result_containers = dict()  # type: Dict[MVACutKeyType, FitResultContainer]
        for mva_cut_signature, d in input_dict.items():
            assert mva_cut_signature not in fit_result_containers.keys(), mva_cut_signature
            assert mva_cut_signature in self.expected_keys, mva_cut_signature
            result_container = FitResultContainer.from_file(file_path=d[fit_result_container_key])
            fit_result_containers.update({mva_cut_signature: result_container})

        self.produce_performance_plots(fit_result_dict=fit_result_containers)
        self.write_evaluation_results(fit_result_dict=fit_result_containers)

    def produce_performance_plots(self, fit_result_dict: Dict[MVACutKeyType, FitResultContainer]) -> None:
        set_matplotlibrc_params()

        best_fit_results = self.get_best_setups(fit_result_dict=fit_result_dict)  # type: Dict[str, Tuple[float, str]]
        overall_best_fit_result = best_fit_results["all"]  # type: Tuple[float, str]
        overall_best_fit_cut_info_tuple = self.base_mva_cut_info.get_cut_var_val_pair_from_signature(
            signature=overall_best_fit_result[1]
        )  # type: CutInfoTupleType

        cut_reco_mode_var_labels = tuple([s for s, _ in overall_best_fit_cut_info_tuple])  # type: Tuple[str, ...]

        for result_type in self.result_types:
            plot_info = self.plot_info_mapping[result_type]  # type: PlotInfoMapping

            _results = self.get_cut_result_pairs(fit_result_dict=fit_result_dict, result_type=result_type)
            results = [
                (self.base_mva_cut_info.get_cut_var_val_pair_from_signature(signature=v[0]), v[1]) for v in _results
            ]  # type: List[Tuple[CutInfoTupleType, float]]

            best_fit_result = best_fit_results[result_type]  # type: Tuple[float, str]
            best_fit_cut_info_tuple = self.base_mva_cut_info.get_cut_var_val_pair_from_signature(
                signature=best_fit_result[1]
            )  # type: CutInfoTupleType

            for var_i, reco_mode_var_label in enumerate(cut_reco_mode_var_labels):
                current_reco_mode_tuple, current_var = self.get_reco_ch_and_var(
                    cut_var_str=reco_mode_var_label
                )  # type: Optional[Tuple[str, ...]], str
                ch_label = ""  # type: str
                if current_reco_mode_tuple is not None:
                    ch_list = [
                        self.fit_case.reco_mode_labels_dict[ch] for ch in current_reco_mode_tuple
                    ]  # type: List[str]
                    ch_str = " or ".join(ch_list)  # type: str
                    ch_label = f" in {ch_str}"

                best_for_others_map = {
                    c: dict(best_fit_cut_info_tuple)[c] for c in cut_reco_mode_var_labels if c != reco_mode_var_label
                }  # type: Dict[str, float]
                overall_best_for_others_map = {
                    c: dict(overall_best_fit_cut_info_tuple)[c]
                    for c in cut_reco_mode_var_labels
                    if c != reco_mode_var_label
                }  # type: Dict[str, float]

                fig, ax = plt.subplots(nrows=1, ncols=1, dpi=200)

                x_vals, y_vals = self.get_x_y_results_for_var(
                    reco_ch_var_label=reco_mode_var_label,
                    result_list=results,
                    others_best_mapping=best_for_others_map,
                )  # type: List[float], List[float]
                x_vals_with_overall_best, y_vals_with_overall_best = self.get_x_y_results_for_var(
                    reco_ch_var_label=reco_mode_var_label,
                    result_list=results,
                    others_best_mapping=overall_best_for_others_map,
                )  # type: List[float], List[float]
                assert x_vals == x_vals_with_overall_best, (x_vals, x_vals_with_overall_best)

                best_x_val = dict(best_fit_cut_info_tuple)[reco_mode_var_label]  # type: float
                best_y_val = best_fit_result[0]  # type: float
                overall_best_x_val = dict(overall_best_fit_cut_info_tuple)[reco_mode_var_label]  # type: float
                overall_best_y_val = dict(_results)[overall_best_fit_result[1]]  # type: float

                plt.plot(
                    x_vals,
                    y_vals,
                    marker=".",
                    linestyle="None",
                    color=KITColors.kit_blue,
                    label="Scan",
                )

                plt.plot(
                    x_vals,
                    y_vals_with_overall_best,
                    marker="o",
                    markersize=0.8,
                    markeredgewidth=1.5,
                    linestyle="None",
                    markeredgecolor=KITColors.kit_orange,
                    markerfacecolor="None",
                    label="Scan with other best other cuts",
                )

                plt.plot(
                    best_x_val,
                    best_y_val,
                    marker=".",
                    linestyle="None",
                    color=KITColors.kit_red,
                    label=f"Best in this Dimension ({best_x_val}, {round(best_y_val, 4)})",
                )

                plt.plot(
                    overall_best_x_val,
                    overall_best_y_val,
                    marker="o",
                    linestyle="None",
                    markeredgecolor=KITColors.kit_green,
                    markerfacecolor="None",
                    markeredgewidth=2,
                    label=f"Best Overall ({overall_best_x_val}, {round(overall_best_y_val, 4)})",
                )

                ax.set_title(plot_info.latex_label)
                ax.set_xlabel(LatexObsNames.mva_sel_output + ch_label)
                ax.set_ylabel(f"Statistical Uncertainty on {plot_info.latex_label}")
                ax.legend(loc=2)
                export(
                    fig=fig,
                    filename=plot_info.key if var_i == 0 else f"{plot_info.key}_{reco_mode_var_label}",
                    target_dir=self.base_dir_path,
                    close_figure=True,
                )

    def get_reco_ch_and_var(self, cut_var_str: str) -> Tuple[Optional[Tuple[str, ...]], str]:
        return self.base_mva_cut_info.split_cut_reco_ch_var_str(reco_ch_var_pair=cut_var_str)

    @staticmethod
    def get_x_y_results_for_var(
        reco_ch_var_label: str, result_list: List[Tuple[CutInfoTupleType, float]], others_best_mapping: Dict[str, float]
    ) -> Tuple[List[float], List[float]]:
        assert reco_ch_var_label not in others_best_mapping.keys(), (reco_ch_var_label, others_best_mapping)
        matching = [
            all(dict(cit)[cv] == res for cv, res in others_best_mapping.items()) for cit, _ in result_list
        ]  # type: List[bool]
        assert len(matching) == len(result_list), (len(matching), len(result_list))

        x_vals = [dict(x)[reco_ch_var_label] for (x, _), m in zip(result_list, matching)]  # type: List[float]
        y_vals = [y for (_, y), m in zip(result_list, matching)]  # type: List[float]
        assert len(x_vals) == len(y_vals), (len(x_vals), len(y_vals))
        return x_vals, y_vals

    def write_evaluation_results(self, fit_result_dict: Dict[MVACutKeyType, FitResultContainer]) -> None:
        tmp_res_path = self.get_path(key=self.optimization_result_key, tmp=True)  # type: PathType

        rd_results = self.get_cut_result_pairs(fit_result_dict=fit_result_dict, result_type=self.rd_result_type)
        rd_star_results = self.get_cut_result_pairs(fit_result_dict=fit_result_dict, result_type=self.rd_star_result_type)

        best_fit_results = self.get_best_setups(fit_result_dict=fit_result_dict)  # type: Dict[str, Tuple[float, str]]

        best_signatures = {k: v[1] for k, v in best_fit_results.items()}  # type: Dict[str, str]
        best_cut_var_val_pair = {
            k: self.base_mva_cut_info.get_cut_var_val_pair_from_signature(signature=v[1])
            for k, v in best_fit_results.items()
        }  # type: Dict[str, CutInfoTupleType]
        setup_map = self.base_mva_cut_info.get_probe_combinations_dict()  # type: Dict[str, MVACutInfo]
        setup_dict_map = {k: ci.as_dict() for k, ci in setup_map.items()}  # type: Dict[str, Dict[str, Any]]

        optimization_results = {
            self.rd_result_key: best_fit_results[self.rd_result_type],
            self.rd_st_result_key: best_fit_results[self.rd_star_result_type],
            "best_all": best_fit_results["all"],
            f"best_{self.rd_result_type}_setup": best_cut_var_val_pair[self.rd_result_type],
            f"best_{self.rd_star_result_type}_setup": best_cut_var_val_pair[self.rd_star_result_type],
            "best_all_setup": best_cut_var_val_pair["all"],
            f"best_{self.rd_result_type}_setup_dict": setup_dict_map[best_signatures[self.rd_result_type]],
            f"best_{self.rd_star_result_type}_setup_dict": setup_dict_map[best_signatures[self.rd_star_result_type]],
            "best_all_setup_dict": setup_dict_map[best_signatures["all"]],
            self.rd_all_key: rd_results,
            self.rd_st_all_key: rd_star_results,
        }

        with open(tmp_res_path, "w") as out_file:
            json.dump(optimization_results, out_file, indent=2)

        os.rename(src=tmp_res_path, dst=self.get_path(key=self.optimization_result_key, tmp=False))

    @staticmethod
    def get_cut_result_pairs(
        fit_result_dict: Dict[MVACutKeyType, FitResultContainer],
        result_type: str,
    ) -> Tuple[Tuple[str, float], ...]:
        assert result_type in FitOptimizationValidator.result_types, result_type

        assert all(r.rd_ratio_result is not None for r in fit_result_dict.values())
        _fit_results = {
            cs: res.rd_ratio_result if result_type == FitOptimizationValidator.rd_result_type else res.rdstar_ratio_result
            for cs, res in fit_result_dict.items()
        }  # type: Dict[MVACutKeyType, Optional[UpDownUValue]]
        assert all(r is not None for r in _fit_results.values())
        fit_results = {k: v for k, v in _fit_results.items() if v is not None}  # type: Dict[MVACutKeyType, UpDownUValue]

        return tuple([(cs, max(res.std_dev_up, res.std_dev_down)) for cs, res in fit_results.items()])

    def get_best_setups(
        self,
        fit_result_dict: Dict[MVACutKeyType, FitResultContainer],
    ) -> Dict[str, Tuple[float, str]]:
        result_dict = dict()  # type: Dict[str, Tuple[float, str]]
        all_result_tuples = []  # type: List[Tuple[Tuple[str, float], ...]]

        for result_type in self.result_types:
            res_tuples = self.get_cut_result_pairs(fit_result_dict=fit_result_dict, result_type=result_type)
            all_result_tuples.append(res_tuples)
            min_val = min(v for _, v in res_tuples)  # type: float
            best_setup_list = [k for k, v in res_tuples if v == min_val]  # type: List[str]
            assert len(best_setup_list) == 1, best_setup_list
            result_dict.update({result_type: (min_val, best_setup_list[0])})

        # Check if the cut info signatures of each matched result are the same
        assert all(all(rts[0][0] == rt[0] for rt in rts) for rts in zip(*all_result_tuples))
        min_val_sum = min(sum([rt[1] for rt in rts]) for rts in zip(*all_result_tuples))  # type: float
        sigs_of_best = [
            rts[0][0] for rts in zip(*all_result_tuples) if sum([rt[1] for rt in rts]) == min_val_sum
        ]  # type: List[str]
        assert len(sigs_of_best) == 1
        signature_of_best = sigs_of_best[0]  # type: str
        result_dict.update({"all": (min_val_sum, signature_of_best)})
        return result_dict

    def get_path(self, key: str, tmp: bool = False) -> PathType:
        tmp_tag = "tmp_" if tmp else ""  # type: str
        if "Plot" in key:
            return os.path.join(self.base_dir_path, f"{tmp_tag}{key}.pdf")
        else:
            return os.path.join(self.base_dir_path, f"{tmp_tag}{key}.json")

    def get_output_dict(self) -> Dict[str, PathType]:
        return {
            self.r_d_plot_key: self.get_path(key=self.r_d_plot_key, tmp=False),
            self.r_dst_plot_key: self.get_path(key=self.r_dst_plot_key, tmp=False),
            self.optimization_result_key: self.get_path(key=self.optimization_result_key, tmp=False),
        }
