"""
Histogram Variables

This file provides the histogram variables used in the analysis.

Felix Metzner 2020
"""


from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable
from rdstar.offline_analysis.observables.e_miss_minus_p_miss import EMissMinusPMissVariable
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.multistage_setups import main_msrc_corrected_mms_column_name
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    e_miss_minus_p_miss_res_corr_setup,
)


__all__ = [
    "RDStarHistVars",
]


main_resolution_corrected_mms_col_name: str = main_msrc_corrected_mms_column_name
alternative_res_corr_mms_col_name: str = Cols.alt_res_corr_missing_mass
old_resolution_corrected_mms_col_name: str = MissingMassResolutionCorrector.get_main_single_stage_correction_column_name(
    use_laplace_smeared_version=False
)


class RDStarHistVars:
    # Missing Mass Squared Definitions
    missing_mass = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=30,
        scope=(-2.0, 6.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_for_main_fit = HistVariable(
        df_label=alternative_res_corr_mms_col_name,
        n_bins=10,
        scope=(-0.5, 9.5),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=42,
        scope=(-0.5, 10.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_alt_res_corr = missing_mass_res_corr.get_modified_histogram_variable(
        df_label=alternative_res_corr_mms_col_name,
    )

    missing_mass_res_corr_for_fit = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=20,
        scope=(-2.0, 10.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_narrow_for_fit = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=20,
        scope=(-1.0, 5.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_wide = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=48,
        scope=(-2.0, 10.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_q2sb = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=24,
        scope=(-0.5, 4.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_q2sb_wide = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=24,
        scope=(-2.0, 4.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_alt_res_corr_q2sb_wide = missing_mass_res_corr_q2sb_wide.get_modified_histogram_variable(
        df_label=alternative_res_corr_mms_col_name,
        n_bins=18,
    )

    missing_mass_for_main_fei_cal = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=20,
        scope=(-0.5, 1.5),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_q2sb_fit = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=12,
        scope=(-0.5, 3.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_q2sb_fit_with_old_res_corr = HistVariable(
        df_label=old_resolution_corrected_mms_col_name,
        n_bins=12,
        scope=(-0.5, 3.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_extreme_fine_binning = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=100,
        scope=(-0.5, 3.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_extreme_fine_binning_wide = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=150,
        scope=(-2.0, 4.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    #
    # Lepton Momentum

    pl_star_signal_fit = HistVariable(
        df_label=Cols.p_star_lepton,
        n_bins=15,
        scope=(0.1, 2.2),
        var_name=LatexObsNames.fitting_p_l_star,
        unit=LatexUnits.gev,
    )

    pl_star_sideband_fit = HistVariable(
        df_label=Cols.p_star_lepton,
        n_bins=15,
        scope=(0.2, 2.3),
        var_name=LatexObsNames.fitting_p_l_star,
        unit=LatexUnits.gev,
    )

    pl_star = HistVariable(
        df_label=Cols.p_star_lepton,
        n_bins=20,
        scope=(0.1, 2.3),
        var_name=LatexObsNames.fitting_p_l_star,
        unit=LatexUnits.gev,
    )

    #
    # E_Miss - p_Miss Definition

    e_miss_minus_p_miss = HistVariable(
        df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_col_name,
        n_bins=33,
        scope=(-5.125, 3.125),
        var_name=LatexObsNames.fitting_e_miss_minus_p_miss,
        unit=LatexUnits.gev,
    )

    e_miss_minus_p_miss_cms = HistVariable(
        df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_cms_col_name,
        n_bins=60,
        scope=(-1.0, 2.0),
        var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
        unit=LatexUnits.gev,
    )

    #
    # E_Miss - p_Miss with resolution correction Definition

    e_miss_minus_p_miss_corr_setup_name: str = e_miss_minus_p_miss_res_corr_setup.name
    e_miss_minus_p_miss_corr_df_label: str = MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=e_miss_minus_p_miss_corr_setup_name,
        use_laplace_smeared_version=False,
    )
    e_miss_minus_p_miss_res_corr_signal_var = HistVariable(
        df_label=e_miss_minus_p_miss_corr_df_label,
        n_bins=20,
        scope=(-1.0, 10.0),
        var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
        unit=LatexUnits.gev,
    )

    e_miss_minus_p_miss_res_corr_q2sb_var = HistVariable(
        df_label=e_miss_minus_p_miss_corr_df_label,
        n_bins=20,
        scope=(-1.0, 2.0),
        var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
        unit=LatexUnits.gev,
    )

    # MVA Output Hist Variable for Fit
    mva_output_for_fit_var = HistVariable(
        df_label="sec_sel_mva_output",
        n_bins=10,
        scope=(0.5, 1.0),
        var_name=LatexObsNames.mva_sel_output,
        unit=LatexUnits.no_unit,
    )

    mva_output_loose_for_fit_var = HistVariable(
        df_label="sec_sel_mva_output",
        n_bins=8,
        scope=(0.2, 1.0),
        var_name=LatexObsNames.mva_sel_output,
        unit=LatexUnits.no_unit,
    )

    #
    # q2 Reco
    q2_reco_in_signal_region = HistVariable(
        df_label=Cols.q_squared_reco,
        n_bins=6,
        scope=(4.0, 10.0),
        var_name=LatexObsNames.fitting_q2,
        unit=LatexUnits.gev_sq,
    )

    q2_reco_in_sideband_region = HistVariable(
        df_label=Cols.q_squared_reco,
        n_bins=8,
        scope=(0.0, 4.0),
        var_name=LatexObsNames.fitting_q2,
        unit=LatexUnits.gev_sq,
    )

    q2_reco_full_region = HistVariable(
        df_label=Cols.q_squared_reco,
        n_bins=10,
        scope=(0.0, 10.0),
        var_name=LatexObsNames.fitting_q2,
        unit=LatexUnits.gev_sq,
    )

    #
    # Sideband Validation Variables

    neextra_for_main_fit = HistVariable(
        df_label=Cols.event_ne_extra,
        n_bins=10,
        scope=(0.0, 1.4),
        var_name=LatexObsNames.roe_n_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    neextra_for_fit = HistVariable(
        df_label=Cols.event_ne_extra,
        n_bins=12,
        scope=(0.0, 2.5),
        var_name=LatexObsNames.roe_n_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    neextra_fine = HistVariable(
        df_label=Cols.event_ne_extra,
        n_bins=40,
        scope=(0.0, 2.5),
        var_name=LatexObsNames.roe_n_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    neextra_narrow_for_fit = HistVariable(
        df_label=Cols.event_ne_extra,
        n_bins=10,
        scope=(0.0, 1.5),
        var_name=LatexObsNames.roe_n_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    neextra_zoom_in_fine = HistVariable(
        df_label=Cols.event_ne_extra,
        n_bins=40,
        scope=(0.0, 0.4),
        var_name=LatexObsNames.roe_n_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    eextra = HistVariable(
        df_label=Cols.event_e_extra,
        n_bins=20,
        scope=(0.0, 3.0),
        var_name=LatexObsNames.roe_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    eextra_zoom_in = HistVariable(
        df_label=Cols.event_e_extra,
        n_bins=30,
        scope=(0.0, 0.6),
        var_name=LatexObsNames.roe_e_ecl_extra,
        unit=LatexUnits.gev,
    )

    mva_output_for_sbv_var = HistVariable(
        df_label="sec_sel_mva_output",
        n_bins=20,
        scope=(0.5, 1.0),
        var_name=LatexObsNames.mva_sel_output,
        unit=LatexUnits.no_unit,
    )

    mva_output_loose_for_sbv_var = HistVariable(
        df_label="sec_sel_mva_output",
        n_bins=21,
        scope=(0.2, 1.0),
        var_name=LatexObsNames.mva_sel_output,
        unit=LatexUnits.no_unit,
    )

    cs_mva_output = HistVariable(
        df_label=Cols.cs_mva_output,
        n_bins=40,
        scope=(0.5, 1.0),
        var_name=LatexObsNames.mva_cs_output,
        unit=LatexUnits.no_unit,
    )

    loose_cs_mva_output = HistVariable(
        df_label=Cols.cs_mva_output,
        n_bins=40,
        scope=(0.2, 1.0),
        var_name=LatexObsNames.mva_cs_output,
        unit=LatexUnits.no_unit,
    )

    cs_mva_output_cs_sideband_var = HistVariable(
        df_label=Cols.cs_mva_output,
        n_bins=30,
        scope=(0.2, 0.5),
        var_name=LatexObsNames.mva_cs_output,
        unit=LatexUnits.no_unit,
    )

    cs_mva_output_cs_new_sideband_var = HistVariable(
        df_label=Cols.cs_mva_output,
        n_bins=65,
        scope=(0.2, 0.85),
        var_name=LatexObsNames.mva_cs_output,
        unit=LatexUnits.no_unit,
    )

    cs_cos_tbto_cs_sideband_var = HistVariable(
        df_label=Cols.cs_var_cos_tbto,
        n_bins=25,
        scope=(0.0, 1.0),
        var_name=LatexObsNames.cs_base_cos_theta_tbto,
        unit=LatexUnits.no_unit,
    )

    cs_cos_tbz_cs_sideband_var = HistVariable(
        df_label=Cols.cs_var_cos_tbz,
        n_bins=25,
        scope=(0.0, 1.0),
        var_name=LatexObsNames.cs_base_cos_theta_tbz,
        unit=LatexUnits.no_unit,
    )

    sig_vertex_fit_prob = HistVariable(
        df_label=Cols.b_sig_vertex_fit_prob,
        n_bins=40,
        scope=(0.0, 1.0),
        var_name=LatexObsNames.b_sig_vertex_prob,
        unit=LatexUnits.no_unit,
    )

    tag_fei_prob = HistVariable(
        df_label=Cols.fei_probability,
        n_bins=40,
        scope=(0.0, 1.0),
        var_name=LatexObsNames.fei_prob,
        unit=LatexUnits.no_unit,
    )

    tag_fei_prob_zoom_in = HistVariable(
        df_label=Cols.fei_probability,
        n_bins=40,
        scope=(0.0, 0.04),
        var_name=LatexObsNames.fei_prob,
        unit=LatexUnits.no_unit,
    )

    pl_star_SidebandValidation = HistVariable(
        df_label=Cols.p_star_lepton,
        n_bins=25,
        scope=(0.1, 2.2),
        var_name=LatexObsNames.fitting_p_l_star,
        unit=LatexUnits.gev,
    )

    missing_mass_q2_SidebandValidation = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=25,
        scope=(-2.0, 4.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_q2sbSidebandValidation = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=25,
        scope=(-2.0, 4.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_cs_mva_SidebandValidation = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=25,
        scope=(-1.5, 6.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_cs_mva_SidebandValidation = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=25,
        scope=(-1.5, 6.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_dss_control_SidebandValidation = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=20,
        scope=(-2.0, 6.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_res_corr_off_resonance_SidebandValidation = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=30,
        scope=(-2.0, 10.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    mbc_SidebandValidation = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=25,
        scope=(5.27, 5.29),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    mbc_full_range_fine_binning_SidebandValidation = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=50,
        scope=(5.22, 5.29),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    mbc_full_range_normal_binning_SidebandValidation = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=35,
        scope=(5.22, 5.29),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    mbc_full_range_rough_binning_SidebandValidation = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=20,
        scope=(5.22, 5.29),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    tag_delta_e = HistVariable(
        df_label=Cols.b_tag_delta_e,
        n_bins=30,
        scope=(-0.05, 0.1),
        var_name=LatexObsNames.b_tag_delta_e,
        unit=LatexUnits.gev,
    )

    mbc_in_mbcsb_SidebandValidation = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=25,
        scope=(5.22, 5.27),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    mbc_in_mbcsb_SidebandValidation_rough_binning = HistVariable(
        df_label=Cols.b_tag_mbc,
        n_bins=15,
        scope=(5.22, 5.27),
        var_name=LatexObsNames.b_tag_mbc,
        unit=LatexUnits.gev,
    )

    #
    # CS Training Input Validation on Sideband Variables (The ones which need extra treatment due to range)

    special_cs_b_tag_ksfw_et = HistVariable(
        df_label="daughter__bo0__cm__spKSFWVariables__boet__bc__bc",
        n_bins=11,
        scope=(5.0, 11.0),
        var_name=LatexObsNames.cs_b_tag_energy_t,
        unit=LatexUnits.no_unit,
    )

    special_cs_b_tag_ksfw_hoo0 = HistVariable(
        df_label="daughter__bo0__cm__spKSFWVariables__bohoo0__bc__bc",
        n_bins=12,
        scope=(0.04, 0.22),
        var_name=LatexObsNames.cs_b_tag_h_oo_0,
        unit=LatexUnits.no_unit,
    )

    special_cs_b_tag_ksfw_mm2 = HistVariable(
        df_label="daughter__bo0__cm__spKSFWVariables__bomm2__bc__bc",
        n_bins=12,
        scope=(-4.0, 4.0),
        var_name=LatexObsNames.cs_b_tag_mm2,
        unit=LatexUnits.no_unit,
    )

    #
    # Selection MVA Training Input Validation on Sideband Variables

    d_mass_for_sb_validation = HistVariable(
        df_label=Cols.d_inv_mass,
        n_bins=25,
        scope=(1.8, 2.1),
        var_name=LatexObsNames.d_w_wo_star_meson_mass,
        unit=LatexUnits.gev,
    )

    d_star_daughter_mass_for_sb_validation = HistVariable(
        df_label=Cols.d_daughter_inv_mass,
        n_bins=15,
        scope=(1.8, 1.95),
        var_name=LatexObsNames.d_w_wo_star_daughter_mass,
        unit=LatexUnits.gev,
    )

    first_daughter_mass_in_d_region_for_sb_validation = HistVariable(
        df_label=Cols.d_daughter_inv_mass,
        n_bins=15,
        scope=(1.8, 1.95),
        var_name=LatexObsNames.d_w_wo_star_daughter_mass,
        unit=LatexUnits.gev,
    )

    first_daughter_mass_in_dstar_region_for_sb_validation = HistVariable(
        df_label=Cols.d_daughter_inv_mass,
        n_bins=15,
        scope=(1.95, 2.1),
        var_name=LatexObsNames.d_w_wo_star_daughter_mass,
        unit=LatexUnits.gev,
    )

    #
    # FSR Correction Validation Variables

    fsr_energy_cms = HistVariable(
        df_label=Cols.fsr_energy_cms,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_cms_energy,
        unit=LatexUnits.gev,
    )

    fsr_energy_rest_frame = HistVariable(
        df_label=Cols.fsr_energy_rest_frame,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_rest_energy,
        unit=LatexUnits.gev,
    )

    fsr_energy_lab = HistVariable(
        df_label=Cols.fsr_energy_lab,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_lab_energy,
        unit=LatexUnits.gev,
    )

    fsr_momentum_cms = HistVariable(
        df_label=Cols.fsr_momentum_cms,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_cms_p,
        unit=LatexUnits.gev,
    )

    fsr_momentum_rest_frame = HistVariable(
        df_label=Cols.fsr_momentum_rest_frame,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_rest_p,
        unit=LatexUnits.gev,
    )

    fsr_momentum_lab = HistVariable(
        df_label=Cols.fsr_momentum_lab,
        n_bins=30,
        scope=(0, 1.0),
        var_name=LatexObsNames.fsr_lab_p,
        unit=LatexUnits.gev,
    )

    #
    # Unused Variables

    # Dstarstar Variables
    missing_mass_dstarstar = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=25,
        scope=(-2.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_rough_binning = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=15,
        scope=(-2.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_dstarstar_mass_ge_1p5 = HistVariable(
        df_label=Cols.original_missing_mass,
        n_bins=25,
        scope=(1.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_corr = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=25,
        scope=(-2.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_rough_binning_corr = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=15,
        scope=(-2.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_dstarstar_mass_ge_1p5_corr = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=25,
        scope=(1.5, 8.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_cal_fit_cut_range = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=16,
        scope=(-0.4, 0.4),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_cal_fit_m02_04_cut_range = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=16,
        scope=(-0.2, 0.4),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_cal_fit_m02_08_cut_range = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=16,
        scope=(-0.2, 0.8),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    missing_mass_dstarstar_cal_fit_m02_20_cut_range = HistVariable(
        df_label=main_resolution_corrected_mms_col_name,
        n_bins=16,
        scope=(-0.2, 2.0),
        var_name=LatexObsNames.fitting_mm2,
        unit=LatexUnits.gev_sq,
    )

    # Other
    diff_dstar_d_mass = HistVariable(
        df_label=Cols.diff_dstar_d_mass,
        n_bins=18,
        scope=(0.1, 1.0),
        var_name=LatexObsNames.d_st_st_delta_m,
        unit=LatexUnits.gev,
    )

    diff_dstar_d_mass_for_dss_cal = HistVariable(
        df_label=Cols.diff_dstar_d_mass,
        n_bins=18,
        scope=(0.2, 1.0),
        var_name=LatexObsNames.d_st_st_delta_m,
        unit=LatexUnits.gev,
    )

    diff_dstar_d_mass_for_dss_cal_with_lt0p3_cut = HistVariable(
        df_label=Cols.diff_dstar_d_mass,
        n_bins=21,
        scope=(0.3, 1.0),
        var_name=LatexObsNames.d_st_st_delta_m,
        unit=LatexUnits.gev,
    )

    missing_mom_cms = HistVariable(
        df_label=Cols.missing_momentum_in_cms,
        n_bins=30,
        scope=(0.0, 1.5),
        var_name=LatexObsNames.fitting_p_miss_star,
        unit=LatexUnits.gev,
    )

    #
    # Analysis Plot Variables

    q2_residual_with_q2reco = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_residual_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_resi,
        unit=LatexUnits.gev_sq,
    )
    q2_residual_with_q2reco2 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_2_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_resi_v2,
        unit=LatexUnits.gev_sq,
    )
    q2_residual_with_q2reco3 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_3_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_resi_v3,
        unit=LatexUnits.gev_sq,
    )
    q2_residual_with_q2reco4 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_residual_with_q2reco_4_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_resi_v4,
        unit=LatexUnits.gev_sq,
    )

    q2_resolution_with_q2reco = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_resolution_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_reso,
        unit=LatexUnits.no_unit,
    )
    q2_resolution_with_q2reco2 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_2_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_reso_v2,
        unit=LatexUnits.no_unit,
    )
    q2_resolution_with_q2reco3 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_3_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_reso_v3,
        unit=LatexUnits.no_unit,
    )
    q2_resolution_with_q2reco4 = HistVariable(
        df_label=QSquaredResidualVariable.q_squared_resolution_with_q2reco_4_col_name,
        n_bins=100,
        scope=(-1.0, 1.0),
        var_name=LatexObsNames.q2_reso_v4,
        unit=LatexUnits.no_unit,
    )
