"""
Extended mc matching information.

Felix Metzner 2023
"""

from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Type

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import (
    MCSigID,
    MatchQueryInfo,
    MCSigIDNameToQueryMappingType,
    MCSigIDToLatexLabelMappingType,
    MCSigIDNameToLatexLabelMappingType,
    RecoModeToExpectedSignalSigIDMappingType,
    SigIDToExpectedSignalForRecoModeMappingType,
)


__all__ = [
    "AbstractBaseMCSigIDsClass",
    "AbstractBaseMCMatchingIDsQueryInfo",
]


class AbstractBaseMCSigIDsClass(ABC):
    @classmethod
    def get_sig_ids_dict(cls) -> Dict[str, MCSigID]:
        _method_names: Tuple[str, ...] = (
            "get_sig_ids_dict",
            "get_all_sig_ids",
            "get_reco_mode_to_sig_id_mapping",
            "get_sig_id_to_reco_mode_mapping",
            "run_checks",
        )

        return {
            attr_name: val
            for attr_name, val in cls.__dict__.items()
            if not ((attr_name in _method_names) or (attr_name.endswith("__") or attr_name.startswith("_")))
        }

    @classmethod
    def get_all_sig_ids(cls) -> Tuple[int, ...]:
        return tuple(sig_id_info.id for sig_id_info in cls.get_sig_ids_dict().values())

    @classmethod
    def get_reco_mode_to_sig_id_mapping(cls) -> RecoModeToExpectedSignalSigIDMappingType:
        relevant_reco_modes: Tuple[int, ...] = tuple(
            set(r for v in cls.get_sig_ids_dict().values() for r in v.get_tuple_with_expected_signal_decay_mode_ids())
        )
        return {
            r: tuple(
                v.id for v in cls.get_sig_ids_dict().values() if r in v.get_tuple_with_expected_signal_decay_mode_ids()
            )
            for r in relevant_reco_modes
        }

    @classmethod
    def get_sig_id_to_reco_mode_mapping(cls) -> SigIDToExpectedSignalForRecoModeMappingType:
        return {
            v.id: v.get_tuple_with_expected_signal_decay_mode_ids()
            for v in cls.get_sig_ids_dict().values()
            if v.expected_signal_in_decay_ids is not None
        }

    @classmethod
    def run_checks(cls) -> None:
        __all_sig_ids: Tuple[int, ...] = cls.get_all_sig_ids()
        __sig_id_dict: Dict[str, MCSigID] = cls.get_sig_ids_dict()

        assert all(isinstance(k, str) for k in __sig_id_dict.keys()), [k for k in __sig_id_dict.keys()]
        assert all(isinstance(v, MCSigID) for v in __sig_id_dict.values()), [(k, v) for k, v in __sig_id_dict.items()]
        assert all(isinstance(v.id, int) for v in __sig_id_dict.values()), [(k, v.id) for k, v in __sig_id_dict.items()]

        assert len(set(__all_sig_ids)) == len(__all_sig_ids), (
            len(__all_sig_ids),
            len(set(__all_sig_ids)),
            set([i for i in set(__all_sig_ids) if __all_sig_ids.count(i) > 1]),
        )

        sig_id_to_reco_mode_dict: RecoModeToExpectedSignalSigIDMappingType = cls.get_sig_id_to_reco_mode_mapping()
        assert all([i in __all_sig_ids for i in sig_id_to_reco_mode_dict.keys()]), (
            [i for i in sig_id_to_reco_mode_dict.keys() if i not in __all_sig_ids],
        )

        reco_mode_to_sig_id_dict: RecoModeToExpectedSignalSigIDMappingType = cls.get_reco_mode_to_sig_id_mapping()

        all_reco_modes: Tuple[int, ...] = RecoModeInfos.get_all_reco_mode_ids()
        assert set(reco_mode_to_sig_id_dict.keys()) == set(all_reco_modes), (
            set(reco_mode_to_sig_id_dict.keys()),
            set(all_reco_modes),
        )

        assert all([i in all_reco_modes for i in reco_mode_to_sig_id_dict.keys()]), (
            [i for i in reco_mode_to_sig_id_dict.keys() if i not in all_reco_modes],
        )

        assert all([i in reco_mode_to_sig_id_dict.keys() for i in all_reco_modes]), (
            [i for i in all_reco_modes if i not in reco_mode_to_sig_id_dict.keys()],
        )


class AbstractBaseMCMatchingIDsQueryInfo(ABC):
    sig_col_name: str = ""

    @classmethod
    @abstractmethod
    def get_sig_col_name(cls) -> str:
        raise NotImplementedError(f"Method 'get_sig_col_name' must be implemented by class {cls.__name__}!")

    @classmethod
    @abstractmethod
    def get_required_columns(cls) -> Tuple[str, ...]:
        raise NotImplementedError(f"Method 'get_required_columns' must be implemented by class {cls.__name__}!")

    @classmethod
    @abstractmethod
    def get_sig_id_class(cls) -> Type[AbstractBaseMCSigIDsClass]:
        raise NotImplementedError(f"Method 'get_sig_col_name' must be implemented by class {cls.__name__}!")

    @classmethod
    def get_mc_sig_id_query_dict(cls) -> MCSigIDNameToQueryMappingType:
        _method_names: Tuple[str, ...] = (
            "sig_col_name",
            "get_sig_col_name",
            "get_required_columns",
            "get_sig_id_class",
            "get_mc_sig_id_query_dict",
            "get_all_sig_ids",
            "get_mc_sig_id_queries",
            "get_sig_id_mapping",
            "get_reco_mode_to_sig_id_mapping",
            "get_sig_id_to_reco_mode_mapping",
            "run_checks",
        )

        _method_name_beginnings: Tuple[str, ...] = (
            "get_sig_id_mapping",
            "get_sig_id_to_latex_label",
            "get_sig_id_name_to_latex_label",
        )

        _query_dict: Dict[str, Any] = {
            attr_name: val
            for attr_name, val in cls.__dict__.items()
            if not (
                (attr_name in _method_names)
                or any(attr_name.startswith(mnb) for mnb in _method_name_beginnings)
                or attr_name.endswith("__")
                or attr_name.startswith("_")
            )
        }
        assert all(isinstance(val, MatchQueryInfo) for val in _query_dict.values()), [
            (k, type(v)) for k, v in _query_dict.items() if not isinstance(v, MatchQueryInfo)
        ]

        query_dict: Dict[str, MatchQueryInfo] = {k: v for k, v in _query_dict.items() if isinstance(v, MatchQueryInfo)}
        return query_dict

    @classmethod
    def get_all_sig_ids(cls) -> Tuple[int, ...]:
        return tuple(v.id for v in cls.get_mc_sig_id_query_dict().values())

    @classmethod
    def get_mc_sig_id_queries(cls) -> Tuple[MatchQueryInfo, ...]:
        return tuple(sig_id for sig_id in cls.get_mc_sig_id_query_dict().values())

    @classmethod
    def get_sig_id_mapping(cls) -> MCSigIDNameToQueryMappingType:
        return cls.get_mc_sig_id_query_dict()

    @classmethod
    def get_sig_id_to_latex_label(cls) -> MCSigIDToLatexLabelMappingType:
        return {v.id: v.latex_label for v in cls.get_mc_sig_id_query_dict().values()}

    @classmethod
    def get_sig_id_name_to_latex_label(cls) -> MCSigIDNameToLatexLabelMappingType:
        return {k: v.latex_label for k, v in cls.get_mc_sig_id_query_dict().items()}

    @classmethod
    def get_reco_mode_to_sig_id_mapping(cls) -> RecoModeToExpectedSignalSigIDMappingType:
        return cls.get_sig_id_class().get_reco_mode_to_sig_id_mapping()

    @classmethod
    def get_sig_id_to_reco_mode_mapping(cls) -> SigIDToExpectedSignalForRecoModeMappingType:
        return cls.get_sig_id_class().get_sig_id_to_reco_mode_mapping()

    @classmethod
    def run_checks(cls) -> None:
        assert cls.sig_col_name, (cls.sig_col_name, cls.__name__)
        assert cls.sig_col_name == cls.get_sig_col_name(), (cls.sig_col_name, cls.get_sig_col_name(), cls.__name__)

        assert isinstance(cls.get_sig_id_class(), type(AbstractBaseMCSigIDsClass)), type(cls.get_sig_id_class()).__name__

        __all_sig_ids_in_query_info: Tuple[int, ...] = cls.get_all_sig_ids()

        assert len(__all_sig_ids_in_query_info) == len(set(__all_sig_ids_in_query_info)), (
            len(__all_sig_ids_in_query_info),
            len(set(__all_sig_ids_in_query_info)),
            set([i for i in set(__all_sig_ids_in_query_info) if __all_sig_ids_in_query_info.count(i) > 1]),
        )

        __all_defined_sig_ids: Tuple[int, ...] = cls.get_sig_id_class().get_all_sig_ids()
        _sig_id_dev_class_name: str = cls.get_sig_id_class().__name__

        assert all([i in __all_defined_sig_ids for i in __all_sig_ids_in_query_info]), (
            f"Sig IDs used in {cls.__name__} that is NOT DEFINED in {_sig_id_dev_class_name} class:",
            [i for i in __all_sig_ids_in_query_info if i not in __all_defined_sig_ids],
        )

        assert all([i in __all_sig_ids_in_query_info for i in __all_defined_sig_ids]), (
            f"Sig IDs defined in {_sig_id_dev_class_name} class that are NOT USED in {cls.__name__}:",
            [i for i in __all_defined_sig_ids if i not in __all_sig_ids_in_query_info],
        )

        _latex_labels: Tuple[str, ...] = tuple(cls.get_sig_id_to_latex_label().values())
        assert len(_latex_labels) == len(set(_latex_labels)), (
            len(_latex_labels),
            len(set(_latex_labels)),
            set([latex_label for latex_label in set(_latex_labels) if _latex_labels.count(latex_label) > 1]),
        )
