"""
Old / Original MC matching information.

Felix Metzner 2023
"""

from typing import Tuple, Type

from templatefitter.plotter import plot_style

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos

from rdstar.offline_analysis.mc_matching.abstract_mc_matching_info import AbstractBaseMCSigIDsClass
from rdstar.offline_analysis.mc_matching.abstract_mc_matching_info import AbstractBaseMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import (
    MCSigID,
    MatchQueryInfo,
    MCSigIDNameToQueryMappingType,
    MCSigIDToLatexLabelMappingType,
    MCSigIDNameToLatexLabelMappingType,
    RecoModeToExpectedSignalSigIDMappingType,
)

from rdstar.offline_analysis.mc_matching.old_mc_matching_queries import OldMCSigQueries


__all__ = [
    "OldMCSigIDs",
    "OldMCMatchingIDsQueryInfo",
]


class OldMCSigIDs(AbstractBaseMCSigIDsClass):
    # fmt: off
    b_to_d_tau_signal: MCSigID = MCSigID(id=1, expected_signal_in_decay_ids=RecoModeInfos.b_to_d_modes.ids, color=plot_style.KITColors.kit_green)
    b_to_dst_tau_signal: MCSigID = MCSigID(id=2, expected_signal_in_decay_ids=RecoModeInfos.b_to_ds_modes.ids, color=plot_style.KITColors.kit_maygreen)
    b_to_d_l_signal: MCSigID = MCSigID(id=3, expected_signal_in_decay_ids=RecoModeInfos.b_to_d_modes.ids, color=plot_style.KITColors.kit_blue)
    b_to_dst_l_signal: MCSigID = MCSigID(id=4, expected_signal_in_decay_ids=RecoModeInfos.b_to_ds_modes.ids, color=plot_style.KITColors.kit_cyan)
    b_to_dstst_l_signal: MCSigID = MCSigID(id=5, expected_signal_in_decay_ids=RecoModeInfos.d_ss_neutral_pion_reco_modes.ids, color=plot_style.KITColors.kit_red)
    b_to_dstst_l_bkg: MCSigID = MCSigID(id=6, expected_signal_in_decay_ids=RecoModeInfos.d_ss_charged_pion_reco_modes.ids, color=plot_style.KITColors.kit_orange)
    hadronic_bkg: MCSigID = MCSigID(id=7, expected_signal_in_decay_ids=None, color=plot_style.KITColors.kit_brown)
    bbar_bkg: MCSigID = MCSigID(id=8, expected_signal_in_decay_ids=None, color=plot_style.KITColors.grey)
    continuum_bkg: MCSigID = MCSigID(id=9, expected_signal_in_decay_ids=None, color=plot_style.KITColors.light_grey)
    b_to_pi_l_nu_bkg: MCSigID = MCSigID(id=10, expected_signal_in_decay_ids=RecoModeInfos.pi_l_nu_reco_modes.ids, color=plot_style.KITColors.kit_yellow)
    # fmt: on


class OldMCMatchingIDsQueryInfo(AbstractBaseMCMatchingIDsQueryInfo):
    sig_col_name: str = Cols.base_sig_id_col
    __required_columns: Tuple[str, ...] = OldMCSigQueries.required_columns
    __sig_ids_definition_class: Type[AbstractBaseMCSigIDsClass] = OldMCSigIDs

    @classmethod
    def get_sig_col_name(cls) -> str:
        return cls.sig_col_name

    @classmethod
    def get_required_columns(cls) -> Tuple[str, ...]:
        return cls.__required_columns

    @classmethod
    def get_sig_id_class(cls) -> Type[AbstractBaseMCSigIDsClass]:
        return cls.__sig_ids_definition_class

    # fmt: off
    Dtau: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_d_tau_signal, query=OldMCSigQueries.b_to_d_tau_signal, latex_label=LatexDecayProcesses.to_d_tau_nu)
    DStau: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_dst_tau_signal, query=OldMCSigQueries.b_to_dst_tau_signal, latex_label=LatexDecayProcesses.to_d_st_tau_nu)
    Dl: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_d_l_signal, query=OldMCSigQueries.b_to_d_l_signal, latex_label=LatexDecayProcesses.to_d_l_nu)
    DSl: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_dst_l_signal, query=OldMCSigQueries.b_to_dst_l_signal, latex_label=LatexDecayProcesses.to_d_st_l_nu)
    DSSl: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_dstst_l_signal, query=OldMCSigQueries.b_to_dstst_l_signal, latex_label=LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero)
    DSSlbkg: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_dstst_l_bkg, query=OldMCSigQueries.b_to_dstst_l_bkg, latex_label=LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm)
    HBKG: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.hadronic_bkg, query=OldMCSigQueries.hadronic_bkg, latex_label=LatexDecayProcesses.to_hadronic_bkg_txt)
    GBKG: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.bbar_bkg, query=OldMCSigQueries.bbar_bkg, latex_label=LatexDecayProcesses.to_b_bbar_bkg)
    PiLNu: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.b_to_pi_l_nu_bkg, query=OldMCSigQueries.b_to_pi_l_nu_signal, latex_label=LatexDecayProcesses.to_pi_l_nu)
    CBKG: MatchQueryInfo = MatchQueryInfo(id_info=OldMCSigIDs.continuum_bkg, query=OldMCSigQueries.continuum_bkg, latex_label=LatexDecayProcesses.to_continuum_txt)
    # fmt: on

    @classmethod
    def get_sig_id_mapping_without_pilnu(cls) -> MCSigIDNameToQueryMappingType:
        return {name: info for name, info in cls.get_sig_id_mapping().items() if info.id != OldMCSigIDs.b_to_pi_l_nu_bkg}

    @classmethod
    def get_sig_id_mapping_without_c_bkg(cls) -> MCSigIDNameToQueryMappingType:
        return {name: info for name, info in cls.get_sig_id_mapping().items() if info.id != OldMCSigIDs.continuum_bkg}

    @classmethod
    def get_sig_id_to_latex_label_without_pilnu(cls) -> MCSigIDToLatexLabelMappingType:
        return {sig_id: l for sig_id, l in cls.get_sig_id_to_latex_label().items() if sig_id != OldMCSigIDs.continuum_bkg}

    @classmethod
    def get_sig_id_name_to_latex_label_without_pilnu(cls) -> MCSigIDNameToLatexLabelMappingType:
        return {
            name: info.latex_label
            for name, info in cls.get_sig_id_mapping().items()
            if info.id != OldMCSigIDs.continuum_bkg
        }


original_decay_mode_to_sig_id: RecoModeToExpectedSignalSigIDMappingType = {
    11: (1, 3),  #  B^0 to D- e+                # noqa: E262
    12: (1, 3),  #  B^0 to D- mu+               # noqa: E262
    13: (1, 3),  #  B^- to D0 e-                # noqa: E262
    14: (1, 3),  #  B^- to D0 mu-               # noqa: E262
    15: (2, 4),  #  B^0 to D*- e+               # noqa: E262
    16: (2, 4),  #  B^0 to D*- mu+              # noqa: E262
    17: (2, 4),  #  B^- to D*0 e-               # noqa: E262
    18: (2, 4),  #  B^- to D*0 mu-              # noqa: E262
    61: (5,),  #     B0 to D- e+ pi0             # noqa: E262
    62: (5,),  #     B0 to D- mu+ pi0            # noqa: E262
    63: (5,),  #     B- to D0 e- pi0             # noqa: E262
    64: (5,),  #     B- to D0 mu- pi0            # noqa: E262
    65: (5,),  #     B0 to D*- e+ pi0            # noqa: E262
    66: (5,),  #     B0 to D*- mu+ pi0           # noqa: E262
    67: (5,),  #     B- to D*0 e- pi0            # noqa: E262
    68: (5,),  #     B- to D*0 mu- pi0           # noqa: E262
    71: (6,),  #     B0 to anti-D0 e+ pi-        # noqa: E262
    72: (6,),  #     B0 to anti-D0 mu+ pi-       # noqa: E262
    73: (6,),  #     B- to D+ e- pi-             # noqa: E262
    74: (6,),  #     B- to D+ mu- pi-            # noqa: E262
    75: (6,),  #     B0 to anti-D*0 e+ pi-       # noqa: E262
    76: (6,),  #     B0 to anti-D*0 mu+ pi-      # noqa: E262
    77: (6,),  #     B- to D*+ e- pi-            # noqa: E262
    78: (6,),  #     B- to D*+ mu- pi-           # noqa: E262
    81: (10,),  #    B0 to pi+ e-                # noqa: E262
    82: (10,),  #    B0 to pi+ mu-               # noqa: E262
    83: (10,),  #    B- to pi0 e-                # noqa: E262
    84: (10,),  #    B- to pi0 mu-               # noqa: E262
}


if __name__ == "__main__":
    # Checks for OldMCSigIDs
    OldMCSigIDs.run_checks()

    # Checks for OldMCMatchingIDs
    OldMCMatchingIDsQueryInfo.run_checks()

    for k, v in OldMCSigIDs.get_sig_ids_dict().items():
        print(f"{k}: {v}")

    print(OldMCSigIDs.get_all_sig_ids())

    new_decay_mode_to_sig_id: RecoModeToExpectedSignalSigIDMappingType = OldMCSigIDs.get_reco_mode_to_sig_id_mapping()
    assert original_decay_mode_to_sig_id == new_decay_mode_to_sig_id

    print(new_decay_mode_to_sig_id)

    for mqi_key, mqi in OldMCMatchingIDsQueryInfo.get_sig_id_mapping().items():
        print(f"{mqi_key}: {mqi}")
