"""
Setups for correction of Squared Missing Mass resolution with iterative shifting and smearing.

Felix Metzner 2021
"""

from collections import Counter
from dataclasses import dataclass
from typing import Optional, Tuple, List, Dict

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.weight_management.weight_information import WeightManager

from rdstar.offline_analysis.mms_resolution_correction import resolution_correction_setups as res_corr_setup
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit_setups import (
    ShiftEvaluationInfo,
    missing_mass_like_for_res_corr_eval_setup,
    missing_mass_with_fei_calib_like_for_res_corr_eval_setup,
)


__all__ = [
    "MultistageResolutionCorrectionSetup",
    "multi_stage_setups_tuple",
    "multi_stage_setups_dict",
    "main_multi_stage_setup_name",
    "main_msrc_corrected_mms_column_name",
]


@dataclass(frozen=True)
class MultistageResolutionCorrectionSetup:
    name: str
    fei_weight_col: Optional[str]
    shift_weight_column: str
    resolution_weight_column: str
    first_shift_fit_case_key: str
    base_first_shift_fit: ShiftEvaluationInfo
    base_first_resolution_corr_setup: res_corr_setup.ResolutionCorrectionSetup
    second_shift_fit_case_key: Optional[str] = None
    base_second_shift_fit: Optional[ShiftEvaluationInfo] = None
    base_second_resolution_corr_setup: Optional[res_corr_setup.ResolutionCorrectionSetup] = None
    use_laplace_smeared_as_final: bool = False

    def __post_init__(self) -> None:
        if int(self.base_second_shift_fit is None) + int(self.second_shift_fit_case_key is None) == 1:
            raise ValueError("If a second iteration is desired, both parts of the shift fir setup must be defined!")
        if int(self.base_second_shift_fit is None) + int(self.base_second_resolution_corr_setup is None) == 1:
            raise ValueError("If a second iteration is desired, both parts of it must be defined, but only one was!")

        assert self.base_first_shift_fit.hist_var.df_label == Cols.original_missing_mass, (
            self.base_first_shift_fit.hist_var.df_label,
            Cols.original_missing_mass,
        )

        assert self.fei_weight_col == self.base_first_shift_fit.fei_weight_col, (
            self.fei_weight_col,
            self.base_first_shift_fit.fei_weight_col,
        )
        assert self.base_first_resolution_corr_setup.fei_weight_col == self.fei_weight_col, (
            self.name,
            self.fei_weight_col,
            self.base_first_resolution_corr_setup.name,
            self.base_first_resolution_corr_setup.fei_weight_col,
        )

        if self.base_second_resolution_corr_setup is not None:
            assert self.base_second_resolution_corr_setup.fei_weight_col == self.fei_weight_col, (
                self.name,
                self.fei_weight_col,
                self.base_second_resolution_corr_setup.name,
                self.base_second_resolution_corr_setup.fei_weight_col,
            )

    @property
    def has_second_iteration(self) -> bool:
        assert int(self.base_second_shift_fit is None) + int(self.second_shift_fit_case_key is None) != 1
        assert int(self.base_second_shift_fit is None) + int(self.base_second_resolution_corr_setup is None) != 1
        return self.base_second_shift_fit is not None

    @property
    def original_col_name(self) -> str:
        return self.base_first_shift_fit.hist_var.df_label

    @property
    def corrected_col_names(self) -> Tuple[str, ...]:
        new_cols = [
            self.first_shift_fit_result_col_name,
            *self.first_res_corr_result_col_names,
        ]  # type: List[str]

        if self.has_second_iteration:
            new_cols.append(self.second_shift_fit_result_col_name)
            new_cols.extend(self.second_res_corr_result_col_names)
        return tuple(new_cols)

    @property
    def final_corrected_col_name(self) -> str:
        res_index = 1 if self.use_laplace_smeared_as_final else 0  # type: int
        if self.has_second_iteration:
            return self.second_res_corr_result_col_names[res_index]
        else:
            return self.first_res_corr_result_col_names[res_index]

    @property
    def first_shift_fit(self) -> ShiftEvaluationInfo:
        return self.base_first_shift_fit.create_copy_with_extended_new_col_name_appendix(new_appendix="after_1stShift")

    @property
    def first_shift_fit_result_col_name(self) -> str:
        return self.first_shift_fit.get_new_col_name(
            base_col_name=self.base_first_shift_fit.hist_var.df_label,
            shift_case_name=self.first_shift_fit_case_key,
        )

    @property
    def first_resolution_corr_setup(self) -> res_corr_setup.ResolutionCorrectionSetup:
        base_col_name = self.base_first_resolution_corr_setup.base_result_col_name  # type: str

        if self.has_second_iteration:
            new_base_col_name = f"{base_col_name}_S1"  # type: str
        else:
            new_base_col_name = f"{self.original_col_name}_{self.name}_multi_stage"
        return self.base_first_resolution_corr_setup.get_with_new_in_and_output_cols(
            new_input_column_name=self.first_shift_fit_result_col_name,
            new_result_column_base_name=new_base_col_name,
        )

    @property
    def first_res_corr_result_col_names(self) -> Tuple[str, str]:
        res_cor_setup = self.first_resolution_corr_setup
        gaussian = res_cor_setup.get_result_col_name(appendix=None)
        laplace = res_cor_setup.get_result_col_name(appendix=MissingMassResolutionCorrector.laplace_smearing_appendix)
        return gaussian, laplace

    @property
    def second_shift_fit(self) -> ShiftEvaluationInfo:
        assert self.has_second_iteration
        assert self.base_second_shift_fit is not None
        return self.base_second_shift_fit.create_copy_with_new_in_and_output_cols(
            new_base_column_name=self.first_res_corr_result_col_names[1],
            new_appendix="after_2ndshift",
        )

    @property
    def second_shift_fit_result_col_name(self) -> str:
        assert self.has_second_iteration
        assert self.second_shift_fit_case_key is not None
        second_shift_fit_info = self.second_shift_fit
        return second_shift_fit_info.get_new_col_name(
            base_col_name=second_shift_fit_info.hist_var.df_label,
            shift_case_name=self.second_shift_fit_case_key,
        )

    @property
    def second_resolution_corr_setup(self) -> res_corr_setup.ResolutionCorrectionSetup:
        assert self.has_second_iteration
        assert self.base_second_resolution_corr_setup is not None

        new_base_col_name = f"{self.original_col_name}_{self.name}_multi_stage"  # type: str
        return self.base_second_resolution_corr_setup.get_with_new_in_and_output_cols(
            new_input_column_name=self.second_shift_fit_result_col_name,
            new_result_column_base_name=new_base_col_name,
        )

    @property
    def second_res_corr_result_col_names(self) -> Tuple[str, str]:
        assert self.has_second_iteration
        res_cor_setup = self.second_resolution_corr_setup
        gaussian = res_cor_setup.get_result_col_name(appendix=None)
        laplace = res_cor_setup.get_result_col_name(appendix=MissingMassResolutionCorrector.laplace_smearing_appendix)
        return gaussian, laplace


_undefined_column_name_str = "ThisColumnNameWillNotExistBecauseItMustBeSetToMatchPreviousStage"  # type: str

second_step_shift_missing_mass_hist_var = HistVariable(
    df_label=_undefined_column_name_str,
    n_bins=48,
    scope=missing_mass_like_for_res_corr_eval_setup.hist_var.scope,  # (-2.0, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

first_res_corr_setup_name = "mbc_based_mms_resolution_correction"
second_res_corr_setup_name = "b_sig_meson_reco_mode_id_based_mms_resolution_correction"

# fmt: off
tag_mbc_and_reco_id_mrcs = MultistageResolutionCorrectionSetup(
    name="tag_mbc_and_reco_id",
    fei_weight_col=None,
    shift_weight_column=WeightManager.final_weight_with_additions_col,
    resolution_weight_column=WeightManager.final_weight_with_additions_col,
    first_shift_fit_case_key="BTagMbc",
    base_first_shift_fit=missing_mass_like_for_res_corr_eval_setup,
    base_first_resolution_corr_setup=res_corr_setup.default_res_cor_setup,
    second_shift_fit_case_key="RecoMode",
    base_second_shift_fit=missing_mass_like_for_res_corr_eval_setup.create_copy_with_other_hist_var(new_hist_var=second_step_shift_missing_mass_hist_var),
    base_second_resolution_corr_setup=res_corr_setup.b_sig_decay_id_based_res_cor_setup,
    use_laplace_smeared_as_final=True,
)

tag_mbc_and_reco_id_mrcs_with_fei_calib = MultistageResolutionCorrectionSetup(
    name="tag_mbc_and_reco_id_with_fei_calib",
    fei_weight_col=WeightManager.used_fei_calib_col,
    shift_weight_column=WeightManager.final_weight_with_additions_col,
    resolution_weight_column=WeightManager.final_weight_with_additions_col,
    first_shift_fit_case_key="BTagMbc",
    base_first_shift_fit=missing_mass_with_fei_calib_like_for_res_corr_eval_setup,
    base_first_resolution_corr_setup=res_corr_setup.default_res_cor_setup.get_with_other_fei_weight_col(
        new_fei_weight_col=WeightManager.used_fei_calib_col,
        new_fei_weight_col_tag="wNewFEIcal",
    ),
    second_shift_fit_case_key="RecoMode",
    base_second_shift_fit=missing_mass_with_fei_calib_like_for_res_corr_eval_setup.create_copy_with_other_hist_var(new_hist_var=second_step_shift_missing_mass_hist_var),
    base_second_resolution_corr_setup=res_corr_setup.b_sig_decay_id_based_res_cor_setup.get_with_other_fei_weight_col(
        new_fei_weight_col=WeightManager.used_fei_calib_col,
        new_fei_weight_col_tag="wNewFEIcal",
    ),
    use_laplace_smeared_as_final=True,
)
# fmt: on


_multi_stage_setups_tuple = tuple(
    [
        tag_mbc_and_reco_id_mrcs,
        tag_mbc_and_reco_id_mrcs_with_fei_calib,
    ]
)  # type: Tuple[MultistageResolutionCorrectionSetup, ...]


assert len(set([v.name for v in _multi_stage_setups_tuple])) == len(_multi_stage_setups_tuple), [
    (k, v) for k, v in Counter([v.name for v in _multi_stage_setups_tuple]).items() if v > 1
]


multi_stage_setups_tuple = _multi_stage_setups_tuple  # type: Tuple[MultistageResolutionCorrectionSetup, ...]

multi_stage_setups_dict = {
    v.name: v for v in multi_stage_setups_tuple
}  # type: Dict[str, MultistageResolutionCorrectionSetup]


main_multi_stage_setup_name = tag_mbc_and_reco_id_mrcs.name  # type: str
main_msrc_corrected_mms_column_name = multi_stage_setups_dict[
    main_multi_stage_setup_name
].final_corrected_col_name  # type: str
