"""
Definition of Missing Mass Squared resolution correction setups

Felix Metzner 2021
"""

from typing import Optional, Tuple, NamedTuple, Dict, Callable

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.observables.e_miss_minus_p_miss import EMissMinusPMissVariable
from rdstar.offline_analysis.observables.missing_mass_variants import MissingMassSquaredWOVertexFitVariable
from rdstar.offline_analysis.observables.kinematic_observables import LeptonPhaseSpaceBinID, KinematicObservables
from rdstar.offline_analysis.observables.reco_modes_observables import (
    RecoIDObservables,
    SimpleDMesonRecoIDDefinition,
    DAndDsMesonRecoIDDefinition,
    BSigMesonRecoIDDefinition,
    BTagMesonRecoIDDefinition,
    U4SRecoIDDefinition,
)
from rdstar.offline_analysis.observables.decay_process_information import (
    DecayProcessVariables,
    BSigTrackCountVariable,
    BSigNeutralCountVariable,
    BTagTrackCountVariable,
    BTagNeutralCountVariable,
)


__all__ = [
    "missing_mass_hist_var",
    "ResolutionCorrectionSetup",
    "ResolutionCorrectionSetupManager",
    # Important setups:
    "default_res_cor_setup",
    "before_vf_mbc_res_cor_setup",
    "e_miss_minus_p_miss_res_corr_setup",
    "lepton_phase_space_based_res_cor_setup",
]

missing_mass_hist_var = HistVariable(
    df_label=Cols.original_missing_mass,
    n_bins=250,
    scope=(-0.5, 2.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

missing_mass_before_vf_hist_var = HistVariable(
    df_label=MissingMassSquaredWOVertexFitVariable.mm2_wo_vf_col_name,
    n_bins=250,
    scope=(-0.5, 2.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

default_mbc_hist_var = HistVariable(
    df_label=Cols.b_tag_mbc,
    n_bins=100,  # Only for visualization, as associated binning will be provided by the dedicated Binning instance.
    scope=(5.270, 5.290),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

default_mbc_binning = Binning(
    bins=(5.2700, 5.2750, 5.2780, 5.2805, 5.2835, 5.290),  # Binning by eye.
    dimensions=1,
    scope=default_mbc_hist_var.scope,
    log_scale=False,
)

w_reco_hist_var = HistVariable(
    df_label=Cols.w_reco,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(1.0, 2.0),
    var_name=LatexObsNames.ff_w,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

w_reco_binning = Binning(
    bins=(1.0, 1.25204834, 1.33214282, 1.39304942, 1.42476468, 1.46135439, 1.50742468, 2.0),  # Eq. fr. binning from mc.
    dimensions=1,
    scope=w_reco_hist_var.scope,
    log_scale=False,
)

q2_reco_hist_var = HistVariable(
    df_label=Cols.q_squared_reco,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(-20.0, 4.0),
    var_name=LatexObsNames.fitting_q2,
    unit=LatexUnits.gev_sq,
    use_log_scale=False,
)

q2_reco_binning = Binning(
    bins=(-20.0, 0.67099225, 1.4066534, 2.01698814, 2.56469315, 3.0785986, 3.57038442, 4.0),  # Eq. fr. binning from mc.
    dimensions=1,
    scope=q2_reco_hist_var.scope,
    log_scale=False,
)

lepton_p_hist_var = HistVariable(
    df_label=Cols.p_lepton_lab_frame,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 2.0),
    var_name=LatexObsNames.lepton_p,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

lepton_p_binning = Binning(
    bins=LeptonPhaseSpaceBinID.lepton_p_5_bins_equal_frequency_binning,  # Equal frequency binning from mc.
    dimensions=1,
    scope=lepton_p_hist_var.scope,
    log_scale=False,
)

lepton_theta_hist_var = HistVariable(
    df_label=Cols.lepton_theta,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 3.0),
    var_name=LatexObsNames.lepton_theta,
    unit=LatexUnits.rad,
    use_log_scale=False,
)

lepton_theta_binning = Binning(
    bins=LeptonPhaseSpaceBinID.lepton_theta_5_bins_equal_frequency_binning,  # Equal frequency binning from mc.
    dimensions=1,
    scope=lepton_theta_hist_var.scope,
    log_scale=False,
)

event_track_count_hist_var = HistVariable(
    df_label=Cols.n_clean_tracks_in_roe,
    n_bins=21,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.5, 20.5),
    var_name=LatexObsNames.decay_info_number_of_tracks_in_event,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

event_track_count_binning = Binning(
    bins=(0.5, 7.5, 8.5, 9.5, 10.5, 20.5),  # Binning by eye.
    dimensions=1,
    scope=event_track_count_hist_var.scope,
    log_scale=False,
)

event_neutral_count_hist_var = HistVariable(
    df_label=Cols.n_neutral_clusters_in_roe,
    n_bins=27,  # Only for visualization; is overwritten by binning instance below.
    scope=(-0.5, 26.5),
    var_name=LatexObsNames.decay_info_number_of_neutrals_in_event,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

event_neutral_count_binning = Binning(
    bins=(-0.5, 2.5, 4.5, 6.5, 8.5, 26.5),  # Binning by eye.
    dimensions=1,
    scope=event_neutral_count_hist_var.scope,
    log_scale=False,
)

fei_probability_hist_var = HistVariable(
    df_label=Cols.fei_probability,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 1.0),
    var_name=LatexObsNames.fei_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

fei_probability_binning = Binning(
    bins=(0.0, 0.00214756, 0.00561211, 0.019896, 0.1106317, 1.0),  # Equal frequency binning determined on mc sample.
    dimensions=1,
    scope=fei_probability_hist_var.scope,
    log_scale=False,
)

b_sig_vertex_fit_hist_var = HistVariable(
    df_label=Cols.b_sig_vertex_fit_prob,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(0.0, 1.0),
    var_name=LatexObsNames.b_sig_vertex_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

b_sig_vertex_fit_binning = Binning(
    bins=(0.0, 4.29092033e-04, 6.24699183e-02, 2.86736459e-01, 6.16876781e-01, 1.0),  # Equal frequency binning from mc.
    dimensions=1,
    scope=b_sig_vertex_fit_hist_var.scope,
    log_scale=False,
)

b_tag_daughter_vertex_fit_hist_var = HistVariable(
    df_label=Cols.b_tag_vertex_fit_prob,
    n_bins=50,  # Only for visualization; is overwritten by binning instance below.
    scope=(-1.0, 1.0),
    var_name=LatexObsNames.b_tag_daughter_vertex_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

b_tag_daughter_vertex_fit_binning = Binning(
    bins=(-1.0, 0.0, 0.4258008, 1.0),  # Equal frequency binning (as good as possible) determined on mc sample.
    dimensions=1,
    scope=b_tag_daughter_vertex_fit_hist_var.scope,
    log_scale=False,
)


class ResolutionCorrectionSetup(NamedTuple):
    name: str
    fei_weight_col: Optional[str]
    latex_title_appendix: str
    corrected_variable_appendix: str
    primary_hist_var: HistVariable
    secondary_hist_var: HistVariable
    secondary_var_binning: Optional[Binning]
    variable_definition: Optional[Callable] = None
    variable_definition_class: Optional[object] = None
    tick_label_mapping: Optional[Dict[int, str]] = None
    use_alternative_gap: bool = False
    overwrite_for_new_column_name: Optional[str] = None

    @staticmethod
    def get_default_correction_col_tag() -> str:
        return "CorrRes"

    def get_with_alternative_gap_setup(self) -> "ResolutionCorrectionSetup":
        assert self.use_alternative_gap is False, self.name

        alt_gap_tag = "_with_alt_gap"  # type: str
        return ResolutionCorrectionSetup(
            name=self.name + alt_gap_tag,
            fei_weight_col=self.fei_weight_col,
            latex_title_appendix=self.latex_title_appendix,
            corrected_variable_appendix=self.corrected_variable_appendix + alt_gap_tag,
            primary_hist_var=self.primary_hist_var,
            secondary_hist_var=self.secondary_hist_var,
            secondary_var_binning=self.secondary_var_binning,
            variable_definition=self.variable_definition,
            variable_definition_class=self.variable_definition_class,
            tick_label_mapping=self.tick_label_mapping,
            use_alternative_gap=True,
            overwrite_for_new_column_name=self.overwrite_for_new_column_name,
        )

    def get_with_other_fei_weight_col(
        self,
        new_fei_weight_col: Optional[str],
        new_fei_weight_col_tag: str,
    ) -> "ResolutionCorrectionSetup":
        assert self.fei_weight_col != new_fei_weight_col, (self.name, self.fei_weight_col, new_fei_weight_col)
        _new_fei_weight_col_tag = f"_{new_fei_weight_col_tag}"  # type: str

        return ResolutionCorrectionSetup(
            name=self.name + _new_fei_weight_col_tag,
            fei_weight_col=new_fei_weight_col,
            latex_title_appendix=self.latex_title_appendix,
            corrected_variable_appendix=self.corrected_variable_appendix + _new_fei_weight_col_tag,
            primary_hist_var=self.primary_hist_var,
            secondary_hist_var=self.secondary_hist_var,
            secondary_var_binning=self.secondary_var_binning,
            variable_definition=self.variable_definition,
            variable_definition_class=self.variable_definition_class,
            tick_label_mapping=self.tick_label_mapping,
            use_alternative_gap=self.use_alternative_gap,
            overwrite_for_new_column_name=self.overwrite_for_new_column_name,
        )

    def get_with_other_primary_hist_val_column_name(
        self,
        new_column_name: str,
        appendix: str,
    ) -> "ResolutionCorrectionSetup":
        new_primary_hist_var = self.get_new_hist_var_for(new_column_name=new_column_name)  # type: HistVariable

        return ResolutionCorrectionSetup(
            name=f"{self.name}_{appendix}",
            fei_weight_col=self.fei_weight_col,
            latex_title_appendix=self.latex_title_appendix,
            corrected_variable_appendix=f"{self.corrected_variable_appendix}_{appendix}",
            primary_hist_var=new_primary_hist_var,
            secondary_hist_var=self.secondary_hist_var,
            secondary_var_binning=self.secondary_var_binning,
            variable_definition=self.variable_definition,
            variable_definition_class=self.variable_definition_class,
            tick_label_mapping=self.tick_label_mapping,
            use_alternative_gap=self.use_alternative_gap,
            overwrite_for_new_column_name=self.overwrite_for_new_column_name,
        )

    def get_with_other_new_column_base_name(self, new_result_column_base_name: str) -> "ResolutionCorrectionSetup":
        assert self.overwrite_for_new_column_name is None, self.overwrite_for_new_column_name
        return ResolutionCorrectionSetup(
            name=self.name,
            fei_weight_col=self.fei_weight_col,
            latex_title_appendix=self.latex_title_appendix,
            corrected_variable_appendix=self.corrected_variable_appendix,
            primary_hist_var=self.primary_hist_var,
            secondary_hist_var=self.secondary_hist_var,
            secondary_var_binning=self.secondary_var_binning,
            variable_definition=self.variable_definition,
            variable_definition_class=self.variable_definition_class,
            tick_label_mapping=self.tick_label_mapping,
            use_alternative_gap=self.use_alternative_gap,
            overwrite_for_new_column_name=new_result_column_base_name,
        )

    def get_with_new_in_and_output_cols(
        self,
        new_input_column_name: str,
        new_result_column_base_name: str,
    ) -> "ResolutionCorrectionSetup":
        assert self.overwrite_for_new_column_name is None, self.overwrite_for_new_column_name
        new_primary_hist_var = self.get_new_hist_var_for(new_column_name=new_input_column_name)  # type: HistVariable

        return ResolutionCorrectionSetup(
            name=self.name,
            fei_weight_col=self.fei_weight_col,
            latex_title_appendix=self.latex_title_appendix,
            corrected_variable_appendix=self.corrected_variable_appendix,
            primary_hist_var=new_primary_hist_var,
            secondary_hist_var=self.secondary_hist_var,
            secondary_var_binning=self.secondary_var_binning,
            variable_definition=self.variable_definition,
            variable_definition_class=self.variable_definition_class,
            tick_label_mapping=self.tick_label_mapping,
            use_alternative_gap=self.use_alternative_gap,
            overwrite_for_new_column_name=new_result_column_base_name,
        )

    def get_new_hist_var_for(self, new_column_name: str) -> HistVariable:
        assert new_column_name != self.primary_col_name, (new_column_name, self.primary_col_name)

        return HistVariable(
            df_label=new_column_name,
            n_bins=self.primary_hist_var.n_bins,
            scope=self.primary_hist_var.scope,
            var_name=self.primary_hist_var.variable_name,
            unit=self.primary_hist_var.unit if self.primary_hist_var.unit else None,
            use_log_scale=self.primary_hist_var.use_log_scale,
        )

    @property
    def primary_col_name(self) -> str:
        return self.primary_hist_var.df_label

    @property
    def base_result_col_name(self) -> str:
        return f"{self.primary_col_name}_{self.corrected_variable_appendix}"

    def get_result_col_name(self, appendix: Optional[str] = None) -> str:
        base_name = self.base_result_col_name  # type: str

        col_end = self.get_default_correction_col_tag()  # type: str
        final_appendix = "" if appendix is None else f"_{appendix}"  # type: str

        if self.overwrite_for_new_column_name is not None:
            assert base_name != self.overwrite_for_new_column_name, (base_name, self.overwrite_for_new_column_name)
            base_name = self.overwrite_for_new_column_name

        return f"{base_name}_{col_end}{final_appendix}"


default_res_cor_setup = ResolutionCorrectionSetup(
    name="mbc_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.b_tag_mbc,
    corrected_variable_appendix="mbc_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=default_mbc_hist_var,
    secondary_var_binning=default_mbc_binning,
)

before_vf_mbc_res_cor_setup = ResolutionCorrectionSetup(
    name="mbc_based_mms_before_vf_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.b_tag_mbc,
    corrected_variable_appendix="mbc_based",
    primary_hist_var=missing_mass_before_vf_hist_var,
    secondary_hist_var=default_mbc_hist_var,
    secondary_var_binning=default_mbc_binning,
)

w_reco_res_cor_setup = ResolutionCorrectionSetup(
    name="w_reco_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.ff_w,
    corrected_variable_appendix="w_reco_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=w_reco_hist_var,
    secondary_var_binning=w_reco_binning,
)

q2_reco_res_cor_setup = ResolutionCorrectionSetup(
    name="q2_reco_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.fitting_q2,
    corrected_variable_appendix="q2_reco_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=q2_reco_hist_var,
    secondary_var_binning=q2_reco_binning,
)

lepton_p_res_cor_setup = ResolutionCorrectionSetup(
    name="lepton_p_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.lepton_p,
    corrected_variable_appendix="lepton_p_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=lepton_p_hist_var,
    secondary_var_binning=lepton_p_binning,
)

lepton_theta_res_cor_setup = ResolutionCorrectionSetup(
    name="lepton_theta_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.lepton_theta,
    corrected_variable_appendix="lep_theta_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=lepton_theta_hist_var,
    secondary_var_binning=lepton_theta_binning,
)

lepton_phase_space_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{LeptonPhaseSpaceBinID.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.lepton_p_theta_bin,
    corrected_variable_appendix="lep_PHS_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=LeptonPhaseSpaceBinID.hist_var,
    secondary_var_binning=LeptonPhaseSpaceBinID.binning,
    variable_definition=KinematicObservables.get_definition_function(name=LeptonPhaseSpaceBinID.new_column_name),
    variable_definition_class=KinematicObservables.get_kinematic_definition(name=LeptonPhaseSpaceBinID.new_column_name),
    tick_label_mapping=LeptonPhaseSpaceBinID.get_tick_label_mapping(),
)

before_vf_lepton_phase_space_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{LeptonPhaseSpaceBinID.new_column_name}_based_mms_before_vf_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.lepton_p_theta_bin,
    corrected_variable_appendix="lep_PHS_based",
    primary_hist_var=missing_mass_before_vf_hist_var,
    secondary_hist_var=LeptonPhaseSpaceBinID.hist_var,
    secondary_var_binning=LeptonPhaseSpaceBinID.binning,
    variable_definition=KinematicObservables.get_definition_function(
        name=LeptonPhaseSpaceBinID.new_column_name, fail_when_exists=False
    ),
    variable_definition_class=KinematicObservables.get_kinematic_definition(name=LeptonPhaseSpaceBinID.new_column_name),
    tick_label_mapping=LeptonPhaseSpaceBinID.get_tick_label_mapping(),
)

d_meson_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{SimpleDMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.reco_d_reco_mode_label,
    corrected_variable_appendix="simple_DRecoID_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=SimpleDMesonRecoIDDefinition.hist_var,
    secondary_var_binning=SimpleDMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=SimpleDMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=SimpleDMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=SimpleDMesonRecoIDDefinition.get_tick_label_mapping(),
)

d_and_dst_meson_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{DAndDsMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.reco_d_or_dst_reco_mode_label,
    corrected_variable_appendix="D_Dst_RecoID_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=DAndDsMesonRecoIDDefinition.hist_var,
    secondary_var_binning=DAndDsMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=DAndDsMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=DAndDsMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=DAndDsMesonRecoIDDefinition.get_tick_label_mapping(),
)

b_sig_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.reco_b_sig_reco_mode_label,
    corrected_variable_appendix="bsig_RecoID_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigMesonRecoIDDefinition.hist_var,
    secondary_var_binning=BSigMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=BSigMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=BSigMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=BSigMesonRecoIDDefinition.get_tick_label_mapping(),
)

signal_neutrals_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigNeutralCountVariable.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_b_sig_short,
    corrected_variable_appendix="sig_neutrals_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigNeutralCountVariable.hist_var,
    secondary_var_binning=BSigNeutralCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BSigNeutralCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BSigNeutralCountVariable.new_column_name
    ),
)

signal_track_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BSigTrackCountVariable.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_b_sig_short,
    corrected_variable_appendix="sig_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BSigTrackCountVariable.hist_var,
    secondary_var_binning=BSigTrackCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BSigTrackCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BSigTrackCountVariable.new_column_name
    ),
)

b_tag_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagMesonRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.reco_b_tag_reco_mode_label,
    corrected_variable_appendix="b_tag_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagMesonRecoIDDefinition.hist_var,
    secondary_var_binning=BTagMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=BTagMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=BTagMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=BTagMesonRecoIDDefinition.get_tick_label_mapping(),
)

tag_neutrals_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagNeutralCountVariable.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_b_tag_short,
    corrected_variable_appendix="tag_neutrals_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagNeutralCountVariable.hist_var,
    secondary_var_binning=BTagNeutralCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BTagNeutralCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BTagNeutralCountVariable.new_column_name
    ),
)

tag_track_count_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{BTagTrackCountVariable.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_b_tag_short,
    corrected_variable_appendix="tag_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagTrackCountVariable.hist_var,
    secondary_var_binning=BTagTrackCountVariable.binning,
    variable_definition=DecayProcessVariables.get_definition_function(name=BTagTrackCountVariable.new_column_name),
    variable_definition_class=DecayProcessVariables.get_decay_info_definition(
        name=BTagTrackCountVariable.new_column_name
    ),
)

u4s_decay_id_based_res_cor_setup = ResolutionCorrectionSetup(
    name=f"{U4SRecoIDDefinition.new_column_name}_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.reco_y4s_reco_mode_label,
    corrected_variable_appendix="u4s_reco_id_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=U4SRecoIDDefinition.hist_var,
    secondary_var_binning=U4SRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=U4SRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=U4SRecoIDDefinition.new_column_name),
    tick_label_mapping=U4SRecoIDDefinition.get_tick_label_mapping(),
)

event_track_count_res_cor_setup = ResolutionCorrectionSetup(
    name="event_track_count_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_tracks_in_event_short,
    corrected_variable_appendix="event_track_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=event_track_count_hist_var,
    secondary_var_binning=event_track_count_binning,
)

event_neutral_count_res_cor_setup = ResolutionCorrectionSetup(
    name="event_neutral_count_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.decay_info_number_of_neutrals_in_event_short,
    corrected_variable_appendix="event_neutral_count_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=event_neutral_count_hist_var,
    secondary_var_binning=event_neutral_count_binning,
)

fei_probability_res_cor_setup = ResolutionCorrectionSetup(
    name="fei_probability_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.fei_prob,
    corrected_variable_appendix="fei_probability_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=fei_probability_hist_var,
    secondary_var_binning=fei_probability_binning,
)

b_sig_vertex_fit_res_cor_setup = ResolutionCorrectionSetup(
    name="b_sig_vertex_fit_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.b_sig_vertex_prob,
    corrected_variable_appendix="b_sig_vertex_fit_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=b_sig_vertex_fit_hist_var,
    secondary_var_binning=b_sig_vertex_fit_binning,
)

b_tag_daughter_vertex_fit_res_cor_setup = ResolutionCorrectionSetup(
    name="b_tag_daughter_vertex_fit_based_mms_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.b_tag_daughter_vertex_prob,
    corrected_variable_appendix="b_tag_daughter_vertex_fit_based",
    primary_hist_var=missing_mass_hist_var,
    secondary_hist_var=b_tag_daughter_vertex_fit_hist_var,
    secondary_var_binning=b_tag_daughter_vertex_fit_binning,
)


# E_Miss - P_Miss resolution correction setups
e_miss_minus_p_miss_cms = HistVariable(
    df_label=EMissMinusPMissVariable.e_miss_minus_p_miss_cms_col_name,
    n_bins=60,
    scope=(-1.0, 2.0),
    var_name=LatexObsNames.fitting_e_miss_minus_p_miss_cms,
    unit=LatexUnits.gev,
)  # type: HistVariable

e_miss_minus_p_miss_res_corr_setup = ResolutionCorrectionSetup(
    name="mbc_based_e_miss_minus_p_miss_resolution_correction",
    fei_weight_col=None,
    latex_title_appendix=LatexObsNames.b_tag_mbc,
    corrected_variable_appendix="mbc_based",
    primary_hist_var=e_miss_minus_p_miss_cms,
    secondary_hist_var=default_mbc_hist_var,
    secondary_var_binning=default_mbc_binning,
)


class ResolutionCorrectionSetupManager:
    setup_name_to_use = default_res_cor_setup.name  # type: str

    # Further setups to be considered:
    #  - D Meson channel with ids boiled down to similar groups of decay processes
    #  - D and Dst Meson channels with ids boiled down to similar groups of decay processes
    #  - B Sig channel with ids boiled down to similar groups of decay processes
    #  - B Tag channel with ids boiled down to similar groups of decay processes
    #  - B Sig + Y(4S) (using the latter to differentiate between normal and wrong combinations) channels ?
    #       -> Only for boiled down B_sig modes, because otherwise we get to many bins!
    #  - Multiplicity variable, which combines tracks and clusters...

    _setup_definitions_tuple = (
        # Default Setup: Mbc
        default_res_cor_setup,
        before_vf_mbc_res_cor_setup,
        #
        # Event kinematic based setups:
        # w_reco_res_cor_setup,
        # q2_reco_res_cor_setup,
        #
        # Lepton based setups:
        # lepton_p_res_cor_setup,
        # lepton_theta_res_cor_setup,
        lepton_phase_space_based_res_cor_setup,
        # before_vf_lepton_phase_space_based_res_cor_setup,
        #
        # D Meson based setups:
        # d_meson_decay_id_based_res_cor_setup,
        # d_and_dst_meson_decay_id_based_res_cor_setup,
        #
        # B Sig Meson based setups:
        b_sig_decay_id_based_res_cor_setup,
        # signal_neutrals_count_based_res_cor_setup,
        # signal_track_count_based_res_cor_setup,
        #
        # B Sig Meson based setups:
        # b_tag_decay_id_based_res_cor_setup0
        # tag_neutrals_count_based_res_cor_setup,
        # tag_track_count_based_res_cor_setup,
        #
        # Y(4S) Meson based setups:
        # u4s_decay_id_based_res_cor_setup,
        #
        # Event based setups:
        # event_track_count_res_cor_setup,
        # event_neutral_count_res_cor_setup,
        #
        # Other setups:
        # fei_probability_res_cor_setup,
        # b_sig_vertex_fit_res_cor_setup,
        # b_tag_daughter_vertex_fit_res_cor_setup,
        #
        #
        # E_Miss - P_Miss resolution correction setups
        e_miss_minus_p_miss_res_corr_setup,
    )  # type: Tuple[ResolutionCorrectionSetup, ...]

    _setup_definitions_with_gap_tuple = tuple(
        [sd.get_with_alternative_gap_setup() for sd in _setup_definitions_tuple]
    )  # type: Tuple[ResolutionCorrectionSetup, ...]
    _all_setup_definitions_tuple = tuple(
        list(_setup_definitions_tuple) + list(_setup_definitions_with_gap_tuple)
    )  # type: Tuple[ResolutionCorrectionSetup, ...]

    setup_definitions = {sd.name: sd for sd in _all_setup_definitions_tuple}  # type: Dict[str, ResolutionCorrectionSetup]

    alternative_primary_variable_mapping = {
        missing_mass_before_vf_hist_var.df_label: missing_mass_hist_var.df_label,
    }  # type: Dict[str, str]

    @staticmethod
    def get_setup_definition_names() -> Tuple[str, ...]:
        return tuple(ResolutionCorrectionSetupManager.setup_definitions.keys())

    @staticmethod
    def get_setup_definition(name: str) -> ResolutionCorrectionSetup:
        return ResolutionCorrectionSetupManager.setup_definitions[name]

    @staticmethod
    def get_n_bin_options_for_setup(setup_name: str) -> Tuple[int, ...]:
        n_bins_options = (
            20,
            # 25,
            50,
            # 75,
            # 125,
            250,
        )  # type: Tuple[int, ...]
        return n_bins_options

    @staticmethod
    def get_settings_key(setup_name: str) -> str:
        corr_setup = ResolutionCorrectionSetupManager.setup_definitions[setup_name]  # type: ResolutionCorrectionSetup

        _prim_var_col = corr_setup.primary_col_name  # type: str
        return ResolutionCorrectionSetupManager.alternative_primary_variable_mapping.get(_prim_var_col, _prim_var_col)

    @staticmethod
    def get_scope_options_for_setup(setup_name: str) -> Dict[str, Tuple[float, float]]:
        v_label = ResolutionCorrectionSetupManager.get_settings_key(setup_name=setup_name)  # type: str

        v1 = missing_mass_hist_var.df_label  # type: str
        v2 = e_miss_minus_p_miss_cms.df_label  # type: str

        valid_vs = (v1, v2)  # type: Tuple[str, ...]

        if v_label not in valid_vs:
            raise KeyError(
                f"No scope settings defined for setup '{setup_name}' with primary variable column name '{v_label}'!\n"
                f"Valid variable column names are {valid_vs}."
            )

        hist_var_scope_settings = {
            "wide1": {v1: (-0.5, 2.0), v2: (-1.0, 1.0)},
            # "wide2": {v1: (-0.5, 1.0), v2: (-1.0, 0.8)},
            # "wide3": {v1: (0.0, 2.0), v2: (-1.2, 1.2)},
            # "wide4": {v1: (-0.1, 2.0), v2: (-1.5, 1.5)},
            # "wide5": {v1: (0.0, 1.0), v2: (-2.0, 2.0)},
            # "wide6": {v1: (-0.1, 1.0), v2: (-0.5, 1.0)},
            # "tight1": {v1: (-0.5, 0.5), v2: (-0.5, 0.5)},
            # "tight2": {v1: (-0.3, 0.3), v2: (-0.3, 0.3)},
            # "tight3": {v1: (-0.3, 0.5), v2: (-0.3, 0.5)},
            # "tight4": {v1: (-0.1, 0.5), v2: (-0.1, 0.5)},
        }  # type: Dict[str, Dict[str, Tuple[float, float]]]

        return {k: v[v_label] for k, v in hist_var_scope_settings.items()}

    @staticmethod
    def get_special_options_for_setup(setup_name: str) -> Dict[Tuple[str, int], Tuple[float, float]]:
        v_label = ResolutionCorrectionSetupManager.get_settings_key(setup_name=setup_name)  # type: str

        v1 = missing_mass_hist_var.df_label  # type: str
        v2 = e_miss_minus_p_miss_cms.df_label  # type: str

        valid_vs = (v1, v2)  # type: Tuple[str, ...]

        if v_label not in valid_vs:
            return {}

        special_option_cases = {
            ("wide1", 20): {v2: (-0.5, 1.0)},
            ("wide6", 50): {v1: (-0.2, 1.8), v2: (-0.2, 1.8)},
        }  # type: Dict[Tuple[str, int], Dict[str, Tuple[float, float]]]

        return {k: v[v_label] for k, v in special_option_cases.items() if v_label in v.keys()}
