"""
Definition of additional binned mc variables.

Felix Metzner 2021
"""

import copy
import numpy as np
import pandas as pd

from typing import Union, Tuple, List, Dict, Callable, Type

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.latex_labels.latex_particle_names import LatexParticleNames

from rdstar.offline_analysis.rdstar_samples import main_samples
from rdstar.offline_analysis.observables.offline_observable_definition_base import BinnedOfflineObservableBase


__all__ = [
    "LeptonMCPDGBinID",
    "DMesonMCPDGBinID",
    "MCDataTypeBinID",
    "Q2TruthSigSide",
    "WTruthSigSide",
    "OfflineBinnedMCObservables",
]


class LeptonMCPDGBinID(BinnedOfflineObservableBase):
    new_column_name = "lepton_mc_pdg_bin_id"  # type: str

    _no_match_value = -999  # type: int
    _other_pdg_value = -998  # type: int
    _abs_no_match_value = abs(_no_match_value)  # type: int

    _pdg_table = {
        _no_match_value: "No Match",
        _other_pdg_value: "Other",
        11: LatexParticleNames.electron_str,
        13: LatexParticleNames.mu_str,
        22: LatexParticleNames.gamma_str,
        130: LatexParticleNames.k_long_str,
        211: LatexParticleNames.pi_pm_str,
        310: LatexParticleNames.k_short_str,
        321: LatexParticleNames.k_pm_str,
        2212: LatexParticleNames.proton_pm_str,
    }  # type: Dict[int, str]

    _pdg_to_bin_mapping = {pdg: i for i, pdg in enumerate(sorted(_pdg_table.keys()))}  # type: Dict[int, int]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(_pdg_to_bin_mapping),
        scope=tuple(np.arange(len(_pdg_to_bin_mapping) + 1) - 0.5)[:: len(_pdg_to_bin_mapping)],
        var_name=LatexObsNames.lepton_pdg_id,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    lepton_mc_matched_pdg = "lepton_mc_matched_pdg"  # type: str

    required_variables = {
        lepton_mc_matched_pdg: "daughter__bo1__cm__spdaughter__bo1__cm__spmatchedMC__boPDG__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        _valid_pdg_codes = [k for k in cls._pdg_table.keys() if k > 0]  # type: List[int]
        valid_pdg_codes = tuple([cls._abs_no_match_value] + _valid_pdg_codes)  # type: Tuple[int, ...]

        lepton_pdg_ids_array = np.abs(df[cls.required_variables[cls.lepton_mc_matched_pdg]].values)  # type: np.array

        is_other_pdg = ~np.isin(element=lepton_pdg_ids_array, test_elements=list(valid_pdg_codes))  # type: np.array
        is_no_match = lepton_pdg_ids_array == cls._abs_no_match_value  # type: np.array

        lepton_pdg_ids_array[is_other_pdg] = cls._other_pdg_value
        lepton_pdg_ids_array[is_no_match] = cls._no_match_value

        assert np.all(
            np.isin(element=np.unique(lepton_pdg_ids_array), test_elements=list(cls._pdg_to_bin_mapping.keys()))
        ), (np.unique(lepton_pdg_ids_array), cls._pdg_to_bin_mapping.keys())

        return np.vectorize(cls._pdg_to_bin_mapping.get)(lepton_pdg_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        return {i: LeptonMCPDGBinID._pdg_table[pdg] for pdg, i in LeptonMCPDGBinID._pdg_to_bin_mapping.items()}


class DMesonMCPDGBinID(BinnedOfflineObservableBase):
    new_column_name = "d_meson_mc_pdg_bin_id"  # type: str

    _no_match_value = -999  # type: int
    _other_pdg_value = -998  # type: int
    _abs_no_match_value = abs(_no_match_value)  # type: int

    _pdg_table = {
        _no_match_value: "No Match",
        _other_pdg_value: "Other",
        411: LatexParticleNames.d_pm_str,
        421: LatexParticleNames.d_zero_str,
        413: LatexParticleNames.d_st_pm_str,
        423: LatexParticleNames.d_st_zero_str,
        415: LatexParticleNames.d_two_st_pm_str,
        425: LatexParticleNames.d_two_st_zero_str,
        10411: LatexParticleNames.d_sszero_st_pm_str,
        10421: LatexParticleNames.d_sszero_st_zero_str,
        10413: LatexParticleNames.d_one_pm_str,
        10423: LatexParticleNames.d_one_zero_str,
        20413: LatexParticleNames.d_prime_one_pm_str,
        20423: LatexParticleNames.d_prime_one_zero_str,
        431: LatexParticleNames.d_strange_pm_str,
        10431: LatexParticleNames.d_strange_zero_str,
        20213: LatexParticleNames.a_one_pm_str,
        20113: LatexParticleNames.a_one_zero_str,
        62: LatexParticleNames.xu_pm_str,
        61: LatexParticleNames.xu_zero_str,
        511: LatexParticleNames.b_zero_str,
        521: LatexParticleNames.b_pm_str,
        300553: LatexParticleNames.y4s_str,
        10022: LatexParticleNames.virtual_gamma_str,
    }  # type: Dict[int, str]

    _pdg_to_bin_mapping = {pdg: i for i, pdg in enumerate(_pdg_table.keys())}  # type: Dict[int, int]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(_pdg_to_bin_mapping),
        scope=tuple(np.arange(len(_pdg_to_bin_mapping) + 1) - 0.5)[:: len(_pdg_to_bin_mapping)],
        var_name=LatexObsNames.d_w_wo_st_pdg_id,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    d_meson_mc_matched_pdg = "d_meson_mc_matched_pdg"  # type: str

    required_variables = {
        d_meson_mc_matched_pdg: "daughter__bo1__cm__spdaughter__bo0__cm__spmatchedMC__boPDG__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        _valid_pdg_codes = [k for k in cls._pdg_table.keys() if k > 0]  # type: List[int]
        valid_pdg_codes = tuple([cls._abs_no_match_value] + _valid_pdg_codes)  # type: Tuple[int, ...]

        d_meson_pdg_ids_array = np.abs(df[cls.required_variables[cls.d_meson_mc_matched_pdg]].values)  # type: np.array

        is_other_pdg = ~np.isin(element=d_meson_pdg_ids_array, test_elements=list(valid_pdg_codes))  # type: np.array
        is_no_match = d_meson_pdg_ids_array == cls._abs_no_match_value  # type: np.array

        d_meson_pdg_ids_array[is_other_pdg] = cls._other_pdg_value
        d_meson_pdg_ids_array[is_no_match] = cls._no_match_value

        assert np.all(
            np.isin(element=np.unique(d_meson_pdg_ids_array), test_elements=list(cls._pdg_to_bin_mapping.keys()))
        ), (np.unique(d_meson_pdg_ids_array), cls._pdg_to_bin_mapping.keys())

        return np.vectorize(cls._pdg_to_bin_mapping.get)(d_meson_pdg_ids_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        return {i: DMesonMCPDGBinID._pdg_table[pdg] for pdg, i in DMesonMCPDGBinID._pdg_to_bin_mapping.items()}


class MCDataTypeBinID(BinnedOfflineObservableBase):
    new_column_name = "mc_data_type_bin_id"  # type: str

    _data_type_list = copy.copy(main_samples)  # type: List[str]
    _data_type_mapping = {data_type_name: i for i, data_type_name in enumerate(_data_type_list)}  # type: Dict[str, int]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=len(_data_type_mapping),
        scope=tuple(np.arange(len(_data_type_mapping) + 1) - 0.5)[:: len(_data_type_mapping)],
        var_name=LatexObsNames.mc_info_data_type,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    data_type_info = "data_type_info"  # type: str

    required_variables = {
        data_type_info: "data_type",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        _valid_data_type_values = list(cls._data_type_mapping.keys())  # type: List[str]

        data_type_array = df[cls.required_variables[cls.data_type_info]].values  # type: np.array
        unique_dt_values_in_df = np.unique(data_type_array)  # type: np.array

        assert np.all(np.isin(element=unique_dt_values_in_df, test_elements=_valid_data_type_values)), (
            unique_dt_values_in_df,
            _valid_data_type_values,
        )

        return np.vectorize(cls._data_type_mapping.get)(data_type_array)

    @staticmethod
    def get_tick_label_mapping() -> Dict[int, str]:
        return {i: data_type_name for data_type_name, i in MCDataTypeBinID._data_type_mapping.items()}


class Q2TruthSigSide(BinnedOfflineObservableBase):
    new_column_name = "q2_truth_signal_side"  # type: str

    __number_of_bins = 20  # type: int
    __scope = (0.0, 4.0)  # type: Tuple[float, float]

    # Adding one bin for no match case, two for negative and high values and one bin for error in observable case:
    __additional_bins = 4  # type: int
    __bin_size = (__scope[1] - __scope[0]) * 1.0 / __number_of_bins  # type: float
    _number_of_bins = __number_of_bins + __additional_bins  # type: int
    _scope = (__scope[0] - 3 * __bin_size, __scope[1] + 1.0 * __bin_size)  # type: Tuple[float, float]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=_number_of_bins,
        scope=_scope,
        var_name=LatexObsNames.q2_truth,
        unit=LatexUnits.gev_sq,
        use_log_scale=False,
    )  # type: HistVariable

    _sig_b_mdst_indices = (1, 2)  # type: Tuple[int, ...]
    _no_match_index = 0  # type: int
    _no_match_value = _scope[0] + 0.5 * __bin_size  # type: float
    _error_in_obs_value = -999  # type: int
    _error_value = _scope[0] + 1.5 * __bin_size  # type: float
    _negative_value = _scope[0] + 2.5 * __bin_size  # type: float
    _overflow_value = _scope[-1] - 0.5 * __bin_size  # type: float

    q2_truth_1st_mc_b = "q2_truth_1st_mc_b"  # type: str
    q2_truth_2nd_mc_b = "q2_truth_2nd_mc_b"  # type: str
    sig_b_mdst_index = "sig_b_mdst_index"  # type: str
    sig_b_via_tau_mdst_index = "sig_b_via_tau_mdst_index"  # type: str

    required_variables = {
        q2_truth_1st_mc_b: "genUpsilon4S__bomcDaughter__bo0__cm__spqSquaredMC__bc__bc",
        q2_truth_2nd_mc_b: "genUpsilon4S__bomcDaughter__bo1__cm__spqSquaredMC__bc__bc",
        sig_b_mdst_index: "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        sig_b_via_tau_mdst_index: "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:

        sig_b_mdst = df[cls.required_variables[cls.sig_b_mdst_index]].values  # type: np.array
        sig_b_via_tau_mdst = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values  # type: np.array

        is_sig_b = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)  # type: np.array
        is_sig_b_via_tau = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)  # type: np.array

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )  # type: np.array

        _unique_indices = np.unique(b_mdst_index_array)  # type: np.array
        _valid_indices = list(cls._sig_b_mdst_indices) + [cls._no_match_index]  # type: List[int]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        q2_truth_from_1st_b = df[cls.required_variables[cls.q2_truth_1st_mc_b]].values  # type: np.array
        q2_truth_from_2nd_b = df[cls.required_variables[cls.q2_truth_2nd_mc_b]].values  # type: np.array

        q2_truth_array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[q2_truth_from_1st_b, q2_truth_from_2nd_b],
            default=cls._no_match_value,
        )  # type: np.array

        is_negative_value = (
            np.isin(element=b_mdst_index_array, test_elements=cls._sig_b_mdst_indices)
            & (q2_truth_array != cls._error_in_obs_value)
            & (q2_truth_array < 0)
        )  # type: np.array

        q2_truth_array[q2_truth_array == cls._error_in_obs_value] = cls._error_value
        q2_truth_array[is_negative_value] = cls._negative_value
        q2_truth_array[q2_truth_array > cls.__scope[-1]] = cls._overflow_value

        assert np.all(q2_truth_array[b_mdst_index_array != cls._no_match_index] >= cls._scope[0] + cls.__bin_size), (
            np.min(q2_truth_array[b_mdst_index_array != cls._no_match_index]),
            cls.__scope,
        )

        if check_range:
            assert np.all(q2_truth_array >= cls.hist_var.scope[0]), (np.min(q2_truth_array), cls.hist_var.scope)
            assert np.all(q2_truth_array <= cls.hist_var.scope[1]), (np.max(q2_truth_array), cls.hist_var.scope)

        return q2_truth_array


class WTruthSigSide(BinnedOfflineObservableBase):
    new_column_name = "w_truth_signal_side"  # type: str

    __number_of_bins = 20  # type: int
    __scope = (1.0, 1.6)  # type: Tuple[float, float]

    # Adding one bin for no match case, two for negative and high values and one bin for error in observable case:
    __additional_bins = 4  # type: int
    __bin_size = (__scope[1] - __scope[0]) * 1.0 / __number_of_bins  # type: float
    _number_of_bins = __number_of_bins + __additional_bins  # type: int
    _scope = (__scope[0] - 3 * __bin_size, __scope[1] + 1.0 * __bin_size)  # type: Tuple[float, float]

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=_number_of_bins,
        scope=_scope,
        var_name=LatexObsNames.ff_w_truth,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    _sig_b_mdst_indices = (1, 2)  # type: Tuple[int, ...]
    _no_match_index = 0  # type: int
    _no_match_value = _scope[0] + 0.5 * __bin_size  # type: float
    _error_in_obs_value = -np.inf  # type: float
    _error_value = _scope[0] + 1.5 * __bin_size  # type: float
    _negative_value = _scope[0] + 2.5 * __bin_size  # type: float
    _overflow_value = _scope[-1] - 0.5 * __bin_size  # type: float

    w_truth_1st_mc_b = "w_truth_1st_mc_b"  # type: str
    w_truth_2nd_mc_b = "w_truth_2nd_mc_b"  # type: str
    sig_b_mdst_index = "sig_b_mdst_index"  # type: str
    sig_b_via_tau_mdst_index = "sig_b_via_tau_mdst_index"  # type: str

    required_variables = {
        w_truth_1st_mc_b: "genUpsilon4S__bomcDaughter__bo0__cm__spwMC__bc__bc",
        w_truth_2nd_mc_b: "genUpsilon4S__bomcDaughter__bo1__cm__spwMC__bc__bc",
        sig_b_mdst_index: "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc",
        sig_b_via_tau_mdst_index: "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:

        sig_b_mdst = df[cls.required_variables[cls.sig_b_mdst_index]].values  # type: np.array
        sig_b_via_tau_mdst = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values  # type: np.array

        is_sig_b = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)  # type: np.array
        is_sig_b_via_tau = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)  # type: np.array

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )  # type: np.array

        _unique_indices = np.unique(b_mdst_index_array)  # type: np.array
        _valid_indices = list(cls._sig_b_mdst_indices) + [cls._no_match_index]  # type: List[int]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        w_truth_from_1st_b = df[cls.required_variables[cls.w_truth_1st_mc_b]].values  # type: np.array
        w_truth_from_2nd_b = df[cls.required_variables[cls.w_truth_2nd_mc_b]].values  # type: np.array

        w_truth_array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[w_truth_from_1st_b, w_truth_from_2nd_b],
            default=cls._no_match_value,
        )  # type: np.array

        is_negative_value = (
            np.isin(element=b_mdst_index_array, test_elements=cls._sig_b_mdst_indices)
            & (w_truth_array != cls._error_in_obs_value)
            & (w_truth_array < 0)
        )  # type: np.array

        w_truth_array[w_truth_array == cls._error_in_obs_value] = cls._error_value
        w_truth_array[is_negative_value] = cls._negative_value
        w_truth_array[w_truth_array > cls.__scope[-1]] = cls._overflow_value

        assert np.all(w_truth_array[b_mdst_index_array != cls._no_match_index] >= cls._scope[0] + cls.__bin_size), (
            np.min(w_truth_array[b_mdst_index_array != cls._no_match_index]),
            cls.__scope,
        )

        if check_range:
            assert np.all(w_truth_array >= cls.hist_var.scope[0]), (np.min(w_truth_array), cls.hist_var.scope)
            assert np.all(w_truth_array <= cls.hist_var.scope[1]), (np.max(w_truth_array), cls.hist_var.scope)

        return w_truth_array


class OfflineBinnedMCObservables:

    definitions = {
        LeptonMCPDGBinID.get_new_column_name(): LeptonMCPDGBinID,
        DMesonMCPDGBinID.get_new_column_name(): DMesonMCPDGBinID,
        MCDataTypeBinID.get_new_column_name(): MCDataTypeBinID,
        Q2TruthSigSide.get_new_column_name(): Q2TruthSigSide,
        WTruthSigSide.get_new_column_name(): WTruthSigSide,
    }  # type: Dict[str, Type[BinnedOfflineObservableBase]]

    required_variables = tuple(
        [c for d in definitions.values() for c in d.get_required_columns()]
    )  # type: Tuple[str, ...]

    @staticmethod
    def add_kinematic_observables_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in OfflineBinnedMCObservables.get_required_columns_list()), [
            c for c in OfflineBinnedMCObservables.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in OfflineBinnedMCObservables.get_new_columns_list()), [
            c for c in OfflineBinnedMCObservables.get_new_columns_list() if c in df.columns
        ]

        for new_col, definition_class in OfflineBinnedMCObservables.definitions.items():
            new_values = definition_class.calculate_new_column_values(df=df)
            assert len(new_values) == len(df), (len(new_values), len(df), new_col)
            df.loc[:, new_col] = new_values

            assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_definition_function(
        name: str,
        fail_when_exists: bool = True,
    ) -> Callable:
        def definition_function(
            df: pd.DataFrame,
            check_range: bool = True,
            accept_existing_column: bool = False,
        ) -> None:
            definition = OfflineBinnedMCObservables.definitions[name]  # type: Type[BinnedOfflineObservableBase]
            new_col_name = definition.get_new_column_name()

            if fail_when_exists and not accept_existing_column:
                assert new_col_name not in df.columns, new_col_name
            else:
                if new_col_name in df.columns:
                    return

            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(OfflineBinnedMCObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(OfflineBinnedMCObservables.definitions.keys())

    @staticmethod
    def get_mc_obs_definition(name: str) -> Type[BinnedOfflineObservableBase]:
        return OfflineBinnedMCObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(OfflineBinnedMCObservables.definitions.keys())
