import copy
import pandas as pd

from functools import partial
from inspect import signature
from dataclasses import dataclass
from typing import Union, Optional, Tuple, Dict, Callable, Type, Any, ClassVar

from rdstar.offline_analysis.selection.selection_evaluation_utilities import map_function_type_to_value_class
from rdstar.offline_analysis.selection.selection_evaluation_utilities import map_function_type_to_value_interpretation


__all__ = [
    "EvalInfoEntry",
    "EvalInfoBase",
    "EvalInfoContainerBase",
    "create_container_class",
]


@dataclass(frozen=True)
class EvalInfoEntry:
    calc: Callable
    add: Callable

    def __post_init__(self) -> None:
        calc_signature = signature(self.calc)
        add_signature = signature(self.add)
        fns = (self.calc.__name__, self.add.__name__)  # type: Tuple[str, str]

        # Check if calc and add functions have matching signatures
        assert calc_signature.return_annotation == add_signature.return_annotation, (
            calc_signature.return_annotation,
            add_signature.return_annotation,
            *fns,
        )

        # Checking if calc function has the expected kw args
        _expected_kw_args = (
            "df",
            "data_type",
        )  # type: Tuple[str, ...]
        calc_fcn_parameters = calc_signature.parameters
        assert len(calc_fcn_parameters) == len(_expected_kw_args), (calc_fcn_parameters.keys(), _expected_kw_args, *fns)
        assert all(kwa in calc_fcn_parameters.keys() for kwa in _expected_kw_args), (
            calc_fcn_parameters.keys(),
            _expected_kw_args,
            *fns,
        )

        # Checking if add function looks valid
        assert len(add_signature.parameters) == 2, (len(add_signature.parameters), add_signature.parameters, *fns)
        add_params = tuple(add_signature.parameters.values())
        assert add_params[0].annotation == add_params[1].annotation, ([param.annotation for param in add_params], *fns)
        assert add_params[0].annotation == add_signature.return_annotation, (
            add_params[0].annotation,
            add_signature.return_annotation,
            *fns,
        )

    def get_value_type(self) -> type:
        return map_function_type_to_value_class(calc_function=self.calc)

    def interpret_value(self, v: Any) -> Any:
        return map_function_type_to_value_interpretation(calc_function=self.calc)(v)


class EvalInfoBase:
    __internal_methods: ClassVar[Tuple[str, ...]] = (
        "check_validity",
        "get_entries_dict",
    )
    __forbidden_names: ClassVar[Tuple[str, ...]] = (
        "as_dict",
        "from_dict",
        "fill",
        "merge",
        "entry_names",
        "entries_dict",
        "values_dict",
        "get_entry_names",
    )

    @classmethod
    def get_entries_dict(cls) -> Dict[str, EvalInfoEntry]:
        assert not any(
            isinstance(v, EvalInfoEntry) and (k in EvalInfoBase.__forbidden_names or k in EvalInfoBase.__internal_methods)
            for k, v in vars(cls).items()
        ), [
            k
            for k, v in vars(cls).items()
            if isinstance(v, EvalInfoEntry)
            and (k in EvalInfoBase.__forbidden_names or k in EvalInfoBase.__internal_methods)
        ]
        return {
            k: v
            for k, v in vars(cls).items()
            if k not in EvalInfoBase.__internal_methods
            and k not in EvalInfoBase.__forbidden_names
            and not k.startswith("_")
        }

    @classmethod
    def get_entry_names(cls) -> Tuple[str, ...]:
        return tuple(cls.get_entries_dict().keys())

    @classmethod
    def check_validity(cls) -> bool:
        return all(isinstance(v, EvalInfoEntry) for k, v in cls.get_entries_dict().items())

    def __init__(self, df: Optional[pd.DataFrame], data_type: Optional[str]) -> None:
        ...

    def as_dict(self) -> Dict[str, Any]:
        ...

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "EvalInfoContainerBase":
        ...

    def merge(self, other: Union["EvalInfoContainerBase", "EvalInfoBase"]) -> "EvalInfoContainerBase":
        ...


class EvalInfoContainerBase:
    @classmethod
    def get_definition(cls, key: str) -> EvalInfoEntry:
        ...

    def get_value(self, key: str) -> Any:
        ...

    entries_dict = {}  # type: Dict[str, EvalInfoEntry]

    def __init__(self, df: Optional[pd.DataFrame], data_type: Optional[str]) -> None:
        if not self.entries_dict:
            raise RuntimeError("No definitions available!")

        self._values_dict = None  # type: Optional[Dict[str, Any]]

        if df is not None:
            assert data_type is not None and isinstance(data_type, str), (data_type, type(data_type))
            self.fill(df=df, data_type=data_type)

    def fill(
        self,
        df: pd.DataFrame,
        data_type: str,
    ) -> None:
        assert self._values_dict is None
        assert isinstance(df, pd.DataFrame), type(df)
        self._values_dict = {k: v.calc(df=df, data_type=data_type) for k, v in self.entries_dict.items()}

    @property
    def values_dict(self) -> Dict[str, Any]:
        assert self._values_dict is not None
        return self._values_dict

    def merge(self, other: Union["EvalInfoContainerBase", "EvalInfoBase"]) -> "EvalInfoContainerBase":
        assert not isinstance(other, EvalInfoBase), type(other)
        assert isinstance(other, self.__class__), (type(other), self.__class__)
        assert set(self.entries_dict.keys()) == set(other.entries_dict.keys()), list(
            set(self.entries_dict.keys()) ^ set(other.entries_dict.keys())
        )
        _new = self.__class__(df=None, data_type=None)
        _new._values_dict = {
            k: v.add(copy.deepcopy(self.get_value(key=k)), copy.deepcopy(other.get_value(key=k)))
            for k, v in self.__class__.entries_dict.items()
        }
        return _new

    @classmethod
    def get_entry_names(cls) -> Tuple[str, ...]:
        return tuple(cls.entries_dict.keys())

    @classmethod
    def get_entry_type_mapping(cls) -> Dict[str, type]:
        return {k: v.get_value_type() for k, v in cls.entries_dict.items()}

    @classmethod
    def get_entry_type(cls, entry_key: str) -> type:
        return cls.entries_dict[entry_key].get_value_type()

    @property
    def entry_names(self) -> Tuple[str, ...]:
        return tuple(self.entries_dict.keys())

    def as_dict(self) -> Dict[str, Any]:
        return {
            "class_name": self.__class__.__name__,
            "values": self.values_dict,
        }

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "EvalInfoContainerBase":
        assert input_dict["class_name"] == cls.__name__, (input_dict["class_name"], cls.__name__)
        _new = cls(df=None, data_type=None)
        _new_values_dict = input_dict["values"]
        new_values_dict = {k: cls.get_definition(key=k).interpret_value(v=v) for k, v in _new_values_dict.items()}
        _new._values_dict = new_values_dict
        return _new

    def __deepcopy__(self, memo) -> "EvalInfoContainerBase":
        _new = self.__class__(df=None, data_type=None)
        _new._values_dict = copy.deepcopy(self._values_dict)
        return _new

    def __copy__(self, memo) -> "EvalInfoContainerBase":
        return self.__deepcopy__(memo)


def create_container_class(base_cls: Type[EvalInfoBase]) -> Type[EvalInfoContainerBase]:
    assert issubclass(base_cls, EvalInfoBase), base_cls.__name__
    assert base_cls.check_validity()

    def get_definition(cls, key: str) -> EvalInfoEntry:
        return cls.entries_dict[key]

    def get_value(self, key: str) -> Any:
        return self.values_dict[key]

    class NewEvalInfoContainerClass(EvalInfoContainerBase):
        entries_dict = base_cls.get_entries_dict()  # type: Dict[str, EvalInfoEntry]

    setattr(NewEvalInfoContainerClass, "get_definition", partial(get_definition, cls=NewEvalInfoContainerClass))
    setattr(NewEvalInfoContainerClass, "get_value", get_value)

    for kw in base_cls.get_entries_dict().keys():
        assert kw not in NewEvalInfoContainerClass.__dict__, kw
        setattr(NewEvalInfoContainerClass, kw, property(fget=partial(get_definition, key=kw)))
        assert f"{kw}_val" not in NewEvalInfoContainerClass.__dict__, (kw, f"{kw}_val")
        setattr(NewEvalInfoContainerClass, f"{kw}_val", property(fget=partial(get_value, key=kw)))

    NewEvalInfoContainerClass.__name__ = f"{base_cls.__name__}Container"
    return NewEvalInfoContainerClass
