"""
Combination of Data Samples

Felix Metzner 2019
"""

import os
import copy
import numpy as np
import pandas as pd
import multiprocessing
import b2luigi as luigi

from functools import partial
from collections import defaultdict, Iterable
from typing import Any, Union, Optional, Tuple, List, Dict, Generator, Sequence

from analysistools.batch_tools.file_handler import Belle_url_mc_types
from analysistools.pid_corrections.pid_weight_evaluator import PIDWeightEvaluation

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.tasks.data_reduction_tasks import DataReductionTask
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask
from rdstar.offline_analysis.selection.selection_evaluation_info_container import SelectionEvaluationContainer


__all__ = [
    "SampleCombinerTask",
    "EvaluationSampleCombinerTask",
    "SelectionInfoCombinerTask",
]


class SampleCombinerTask(utilities.BaseOfflineTask):
    combo_name: str = luigi.Parameter(positional=False)
    working_dir: PathType = luigi.Parameter(positional=False)
    selection_name: str = luigi.Parameter(positional=False)
    config_params: Sequence[Dict[str, Any]] = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name: str = "CombinedSamples"

    stream_column_name: str = Cols.stream
    data_type_column_name: str = Cols.data_type

    @property
    def resources(self) -> Dict[str, int]:
        if self.combo_name in ["Charm", "UDS", "Charged", "Mixed"]:
            return {utilities.offline_memory_resource_name: 6}
        else:
            return {utilities.offline_memory_resource_name: 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path: PathType = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)

        self.cs_tag: str = self.config_params[0]["cs_tag"]

        self.pid_weight_info_key: str = "pid_weight_info"
        self.ff_d_weight_info_key: str = "ff_d_weight_info"
        self.ff_ds_weight_info_key: str = "ff_ds_weight_info"
        self.ff_dss_weight_info_key: str = "ff_dss_weight_info"

        self.evt_id_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()
        self.c_id_col: str = "unique_candidate_id"
        self.id_col: str = "combined_candidate_id"

        self.pid_sample_random_seed: int = 47471337

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @staticmethod
    def get_data_combo_dict() -> Dict[str, List[str]]:
        return rdstar_samples.data_type_combinations

    def get_input_file_info(self) -> Generator[Tuple[str, str, int], None, None]:
        for params in self.config_params:
            stream = params["stream"]
            stream_tag = f"_{stream}"
            if stream == -999:
                stream_tag = ""
            yield f"{params['data_type']}{stream_tag}_{params['tuple_type']}", params["data_type"], stream

    def get_paths_for_key(self, input_key: str) -> Generator[Tuple[str, str, int, PathType], None, None]:
        for base_key, data_type, stream in self.get_input_file_info():
            for key, path in self.input_dict.items():
                is_hdf5_file = os.path.splitext(path)[1] == ".h5"
                if key.endswith(input_key) and key.startswith(base_key) and is_hdf5_file:
                    yield key, data_type, stream, path

    def load_and_combine(self, input_key: str) -> pd.DataFrame:
        assert not self.combo_name.startswith("Data"), self.combo_name
        initial_length: int = 0
        dfs: List[pd.DataFrame] = []
        scale_weight_dict = defaultdict(list)
        for key, data_type, stream, path in self.get_paths_for_key(input_key=input_key):
            df = pd.read_hdf(path)
            assert isinstance(df, pd.DataFrame)
            if len(df.index) > 0:
                initial_length += len(df.index)

                df.loc[:, self.stream_column_name] = stream
                df.loc[:, self.data_type_column_name] = data_type

                if input_key == self.pid_weight_info_key:
                    if len(df.index) != len(np.unique(df.index)):
                        df.reset_index(drop=True, inplace=True)
                    self.add_unique_id(df=df)
                    selection_conds = self.get_selection_conditions(data_type=data_type, stream=stream)
                    evt_cond = df[self.evt_id_col].isin(selection_conds[0])
                    id_cond = df.loc[evt_cond, self.id_col].isin(selection_conds[1])
                    selected_indices: pd.Index = df.loc[evt_cond].loc[id_cond].index
                    assert len(selected_indices) == len(set(selected_indices)), (
                        len(selected_indices),
                        len(set(selected_indices)),
                    )

                    sampled_indices: pd.Index = (
                        df.loc[~df.index.isin(selected_indices)]
                        .sample(frac=0.1, random_state=self.pid_sample_random_seed)
                        .index
                    )
                    assert len(sampled_indices) == len(set(sampled_indices)), (
                        len(sampled_indices),
                        len(set(sampled_indices)),
                    )

                    n_combined_indices = len(set(selected_indices) | set(sampled_indices))
                    assert n_combined_indices == len(selected_indices) + len(sampled_indices), (
                        n_combined_indices,
                        len(selected_indices) + len(sampled_indices),
                    )

                    chosen_indices: pd.Index = selected_indices.union(sampled_indices)

                    assert len(chosen_indices) == len(selected_indices) + len(sampled_indices), (
                        len(chosen_indices),
                        len(selected_indices) + len(sampled_indices),
                    )

                    df = df.iloc[chosen_indices]
                    del df[self.id_col]
                    utilities.decompose_event_id(df=df, remove_original_column=True)

                if data_type in Belle_url_mc_types.keys() and input_key not in self.get_info_keys():
                    scale_weight = df[scale_weight_column_name].unique()
                    scale_correction = df[DataPreparationTask.scale_corr_column].unique()
                    if len(scale_weight) == 1 and len(scale_correction) == 1:
                        scale_weight_dict[data_type].append((1.0 / scale_weight[0]) * (1.0 - scale_correction[0]))
                    else:
                        raise RuntimeError(
                            f"Unexpected scaling_weight = {scale_weight}, scale_correction = {scale_correction} pair."
                        )
            else:
                df = df.assign(**{self.stream_column_name: stream})
                df = df.assign(**{self.data_type_column_name: data_type})

            dfs.append(df)

        combined_df = pd.concat(dfs, axis=0, ignore_index=True, sort=False)

        for data_type in scale_weight_dict.keys():
            scaling_weight = 1.0 / sum(scale_weight_dict[data_type])
            print(
                f"Extracted the scaling weights {scale_weight_dict[data_type]} for {data_type} and applying the "
                f"weight factor 1. / {sum(scale_weight_dict[data_type])} = {scaling_weight}"
            )

            combined_df.loc[combined_df.data_type == data_type, scale_weight_column_name] = scaling_weight

        if input_key != self.pid_weight_info_key:
            assert len(combined_df.index) == initial_length, (len(combined_df.index), initial_length)

        if input_key == self.pid_weight_info_key:
            combined_df.loc[:, self.data_type_column_name] = combined_df[self.data_type_column_name].astype("category")
            pid_weight_type_col = PIDWeightEvaluation.weight_type_column_name
            combined_df.loc[:, pid_weight_type_col] = combined_df[pid_weight_type_col].astype("category")

        return combined_df

    def add_unique_id(self, df: pd.DataFrame, check_if_unique: bool = False) -> np.ndarray:
        id_values = utilities.add_unique_candidate_id(
            df=df,
            id_column_name=self.id_col,
            event_id_name=self.evt_id_col,
            candidate_id_name=self.c_id_col,
            check_if_unique=check_if_unique,
        )
        return id_values

    def get_selection_conditions(self, data_type: str, stream: Union[int, float]) -> Tuple[np.ndarray, np.ndarray]:
        selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
        type_info = f"{data_type}" if stream == -999 else f"{data_type}_{stream}"
        df_paths: List[PathType] = [
            path
            for selection_tag in selection_tags
            for key, path in self.input_dict.items()
            if key.endswith(selection_tag) and key.startswith(type_info) and os.path.splitext(path)[1] == ".h5"
        ]
        assert len(df_paths) == len(selection_tags), (len(selection_tags), len(df_paths), selection_tags, df_paths)

        cols = [self.evt_id_col, self.c_id_col]
        df = pd.concat(
            [df.loc[:, cols] for path in df_paths for df in [pd.read_hdf(path)] if isinstance(df, pd.DataFrame)],
            axis=0,
            ignore_index=True,
            sort=False,
        )
        unique_candidate_id = self.add_unique_id(df=df)

        return df[self.evt_id_col].unique(), np.unique(unique_candidate_id)

    def get_data_path(self, input_key: str) -> PathType:
        assert self.combo_name.startswith("Data"), self.combo_name
        paths: List[PathType] = []
        for key, data_type, stream, path in self.get_paths_for_key(input_key=input_key):
            paths.append(path)
        assert len(paths) == 1
        return paths[0]

    def get_info_keys(self) -> List[str]:
        return [
            self.pid_weight_info_key,
            self.ff_d_weight_info_key,
            self.ff_ds_weight_info_key,
            self.ff_dss_weight_info_key,
        ]

    def get_subselection_keys(self, selection_keys: List[str]) -> List[str]:
        subselection_keys = []
        possible_extensions = DataSelectionTask.get_possible_subsample_extensions(input_dict=self.input_dict)
        for selection_key in selection_keys:
            for ext in possible_extensions:
                tag = f"{selection_key}{ext}"
                if any(key.endswith(tag) for key in self.input_dict.keys()):
                    subselection_keys.append(tag)

        return subselection_keys

    def get_keys(self) -> List[str]:
        _selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
        selection_keys = list(_selection_tags)
        selection_keys.extend(self.get_subselection_keys(selection_keys))

        if self.combo_name.startswith("Data"):
            return selection_keys

        keys = self.get_info_keys()
        keys.extend(selection_keys)

        if self.combo_name.startswith("Special"):
            keys.remove(self.ff_dss_weight_info_key)
        return keys

    def combine_sample_with_key(self, key: str, dry_run: bool) -> Tuple[str, PathType]:
        output_key = f"{self.combo_name}_{key}"
        output_path = os.path.join(self.base_path, f"{output_key}.h5")
        temp_output_path = os.path.join(self.base_path, f"temp_{output_key}.h5")

        if not dry_run:
            if self.combo_name.startswith("Data"):
                if os.path.exists(output_path):
                    os.remove(output_path)
                os.symlink(src=self.get_data_path(input_key=key), dst=output_path)
            else:
                df = self.load_and_combine(input_key=key)
                df.to_hdf(
                    temp_output_path, key=output_key, format="table" if key == self.pid_weight_info_key else "fixed"
                )
                os.rename(temp_output_path, output_path)

        return output_key, output_path

    def combine_samples(self, dry_run: bool = False) -> List[Tuple[str, PathType]]:
        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

        with multiprocessing.Pool(processes=4) as pool:
            output_tuples = pool.map(
                partial(SampleCombinerTask.combine_sample_with_key, self, dry_run=dry_run), self.get_keys()
            )

        return output_tuples

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.combine_samples()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.combine_samples(dry_run=True):
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    @staticmethod
    def get_reduced_params_for_data_preparation_task(params_dict: Dict[str, Any]) -> Dict[str, Any]:
        exclude_list = [
            "cs_tag",
            "cs_fastbdt_settings",
        ]
        return {k: v for k, v in params_dict.items() if k not in exclude_list}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_params, Iterable)
        for params in self.config_params:
            yield DataSelectionTask(selection_name=self.selection_name, **params)
            yield DataPreparationMergerTask(**self.get_reduced_params_for_data_preparation_task(params))


class EvaluationSampleCombinerTask(utilities.BaseOfflineTask):
    combo_name: str = luigi.Parameter(positional=False)
    working_dir: PathType = luigi.Parameter(positional=False)
    config_params: Sequence[Dict[str, Any]] = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name: str = "CombinedEvaluationSamples"
    validation_output_tag: str = "evaluation_sample"

    stream_column_name: str = SampleCombinerTask.stream_column_name
    data_type_column_name: str = SampleCombinerTask.data_type_column_name

    @property
    def resources(self) -> Dict[str, int]:
        if self.combo_name in ["Charm", "UDS", "Charged", "Mixed"]:
            return {utilities.offline_memory_resource_name: 6}
        else:
            return {utilities.offline_memory_resource_name: 1}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path: PathType = os.path.join(self.working_dir, self.base_dir_name)
        self.cs_tag: str = self.config_params[0]["cs_tag"]

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_input_file_info(self) -> Generator[Tuple[str, str, int], None, None]:
        for params in self.config_params:
            stream = params["stream"]
            stream_tag: str = "" if stream == -999 else f"_{stream}"
            data_type: str = params["data_type"]
            base_key_start: str = f"{data_type}{stream_tag}_{params['tuple_type']}_with_CS_{self.cs_tag}"
            yield base_key_start, data_type, stream

    def get_eval_sample_paths(self) -> Generator[Tuple[str, str, int, PathType], None, None]:
        eval_sample_key_tag: str = DataReductionTask.evaluation_sample_tag

        for base_key, data_type, stream in self.get_input_file_info():
            for key, path in self.input_dict.items():
                is_hdf5_file = os.path.splitext(path)[1] == ".h5"
                if key.endswith(eval_sample_key_tag) and key.startswith(base_key) and is_hdf5_file:
                    yield key, data_type, stream, path

    def load_and_combine(self) -> pd.DataFrame:
        assert not self.combo_name.startswith("Data"), self.combo_name
        initial_length: int = 0
        dfs: List[pd.DataFrame] = []
        scale_weight_dict = defaultdict(list)
        for key, data_type, stream, path in self.get_eval_sample_paths():
            df = pd.read_hdf(path)
            assert isinstance(df, pd.DataFrame)
            if len(df.index) > 0:
                initial_length += len(df.index)

                df.loc[:, self.stream_column_name] = stream
                df.loc[:, self.data_type_column_name] = data_type

                if data_type in Belle_url_mc_types.keys():
                    scale_weight = df[scale_weight_column_name].unique()
                    scale_correction = df[DataPreparationTask.scale_corr_column].unique()
                    if len(scale_weight) == 1 and len(scale_correction) == 1:
                        scale_weight_dict[data_type].append((1.0 / scale_weight[0]) * (1.0 - scale_correction[0]))
                    else:
                        raise RuntimeError(
                            f"Unexpected scaling_weight = {scale_weight}, scale_correction = {scale_correction} pair."
                        )
            else:
                df = df.assign(**{self.stream_column_name: stream})
                df = df.assign(**{self.data_type_column_name: data_type})

            dfs.append(df)

        combined_df = pd.concat(dfs, axis=0, ignore_index=True, sort=False)

        for data_type in scale_weight_dict.keys():
            scaling_weight = 1.0 / sum(scale_weight_dict[data_type])
            print(
                f"Extracted the scaling weights {scale_weight_dict[data_type]} for {data_type} and applying the "
                f"weight factor 1. / {sum(scale_weight_dict[data_type])} = {scaling_weight}"
            )

            combined_df.loc[combined_df.data_type == data_type, scale_weight_column_name] = scaling_weight

        assert len(combined_df.index) == initial_length, (len(combined_df.index), initial_length)
        return combined_df

    def get_data_path(self) -> PathType:
        assert self.combo_name.startswith("Data"), self.combo_name
        paths: List[PathType] = []
        for key, data_type, stream, path in self.get_eval_sample_paths():
            paths.append(path)
        assert len(paths) == 1
        return paths[0]

    def combine_validation_samples(self, dry_run: bool = False) -> Tuple[str, PathType]:
        output_key: str = f"{self.combo_name}_{self.validation_output_tag}"
        output_path: PathType = os.path.join(self.base_path, f"{output_key}.h5")
        temp_output_path: PathType = os.path.join(self.base_path, f"tmp_{output_key}.h5")

        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

            if self.combo_name.startswith("Data"):
                if os.path.exists(output_path):
                    os.remove(output_path)
                os.symlink(src=self.get_data_path(), dst=output_path)
            else:
                df = self.load_and_combine()
                df.to_hdf(temp_output_path, key=output_key, format="fixed")
                os.rename(temp_output_path, output_path)

        return output_key, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.combine_validation_samples()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_key, target_path = self.combine_validation_samples(dry_run=True)  # type: str, PathType
        return {output_key: luigi.LocalTarget(target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_params, Iterable)
        for params in self.config_params:
            yield DataReductionTask(**params)


class SelectionInfoCombinerTask(utilities.BaseOfflineTask):
    working_dir: PathType = luigi.Parameter(positional=False)
    selection_name: str = luigi.Parameter(positional=False)
    config_combinations: utilities.ConfigCombinationsType = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name: str = "CombinedSelectionInfo"

    _info_combos: Dict[str, List[str]] = {
        "main": rdstar_samples.main_samples,
        "eff": rdstar_samples.efficiency_samples,
        "data": rdstar_samples.recorded_data_samples,
        "allMC": rdstar_samples.main_samples + rdstar_samples.efficiency_samples,
    }
    _info_combos.update({k: [k] for k in rdstar_samples.recorded_data_samples})
    info_combos: Dict[str, List[str]] = _info_combos
    info_names: List[str] = [k for k in info_combos.keys()]
    main_info_name: str = "main"
    data_info_name: str = "DataY4S"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path: PathType = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)

        self.cs_tag: str = self.config_combinations[0][2][0]["cs_tag"]
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_selection_info_files_per_tag(self) -> Dict[str, List[PathType]]:
        file_dict = defaultdict(list)
        type_dict = defaultdict(list)
        selection_tags = []
        for key, path in self.input_dict.items():
            has_prefix = os.path.basename(path).startswith(DataSelectionTask.evaluation_info_prefix)
            is_json = os.path.splitext(path)[1] == ".json"
            if not (has_prefix and is_json):
                continue

            data_type = key.split("_", 1)[0]
            selection_tag = key.split(f"{DataSelectionTask.evaluation_info_prefix}_", 1)[1]
            if selection_tag not in selection_tags:
                selection_tags.append(selection_tag)

            for combo_name, combo_keys in self.info_combos.items():
                if data_type in combo_keys:
                    type_dict[f"{combo_name}_{selection_tag}"].append(data_type)
                    file_dict[f"{combo_name}_{selection_tag}"].append(path)

        for combo_name in self.info_combos:
            keys = [f"{combo_name}_{st}" for st in selection_tags]
            assert all(len(type_dict[t]) == len(type_dict[keys[0]]) for t in keys)
        assert all(len(file_dict[t]) == len(type_dict[t]) for t in file_dict.keys())

        for data_type in rdstar_samples.recorded_data_samples:
            assert all(len(v) == 1 for k, v in file_dict.items() if k.startswith(data_type))

        assert all(len(file_dict[t]) == len(set(file_dict[t])) for t in file_dict.keys())
        assert all(os.path.exists(p) for t in file_dict.keys() for p in file_dict[t])

        return file_dict

    @staticmethod
    def merge_selection_infos(file_paths: List[PathType]) -> SelectionEvaluationContainer:
        first_file_path, other_file_paths = file_paths[0], file_paths[1:]  # type: PathType, List[PathType]
        combined_container = SelectionEvaluationContainer.read(file_path=first_file_path)
        for path in other_file_paths:
            previous_container: SelectionEvaluationContainer = copy.deepcopy(combined_container)
            new_container = SelectionEvaluationContainer.read(file_path=path)
            combined_container = previous_container.merge(other=new_container)

        return combined_container

    def combine_selection_infos(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = {}

        selection_tags: Tuple[str, ...] = SelectionSetups.get_selection_tags_with_eval_info(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )

        result_keys: List[str] = [f"{combo}_{sel_tag}" for sel_tag in selection_tags for combo in self.info_combos]

        for key in result_keys:
            output_key: str = f"{DataSelectionTask.evaluation_info_prefix}_{key}"
            output_file_path: PathType = os.path.join(self.base_path, f"{output_key}.json")
            output_dict.update({output_key: output_file_path})

        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

            selection_info_files_per_tag_dict = self.get_selection_info_files_per_tag()
            assert len(result_keys) == len(selection_info_files_per_tag_dict)
            assert set(result_keys) == set(selection_info_files_per_tag_dict.keys())

            for result_key, in_files in selection_info_files_per_tag_dict.items():
                assert result_key in result_keys, result_key
                assert len(in_files) > 0, (result_key, in_files)
                output_key = f"{DataSelectionTask.evaluation_info_prefix}_{result_key}"

                output_path: PathType = output_dict[output_key]

                if not dry_run:
                    if len(in_files) == 1:
                        if os.path.exists(output_path):
                            os.remove(output_path)
                        os.symlink(in_files[0], output_path)
                    else:
                        merged_selection_info = self.merge_selection_infos(file_paths=in_files)
                        merged_selection_info.write(file_path=output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.combine_selection_infos()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.combine_selection_infos(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for params in [c for _, _, rcs in self.config_combinations for c in rcs]:
            yield DataSelectionTask(selection_name=self.selection_name, **params)
