"""
Tasks producing the Validation Results for the Analysis

Felix Metzner 2019
"""

import os
import b2luigi as luigi

from luigi.util import inherits
from typing import Tuple, Generator

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.latex_tasks import LatexTasks
from rdstar.offline_analysis.tasks.cs_tasks import CSValidationTask
from rdstar.offline_analysis.tasks.result_tasks import EfficiencyEvaluationTask
from rdstar.offline_analysis.tasks.fitting_tasks import FittingTasksWrapperTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightValidationWrapperTask
from rdstar.offline_analysis.tasks.flow_validation_tasks import AnalysisEvaluationTask
from rdstar.offline_analysis.tasks.mva_selector_tasks import MVASelectionValidatorTask
from rdstar.offline_analysis.tasks.analysis_plotter_tasks import AnalysisPlotManagerTask
from rdstar.offline_analysis.tasks.mm2_shift_evaluation_tasks import MM2ShiftEvaluationTask
from rdstar.offline_analysis.tasks.sideband_validation_tasks import SidebandEvaluationWrapperTask
from rdstar.offline_analysis.tasks.efficiency_validation_task import EfficiencyValidationWrapperTask
from rdstar.offline_analysis.tasks.pre_selection_validation_task import PreSelectionValidationManagerTask
from rdstar.offline_analysis.tasks.fsr_correction_validation_tasks import FSRCorrectionValidationManagerTask
from rdstar.offline_analysis.tasks.variable_impact_validation_task import VariableImpactValidationWrapperTask
from rdstar.offline_analysis.tasks.resolution_correction_task import (
    ResolutionCorrectionValidationManagerTask,
    ResolutionCorrectionWrapperTask,
)

from rdstar.offline_analysis.dss_fitting.tasks.fitting_main_task import FittingMaster


__all__ = [
    "ValidationWrapperTask",
    "SelectionManagerTask",
    "GeneralManagerTask",
]


@inherits(
    AnalysisEvaluationTask,
    SidebandEvaluationWrapperTask,
    ResolutionCorrectionValidationManagerTask,
)
class ValidationWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        # yield self.clone(AnalysisEvaluationTask)  # TODO: Needs debugging
        yield self.clone(SidebandEvaluationWrapperTask)
        # yield self.clone(PreSelectionValidationManagerTask)  # TODO: Needs debugging

        # # Commented out, because this takes a lot of time and is replaced by the MultistageResolutionEvaluator!
        # if self.selection_name in SelectionSetups.rdstar_selections:
        #     yield self.clone(MM2ShiftEvaluationTask)

        # TODO: Commented out, to save time during debugging.
        # if self.selection_name in SelectionSetups.rdstar_selections:
        #     yield self.clone(EfficiencyValidationWrapperTask)

        if self.selection_name == SelectionSetups.used_selection_name:
            yield self.clone(VariableImpactValidationWrapperTask)

        if self.selection_name in SelectionSetups.resolution_correction_mapping.values():
            yield self.clone(ResolutionCorrectionValidationManagerTask)

        yield self.clone(WeightValidationWrapperTask)
        yield self.clone(FSRCorrectionValidationManagerTask)
        yield self.clone(AnalysisPlotManagerTask)


class SelectionManagerTask(luigi.WrapperTask):
    selection_setup_names = luigi.TupleParameter(positional=False, hashed=True)  # type: Tuple[str, ...]
    config_combinations = luigi.ListParameter(positional=False, hashed=True)  # type: utilities.ConfigCombinationsType
    working_dir = luigi.Parameter(positional=False)  # type: PathType
    cs_tag = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for selection_name in self.selection_setup_names:
            if not SelectionSetups.selection_is_evaluated(selection_name=selection_name):
                continue

            # yield self.clone(ValidationWrapperTask, selection_name=selection_name)
            # yield self.clone(EfficiencyEvaluationTask, selection_name=selection_name)

            if selection_name in SelectionSetups.rdstar_selections:
                yield self.clone(FittingTasksWrapperTask, selection_name=selection_name)
            elif selection_name == DssFittingManager.dss_fit_selection:
                yield self.clone(DssFittingManager, dss_selection_name=selection_name)
            else:
                yield ResolutionCorrectionWrapperTask(
                    working_dir=self.working_dir,
                    selection_name=selection_name,
                    cs_tag=self.cs_tag,
                    config_combinations=self.config_combinations,
                )


class GeneralManagerTask(luigi.WrapperTask):
    offline_config_path = luigi.Parameter(positional=False)  # type: PathType

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(LatexTasks)
        yield self.clone(CSValidationTask)
        yield MVASelectionValidatorTask(main_offline_config_path=self.offline_config_path)


@inherits(SelectionManagerTask)
class DssFittingManager(luigi.WrapperTask):
    dss_fit_selection = FittingMaster.dss_fit_selection  # type: str

    dss_selection_name = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert self.dss_selection_name == self.dss_fit_selection, (self.dss_selection_name, self.dss_fit_selection)
        self.dss_working_dir = os.path.join(
            self.working_dir,
            FittingMaster.dss_sub_dir_name,
            self.dss_fit_selection,
        )  # type: PathType
        conf = utilities.load_yaml_config(yaml_config_path=self.config_combinations[0][2][0]["offline_config_path"])
        self.dss_production_key = "data_set_" + conf["production_tag"].rsplit("_stream", 1)[0]  # type: str

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def complete(self):
        return all(r.complete() for r in list(self.requires()) + list(self.runtime_requirements()))

    def run(self) -> Generator[utilities.TaskType, None, None]:
        for runtime_requirement in self.runtime_requirements():
            yield runtime_requirement

    def runtime_requirements(self) -> Generator[utilities.TaskType, None, None]:
        yield FittingMaster(
            selection=self.dss_selection_name,
            production_key=self.dss_production_key,
            working_dir=self.dss_working_dir,
        )

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield ResolutionCorrectionWrapperTask(
            working_dir=self.working_dir,
            selection_name=self.dss_selection_name,
            cs_tag=self.cs_tag,
            config_combinations=self.config_combinations,
        )
