"""
Helpers for correlation calculation

Felix Metzner 2021
"""

import pandas as pd

from typing import Union
from dataclasses import dataclass

from templatefitter.plotter.histogram_variable import HistVariable
from analysistools.statistics.weighted_correlation import weighted_corr, valid_weight_corr_calc_methods


__all__ = [
    "corr_eval_info_additional_info_key",
    "CorrCalcInfo",
]


corr_eval_info_additional_info_key = "corr_eval_info"  # type: str


@dataclass(frozen=True)
class CorrCalcInfo:
    var_col_name: str
    var_latex_label: str
    weighted: bool
    corr_calc_method: str = "pearson"

    def __post_init__(self) -> None:
        if self.corr_calc_method not in valid_weight_corr_calc_methods:
            raise ValueError(
                f"Argument corr_calc_method should be one of {valid_weight_corr_calc_methods},"
                f"but '{self.corr_calc_method}' was provided!"
            )

    def calc_corr(
        self,
        df: pd.DataFrame,
        is_data: bool,
        weight_col_name: str,
        base_hist_var: Union[HistVariable, str],
    ) -> float:
        base_col_name = base_hist_var.df_label if isinstance(base_hist_var, HistVariable) else base_hist_var  # type: str
        assert base_col_name in df.columns, base_hist_var
        assert self.var_col_name in df.columns, self.var_col_name

        if (not self.weighted) or is_data:
            corr = df[base_col_name].corr(df[self.var_col_name], method=self.corr_calc_method)  # type: float
        else:
            assert weight_col_name in df.columns, weight_col_name
            corr = weighted_corr(
                x=df[base_col_name].values,
                y=df[self.var_col_name].values,
                w=df[weight_col_name].values,
                method=self.corr_calc_method,
            )

        return corr
