"""
Evaluation of impact of CS MVA cuts on data MC agreement.

Felix Metzner 2021
"""

import os
import copy
import numpy as np
import pandas as pd

from dataclasses import dataclass
from collections import defaultdict
from typing import Optional, Tuple, List, Dict

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.comparison_tools.data_mc_comparer import DataMCComparer
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.correlation_calculator import CorrCalcInfo, corr_eval_info_additional_info_key


__all__ = [
    "ValidationVariableInfo",
    "ValidationVariableSetups",
    "VariableCutImpactValidator",
]


logger = log_to_default(name=__name__)


sideband_data_component_label = "Sideband Data"  # type: str

q2_sideband_key_for_cs_mva_validation = "q2_sideband_without_cut_2"  # type: str
q2_sideband_key_for_sel_mva_validation = "q2_sideband"  # type: str

q2_sideband_title = rf"{LatexObsNames.fitting_q2} Sideband"  # type: str
q2_sideband_text = LatexObsNames.fitting_q2[:-1] + r" < 4.0" + LatexUnits.gev_sq[1:]  # type: str


cs_mva_corr_eval_info = CorrCalcInfo(
    var_col_name="cs_mva_output",
    var_latex_label=LatexObsNames.mva_cs_output,
    weighted=True,
    corr_calc_method="pearson",
)  # type: CorrCalcInfo


_cs_impact_sideband_validation_infos = (
    SidebandInfo(
        key=q2_sideband_key_for_cs_mva_validation,
        hist_var=RDStarHistVars.missing_mass_q2_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.3, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            corr_eval_info_additional_info_key: cs_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_cs_mva_validation,
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.3, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            corr_eval_info_additional_info_key: cs_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_cs_mva_validation,
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.4, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            corr_eval_info_additional_info_key: cs_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_cs_mva_validation,
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.8, 0.97),
        text_h_pos=(0.03, 0.7),
        y_scale=(1.3, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        file_name_appendix="loose_cut",
        requires_selection=(SelectionSetups.used_selection_name,),
        data_component_label=sideband_data_component_label,
        additional_info={
            corr_eval_info_additional_info_key: cs_mva_corr_eval_info,
        },
    ),
)  # type: Tuple[SidebandInfo, ...]


_cs_impact_sideband_validation_infos_with_full_sys = SidebandInfo.create_comparer_infos_with_systematics(
    comparer_info_list=list(_cs_impact_sideband_validation_infos),
    systematics_collection=full_systematics_cols,
    sys_name_tag=full_systematics_name,
)  # type: List[SidebandInfo]

cs_impact_sideband_validation_infos = tuple(
    list(_cs_impact_sideband_validation_infos) + _cs_impact_sideband_validation_infos_with_full_sys
)  # type: Tuple[SidebandInfo, ...]


__missing_mass_for_sel_mva_impact_validation = HistVariable(
    df_label=RDStarHistVars.missing_mass_q2_SidebandValidation.df_label,
    n_bins=18,
    scope=(-1.5, 3.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)
__missing_mass_res_corr_for_sel_mva_impact_validation = HistVariable(
    df_label=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation.df_label,
    n_bins=__missing_mass_for_sel_mva_impact_validation.n_bins,
    scope=__missing_mass_for_sel_mva_impact_validation.scope,
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)
__mm2_binning = tuple(round(x, 2) for x in np.linspace(*__missing_mass_for_sel_mva_impact_validation.scope, 19))


sel_mva_corr_eval_info = CorrCalcInfo(
    var_col_name="sec_sel_mva_output",
    var_latex_label=LatexObsNames.mva_sel_output_short,
    weighted=True,
    corr_calc_method="pearson",
)  # type: CorrCalcInfo

_sel_mva_impact_sideband_validation_infos = (
    SidebandInfo(
        key=q2_sideband_key_for_sel_mva_validation,
        hist_var=__missing_mass_for_sel_mva_impact_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.3, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            "special_binning": __mm2_binning,
            corr_eval_info_additional_info_key: sel_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_sel_mva_validation,
        hist_var=__missing_mass_res_corr_for_sel_mva_impact_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.3, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            "special_binning": __mm2_binning,
            corr_eval_info_additional_info_key: sel_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_sel_mva_validation,
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.64, 0.7),
        y_scale=(1.4, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            "special_binning": tuple(round(x, 2) for x in np.linspace(*(0.1, 2.2), 22)),
            corr_eval_info_additional_info_key: sel_mva_corr_eval_info,
        },
    ),
    SidebandInfo(
        key=q2_sideband_key_for_sel_mva_validation,
        hist_var=RDStarHistVars.cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.8, 0.97),
        text_h_pos=(0.03, 0.7),
        y_scale=(1.2, 2.0),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            corr_eval_info_additional_info_key: sel_mva_corr_eval_info,
        },
    ),
)  # type: Tuple[SidebandInfo, ...]


_sel_mva_impact_sideband_validation_infos_with_full_sys = SidebandInfo.create_comparer_infos_with_systematics(
    comparer_info_list=list(_sel_mva_impact_sideband_validation_infos),
    systematics_collection=full_systematics_cols,
    sys_name_tag=full_systematics_name,
)  # type: List[SidebandInfo]

sel_mva_impact_sideband_validation_infos = tuple(
    list(_sel_mva_impact_sideband_validation_infos) + _sel_mva_impact_sideband_validation_infos_with_full_sys
)  # type: Tuple[SidebandInfo, ...]


@dataclass(frozen=True)
class ValidationVariableInfo:
    hist_var: HistVariable
    cut_values: Tuple[float, ...]
    validation_sideband_infos: Tuple[SidebandInfo, ...]

    def __post_init__(self) -> None:
        s_min = self.hist_var.scope[0]  # type: float
        s_max = self.hist_var.scope[1]  # type: float
        assert all(c >= s_min for c in self.cut_values), [c for c in self.cut_values if c < s_min]
        assert all(c <= s_max for c in self.cut_values), [c for c in self.cut_values if c > s_max]

    @property
    def df_label(self) -> str:
        return self.hist_var.df_label


class ValidationVariableSetups:
    __cs_mva_cut_values = tuple(
        set(
            [round(x, 2) for x in np.linspace(*RDStarHistVars.loose_cs_mva_output.scope, 17)[:-1]]
            + [round(x, 2) for x in np.linspace(0.8, 0.9, 6)[:-1]]
            + [round(x, 2) for x in np.linspace(0.9, RDStarHistVars.loose_cs_mva_output.scope[1], 10)[:-1]]
        )
    )  # type: Tuple[float, ...]
    cs_mva_cut_based = ValidationVariableInfo(
        hist_var=RDStarHistVars.loose_cs_mva_output,
        cut_values=__cs_mva_cut_values,
        validation_sideband_infos=cs_impact_sideband_validation_infos,
    )

    __sel_mva_cut_values = (
        0.5,
        0.575,
        0.65,
        0.75,
        0.808,
        0.85,
        0.88,
        0.9,
        0.91,
        0.92,
        0.93,
        0.94,
        0.95,
    )  # type: Tuple[float, ...]
    sel_mva_cut_based = ValidationVariableInfo(
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        cut_values=__sel_mva_cut_values,
        validation_sideband_infos=sel_mva_impact_sideband_validation_infos,
    )

    @staticmethod
    def get_setup_mapping() -> Dict[str, ValidationVariableInfo]:
        _method_names = [
            "get_setup_mapping",
            "get_setup",
        ]  # type: List[str]
        class_prefix = f"_{ValidationVariableSetups.__name__}"  # type: str
        prefixes = (class_prefix, "__")  # type: Tuple[str, ...]

        setup_dict = {
            attr_name: val
            for attr_name, val in ValidationVariableSetups.__dict__.items()
            if not ((attr_name in _method_names) or attr_name.endswith("__") or attr_name.startswith(prefixes))
        }  # type: Dict[str, ValidationVariableInfo]
        assert all(isinstance(v, ValidationVariableInfo) for v in setup_dict.values()), [
            k for k, v in setup_dict.items() if not isinstance(v, ValidationVariableInfo)
        ]
        return setup_dict

    @staticmethod
    def get_setup(setup_name: str) -> ValidationVariableInfo:
        return ValidationVariableSetups.get_setup_mapping()[setup_name]


class VariableCutImpactValidator:
    base_dir_name = "VariableCutImpactValidator"  # type: str

    default_fig_size = (6.5, 5.0)  # type: Tuple[float, float]

    def __init__(
        self,
        validation_variable: ValidationVariableInfo,
        result_dir_path: PathType,
    ) -> None:

        self.validation_variable = validation_variable  # type: ValidationVariableInfo
        self.sideband_infos = self.validation_variable.validation_sideband_infos  # type: Tuple[SidebandInfo, ...]

        self.base_dir_path = os.path.join(result_dir_path, self.validation_variable.df_label)  # type: PathType

    @staticmethod
    def get_sideband_infos_by_key_mapping(validation_setup_name: str) -> Dict[str, List[SidebandInfo]]:
        mapping = defaultdict(list)  # type: Dict[str, List[SidebandInfo]]

        val_setup = ValidationVariableSetups.get_setup(setup_name=validation_setup_name)  # type: ValidationVariableInfo
        relevant_sideband_infos = val_setup.validation_sideband_infos  # type: Tuple[SidebandInfo, ...]

        for sideband_info in relevant_sideband_infos:
            mapping[sideband_info.sideband_sample_key].append(copy.copy(sideband_info))

        assert sum(len(v) for v in mapping.values()) == len(relevant_sideband_infos), (
            sum(len(v) for v in mapping.values()),
            len(relevant_sideband_infos),
        )
        return mapping

    def get_cut_info_text(self, current_cut_value: float) -> str:
        cut_h_var = self.validation_variable.hist_var  # type: HistVariable
        cut_var_unit_str = r"$" if not cut_h_var.unit else " " + cut_h_var.unit[1:]  # type: str
        return cut_h_var.variable_name[:-1] + f" > {current_cut_value}" + cut_var_unit_str

    def perform_impact_validation(
        self,
        mc_sample: Optional[pd.DataFrame],
        data_sample: Optional[pd.DataFrame],
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        for current_cut_value in self.validation_variable.cut_values:
            sub_output_dict = dict()  # type: Dict[str, PathType]

            if dry_run:
                mc_sub_sample = mc_sample  # type: Optional[pd.DataFrame]
                data_sub_sample = data_sample  # type: Optional[pd.DataFrame]
            else:
                assert isinstance(mc_sample, pd.DataFrame), type(mc_sample).__name__
                assert isinstance(data_sample, pd.DataFrame), type(data_sample).__name__
                mc_sub_sample = mc_sample.loc[mc_sample[self.validation_variable.df_label] >= current_cut_value]
                data_sub_sample = data_sample.loc[data_sample[self.validation_variable.df_label] >= current_cut_value]

            for sideband_info in self.sideband_infos:
                cut_info_text = self.get_cut_info_text(current_cut_value=current_cut_value)  # type: str
                info_text_w_cut = f"{sideband_info.info_text}\n{cut_info_text}"  # type: str

                cut_value_str = str(current_cut_value).replace(".", "p")  # type: str
                if sideband_info.file_name_appendix:
                    new_file_name_appendix = f"{sideband_info.file_name_appendix}_{cut_value_str}"  # type: str
                else:
                    new_file_name_appendix = cut_value_str

                sideband_info_w_cut_str = SidebandInfo.create_modified_comparer_info(
                    comparer_info=sideband_info,
                    info_text=info_text_w_cut,
                    file_name_appendix=new_file_name_appendix,
                )

                eval_output = self.produce_sideband_plot_with_check(
                    mc_df=mc_sub_sample,
                    data_df=data_sub_sample,
                    sideband_info=sideband_info_w_cut_str,
                    cut_value=current_cut_value,
                    dry_run=dry_run,
                    overwrite_existing=False,
                )

                assert all(k not in output_dict.keys() for k in eval_output.keys()), [
                    k for k in eval_output.keys() if k in output_dict.keys()
                ]
                assert all(v not in output_dict.values() for v in eval_output.values()), [
                    v for v in eval_output.values() if v in output_dict.values()
                ]
                output_dict.update(eval_output)
                sub_output_dict.update(eval_output)

            assert len(sub_output_dict) % len(self.sideband_infos) == 0, (len(sub_output_dict), len(self.sideband_infos))

        return output_dict

    def produce_sideband_plot_with_check(
        self,
        mc_df: Optional[pd.DataFrame],
        data_df: Optional[pd.DataFrame],
        sideband_info: SidebandInfo,
        cut_value: float,
        dry_run: bool,
        overwrite_existing: bool = False,
    ) -> Dict[str, PathType]:

        if not overwrite_existing and not dry_run:
            eval_output = self.produce_validation_plots(
                mc_df=mc_df,
                data_df=data_df,
                sideband_info=sideband_info,
                cut_value=cut_value,
                dry_run=True,
            )  # type: Dict[str, PathType]
            if all(os.path.exists(p) for p in eval_output.values()):
                return eval_output

        eval_output = self.produce_validation_plots(
            mc_df=mc_df,
            data_df=data_df,
            sideband_info=sideband_info,
            cut_value=cut_value,
            dry_run=dry_run,
        )
        return eval_output

    def produce_validation_plots(
        self,
        mc_df: Optional[pd.DataFrame],
        data_df: Optional[pd.DataFrame],
        sideband_info: SidebandInfo,
        cut_value: float,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_path = os.path.join(
            self.base_dir_path,
            sideband_info.hist_var.df_label,
            sideband_info.systematics_name_tag,
            self.get_cut_value_string(cut_value=cut_value),
        )  # type: PathType

        if not dry_run:
            assert isinstance(mc_df, pd.DataFrame), type(mc_df)
            assert isinstance(data_df, pd.DataFrame), type(data_df)
            os.makedirs(name=output_path, exist_ok=True)

        data_mc_comparer = DataMCComparer(comparer_info=sideband_info)

        output_dict = dict()  # type: Dict[str, PathType]
        for style in sideband_info.styles:
            try:
                plot_output_dict = data_mc_comparer.plot_data_mc_comparison(
                    mc_df=mc_df,
                    data_df=data_df,
                    dry_run=dry_run,
                    weight_col=WeightManager.final_weight_with_additions_col,
                    style=style,
                    fig_size=self.default_fig_size,
                    normalize_mc_to_data=True,
                    ratio_type="vs_uncert",
                    gof_check_method="toys",
                    adaptive_binning=True,
                    export_path=output_path,
                )
                assert plot_output_dict is not None
                assert all(k not in output_dict.keys() for k in plot_output_dict.keys()), [
                    k for k in plot_output_dict.keys() if k in output_dict.keys()
                ]
                assert all(p not in output_dict.values() for p in plot_output_dict.values()), [
                    p for p in plot_output_dict.values() if p in output_dict.values()
                ]
                output_dict.update(plot_output_dict)
            except Exception as e:
                logger.error(
                    f"Exception occurred for variable {sideband_info.hist_var.df_label} "
                    f"and validation variable name {self.validation_variable.df_label}!"
                )
                raise e

        return output_dict

    def get_cut_value_string(self, cut_value: float) -> str:
        return f"{self.validation_variable.df_label}_ge_" + str(cut_value).replace(".", "p") + "_cut"
