"""
Steering file for skim study for the B -> D(*) tau nu analysis.

Felix Metzner 2019
"""

import basf2

from modularAnalysis import setAnalysisConfigParams, statistics

from analysistools.processing import job_interface
from analysistools.processing import path_utilities

from rdstar.online_analysis import b2bii_utils
from rdstar.online_analysis.decay_utilities import Reconstructor
from rdstar.online_analysis import decay_utilities, decay_channels


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(
    configParametersAndValues={"mcMatchingVersion": "Belle"},
    path=main_path,
)


# Get Final State Particles
main_path.add_module(module="CurlerCloneMarker")
decay_utilities.fsp.fsp_collection.get_final_state_particles(
    path=main_path,
    use_mc_info=cmd_args.config["use_mc_info"],
)

reconstructor = Reconstructor(
    path=main_path,
    config=cmd_args.config,
    decay_channels_module=decay_channels,
)

reconstructor.reconstruct_tau_for_gen()


# Build intermediate D mesons
reconstructor.reconstruct_D()
reconstructor.reconstruct_Dstar()
reconstructor.reconstruct_D2star_for_gen()


# Build signal side B meson
reconstructor.reconstruct_sig_B()


##########
# Output #
##########

p_lists = {
    "B+:sig": "Bplus",
    "B0:sig": "Bzero",
}

for p_list, file_tag in p_lists.items():
    main_path.add_module(
        module="VariablesToNtuple",
        fileName=cmd_args.output_dict[f"{file_tag}_ntuple"],
        particleList=p_list,
        variables=[
            "nCleanedTracks(dr < 2 and abs(dz) < 4)",
            "countInList(gamma:fsp)",
            "ROE_deltae(CleanROE)",
            "ROE_mbc(CleanROE)",
            "ROE_charge(CleanROE)",
            "nROE_Tracks(CleanROE)",
            "ROE_eextra(CleanROE)",
            "ROE_neextra(CleanROE)",
            "charge",
            "mdstIndex",
            "extraInfo(decayModeID)",
            "matchedMC(PDG)",
            "mcMother(PDG)",
            "isSignal",
            "isExtendedSignal",
            "isPrimarySignal",
            "isSignalAcceptMissingNeutrino",
            "isContinuumEvent",
            "isNotContinuumEvent",
            "genUpsilon4S(PDG)",
            "genUpsilon4S(mcDaughter(0, PDG))",
            "genUpsilon4S(mcDaughter(0, mdstIndex))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(0,PDG)))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(0,mcDaughter(0,PDG))))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(0,mcDaughter(1,PDG))))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(1,PDG)))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(1,mcDaughter(0,PDG))))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(1,mcDaughter(1,PDG))))",
            "genUpsilon4S(mcDaughter(1, PDG))",
            "genUpsilon4S(mcDaughter(1, mdstIndex))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(0,PDG)))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(0,mcDaughter(0,PDG))))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(0,mcDaughter(1,PDG))))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(1,PDG)))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(1,mcDaughter(0,PDG))))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(1,mcDaughter(1,PDG))))",
            "genUpsilon4S(mcDaughter(0, charge))",
            "genUpsilon4S(mcDaughter(0, E))",
            "genUpsilon4S(mcDaughter(0, qSquaredMC))",
            "genUpsilon4S(mcDaughter(0, wMC))",
            "genUpsilon4S(mcDaughter(0, costhetalMC))",
            "genUpsilon4S(mcDaughter(0, costhetavMC))",
            "genUpsilon4S(mcDaughter(0, chi_primeMC))",
            "genUpsilon4S(mcDaughter(1, charge))",
            "genUpsilon4S(mcDaughter(1, E))",
            "genUpsilon4S(mcDaughter(1, qSquaredMC))",
            "genUpsilon4S(mcDaughter(1, wMC))",
            "genUpsilon4S(mcDaughter(1, costhetalMC))",
            "genUpsilon4S(mcDaughter(1, costhetavMC))",
            "genUpsilon4S(mcDaughter(1, chi_primeMC))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(2,PDG)))",
            "genUpsilon4S(mcDaughter(0, mcDaughter(3,PDG)))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(2,PDG)))",
            "genUpsilon4S(mcDaughter(1, mcDaughter(3,PDG)))",
        ],
    )

basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
