"""
This file updates the final state particles cut infos for the low level cut study.

Patrick Ecker 2020
"""

from typing import List, Dict

from analysistools.processing.final_state_particles_utils import FinalStateParticleCollection

from rdstar.cut_information_objects.cut_info import CutInfo
import rdstar.online_analysis.final_state_particles as fsp_base_pkg

__all__ = [
    "fsp_collection",
    "fsp_cut_sets",
]


default_scale_factor = 1.0  # type: float


def scale_cut_list(cut_list: List[CutInfo], scale_factor: float) -> List[CutInfo]:
    return [
        cut_info.scale_cut(percentage=scale_factor)
        for cut_info in cut_list
        if cut_info.scale_cut(percentage=scale_factor).cut_str
    ]


study_electron_cuts = scale_cut_list(cut_list=fsp_base_pkg.electron_cuts, scale_factor=default_scale_factor)
study_muon_cuts = scale_cut_list(cut_list=fsp_base_pkg.muon_cuts, scale_factor=default_scale_factor)
study_charged_pion_cuts = scale_cut_list(cut_list=fsp_base_pkg.charged_pion_cuts, scale_factor=default_scale_factor)
study_neutral_pion_cuts = scale_cut_list(cut_list=fsp_base_pkg.neutral_pion_cuts, scale_factor=default_scale_factor)
study_photon_cuts = scale_cut_list(cut_list=fsp_base_pkg.photon_cuts, scale_factor=default_scale_factor)
study_charged_kaon_cuts = scale_cut_list(cut_list=fsp_base_pkg.charged_kaon_cuts, scale_factor=default_scale_factor)
study_neutral_kaon_cuts = scale_cut_list(cut_list=fsp_base_pkg.neutral_kaon_cuts, scale_factor=default_scale_factor)


# Particle Definition
electron = fsp_base_pkg.generate_fsp(
    label="e",
    particle_list="e-:fsp",
    origin="reco",
    cut_infos=study_electron_cuts,
    electron_fsr_correction=fsp_base_pkg.electron_fsr_correction_settings,
)

muon = fsp_base_pkg.generate_fsp(
    label="mu",
    particle_list="mu-:fsp",
    origin="reco",
    cut_infos=study_muon_cuts,
)

charged_pion = fsp_base_pkg.generate_fsp(
    label="pi+",
    particle_list="pi+:fsp",
    origin="reco",
    cut_infos=study_charged_pion_cuts,
)

neutral_pion = fsp_base_pkg.generate_fsp(
    label="pi0",
    particle_list="pi0:fsp",
    origin=["pi0:mdst"],
    cut_infos=study_neutral_pion_cuts,
)

photon = fsp_base_pkg.generate_fsp(
    label="gamma",
    particle_list="gamma:fsp",
    origin=["gamma:mdst", "gamma:v0mdst"],
    cut_infos=study_photon_cuts,
)

charged_kaon = fsp_base_pkg.generate_fsp(
    label="K+",
    particle_list="K+:fsp",
    origin="reco",
    cut_infos=study_charged_kaon_cuts,
)

neutral_kaon = fsp_base_pkg.generate_fsp(
    label="K_S0",
    particle_list="K_S0:fsp",
    origin=["K_S0:mdst"],
    cut_infos=study_neutral_kaon_cuts,
)


# Add Particles to FSPCollection
fsp_collection = FinalStateParticleCollection()
fsp_collection.add_fsps(
    electron=electron,
    muon=muon,
    charged_pion=charged_pion,
    neutral_pion=neutral_pion,
    photon=photon,
    charged_kaon=charged_kaon,
    neutral_kaon=neutral_kaon,
)

fsp_cut_sets = {
    electron.label: study_electron_cuts,
    muon.label: study_muon_cuts,
    charged_pion.label: study_charged_pion_cuts,
    neutral_pion.label: study_neutral_pion_cuts,
    photon.label: study_photon_cuts,
    charged_kaon.label: study_charged_kaon_cuts,
    neutral_kaon.label: study_neutral_kaon_cuts,
}  # type: Dict[str, List[CutInfo]]

if __name__ == "__main__":
    print(fsp_collection)
