"""
Provides information about charged B meson decays as used in Belle.

Felix Metzner 2020
"""

import numpy as np

from typing import Tuple, List, Set

from rdstar.utilities.logging import log_to_default
from rdstar.branching_fractions.bf_utilities import BelleBBFEntry

logger = log_to_default(name=__name__)

__all__ = [
    "sum_belle_decay_b_plus",
    "sum_belle_decay_b_minus",
    "belle_bf_bp_to_d_l_nu",
    "belle_bf_bp_to_d_st_l_nu",
    "belle_bf_bp_to_d_tau_nu",
    "belle_bf_bp_to_d_st_tau_nu",
    "belle_bf_bp_to_dzero_st_l_nu",
    "belle_bf_bp_to_done_l_nu",
    "belle_bf_bp_to_done_prime_l_nu",
    "belle_bf_bp_to_dtwo_st_l_nu",
    "belle_bf_bp_to_dzero_st_tau_nu",
    "belle_bf_bp_to_done_tau_nu",
    "belle_bf_bp_to_done_prime_tau_nu",
    "belle_bf_bp_to_dtwo_st_tau_nu",
    # Break up of Done cases:
    "belle_bf_bp_to_done_to_dst_pi_l_nu",
    "belle_bf_bp_to_done_to_d_pi_pi_l_nu",
    "belle_bf_bp_to_done_to_dst_pi_tau_nu",
    "belle_bf_bp_to_done_to_d_pi_pi_tau_nu",
    # Sum of old B to Dstst l nu bfs
    "sum_of_belle_bp_to_dss_l_nu_bfs",
    # Removed BFs:
    "belle_bf_bp_to_d_2s_l_nu",
    "belle_bf_bp_to_dst_2s_l_nu",
    "belle_bf_bp_to_dst_pi_zero_l_nu",
    "belle_bf_bp_to_dst_pi_minus_l_nu",
    "belle_bf_bp_to_d_pi_zero_l_nu",
    "belle_bf_bp_to_d_pi_minus_l_nu",
    "sum_of_belle_removed_bp_bfs",
]


# fmt: off
belle_decay_infos_b_plus = (
    BelleBBFEntry(decay="anti-D*0     e+   nu_e", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0579),
    BelleBBFEntry(decay="anti-D0      e+   nu_e", model="PHOTOS  HQET2 1.15;", bf=0.0231),
    BelleBBFEntry(decay="anti-D_10    e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="anti-D_0*0   e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0091),
    BelleBBFEntry(decay="anti-D'_10   e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="anti-D_2*0   e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0039),
    BelleBBFEntry(decay="anti-D(2S)0  e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="anti-D*(2S)0 e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0 pi0 e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*-      pi+ e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0007),
    BelleBBFEntry(decay="anti-D0  pi0 e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D-       pi+ e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0011),
    BelleBBFEntry(decay="anti-D*0     mu+  nu_mu", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0579),
    BelleBBFEntry(decay="anti-D0      mu+  nu_mu", model="PHOTOS  HQET2 1.15;", bf=0.0231),
    BelleBBFEntry(decay="anti-D_10    mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="anti-D_0*0   mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0091),
    BelleBBFEntry(decay="anti-D'_10   mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="anti-D_2*0   mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0039),
    BelleBBFEntry(decay="anti-D(2S)0  mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="anti-D*(2S)0 mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0 pi0 mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*-      pi+ mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0007),
    BelleBBFEntry(decay="anti-D0  pi0 mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D-       pi+ mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0011),
    BelleBBFEntry(decay="anti-D*0     tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0154),
    BelleBBFEntry(decay="anti-D0      tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0077),
    BelleBBFEntry(decay="anti-D_10    tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0013),
    BelleBBFEntry(decay="anti-D_0*0   tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0016),
    BelleBBFEntry(decay="anti-D'_10   tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="anti-D_2*0   tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0008),
    BelleBBFEntry(decay="eta_c K+", model="PHOTOS    PHSP;", bf=0.001030),
    BelleBBFEntry(decay="K*+ eta_c", model="PHOTOS  SVS;", bf=0.001700),
    BelleBBFEntry(decay="eta_c K0 pi+", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c K+ pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c K+ pi+ pi-", model="PHOTOS    PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c K+ pi0 pi0", model="PHOTOS    PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c K0 pi+ pi0", model="PHOTOS PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c anti-Lambda0 p+", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c Sigma+ n0", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c Xsu", model="PHSP;", bf=0.001348),
    BelleBBFEntry(decay="eta_c(2S) K+", model="PHOTOS  PHSP;", bf=0.000360),
    BelleBBFEntry(decay="K*+ eta_c(2S)", model="PHOTOS   SVS;", bf=0.000680),
    BelleBBFEntry(decay="eta_c(2S) K0 pi+", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K+ pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K+ pi+ pi-", model="PHOTOS    PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S) K+ pi0 pi0", model="PHOTOS    PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S) K0 pi+ pi0", model="PHOTOS PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S)  Xsu", model="PHSP;", bf=0.000544),
    BelleBBFEntry(decay="J/psi K+", model="PHOTOS     SVS;", bf=0.000990),
    BelleBBFEntry(decay="J/psi K*+", model="PHOTOS    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001350),
    BelleBBFEntry(decay="J/psi K_1+", model="PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001800),
    BelleBBFEntry(decay="J/psi K'_1+", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K+ pi+ pi-", model="PHOTOS  PHSP;", bf=0.000770),
    BelleBBFEntry(decay="J/psi phi K+", model="PHOTOS  PHSP;", bf=0.000052),
    BelleBBFEntry(decay="J/psi pi+", model="PHOTOS   SVS;", bf=0.000048),
    BelleBBFEntry(decay="J/psi rho+", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000017),
    BelleBBFEntry(decay="J/psi a_1+", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K_0*+", model="PHOTOS  SVS;", bf=0.000050),
    BelleBBFEntry(decay="J/psi K_2*+", model="PHOTOS  PHSP;", bf=0.000260),
    BelleBBFEntry(decay="J/psi K''*+", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000005),
    BelleBBFEntry(decay="J/psi K0 pi+", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K+ pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K+ pi0 pi0", model="PHOTOS    PHSP;", bf=0.001120),
    BelleBBFEntry(decay="J/psi K0 pi+ pi0", model="PHOTOS PHSP;", bf=0.001120),
    BelleBBFEntry(decay="J/psi anti-Lambda0 p+", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi Sigma+ n0", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi Xsu", model="PHSP;", bf=0.000512),
    BelleBBFEntry(decay="psi(2S) K+", model="PHOTOS  SVS;", bf=0.000680),
    BelleBBFEntry(decay="psi(2S) K*+", model="PHOTOS  SVV_HELAMP  PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000920),
    BelleBBFEntry(decay="psi(2S) K+ pi+ pi-", model="PHOTOS PHSP;", bf=0.001552),
    BelleBBFEntry(decay="chi_c0 K+", model="PHOTOS  PHSP;", bf=0.000300),
    BelleBBFEntry(decay="K*+ chi_c0", model="PHOTOS SVS;", bf=0.000700),
    BelleBBFEntry(decay="chi_c0 Xsu", model="PHSP;", bf=0.001291),
    BelleBBFEntry(decay="chi_c1 K+", model="PHOTOS  SVS;", bf=0.000680),
    BelleBBFEntry(decay="chi_c1 K*+", model="PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000567),
    BelleBBFEntry(decay="chi_c1 pi+", model="PHOTOS  SVS;", bf=0.000022),
    BelleBBFEntry(decay="chi_c1  Xsu", model="PHSP;", bf=0.002207),
    BelleBBFEntry(decay="chi_c2 K+", model="PHOTOS       STS;", bf=0.000084),
    BelleBBFEntry(decay="chi_c2 K*+", model="PHOTOS       PHSP;", bf=0.000126),
    BelleBBFEntry(decay="chi_c2 Xsu", model="PHSP;", bf=0.001508),
    BelleBBFEntry(decay="h_c K+", model="PHOTOS       SVS;", bf=0.000200),
    BelleBBFEntry(decay="h_c K*+", model="PHOTOS      PHSP;", bf=0.000229),
    BelleBBFEntry(decay="h_c Xsu", model="PHSP;", bf=0.001020),
    BelleBBFEntry(decay="X(3872) K+", model="PHOTOS SVS;", bf=0.000080),
    BelleBBFEntry(decay="Y(3940)  K+", model="PHOTOS PHSP;", bf=0.000071),
    BelleBBFEntry(decay="anti-D0  D_s+", model="PHOTOS PHSP;", bf=0.01000),
    BelleBBFEntry(decay="anti-D*0  D_s+", model="PHOTOS SVS;", bf=0.00820),
    BelleBBFEntry(decay="D_s*+ anti-D0", model="PHOTOS SVS;", bf=0.00760),
    BelleBBFEntry(decay="D_s*+ anti-D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01710),
    BelleBBFEntry(decay="D_s+ anti-D_0*0", model="PHOTOS PHSP;", bf=0.00033),
    BelleBBFEntry(decay="D_s*+ anti-D_0*0", model="PHOTOS SVS;", bf=0.00067),
    BelleBBFEntry(decay="anti-D'_10  D_s+", model="PHOTOS SVS;", bf=0.00034),
    BelleBBFEntry(decay="anti-D'_10  D_s*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00068),
    BelleBBFEntry(decay="anti-D_10  D_s+", model="PHOTOS SVS;", bf=0.00068),
    BelleBBFEntry(decay="anti-D_10  D_s*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00135),
    BelleBBFEntry(decay="anti-D_2*0  D_s+", model="PHOTOS STS;", bf=0.00181),
    BelleBBFEntry(decay="anti-D_2*0  D_s*+", model="PHOTOS PHSP;", bf=0.00363),
    BelleBBFEntry(decay="anti-D0 D_s0*+", model="PHOTOS PHSP;", bf=0.00073),
    BelleBBFEntry(decay="anti-D*0 D_s0*+", model="PHOTOS SVS;", bf=0.00090),
    BelleBBFEntry(decay="D'_s1+ anti-D0", model="PHOTOS SVS;", bf=0.00310),
    BelleBBFEntry(decay="D'_s1+ anti-D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01200),
    BelleBBFEntry(decay="D_s+  D-  pi+", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s+  anti-D0 pi0", model="PHOTOS PHSP;", bf=0.00200),
    BelleBBFEntry(decay="D_s*+ D-  pi+", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*+ anti-D0 pi0", model="PHOTOS PHSP;", bf=0.00200),
    BelleBBFEntry(decay="D_s+  D-  pi+  pi0", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s+  anti-D0 pi+  pi-", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s+  anti-D0 pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*+ D-  pi+  pi0", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s*+ anti-D0 pi+  pi-", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s*+ anti-D0 pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s-  pi+  K+", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*- pi+  K+", model="PHOTOS PHSP;", bf=0.00096),
    BelleBBFEntry(decay="D_s-  pi+  K+  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*- pi+  K+  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s-  pi+  K0  pi+", model="PHOTOS PHSP;", bf=0.00067),
    BelleBBFEntry(decay="D_s*- pi+  K0  pi+", model="PHOTOS PHSP;", bf=0.00067),
    BelleBBFEntry(decay="D_s-  pi+  K+  pi0 pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*- pi+  K+  pi0 pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s-  pi+  K+  pi+ pi-", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*- pi+  K+  pi+ pi-", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s-  pi+  K0  pi+ pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*- pi+  K0  pi+ pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D+   anti-D0", model="PHOTOS PHSP;", bf=0.00042),
    BelleBBFEntry(decay="anti-D*0  D+", model="PHOTOS SVS;", bf=0.00063),
    BelleBBFEntry(decay="D*+  anti-D0", model="PHOTOS SVS;", bf=0.00039),
    BelleBBFEntry(decay="D*+  anti-D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.00081),
    BelleBBFEntry(decay="D'_1+  anti-D0", model="PHOTOS SVS;", bf=0.000018),
    BelleBBFEntry(decay="anti-D'_10  D+", model="PHOTOS SVS;", bf=0.000018),
    BelleBBFEntry(decay="D'_1+  anti-D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000037),
    BelleBBFEntry(decay="anti-D'_10  D*+", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000037),
    BelleBBFEntry(decay="D_1+  anti-D0", model="PHOTOS SVS;", bf=0.000037),
    BelleBBFEntry(decay="anti-D_10  D+", model="PHOTOS SVS;", bf=0.000037),
    BelleBBFEntry(decay="D_1+  anti-D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000075),
    BelleBBFEntry(decay="anti-D_10  D*+", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000075),
    BelleBBFEntry(decay="D_2*+  anti-D0", model="PHOTOS STS;", bf=0.000122),
    BelleBBFEntry(decay="anti-D_2*0  D+", model="PHOTOS STS;", bf=0.000122),
    BelleBBFEntry(decay="D_2*+  anti-D*0", model="PHOTOS PHSP;", bf=0.000244),
    BelleBBFEntry(decay="anti-D_2*0  D*+", model="PHOTOS PHSP;", bf=0.000244),
    BelleBBFEntry(decay="D_s1+ anti-D0", model="PHOTOS SVS;", bf=0.00045),
    BelleBBFEntry(decay="D_s1+ anti-D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00094),
    BelleBBFEntry(decay="D_s2*+ anti-D0", model="PHOTOS STS;", bf=0.00036),
    BelleBBFEntry(decay="D_s2*+ anti-D*0", model="PHOTOS PHSP;", bf=0.00036),
    BelleBBFEntry(decay="D_sj(2700)+ anti-D0", model="PHOTOS SVS;", bf=0.00113),
    BelleBBFEntry(decay="D_sj(2700)+ anti-D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00036),
    BelleBBFEntry(decay="psi(3770) K+", model="PHOTOS SVS;", bf=0.00049),
    BelleBBFEntry(decay="psi(3770) K*+", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.00068),
    BelleBBFEntry(decay="psi(3770) K0 pi+", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) K+ pi0", model="PHOTOS PHSP;", bf=0.00031),
    BelleBBFEntry(decay="psi(3770) K+ pi0 pi0", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="psi(3770) K0 pi+ pi0", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="psi(4040) K+", model="PHOTOS SVS;", bf=0.001548),
    BelleBBFEntry(decay="psi(4040) K*+", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.002322),
    BelleBBFEntry(decay="psi(4040) K0 pi+", model="PHOTOS PHSP;", bf=0.000811),
    BelleBBFEntry(decay="psi(4040) K+ pi0", model="PHOTOS PHSP;", bf=0.000405),
    BelleBBFEntry(decay="psi(4160) K+", model="PHOTOS SVS;", bf=0.001548),
    BelleBBFEntry(decay="psi(4160) K*+", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.002322),
    BelleBBFEntry(decay="psi(4160) K0 pi+", model="PHOTOS PHSP;", bf=0.000811),
    BelleBBFEntry(decay="psi(4160) K+ pi0", model="PHOTOS PHSP;", bf=0.000405),
    BelleBBFEntry(decay="D+   anti-D0   pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   D-   pi+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D0   anti-D0   pi+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   anti-D0   pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*+   D-   pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*0   anti-D0   pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+   anti-D*0   pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+   D*-   pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D0   anti-D*0   pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*+   anti-D*0   pi0", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*+   D*-   pi+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*0   anti-D*0   pi+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D+   anti-D0   rho0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   D-   rho+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   anti-D0   rho+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*+   anti-D0   rho0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   D-   rho+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*0   anti-D0   rho+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   anti-D*0   rho0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   D*-   rho+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D0   anti-D*0   rho+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   anti-D*0   rho0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*+   D*-   rho+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*0   anti-D*0   rho+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+   anti-D0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   anti-D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   D-   pi+  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   anti-D0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   D-   pi+  pi+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*+   anti-D0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   anti-D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   D-   pi+  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*0   anti-D0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*0   D-   pi+  pi+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   anti-D*0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   anti-D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D+   D*-   pi+  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D0   anti-D*0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D0   D*-   pi+  pi+", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*+   anti-D*0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*+   anti-D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*+   D*-   pi+  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*0   anti-D*0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*0   D*-   pi+  pi+", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="anti-D0  D0  K+", model="PHOTOS PHSP;", bf=0.001370),
    BelleBBFEntry(decay="anti-D0  D+   K0", model="PHOTOS PHSP;", bf=0.001700),
    BelleBBFEntry(decay="anti-D0  D0  K*+", model="PHOTOS PHSP;", bf=0.000430),
    BelleBBFEntry(decay="anti-D0  D+   K*0", model="PHOTOS PHSP;", bf=0.000430),
    BelleBBFEntry(decay="anti-D*0 D0 K+", model="PHOTOS PHSP;", bf=0.006500),
    BelleBBFEntry(decay="anti-D*0 D+ K0", model="PHOTOS PHSP;", bf=0.003100),
    BelleBBFEntry(decay="anti-D0  D*0  K+", model="PHOTOS PHSP;", bf=0.004700),
    BelleBBFEntry(decay="anti-D0  D*+  K0", model="PHOTOS PHSP;", bf=0.005200),
    BelleBBFEntry(decay="anti-D*0 D0   K*+", model="PHOTOS PHSP;", bf=0.001721),
    BelleBBFEntry(decay="anti-D*0 D+   K*0", model="PHOTOS PHSP;", bf=0.000860),
    BelleBBFEntry(decay="anti-D0  D*0  K*+", model="PHOTOS PHSP;", bf=0.001290),
    BelleBBFEntry(decay="anti-D0  D*+  K*0", model="PHOTOS PHSP;", bf=0.001290),
    BelleBBFEntry(decay="anti-D*0  D*0  K+", model="PHOTOS PHSP;", bf=0.005300),
    BelleBBFEntry(decay="anti-D*0  D*+  K0", model="PHOTOS PHSP;", bf=0.007800),
    BelleBBFEntry(decay="anti-D*0  D*0  K*+", model="PHOTOS PHSP;", bf=0.002237),
    BelleBBFEntry(decay="anti-D*0  D*+  K*0", model="PHOTOS PHSP;", bf=0.003011),
    BelleBBFEntry(decay="anti-D_10 D0  K+", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D_10 D+   K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D0  D_10  K+", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D0  D_1+  K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D_10  D*0  K+", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D_10  D*+  K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D*0  D_10  K+", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D*0  D_1+  K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D_2*0 D0  K+", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D_2*0 D+   K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D0  D_2*0  K+", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D0  D_2*+  K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="anti-D_2*0  D*0  K+", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D_2*0  D*+  K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D*0  D_2*0  K+", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D*0  D_2*+  K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="anti-D0  D0  K+  pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-  D0  K+  pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D0  D0  K0  pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D0  D+  K+  pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D0  D+  K0  pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-  D+  K0  pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D*0  D0  K+  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*-  D0  K+  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D*0  D0  K0  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D*0  D+  K+  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D*0  D+  K0  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*-  D+  K0  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D0  D*0  K+  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D-  D*0  K+  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D0  D*0  K0  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D0  D*+  K+  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D0  D*+  K0  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D-  D*+  K0  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="anti-D*0  D*0  K+  pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="D*-  D*0  K+  pi+", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="anti-D*0  D*0  K0  pi+", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="anti-D*0  D*+  K+  pi-", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="anti-D*0  D*+  K0  pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="D*-  D*+  K0  pi+", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="anti-D0  D0  K*+  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D0  K*+  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D0  K*0  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+  K*+  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+  K*0  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+  K*0  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D0  K+  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D0  K+  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D0  K0  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+  K+  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+  K0  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+  K0  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+  K+", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D-  D+  K*+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*-  D+  K+", model="PHOTOS PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D*-  D+  K*+", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D-  D*+  K+", model="PHOTOS PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D-  D*+  K*+", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D*-  D*+  K+", model="PHOTOS PHSP;", bf=0.002409),
    BelleBBFEntry(decay="D*-  D*+  K*+", model="PHOTOS PHSP;", bf=0.000774),
    BelleBBFEntry(decay="D_1-  D+  K+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-  D_1+  K+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D_1-  D*+  K+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*-  D_1+  K+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_2*-  D+  K+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-  D_2*+  K+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D_2*-  D*+  K+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*-  D_2*+  K+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="anti-D0   pi+", model="PHOTOS PHSP;", bf=0.00484),
    BelleBBFEntry(decay="anti-D0  pi0  pi+", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="rho+   anti-D0", model="PHOTOS SVS;", bf=0.0134),
    BelleBBFEntry(decay="anti-D0  pi-  pi+  pi+", model="PHOTOS PHSP;", bf=0.0046),
    BelleBBFEntry(decay="anti-D0  rho0 pi+", model="PHOTOS PHSP;", bf=0.0039),
    BelleBBFEntry(decay="a_1+  anti-D0", model="PHOTOS SVS;", bf=0.00365),
    BelleBBFEntry(decay="b_1+  anti-D0", model="PHOTOS SVS;", bf=0.0000),
    BelleBBFEntry(decay="anti-D0  pi-  pi+  pi+  pi0", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="anti-D0  omega  pi+", model="PHOTOS PHSP;", bf=0.0041),
    BelleBBFEntry(decay="anti-D*0 pi+", model="PHOTOS SVS;", bf=0.0052),
    BelleBBFEntry(decay="anti-D*0 pi0  pi+", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0  rho+", model="PHOTOS SVV_HELAMP 0.228 0.95 0.932 0 0.283 1.13;", bf=0.0155),
    BelleBBFEntry(decay="anti-D*0 pi-  pi+  pi+", model="PHOTOS PHSP;", bf=0.0103),
    BelleBBFEntry(decay="anti-D*0 rho0 pi+", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="anti-D*0  a_1+", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0190),
    BelleBBFEntry(decay="anti-D*0  b_1+", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0000),
    BelleBBFEntry(decay="anti-D*0  pi+  pi0  pi0", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0  rho+  pi0", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0  pi-  pi+  pi+  pi0", model="PHOTOS PHSP;", bf=0.0180),
    BelleBBFEntry(decay="anti-D*0  omega  pi+", model="PHOTOS PHSP;", bf=0.0045),
    BelleBBFEntry(decay="D-  pi+  pi+", model="PHOTOS PHSP;", bf=0.0017),
    BelleBBFEntry(decay="D-  pi0   pi+  pi+", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="D-  rho+  pi+", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="D*- pi+  pi+", model="PHOTOS PHSP;", bf=0.00001),
    BelleBBFEntry(decay="D*- pi0   pi+  pi+", model="PHOTOS PHSP;", bf=0.0150),
    BelleBBFEntry(decay="D*- rho+  pi+", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="anti-D_0*0 pi+", model="PHOTOS PHSP;", bf=0.00083),
    BelleBBFEntry(decay="anti-D_10 pi+", model="PHOTOS SVS;", bf=0.00093),
    BelleBBFEntry(decay="anti-D'_10 pi+", model="PHOTOS SVS;", bf=0.00062),
    BelleBBFEntry(decay="anti-D_2*0 pi+", model="PHOTOS STS;", bf=0.00085),
    BelleBBFEntry(decay="rho+ anti-D_0*0", model="PHOTOS SVS;", bf=0.00083),
    BelleBBFEntry(decay="anti-D_10 rho+", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="anti-D'_10 rho+", model="PHOTOS PHSP;", bf=0.00062),
    BelleBBFEntry(decay="anti-D_2*0 rho+", model="PHOTOS PHSP;", bf=0.00085),
    BelleBBFEntry(decay="anti-D0  K+", model="PHOTOS PHSP;", bf=0.00040),
    BelleBBFEntry(decay="K*+ anti-D0", model="PHOTOS SVS;", bf=0.00053),
    BelleBBFEntry(decay="anti-D*0 K+", model="PHOTOS SVS;", bf=0.00042),
    BelleBBFEntry(decay="anti-D*0 K*+", model="PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00081),
    BelleBBFEntry(decay="u  anti-s  anti-c  u", model="PHOTOS PYTHIA 48;", bf=0.02234),
    BelleBBFEntry(decay="u  anti-d  anti-c  u", model="PHOTOS PYTHIA 48;", bf=0.31234),
    BelleBBFEntry(decay="anti-Lambda_c- Delta++", model="PHOTOS PHSP;", bf=0.000020),
    BelleBBFEntry(decay="anti-Sigma_c-  Delta++", model="PHOTOS PHSP;", bf=0.000020),
    BelleBBFEntry(decay="anti-Sigma_c0  Delta+", model="PHOTOS PHSP;", bf=0.000010),
    BelleBBFEntry(decay="anti-Sigma_c0  p+", model="PHOTOS PHSP;", bf=0.000010),
    BelleBBFEntry(decay="anti-cd_1       uu_1", model="PHOTOS PYTHIA 23;", bf=0.034000),
    BelleBBFEntry(decay="anti-cd_1       uu_1", model="PHOTOS PYTHIA 24;", bf=0.012668),
    BelleBBFEntry(decay="anti-Xi_c-     Sigma_c++", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-Xi_c0     Lambda_c+", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-Xi_c0     Sigma_c+", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-cs_0       cu_0", model="PHOTOS PYTHIA 23;", bf=0.008500),
    BelleBBFEntry(decay="anti-Xi_c-     Delta++", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="anti-Xi_c0     p+", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="anti-cs_1       uu_1", model="PHOTOS PYTHIA 23;", bf=0.001800),
    BelleBBFEntry(decay="anti-cs_1       uu_1", model="PHOTOS PYTHIA 24;", bf=0.000450),
    BelleBBFEntry(decay="anti-Lambda_c-  Sigma_c++", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Sigma_c-   Sigma_c++", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Sigma_c0   Sigma_c+", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-cd_0       cu_0", model="PHOTOS PYTHIA 23;", bf=0.000400),
)  # type: Tuple[BelleBBFEntry, ...]


belle_decay_infos_b_minus = (
    BelleBBFEntry(decay="D*0       e-   anti-nu_e", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0579),
    BelleBBFEntry(decay="D0        e-   anti-nu_e", model="PHOTOS  HQET2 1.15;", bf=0.0231),
    BelleBBFEntry(decay="D_10      e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="D_0*0     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0091),
    BelleBBFEntry(decay="D'_10     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="D_2*0     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0039),
    BelleBBFEntry(decay="D(2S)0    e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)0   e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*0  pi0  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*+  pi-  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0007),
    BelleBBFEntry(decay="D0   pi0  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D+   pi-  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0011),
    BelleBBFEntry(decay="D*0       mu-  anti-nu_mu", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0579),
    BelleBBFEntry(decay="D0        mu-  anti-nu_mu", model="PHOTOS  HQET2 1.15;", bf=0.0231),
    BelleBBFEntry(decay="D_10      mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="D_0*0     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0091),
    BelleBBFEntry(decay="D'_10     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0081),
    BelleBBFEntry(decay="D_2*0     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0039),
    BelleBBFEntry(decay="D(2S)0    mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)0   mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*0  pi0  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*+  pi-  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0007),
    BelleBBFEntry(decay="D0   pi0  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D+   pi-  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0011),
    BelleBBFEntry(decay="D*0       tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0154),
    BelleBBFEntry(decay="D0        tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0077),
    BelleBBFEntry(decay="D_10      tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0013),
    BelleBBFEntry(decay="D_0*0     tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0016),
    BelleBBFEntry(decay="D'_10     tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="D_2*0     tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0008),
    BelleBBFEntry(decay="eta_c K-", model="PHOTOS    PHSP;", bf=0.001030),
    BelleBBFEntry(decay="K*- eta_c", model="PHOTOS  SVS;", bf=0.001700),
    BelleBBFEntry(decay="eta_c anti-K0 pi-", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c K- pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c K- pi+ pi-", model="PHOTOS    PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c K- pi0 pi0", model="PHOTOS    PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c anti-K0 pi- pi0", model="PHOTOS PHSP;", bf=0.000457),
    BelleBBFEntry(decay="eta_c Lambda0 anti-p-", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c Sigma- anti-n0", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c anti-Xsu", model="PHSP;", bf=0.001348),
    BelleBBFEntry(decay="eta_c(2S) K-", model="PHOTOS  PHSP;", bf=0.000360),
    BelleBBFEntry(decay="K*- eta_c(2S)", model="PHOTOS   SVS;", bf=0.000680),
    BelleBBFEntry(decay="eta_c(2S) anti-K0 pi-", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K- pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K- pi+ pi-", model="PHOTOS    PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S) K- pi0 pi0", model="PHOTOS    PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S) anti-K0 pi- pi0", model="PHOTOS PHSP;", bf=0.000202),
    BelleBBFEntry(decay="eta_c(2S)  anti-Xsu", model="PHSP;", bf=0.000544),
    BelleBBFEntry(decay="J/psi K-", model="PHOTOS     SVS;", bf=0.000990),
    BelleBBFEntry(decay="J/psi K*-", model="PHOTOS    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001350),
    BelleBBFEntry(decay="J/psi K_1-", model="PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001800),
    BelleBBFEntry(decay="J/psi K'_1-", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K- pi+ pi-", model="PHOTOS  PHSP;", bf=0.000770),
    BelleBBFEntry(decay="J/psi phi K-", model="PHOTOS  PHSP;", bf=0.000052),
    BelleBBFEntry(decay="J/psi pi-", model="PHOTOS   SVS;", bf=0.000048),
    BelleBBFEntry(decay="J/psi rho-", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000017),
    BelleBBFEntry(decay="J/psi a_1-", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K_0*-", model="PHOTOS  SVS;", bf=0.000050),
    BelleBBFEntry(decay="J/psi K_2*-", model="PHOTOS  PHSP;", bf=0.000260),
    BelleBBFEntry(decay="J/psi K''*-", model="PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000005),
    BelleBBFEntry(decay="J/psi anti-K0 pi-", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K- pi0", model="PHOTOS     PHSP;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K- pi0 pi0", model="PHOTOS    PHSP;", bf=0.001120),
    BelleBBFEntry(decay="J/psi anti-K0 pi- pi0", model="PHOTOS PHSP;", bf=0.001120),
    BelleBBFEntry(decay="J/psi Lambda0 anti-p-", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi Sigma- anti-n0", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi anti-Xsu", model="PHSP;", bf=0.000512),
    BelleBBFEntry(decay="psi(2S) K-", model="PHOTOS  SVS;", bf=0.000680),
    BelleBBFEntry(decay="psi(2S) K*-", model="PHOTOS  SVV_HELAMP  PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000920),
    BelleBBFEntry(decay="psi(2S) K- pi+ pi-", model="PHOTOS PHSP;", bf=0.001552),
    BelleBBFEntry(decay="chi_c0 K-", model="PHOTOS  PHSP;", bf=0.000300),
    BelleBBFEntry(decay="K*- chi_c0", model="PHOTOS SVS;", bf=0.000700),
    BelleBBFEntry(decay="chi_c0 anti-Xsu", model="PHSP;", bf=0.001291),
    BelleBBFEntry(decay="chi_c1 K-", model="PHOTOS  SVS;", bf=0.000680),
    BelleBBFEntry(decay="chi_c1 K*-", model="PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000567),
    BelleBBFEntry(decay="chi_c1 pi-", model="PHOTOS  SVS;", bf=0.000022),
    BelleBBFEntry(decay="chi_c1  anti-Xsu", model="PHSP;", bf=0.002207),
    BelleBBFEntry(decay="chi_c2 K-", model="PHOTOS       STS;", bf=0.000084),
    BelleBBFEntry(decay="chi_c2 K*-", model="PHOTOS       PHSP;", bf=0.000126),
    BelleBBFEntry(decay="chi_c2 anti-Xsu", model="PHSP;", bf=0.001508),
    BelleBBFEntry(decay="h_c K-", model="PHOTOS       SVS;", bf=0.000200),
    BelleBBFEntry(decay="h_c K*-", model="PHOTOS      PHSP;", bf=0.000229),
    BelleBBFEntry(decay="h_c anti-Xsu", model="PHSP;", bf=0.001020),
    BelleBBFEntry(decay="X(3872) K-", model="PHOTOS SVS;", bf=0.000080),
    BelleBBFEntry(decay="Y(3940)  K-", model="PHOTOS PHSP;", bf=0.000071),
    BelleBBFEntry(decay="D0   D_s-", model="PHOTOS PHSP;", bf=0.01000),
    BelleBBFEntry(decay="D*0  D_s-", model="PHOTOS SVS;", bf=0.00820),
    BelleBBFEntry(decay="D_s*- D0", model="PHOTOS SVS;", bf=0.00760),
    BelleBBFEntry(decay="D_s*- D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01710),
    BelleBBFEntry(decay="D_s- D_0*0", model="PHOTOS PHSP;", bf=0.00033),
    BelleBBFEntry(decay="D_s*- D_0*0", model="PHOTOS SVS;", bf=0.00067),
    BelleBBFEntry(decay="D'_10  D_s-", model="PHOTOS SVS;", bf=0.00034),
    BelleBBFEntry(decay="D'_10  D_s*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00068),
    BelleBBFEntry(decay="D_10  D_s-", model="PHOTOS SVS;", bf=0.00068),
    BelleBBFEntry(decay="D_10  D_s*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00135),
    BelleBBFEntry(decay="D_2*0  D_s-", model="PHOTOS STS;", bf=0.00181),
    BelleBBFEntry(decay="D_2*0  D_s*-", model="PHOTOS PHSP;", bf=0.00363),
    BelleBBFEntry(decay="D0 D_s0*-", model="PHOTOS PHSP;", bf=0.00073),
    BelleBBFEntry(decay="D*0 D_s0*-", model="PHOTOS SVS;", bf=0.00090),
    BelleBBFEntry(decay="D'_s1- D0", model="PHOTOS SVS;", bf=0.00310),
    BelleBBFEntry(decay="D'_s1- D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01200),
    BelleBBFEntry(decay="D_s-  D+  pi-", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s-  D0  pi0", model="PHOTOS PHSP;", bf=0.00200),
    BelleBBFEntry(decay="D_s*- D+  pi-", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*- D0  pi0", model="PHOTOS PHSP;", bf=0.00200),
    BelleBBFEntry(decay="D_s-  D+  pi-  pi0", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s-  D0  pi-  pi+", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s-  D0  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*- D+  pi-  pi0", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s*- D0  pi-  pi+", model="PHOTOS PHSP;", bf=0.00160),
    BelleBBFEntry(decay="D_s*- D0  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s+  pi-  K-", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*+ pi-  K-", model="PHOTOS PHSP;", bf=0.00096),
    BelleBBFEntry(decay="D_s+  pi-  K-  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*+ pi-  K-  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.00067),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.00067),
    BelleBBFEntry(decay="D_s+  pi-  K-  pi0 pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*+ pi-  K-  pi0 pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s+  pi-  K-  pi+ pi-", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*+ pi-  K-  pi+ pi-", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0  pi- pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0  pi- pi0", model="PHOTOS PHSP;", bf=0.00018),
    BelleBBFEntry(decay="D-   D0", model="PHOTOS PHSP;", bf=0.00042),
    BelleBBFEntry(decay="D*0  D-", model="PHOTOS SVS;", bf=0.00063),
    BelleBBFEntry(decay="D*-  D0", model="PHOTOS SVS;", bf=0.00039),
    BelleBBFEntry(decay="D*-  D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.00081),
    BelleBBFEntry(decay="D'_1-  D0", model="PHOTOS SVS;", bf=0.000018),
    BelleBBFEntry(decay="D'_10  D-", model="PHOTOS SVS;", bf=0.000018),
    BelleBBFEntry(decay="D'_1-  D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000037),
    BelleBBFEntry(decay="D'_10  D*-", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000037),
    BelleBBFEntry(decay="D_1-  D0", model="PHOTOS SVS;", bf=0.000037),
    BelleBBFEntry(decay="D_10  D-", model="PHOTOS SVS;", bf=0.000037),
    BelleBBFEntry(decay="D_1-  D*0", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000075),
    BelleBBFEntry(decay="D_10  D*-", model="PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;", bf=0.000075),
    BelleBBFEntry(decay="D_2*-  D0", model="PHOTOS STS;", bf=0.000122),
    BelleBBFEntry(decay="D_2*0  D-", model="PHOTOS STS;", bf=0.000122),
    BelleBBFEntry(decay="D_2*-  D*0", model="PHOTOS PHSP;", bf=0.000244),
    BelleBBFEntry(decay="D_2*0  D*-", model="PHOTOS PHSP;", bf=0.000244),
    BelleBBFEntry(decay="D_s1- D0", model="PHOTOS SVS;", bf=0.00045),
    BelleBBFEntry(decay="D_s1- D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00094),
    BelleBBFEntry(decay="D_s2*- D0", model="PHOTOS STS;", bf=0.00036),
    BelleBBFEntry(decay="D_s2*- D*0", model="PHOTOS PHSP;", bf=0.00036),
    BelleBBFEntry(decay="D_sj(2700)- D0", model="PHOTOS SVS;", bf=0.00113),
    BelleBBFEntry(decay="D_sj(2700)- D*0", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00036),
    BelleBBFEntry(decay="psi(3770) K-", model="PHOTOS SVS;", bf=0.00049),
    BelleBBFEntry(decay="psi(3770) K*-", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.00068),
    BelleBBFEntry(decay="psi(3770) anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) K- pi0", model="PHOTOS PHSP;", bf=0.00031),
    BelleBBFEntry(decay="psi(3770) K- pi0 pi0", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="psi(3770) anti-K0 pi- pi0", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="psi(4040) K-", model="PHOTOS SVS;", bf=0.001548),
    BelleBBFEntry(decay="psi(4040) K*-", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.002322),
    BelleBBFEntry(decay="psi(4040) anti-K0 pi-", model="PHOTOS PHSP;", bf=0.000811),
    BelleBBFEntry(decay="psi(4040) K- pi0", model="PHOTOS PHSP;", bf=0.000405),
    BelleBBFEntry(decay="psi(4160) K-", model="PHOTOS SVS;", bf=0.001548),
    BelleBBFEntry(decay="psi(4160) K*-", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.002322),
    BelleBBFEntry(decay="psi(4160) anti-K0 pi-", model="PHOTOS PHSP;", bf=0.000811),
    BelleBBFEntry(decay="psi(4160) K- pi0", model="PHOTOS PHSP;", bf=0.000405),
    BelleBBFEntry(decay="D-   D0   pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D+   pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D0   D0   pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D0   pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*-   D+   pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D*0   D0   pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-   D*0   pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-   D*+   pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D0   D*0   pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*-   D*0   pi0", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*-   D*+   pi-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="anti-D*0   D*0   pi-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D-   D0   rho0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D+   rho-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D0   rho-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*-   D0   rho0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D+   rho-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D*0   D0   rho-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D*0   rho0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D*+   rho-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D0   D*0   rho-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D*0   rho0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*-   D*+   rho-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="anti-D*0   D*0   rho-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D-   D0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D0   pi-  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D+   pi-  pi-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*-   D0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D*0   D0   pi-  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D*0   D+   pi-  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D*0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D-   D*+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D0   D*0   pi-  pi0", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="anti-D0   D*+   pi-  pi-", model="PHOTOS PHSP;", bf=0.000026),
    BelleBBFEntry(decay="D*-   D*0   pi0  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*-   D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D*-   D*+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="anti-D*0   D*0   pi-  pi0", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="anti-D*0   D*+   pi-  pi-", model="PHOTOS PHSP;", bf=0.000077),
    BelleBBFEntry(decay="D0  anti-D0  K-", model="PHOTOS PHSP;", bf=0.001370),
    BelleBBFEntry(decay="D0  D-   anti-K0", model="PHOTOS PHSP;", bf=0.001700),
    BelleBBFEntry(decay="D0  anti-D0  K*-", model="PHOTOS PHSP;", bf=0.000430),
    BelleBBFEntry(decay="D0  D-   anti-K*0", model="PHOTOS PHSP;", bf=0.000430),
    BelleBBFEntry(decay="D*0 anti-D0  K-", model="PHOTOS PHSP;", bf=0.006500),
    BelleBBFEntry(decay="D*0 D-   anti-K0", model="PHOTOS PHSP;", bf=0.003100),
    BelleBBFEntry(decay="D0  anti-D*0  K-", model="PHOTOS PHSP;", bf=0.004700),
    BelleBBFEntry(decay="D0  D*-  anti-K0", model="PHOTOS PHSP;", bf=0.005200),
    BelleBBFEntry(decay="D*0 anti-D0  K*-", model="PHOTOS PHSP;", bf=0.001721),
    BelleBBFEntry(decay="D*0 D-   anti-K*0", model="PHOTOS PHSP;", bf=0.000860),
    BelleBBFEntry(decay="D0  anti-D*0  K*-", model="PHOTOS PHSP;", bf=0.001290),
    BelleBBFEntry(decay="D0  D*-  anti-K*0", model="PHOTOS PHSP;", bf=0.001290),
    BelleBBFEntry(decay="D*0  anti-D*0  K-", model="PHOTOS PHSP;", bf=0.005300),
    BelleBBFEntry(decay="D*0  D*-  anti-K0", model="PHOTOS PHSP;", bf=0.007800),
    BelleBBFEntry(decay="D*0  anti-D*0  K*-", model="PHOTOS PHSP;", bf=0.002237),
    BelleBBFEntry(decay="D*0  D*-  anti-K*0", model="PHOTOS PHSP;", bf=0.003011),
    BelleBBFEntry(decay="D_10 anti-D0  K-", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D_10 D-   anti-K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D0  anti-D_10  K-", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D0  D_1-  anti-K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D_10  anti-D*0  K-", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D_10  D*-  anti-K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D*0  anti-D_10  K-", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D*0  D_1-  anti-K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D_2*0 anti-D0  K-", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D_2*0 D-   anti-K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D0  anti-D_2*0  K-", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D0  D_2*-  anti-K0", model="PHOTOS PHSP;", bf=0.000112),
    BelleBBFEntry(decay="D_2*0  anti-D*0  K-", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D_2*0  D*-  anti-K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D*0  anti-D_2*0  K-", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D*0  D_2*-  anti-K0", model="PHOTOS PHSP;", bf=0.000336),
    BelleBBFEntry(decay="D0  anti-D0  K-  pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+  anti-D0  K-  pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D0  anti-D0  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D0  D-  K-  pi+", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D0  D-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+  D-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*0  anti-D0  K-  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*+  anti-D0  K-  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*0  anti-D0  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*0  D-  K-  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*0  D-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*+  D-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D0  anti-D*0  K-  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D+  anti-D*0  K-  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D0  anti-D*0  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D0  D*-  K-  pi+", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D0  D*-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D+  D*-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000172),
    BelleBBFEntry(decay="D*0  anti-D*0  K-  pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="D*+  anti-D*0  K-  pi-", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="D*0  anti-D*0  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="D*0  D*-  K-  pi+", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="D*0  D*-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="D*+  D*-  anti-K0  pi-", model="PHOTOS PHSP;", bf=0.000309),
    BelleBBFEntry(decay="D0  anti-D0  K*-  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  anti-D0  K*-  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  anti-K*0  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-  K*-  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-  anti-K*0  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-  anti-K*0  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  K-  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  anti-D0  K-  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  anti-K0  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-  K-  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-  anti-K0  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-  anti-K0  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-  K-", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D+  D-  K*-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D*+  D-  K-", model="PHOTOS PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D*+  D-  K*-", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D+  D*-  K-", model="PHOTOS PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D+  D*-  K*-", model="PHOTOS PHSP;", bf=0.000344),
    BelleBBFEntry(decay="D*+  D*-  K-", model="PHOTOS PHSP;", bf=0.002409),
    BelleBBFEntry(decay="D*+  D*-  K*-", model="PHOTOS PHSP;", bf=0.000774),
    BelleBBFEntry(decay="D_1+  D-  K-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+  D_1-  K-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D_1+  D*-  K-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*+  D_1-  K-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_2*+  D-  K-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D+  D_2*-  K-", model="PHOTOS PHSP;", bf=0.000086),
    BelleBBFEntry(decay="D_2*+  D*-  K-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D*+  D_2*-  K-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D0   pi-", model="PHOTOS PHSP;", bf=0.00484),
    BelleBBFEntry(decay="D0  pi0  pi-", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="rho-   D0", model="PHOTOS SVS;", bf=0.0134),
    BelleBBFEntry(decay="D0  pi+  pi-  pi-", model="PHOTOS PHSP;", bf=0.0046),
    BelleBBFEntry(decay="D0  rho0 pi-", model="PHOTOS PHSP;", bf=0.0039),
    BelleBBFEntry(decay="a_1- D0", model="PHOTOS SVS;", bf=0.00365),
    BelleBBFEntry(decay="b_1- D0", model="PHOTOS SVS;", bf=0.0000),
    BelleBBFEntry(decay="D0   pi+  pi-  pi-  pi0", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="D0  omega  pi-", model="PHOTOS PHSP;", bf=0.0041),
    BelleBBFEntry(decay="D*0 pi-", model="PHOTOS SVS;", bf=0.0052),
    BelleBBFEntry(decay="D*0 pi0  pi-", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="D*0  rho-", model="PHOTOS SVV_HELAMP 0.228 0.95 0.932 0 0.283 1.13;", bf=0.0155),
    BelleBBFEntry(decay="D*0 pi+  pi-  pi-", model="PHOTOS PHSP;", bf=0.0103),
    BelleBBFEntry(decay="D*0 rho0 pi-", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="D*0 a_1-", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0190),
    BelleBBFEntry(decay="D*0 b_1-", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0000),
    BelleBBFEntry(decay="D*0 pi-   pi0  pi0", model="PHOTOS  PHSP;", bf=0.0005),
    BelleBBFEntry(decay="D*0 rho-  pi0", model="PHOTOS PHSP;", bf=0.0005),
    BelleBBFEntry(decay="D*0  pi+  pi-  pi-  pi0", model="PHOTOS PHSP;", bf=0.0180),
    BelleBBFEntry(decay="D*0  omega  pi-", model="PHOTOS PHSP;", bf=0.0045),
    BelleBBFEntry(decay="D+  pi-  pi-", model="PHOTOS PHSP;", bf=0.0017),
    BelleBBFEntry(decay="D+  pi0   pi-  pi-", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="D+  rho-  pi-", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="D*+ pi-  pi-", model="PHOTOS PHSP;", bf=0.00001),
    BelleBBFEntry(decay="D*+ pi0   pi-  pi-", model="PHOTOS PHSP;", bf=0.0150),
    BelleBBFEntry(decay="D*+ rho-  pi-", model="PHOTOS PHSP;", bf=0.0020),
    BelleBBFEntry(decay="D_0*0 pi-", model="PHOTOS PHSP;", bf=0.00083),
    BelleBBFEntry(decay="D_10 pi-", model="PHOTOS SVS;", bf=0.00093),
    BelleBBFEntry(decay="D'_10 pi-", model="PHOTOS SVS;", bf=0.00062),
    BelleBBFEntry(decay="D_2*0 pi-", model="PHOTOS STS;", bf=0.00085),
    BelleBBFEntry(decay="rho- D_0*0", model="PHOTOS SVS;", bf=0.00083),
    BelleBBFEntry(decay="D_10 rho-", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="D'_10 rho-", model="PHOTOS PHSP;", bf=0.00062),
    BelleBBFEntry(decay="D_2*0 rho-", model="PHOTOS PHSP;", bf=0.00085),
    BelleBBFEntry(decay="D0  K-", model="PHOTOS PHSP;", bf=0.00040),
    BelleBBFEntry(decay="K*- D0", model="PHOTOS SVS;", bf=0.00053),
    BelleBBFEntry(decay="D*0 K-", model="PHOTOS SVS;", bf=0.00042),
    BelleBBFEntry(decay="D*0 K*-", model="PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00081),
    BelleBBFEntry(decay="anti-u  s  c  anti-u", model="PHOTOS PYTHIA 48;", bf=0.02234),
    BelleBBFEntry(decay="anti-u  d  c  anti-u", model="PHOTOS PYTHIA 48;", bf=0.31234),
    BelleBBFEntry(decay="Lambda_c+ anti-Delta--", model="PHOTOS PHSP;", bf=0.000020),
    BelleBBFEntry(decay="Sigma_c+  anti-Delta--", model="PHOTOS PHSP;", bf=0.000020),
    BelleBBFEntry(decay="Sigma_c0  anti-Delta-", model="PHOTOS PHSP;", bf=0.000010),
    BelleBBFEntry(decay="Sigma_c0  anti-p-", model="PHOTOS PHSP;", bf=0.000010),
    BelleBBFEntry(decay="cd_1       anti-uu_1", model="PHOTOS PYTHIA 23;", bf=0.034000),
    BelleBBFEntry(decay="cd_1       anti-uu_1", model="PHOTOS PYTHIA 24;", bf=0.012668),
    BelleBBFEntry(decay="Xi_c+     anti-Sigma_c--", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="Xi_c0     anti-Lambda_c-", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="Xi_c0     anti-Sigma_c-", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="cs_0       anti-cu_0", model="PHOTOS PYTHIA 23;", bf=0.008500),
    BelleBBFEntry(decay="Xi_c+     anti-Delta--", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="Xi_c0     anti-p-", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="cs_1       anti-uu_1", model="PHOTOS PYTHIA 23;", bf=0.001800),
    BelleBBFEntry(decay="cs_1       anti-uu_1", model="PHOTOS PYTHIA 24;", bf=0.000450),
    BelleBBFEntry(decay="Lambda_c+   anti-Sigma_c--", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Sigma_c+   anti-Sigma_c--", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Sigma_c0    anti-Sigma_c-", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="cd_0       anti-cu_0", model="PHOTOS PYTHIA 23;", bf=0.000400),
)  # type: Tuple[BelleBBFEntry, ...]
# fmt: on

sum_belle_decay_b_plus = sum([bf_e.bf for bf_e in belle_decay_infos_b_plus])  # type: float
sum_belle_decay_b_minus = sum([bf_e.bf for bf_e in belle_decay_infos_b_minus])  # type: float

assert sum_belle_decay_b_minus == sum_belle_decay_b_plus, (sum_belle_decay_b_minus, sum_belle_decay_b_plus)
assert np.isclose(a=sum_belle_decay_b_minus, b=1.0, rtol=1e-5, atol=1e-12), sum_belle_decay_b_minus

# print(f"Sum Decay B+: {sum_belle_decay_b_plus}")
# print(f"Sum Decay B-: {sum_belle_decay_b_minus}")

__comb_list = tuple(list(belle_decay_infos_b_plus) + list(belle_decay_infos_b_minus))  # type: Tuple[BelleBBFEntry, ...]

__belle_bf_bp_to_d_l_nu_keys = (
    "D0 e- anti-nu_e",
    "D0 mu- anti-nu_mu",
    "anti-D0 e+ nu_e",
    "anti-D0 mu+ nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_d_st_l_nu_keys = (
    "D*0 e- anti-nu_e",
    "D*0 mu- anti-nu_mu",
    "anti-D*0 e+ nu_e",
    "anti-D*0 mu+ nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_d_tau_nu_keys = ("D0 tau- anti-nu_tau", "anti-D0 tau+ nu_tau")  # type: Tuple[str, ...]
__belle_bf_bp_to_d_st_tau_nu_keys = ("D*0 tau- anti-nu_tau", "anti-D*0 tau+ nu_tau")  # type: Tuple[str, ...]

__belle_bf_bp_to_d_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_d_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_st_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_d_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_tau_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_d_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_st_tau_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_bp_to_d_l_nu_set) == 1, list(__belle_bf_bp_to_d_l_nu_set)
assert len(__belle_bf_bp_to_d_st_l_nu_set) == 1, list(__belle_bf_bp_to_d_st_l_nu_set)
assert len(__belle_bf_bp_to_d_tau_nu_set) == 1, list(__belle_bf_bp_to_d_tau_nu_set)
assert len(__belle_bf_bp_to_d_st_tau_nu_set) == 1, list(__belle_bf_bp_to_d_st_tau_nu_set)


belle_bf_bp_to_d_l_nu = list(__belle_bf_bp_to_d_l_nu_set)[0]  # type: float
belle_bf_bp_to_d_st_l_nu = list(__belle_bf_bp_to_d_st_l_nu_set)[0]  # type: float
belle_bf_bp_to_d_tau_nu = list(__belle_bf_bp_to_d_tau_nu_set)[0]  # type: float
belle_bf_bp_to_d_st_tau_nu = list(__belle_bf_bp_to_d_st_tau_nu_set)[0]  # type: float


__belle_bf_bp_to_dzero_st_l_nu_keys = (
    "D_0*0 e- anti-nu_e",
    "D_0*0 mu- anti-nu_mu",
    "anti-D_0*0 e+ nu_e",
    "anti-D_0*0 mu+ nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_done_l_nu_keys = (
    "D_10 e- anti-nu_e",
    "D_10 mu- anti-nu_mu",
    "anti-D_10 e+ nu_e",
    "anti-D_10 mu+ nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_done_prime_l_nu_keys = (
    "D'_10 e- anti-nu_e",
    "D'_10 mu- anti-nu_mu",
    "anti-D'_10 e+ nu_e",
    "anti-D'_10 mu+ nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_dtwo_st_l_nu_keys = (
    "D_2*0 e- anti-nu_e",
    "D_2*0 mu- anti-nu_mu",
    "anti-D_2*0 e+ nu_e",
    "anti-D_2*0 mu+ nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_bp_to_dzero_st_tau_nu_keys = (
    "D_0*0 tau- anti-nu_tau",
    "anti-D_0*0 tau+ nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_done_tau_nu_keys = (
    "D_10 tau- anti-nu_tau",
    "anti-D_10 tau+ nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_done_prime_tau_nu_keys = (
    "D'_10 tau- anti-nu_tau",
    "anti-D'_10 tau+ nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_dtwo_st_tau_nu_keys = (
    "D_2*0 tau- anti-nu_tau",
    "anti-D_2*0 tau+ nu_tau",
)  # type: Tuple[str, ...]


__belle_bf_bp_to_dzero_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dzero_st_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_done_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_done_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_done_prime_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_done_prime_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_dtwo_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dtwo_st_l_nu_keys]
)  # type: Set[float]

__belle_bf_bp_to_dzero_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dzero_st_tau_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_done_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_done_tau_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_done_prime_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_done_prime_tau_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_dtwo_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dtwo_st_tau_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_bp_to_dzero_st_l_nu_set) == 1, list(__belle_bf_bp_to_dzero_st_l_nu_set)
assert len(__belle_bf_bp_to_done_l_nu_set) == 1, list(__belle_bf_bp_to_done_l_nu_set)
assert len(__belle_bf_bp_to_done_prime_l_nu_set) == 1, list(__belle_bf_bp_to_done_prime_l_nu_set)
assert len(__belle_bf_bp_to_dtwo_st_l_nu_set) == 1, list(__belle_bf_bp_to_dtwo_st_l_nu_set)

assert len(__belle_bf_bp_to_dzero_st_tau_nu_set) == 1, list(__belle_bf_bp_to_dzero_st_tau_nu_set)
assert len(__belle_bf_bp_to_done_tau_nu_set) == 1, list(__belle_bf_bp_to_done_tau_nu_set)
assert len(__belle_bf_bp_to_done_prime_tau_nu_set) == 1, list(__belle_bf_bp_to_done_prime_tau_nu_set)
assert len(__belle_bf_bp_to_dtwo_st_tau_nu_set) == 1, list(__belle_bf_bp_to_dtwo_st_tau_nu_set)

belle_bf_bp_to_dzero_st_l_nu = list(__belle_bf_bp_to_dzero_st_l_nu_set)[0]  # type: float
belle_bf_bp_to_done_l_nu = list(__belle_bf_bp_to_done_l_nu_set)[0]  # type: float
belle_bf_bp_to_done_prime_l_nu = list(__belle_bf_bp_to_done_prime_l_nu_set)[0]  # type: float
belle_bf_bp_to_dtwo_st_l_nu = list(__belle_bf_bp_to_dtwo_st_l_nu_set)[0]  # type: float

belle_bf_bp_to_dzero_st_tau_nu = list(__belle_bf_bp_to_dzero_st_tau_nu_set)[0]  # type: float
belle_bf_bp_to_done_tau_nu = list(__belle_bf_bp_to_done_tau_nu_set)[0]  # type: float
belle_bf_bp_to_done_prime_tau_nu = list(__belle_bf_bp_to_done_prime_tau_nu_set)[0]  # type: float
belle_bf_bp_to_dtwo_st_tau_nu = list(__belle_bf_bp_to_dtwo_st_tau_nu_set)[0]  # type: float

sum_of_belle_bp_to_dss_l_nu_bfs = sum(
    [
        2 * belle_bf_bp_to_dzero_st_l_nu,
        2 * belle_bf_bp_to_done_l_nu,
        2 * belle_bf_bp_to_done_prime_l_nu,
        2 * belle_bf_bp_to_dtwo_st_l_nu,
        belle_bf_bp_to_dzero_st_tau_nu,
        belle_bf_bp_to_done_tau_nu,
        belle_bf_bp_to_done_prime_tau_nu,
        belle_bf_bp_to_dtwo_st_tau_nu,
    ]
)  # type: float


belle_decay_infos_done_charged = (
    BelleBBFEntry(decay="D*+ pi0", model="PHOTOS VVS_PWAVE  0.0 0.0 0.0 0.0 1.0 0.0;", bf=0.1997),
    BelleBBFEntry(decay="D*0 pi+", model="PHOTOS VVS_PWAVE  0.0 0.0 0.0 0.0 1.0 0.0;", bf=0.3994),
    BelleBBFEntry(decay="D+ pi+ pi-", model="PHOTOS PHSP;", bf=0.1719),
    BelleBBFEntry(decay="D+ pi0 pi0", model="PHOTOS PHSP;", bf=0.1145),
    BelleBBFEntry(decay="D0 pi+ pi0", model="PHOTOS PHSP;", bf=0.1145),
)  # type: Tuple[BelleBBFEntry, ...]


belle_charged_done_regular_entries = [
    e for e in belle_decay_infos_done_charged if e.decay.count(" pi") == 1
]  # type: List[BelleBBFEntry]
belle_charged_done_gap_entries = [
    e for e in belle_decay_infos_done_charged if e.decay.count(" pi") == 2
]  # type: List[BelleBBFEntry]
belle_charged_done_regular_fraction = sum([e.bf for e in belle_charged_done_regular_entries])  # type: float
belle_charged_done_gap_fraction = sum([e.bf for e in belle_charged_done_gap_entries])  # type: float


belle_bf_bp_to_done_to_dst_pi_l_nu = belle_bf_bp_to_done_l_nu * belle_charged_done_regular_fraction  # type: float
belle_bf_bp_to_done_to_d_pi_pi_l_nu = belle_bf_bp_to_done_l_nu * belle_charged_done_gap_fraction  # type: float
belle_bf_bp_to_done_to_dst_pi_tau_nu = belle_bf_bp_to_done_tau_nu * belle_charged_done_regular_fraction  # type: float
belle_bf_bp_to_done_to_d_pi_pi_tau_nu = belle_bf_bp_to_done_tau_nu * belle_charged_done_gap_fraction  # type: float


__belle_bf_bp_to_d_2s_l_nu_keys = (
    "anti-D(2S)0 e+ nu_e",
    "anti-D(2S)0 mu+ nu_mu",
    "D(2S)0 e- anti-nu_e",
    "D(2S)0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_dst_2s_l_nu_keys = (
    "anti-D*(2S)0 e+ nu_e",
    "anti-D*(2S)0 mu+ nu_mu",
    "D*(2S)0 e- anti-nu_e",
    "D*(2S)0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_bp_to_d_2s_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_2s_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_dst_2s_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dst_2s_l_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_bp_to_d_2s_l_nu_set) == 1, list(__belle_bf_bp_to_d_2s_l_nu_set)
assert len(__belle_bf_bp_to_dst_2s_l_nu_set) == 1, list(__belle_bf_bp_to_dst_2s_l_nu_set)

belle_bf_bp_to_d_2s_l_nu = list(__belle_bf_bp_to_d_2s_l_nu_set)[0]  # type: float
belle_bf_bp_to_dst_2s_l_nu = list(__belle_bf_bp_to_dst_2s_l_nu_set)[0]  # type: float


__belle_bf_bp_to_dst_pi_zero_l_nu_keys = (
    "anti-D*0 pi0 e+ nu_e",
    "anti-D*0 pi0 mu+ nu_mu",
    "D*0 pi0 e- anti-nu_e",
    "D*0 pi0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_dst_pi_minus_l_nu_keys = (
    "D*- pi+ e+ nu_e",
    "D*- pi+ mu+ nu_mu",
    "D*+ pi- e- anti-nu_e",
    "D*+ pi- mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_d_pi_zero_l_nu_keys = (
    "anti-D0 pi0 e+ nu_e",
    "anti-D0 pi0 mu+ nu_mu",
    "D0 pi0 e- anti-nu_e",
    "D0 pi0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_bp_to_d_pi_minus_l_nu_keys = (
    "D- pi+ e+ nu_e",
    "D- pi+ mu+ nu_mu",
    "D+ pi- e- anti-nu_e",
    "D+ pi- mu- anti-nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_bp_to_dst_pi_zero_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dst_pi_zero_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_dst_pi_minus_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_dst_pi_minus_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_d_pi_zero_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_pi_zero_l_nu_keys]
)  # type: Set[float]
__belle_bf_bp_to_d_pi_minus_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_bp_to_d_pi_minus_l_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_bp_to_dst_pi_zero_l_nu_set) == 1, list(__belle_bf_bp_to_dst_pi_zero_l_nu_set)
assert len(__belle_bf_bp_to_dst_pi_minus_l_nu_set) == 1, list(__belle_bf_bp_to_dst_pi_minus_l_nu_set)
assert len(__belle_bf_bp_to_d_pi_zero_l_nu_set) == 1, list(__belle_bf_bp_to_d_pi_zero_l_nu_set)
assert len(__belle_bf_bp_to_d_pi_minus_l_nu_set) == 1, list(__belle_bf_bp_to_d_pi_minus_l_nu_set)

belle_bf_bp_to_dst_pi_zero_l_nu = list(__belle_bf_bp_to_dst_pi_zero_l_nu_set)[0]  # type: float
belle_bf_bp_to_dst_pi_minus_l_nu = list(__belle_bf_bp_to_dst_pi_minus_l_nu_set)[0]  # type: float
belle_bf_bp_to_d_pi_zero_l_nu = list(__belle_bf_bp_to_d_pi_zero_l_nu_set)[0]  # type: float
belle_bf_bp_to_d_pi_minus_l_nu = list(__belle_bf_bp_to_d_pi_minus_l_nu_set)[0]  # type: float


sum_of_belle_removed_bp_bfs = sum(
    [
        belle_bf_bp_to_d_2s_l_nu,
        belle_bf_bp_to_dst_2s_l_nu,
        belle_bf_bp_to_dst_pi_zero_l_nu,
        belle_bf_bp_to_dst_pi_minus_l_nu,
        belle_bf_bp_to_d_pi_zero_l_nu,
        belle_bf_bp_to_d_pi_minus_l_nu,
    ]
)  # type: float


original_b_plus_decay_lines = [
    "Decay B+",
    "# B->DXlnu",
    "#     B+ -> X_c l nu, sum = 11.35% (using: tau_B+/tau_B0 = 1.09)",
    "#     2005/04/02 H.Kakuno, Update using the following numbers:",
    "#       B->Xlnu     Br=10.9%               (HFAG summer 2004)",
    "#       B0->Dlnu    Br=2.14%,  rho_D=1.15  (HFAG summer 2004)",
    "#       B0->D*lnu   Br=5.33%               (HFAG summer 2004),",
    "#                    rho_A1=1.3            (hepex/0408027),",
    "#                    R1=1.18,  R2=0.71     (PRL76,3893)",
    "#       B->D**lnu   Br=3.04%          (CERN-EP/2001-050)",
    "#       B->D_1 lnu  Br=0.74%          (PDG 2004)",
    "#       B->D_2* lnu Br<0.65%          (PDG 2004)",
    "#       B->D pi lnu Br=1.54%          (PLB475, 407)",
    "#       B->D*pi lnu Br=1.86%          (PLB475, 407)",
    "#       referred also D0 preliminary result from CKM2005 for D**lnu",
    "0.0579   anti-D*0     e+   nu_e     PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0231   anti-D0      e+   nu_e     PHOTOS  HQET2 1.15;",
    "0.0081   anti-D_10    e+   nu_e     PHOTOS  ISGW2;",
    "0.0091   anti-D_0*0   e+   nu_e     PHOTOS  ISGW2;",
    "0.0081   anti-D'_10   e+   nu_e     PHOTOS  ISGW2;",
    "0.0039   anti-D_2*0   e+   nu_e     PHOTOS  ISGW2;",
    "0.0002   anti-D(2S)0  e+   nu_e     PHOTOS  ISGW2;",
    "0.0005   anti-D*(2S)0 e+   nu_e     PHOTOS  ISGW2;",
    "0.0003   anti-D*0 pi0 e+   nu_e     PHOTOS  GOITY_ROBERTS;",
    "0.0007   D*-      pi+ e+   nu_e     PHOTOS  GOITY_ROBERTS;",
    "0.0005   anti-D0  pi0 e+   nu_e     PHOTOS  GOITY_ROBERTS;",
    "0.0011   D-       pi+ e+   nu_e     PHOTOS  GOITY_ROBERTS;",
    "0.0579   anti-D*0     mu+  nu_mu    PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0231   anti-D0      mu+  nu_mu    PHOTOS  HQET2 1.15;",
    "0.0081   anti-D_10    mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0091   anti-D_0*0   mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0081   anti-D'_10   mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0039   anti-D_2*0   mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0002   anti-D(2S)0  mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0005   anti-D*(2S)0 mu+  nu_mu    PHOTOS  ISGW2;",
    "0.0003   anti-D*0 pi0 mu+  nu_mu    PHOTOS  GOITY_ROBERTS;",
    "0.0007   D*-      pi+ mu+  nu_mu    PHOTOS  GOITY_ROBERTS;",
    "0.0005   anti-D0  pi0 mu+  nu_mu    PHOTOS  GOITY_ROBERTS;",
    "0.0011   D-       pi+ mu+  nu_mu    PHOTOS  GOITY_ROBERTS;",
    "# B->DXtaunu",
    "#                    B- -> X_c tau nu, sum = 2.7% (using: tau_B+/tau_B0 = 1.09)",
    "#                    H.Kakuno 2004/04/02",
    "0.0154   anti-D*0     tau+ nu_tau           PHOTOS ISGW2;",
    "0.0077   anti-D0      tau+ nu_tau           PHOTOS ISGW2;",
    "0.0013   anti-D_10    tau+ nu_tau           PHOTOS ISGW2;",
    "0.0016   anti-D_0*0   tau+ nu_tau           PHOTOS ISGW2;",
    "0.0012   anti-D'_10   tau+ nu_tau           PHOTOS ISGW2;",
    "0.0008   anti-D_2*0   tau+ nu_tau           PHOTOS ISGW2;",
    "# B->ccX",
    "#	BF[%]		B0	B+",
    "#	eta_c		0.5034	0.5455",
    "#	eta_c(2S)	0.2015	0.2190",
    "#	J/psi		0.7539	0.8118",
    "#	psi(2S)		0.2967	0.3152",
    "#	chi_c0		0.2127	0.2291",
    "#	chi_c1		0.3228	0.3476",
    "#	chi_c2		0.1595	0.1718",
    "#	h_c		0.1345	0.1449",
    "#	X(3872)		0.0080	0.0080",
    "#	Y(3940)		0.0070	0.0071",
    "# -------------------------------------",
    "#	sum		2.6000	2.8000",
    "# B+ --> charmonium X,  2.8000%",
    "# B -> cc= s",
    "# eta_c inclusive",
    "# PDG 2005: < 0.009",
    "# Guess 0.005455",
    "# B+ --> eta_c K+",
    "# PDG 2005: 0.00103 +/- 0.00021",
    "0.001030   eta_c K+ PHOTOS    PHSP;",
    "# B+ --> K*- eta_c",
    "# Guess 0.001700",
    "0.001700   K*+ eta_c  PHOTOS  SVS;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   eta_c K0 pi+ PHOTOS     PHSP;",
    "0.000000   eta_c K+ pi0      PHOTOS     PHSP;",
    "0.000457   eta_c K+ pi+ pi-   PHOTOS    PHSP;",
    "0.000457   eta_c K+ pi0 pi0   PHOTOS    PHSP;",
    "0.000457   eta_c K0 pi+ pi0 PHOTOS PHSP;",
    "# baryonic decays",
    "0.000003  eta_c anti-Lambda0 p+      PHOTOS PHSP;",
    "0.000003  eta_c Sigma+ n0  PHOTOS PHSP;",
    "# Fill up by string fragmentation",
    "0.001348  eta_c Xsu PHSP;",
    "# Sum of eta_c: 0.005455",
    "# eta_c(2S) Inclusive",
    "# guess 0.002190 (40% of eta_c)",
    "# B+ --> eta_c(2S) K+",
    "# Guess 0.00036",
    "0.000360   eta_c(2S) K+  PHOTOS  PHSP;",
    "# B+ --> eta_c(2S) K*-",
    "# Guess 0.00068",
    "0.000680   K*+ eta_c(2S) PHOTOS   SVS;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   eta_c(2S) K0 pi+ PHOTOS     PHSP;",
    "0.000000   eta_c(2S) K+ pi0      PHOTOS     PHSP;",
    "0.000202   eta_c(2S) K+ pi+ pi-   PHOTOS    PHSP;",
    "0.000202   eta_c(2S) K+ pi0 pi0   PHOTOS    PHSP;",
    "0.000202   eta_c(2S) K0 pi+ pi0 PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.000544   eta_c(2S)  Xsu  PHSP;",
    "# Sum of eta_c(2S): 0.002190",
    "# J/psi inclusive (direct)",
    "# PDG 2005:0.008118",
    "# B+ --> J/psi K+",
    "# PDG 2003:  0.00099 +/- 0.00004",
    "0.000990   J/psi K+  PHOTOS     SVS;",
    "# B+ --> J/psi K*+",
    "# PDG 2005:  0.00135 +/- 0.00010",
    "0.001350   J/psi K*+  PHOTOS    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B+ --> J/psi K(1270)+",
    "# PDG 2005:  0.0018  +/- 0.0005",
    "0.001800   J/psi K_1+  PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B+ --> J/psi K(1400)-",
    "# PDG 2005:  < 0.0005",
    "0.000000   J/psi K'_1+  PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# This is not part of K1(1270)?",
    "# B+ --> J/psi K+ pi+ pi-",
    "# PDG 2005:  0.00077 +/- 0.00020",
    "0.000770   J/psi K+ pi+ pi- PHOTOS  PHSP;",
    "# Omit B+ --> J/psi eta K+, because it comes from Psi(2S) K+.",
    "# B+ --> J/psi phi K+",
    "# PDG 2005:  0.000052 +/- 0.000017",
    "0.000052   J/psi phi K+ PHOTOS  PHSP;",
    "# B+ --> J/psi pi+",
    "# PDG 2005:  0.000048 +/- 0.000006",
    "0.000048   J/psi pi+   PHOTOS   SVS;",
    "# B+ --> J/psi rho-",
    "# PDG 2005:  < 0.00077",
    "# Guess 0.000017",
    "0.000017   J/psi rho+   PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B+ --> J/psi a1(1260)+",
    "# PDG 2005: < 0.0012",
    "0.000000   J/psi a_1+   PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Belle 0.000050   Belle Nitesh K0(1430) (Kpi [93+-10]%)",
    "0.000050   J/psi K_0*+  PHOTOS  SVS;",
    "# Belle 0.000260   Belle Nitesh K2*(1430) (Kpi [49.9+-1.2]%)  STV not implemented?",
    "0.000260   J/psi K_2*+  PHOTOS  PHSP;",
    "# Belle ?   Belle Nitesh K*(1680) (Kpi [38.7 +- 2.5]%)",
    "0.000005   J/psi K''*+  PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Fill up inclusive rate",
    "# kpipi replaced by K1,",
    "# little evidence for Kpi other than K*, K*(1430)",
    "0.000000   J/psi K0 pi+ PHOTOS     PHSP;",
    "0.000000   J/psi K+ pi0      PHOTOS     PHSP;",
    "0.001120   J/psi K+ pi0 pi0   PHOTOS    PHSP;",
    "0.001120   J/psi K0 pi+ pi0 PHOTOS PHSP;",
    "# baryonic decays",
    "0.000012  J/psi anti-Lambda0 p+      PHOTOS PHSP;",
    "0.000012  J/psi Sigma+ n0  PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.000512   J/psi Xsu PHSP;",
    "# Sum of J/psi: 0.008118",
    "# Psi(2S) inclusive",
    "# PDG 2005: 0.003152",
    "# B+ --> Psi(2S) K+",
    "# PDG 2005: 0.00068 +/- 0.00004",
    "0.000680   psi(2S) K+   PHOTOS  SVS;",
    "# B+ --> Psi(2S) K*(892)-",
    "# PDG 2003: 0.00092 +/- 0.00022",
    "# Belle-Conf 0349 (LP03): 0.000813 +/- 0.000077 +/- 0.000089",
    "0.000920   psi(2S) K*+  PHOTOS  SVV_HELAMP  PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B+ --> Psi(2S) K+ pi+ pi-",
    "# PDG 2005: (1.9+-1.2)*10^-3, adjusted to match inclusive Br.",
    "0.001552   psi(2S) K+ pi+ pi- PHOTOS PHSP;",
    "# Omit multi-body decays.",
    "#0.000000   psi(2S) K0 pi+  PHOTOS PHSP;",
    "#0.000000   psi(2S) K+ pi0 PHOTOS  PHSP;",
    "#0.000532   psi(2S) K+ pi0 pi0 PHOTOS PHSP;",
    "#0.000533   psi(2S) K0 pi+ pi0 PHOTOS PHSP;",
    "# No room to put string fragmenataion.",
    "# Sum of Psi(2S): 0.003152",
    "# chi_c0 inclusive",
    "# Guess 0.2291",
    "# B+ --> chi_c0 K+",
    "# PDG 2005: 0.000300 +- 0.000060",
    "0.000300   chi_c0 K+  PHOTOS  PHSP;",
    "# B+ --> chi_c0 K*+",
    "# Guess 0.000700",
    "0.000700   K*+ chi_c0  PHOTOS SVS;",
    "# Guess 0.001291  Fill up inclusive rate with low momentum",
    "#0.000118   chi_c0 anti-K0 pi- PHOTOS     PHSP;",
    "#0.000118   chi_c0 K+ pi0       PHOTOS    PHSP;",
    "#0.000351   chi_c0 K+ pi+ pi-    PHOTOS   PHSP;",
    "#0.000352   chi_c0 K+ pi0 pi0    PHOTOS   PHSP;",
    "#0.000352   chi_c0 anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001291  chi_c0 Xsu PHSP;",
    "# Sum of chi_c0: 0.2291",
    "# chi_c1 inclusive (direct)",
    "# PDG 2005: 0.003476",
    "# B+ --> chi_c1 K+",
    "# PDG 2005: 0.00068 +/- 0.00011",
    "0.000680   chi_c1 K+  PHOTOS  SVS;",
    "# B+ --> chi_c1 K*+",
    "# PDG 2005: < 0.0021",
    "# Weighted average Belle(PLB634(255)155) and BaBar(PRL94(2005)141801) is",
    "# (5.67+-0.97)*10^-4.",
    "0.000567   chi_c1 K*+ PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Belle chi_c1 pi+ (2.21+-0.42+-0.32)*10^-5 by Rajeev(BN876).",
    "0.000022   chi_c1 pi+  PHOTOS  SVS;",
    "# Fill up inclusive rate with low momentum  0.002207",
    "#0.000000   chi_c1 K0 pi+ PHOTOS     PHSP;",
    "#0.000000   chi_c1 K+ pi0     PHOTOS      PHSP;",
    "#0.000735   chi_c1 K+ pi+ pi-  PHOTOS     PHSP;",
    "#0.000736   chi_c1 K+ pi0 pi0   PHOTOS    PHSP;",
    "#0.000736   chi_c1 K0 pi+ pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation",
    "0.002207   chi_c1  Xsu PHSP;",
    "# Sum of chi_c1: 0.003476",
    "# chi_c2 inclusive (direct)",
    "# PDG 2005: 0.001718",
    "# No real evidence for these decays",
    "# B+ --> chi_c2 K+",
    "# Guess 0.000084",
    "0.000084   chi_c2 K+        PHOTOS       STS;",
    "# B+ --> chi_c2 K*+",
    "# Guess 0.000126",
    "0.000126   chi_c2 K*+       PHOTOS       PHSP;",
    "# Fill up inclusive rate with low momentum",
    "#0.000000   chi_c2 K0 pi+  PHOTOS    PHSP;",
    "#0.000000   chi_c2 K+ pi0      PHOTOS     PHSP;",
    "#0.000503   chi_c2 K+ pi+ pi-   PHOTOS    PHSP;",
    "#0.000502   chi_c2 K+ pi0 pi0     PHOTOS  PHSP;",
    "#0.000503   chi_c2 K0 pi+ pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001508    chi_c2 Xsu  PHSP;",
    "# Sum of chi_c2: 0.001718",
    "# h_c Inclusive",
    "# Guess 0.001449",
    "# Guess 0.000200",
    "0.000200   h_c K+        PHOTOS       SVS;",
    "# Guess 0.000240",
    "0.000229   h_c K*+        PHOTOS      PHSP;",
    "# Guess 0.001020",
    "#0.000000   h_c K0 pi+  PHOTOS    PHSP;",
    "#0.000000   h_c K+ pi0       PHOTOS    PHSP;",
    "#0.000340   h_c K+ pi+ pi-   PHOTOS    PHSP;",
    "#0.000340   h_c K+ pi0 pi0    PHOTOS   PHSP;",
    "#0.000340   h_c K0 pi+ pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001020   h_c Xsu PHSP;",
    "# Sum of h_c: 0.001449",
    "# X(3872), register B+ -> X(3872) K+ only.",
    "# ?",
    "0.000080   X(3872) K+ PHOTOS SVS;",
    "# Omit B -> X(3872) K*, because no significant signal (Karim's work BAM200509)",
    "# 0.000130   X(3872) K*- PHOTOS SVV_HELAMP PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# sum of X(3872) 0.000080",
    "# Register  B+ -> Y(3940) K+",
    "0.000071  Y(3940)  K+  PHOTOS PHSP;",
    "# Sum of B+ --> charmonium X: 0.0280",
    "# B->DsX",
    "#   b -> c (sc=) -> D Ds X    Sum = 8.9%",
    "#   BF(Ds->phipi)=4.38(Babar2005+PDG) is assumed.",
    "#   Updated on 29/01/2008",
    "#   based on PDG update (Ref: W.-M. Yao et al., J. Phys. G 33, 1 (2006) and 2007 partial update for the 2008)",
    "#   by Nikhil J Joshi (njoshi@bpost)",
    "## Upper Vertex Ds",
    "0.01000    anti-D0  D_s+             	PHOTOS PHSP;",
    "0.00820    anti-D*0  D_s+              	PHOTOS SVS;",
    "0.00760    D_s*+ anti-D0              	PHOTOS SVS;",
    "0.01710   D_s*+ anti-D*0               	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00033   D_s+ anti-D_0*0              	PHOTOS PHSP;",
    "0.00067   D_s*+ anti-D_0*0             	PHOTOS SVS;",
    "0.00034   anti-D'_10  D_s+             	PHOTOS SVS;",
    "0.00068   anti-D'_10  D_s*+            	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00068   anti-D_10  D_s+              	PHOTOS SVS;",
    "0.00135   anti-D_10  D_s*+             	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00181   anti-D_2*0  D_s+             	PHOTOS STS;",
    "0.00363   anti-D_2*0  D_s*+            	PHOTOS PHSP;",
    "# DsJ(2317)=D_s0*-  and  DsJ(2457)=D'_s1-",
    "0.00073   anti-D0 D_s0*+              	PHOTOS PHSP;",
    "0.00090   anti-D*0 D_s0*+              	PHOTOS SVS;",
    "0.00310   D'_s1+ anti-D0               	PHOTOS SVS;",
    "0.01200   D'_s1+ anti-D*0              	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "#",
    "0.00100    D_s+  D-  pi+              	PHOTOS PHSP;",
    "0.00200    D_s+  anti-D0 pi0          	PHOTOS PHSP;",
    "0.00100    D_s*+ D-  pi+              	PHOTOS PHSP;",
    "0.00200    D_s*+ anti-D0 pi0         	PHOTOS PHSP;",
    "0.00160    D_s+  D-  pi+  pi0        	PHOTOS PHSP;",
    "0.00160    D_s+  anti-D0 pi+  pi-    	PHOTOS PHSP;",
    "0.00070    D_s+  anti-D0 pi0  pi0     	PHOTOS PHSP;",
    "0.00160    D_s*+ D-  pi+  pi0         	PHOTOS PHSP;",
    "0.00160    D_s*+ anti-D0 pi+  pi-     	PHOTOS PHSP;",
    "0.00070    D_s*+ anti-D0 pi0  pi0     	PHOTOS PHSP;",
    "## End of upper vertex",
    "## Lower vertex Ds",
    "0.00070   D_s-  pi+  K+             	PHOTOS PHSP;",
    "0.00096   D_s*- pi+  K+              	PHOTOS PHSP;",
    "0.00100   D_s-  pi+  K+  pi0          	PHOTOS PHSP;",
    "0.00100   D_s*- pi+  K+  pi0          	PHOTOS PHSP;",
    "0.00067  D_s-  pi+  K0  pi+           	PHOTOS PHSP;",
    "0.00067  D_s*- pi+  K0  pi+           	PHOTOS PHSP;",
    "0.00018   D_s-  pi+  K+  pi0 pi0       	PHOTOS PHSP;",
    "0.00018   D_s*- pi+  K+  pi0 pi0       	PHOTOS PHSP;",
    "0.00018   D_s-  pi+  K+  pi+ pi-       	PHOTOS PHSP;",
    "0.00018   D_s*- pi+  K+  pi+ pi-       	PHOTOS PHSP;",
    "0.00018   D_s-  pi+  K0  pi+ pi0       	PHOTOS PHSP;",
    "0.00018   D_s*- pi+  K0  pi+ pi0       	PHOTOS PHSP;",
    "# B+ ->DDX",
    "# updated by H.Miyake 20021006",
    "# updated by H.Miyake 20031019",
    "# updated by J.Brodzicka (JB) 20080130 basing on PDG07",
    "# b-> c (dc=)",
    "# JB PDG07",
    "0.00042      D+   anti-D0           PHOTOS PHSP;",
    "# JB PDG07",
    "0.00063      anti-D*0  D+           PHOTOS SVS;",
    "# JB PDG07",
    "0.00039      D*+  anti-D0           PHOTOS SVS;",
    "# JB PDG07",
    "0.00081      D*+  anti-D*0          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "# fudge factor to tune DDX: ~0.860298",
    "# guess from neutral mode",
    "0.000018   D'_1+  anti-D0           PHOTOS SVS;",
    "0.000018   anti-D'_10  D+           PHOTOS SVS;",
    "0.000037   D'_1+  anti-D*0          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000037   anti-D'_10  D*+          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000037   D_1+  anti-D0            PHOTOS SVS;",
    "0.000037   anti-D_10  D+            PHOTOS SVS;",
    "0.000075   D_1+  anti-D*0           PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000075   anti-D_10  D*+           PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000122   D_2*+  anti-D0           PHOTOS STS;",
    "0.000122   anti-D_2*0  D+           PHOTOS STS;",
    "0.000244   D_2*+  anti-D*0          PHOTOS PHSP;",
    "0.000244   anti-D_2*0  D*+          PHOTOS PHSP;",
    "# Ds1(2536)+=D_s1+  and  Ds2(2573)+=D_s2*+",
    "# DsJ(2700)+=D_sj(2700)+",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00045   D_s1+ anti-D0                       PHOTOS SVS;",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00094   D_s1+ anti-D*0                      PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00036   D_s2*+ anti-D0                      PHOTOS STS;",
    "0.00036   D_s2*+ anti-D*0                     PHOTOS PHSP;",
    "# JB hep-ex/0707.3491",
    "0.00113   D_sj(2700)+ anti-D0                 PHOTOS SVS;",
    "0.00036   D_sj(2700)+ anti-D*0                PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "# JB PDG07",
    "0.00049   psi(3770) K+   PHOTOS SVS;",
    "0.00068   psi(3770) K*+  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.00064   psi(3770) K0 pi+ PHOTOS PHSP;",
    "0.00031   psi(3770) K+ pi0 PHOTOS PHSP;",
    "0.00093   psi(3770) K+ pi0 pi0 PHOTOS PHSP;",
    "0.00093   psi(3770) K0 pi+ pi0 PHOTOS PHSP;",
    "# JB fudge factor to tune DDX: ~0.737246",
    "0.001548   psi(4040) K+   PHOTOS SVS;",
    "0.002322   psi(4040) K*+  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000811   psi(4040) K0 pi+ PHOTOS PHSP;",
    "0.000405   psi(4040) K+ pi0 PHOTOS PHSP;",
    "0.001548   psi(4160) K+   PHOTOS SVS;",
    "0.002322   psi(4160) K*+  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000811   psi(4160) K0 pi+ PHOTOS PHSP;",
    "0.000405   psi(4160) K+ pi0 PHOTOS PHSP;",
    "#not included in qq98",
    "0.000026     D+   anti-D0   pi0     PHOTOS PHSP;",
    "0.000026     D+   D-   pi+          PHOTOS PHSP;",
    "0.000026     D0   anti-D0   pi+          PHOTOS PHSP;",
    "0.000086     D*+   anti-D0   pi0     PHOTOS PHSP;",
    "0.000086     D*+   D-   pi+          PHOTOS PHSP;",
    "0.000086     D*0   anti-D0   pi+          PHOTOS PHSP;",
    "0.000086     D+   anti-D*0   pi0     PHOTOS PHSP;",
    "0.000086     D+   D*-   pi+          PHOTOS PHSP;",
    "0.000086     D0   anti-D*0   pi+          PHOTOS PHSP;",
    "0.000258     D*+   anti-D*0   pi0     PHOTOS PHSP;",
    "0.000258     D*+   D*-   pi+          PHOTOS PHSP;",
    "0.000258     D*0   anti-D*0   pi+          PHOTOS PHSP;",
    "0.000009     D+   anti-D0   rho0     PHOTOS PHSP;",
    "0.000009     D+   D-   rho+          PHOTOS PHSP;",
    "0.000009     D0   anti-D0   rho+          PHOTOS PHSP;",
    "0.000026     D*+   anti-D0   rho0     PHOTOS PHSP;",
    "0.000026     D*+   D-   rho+          PHOTOS PHSP;",
    "0.000026     D*0   anti-D0   rho+          PHOTOS PHSP;",
    "0.000026     D+   anti-D*0   rho0     PHOTOS PHSP;",
    "0.000026     D+   D*-   rho+          PHOTOS PHSP;",
    "0.000026     D0   anti-D*0   rho+          PHOTOS PHSP;",
    "0.000086     D*+   anti-D*0   rho0     PHOTOS PHSP;",
    "0.000086     D*+   D*-   rho+          PHOTOS PHSP;",
    "0.000086     D*0   anti-D*0   rho+          PHOTOS PHSP;",
    "0.000009    D+   anti-D0   pi0  pi0     PHOTOS PHSP;",
    "0.000009    D+   anti-D0   pi+  pi-     PHOTOS PHSP;",
    "0.000009    D+   D-   pi+  pi0     PHOTOS PHSP;",
    "0.000009    D0   anti-D0   pi+  pi0     PHOTOS PHSP;",
    "0.000009    D0   D-   pi+  pi+     PHOTOS PHSP;",
    "0.000026    D*+   anti-D0   pi0  pi0     PHOTOS PHSP;",
    "0.000026    D*+   anti-D0   pi+  pi-     PHOTOS PHSP;",
    "0.000026    D*+   D-   pi+  pi0     PHOTOS PHSP;",
    "0.000026    D*0   anti-D0   pi+  pi0     PHOTOS PHSP;",
    "0.000026    D*0   D-   pi+  pi+     PHOTOS PHSP;",
    "0.000026    D+   anti-D*0   pi0  pi0     PHOTOS PHSP;",
    "0.000026    D+   anti-D*0   pi+  pi-     PHOTOS PHSP;",
    "0.000026    D+   D*-   pi+  pi0     PHOTOS PHSP;",
    "0.000026    D0   anti-D*0   pi+  pi0     PHOTOS PHSP;",
    "0.000026    D0   D*-   pi+  pi+     PHOTOS PHSP;",
    "0.000077    D*+   anti-D*0   pi0  pi0     PHOTOS PHSP;",
    "0.000077    D*+   anti-D*0   pi+  pi-     PHOTOS PHSP;",
    "0.000077    D*+   D*-   pi+  pi0     PHOTOS PHSP;",
    "0.000077    D*0   anti-D*0   pi+  pi0     PHOTOS PHSP;",
    "0.000077    D*0   D*-   pi+  pi+     PHOTOS PHSP;",
    "#  b -> c (sc=) -> D D= K X    Sum = %",
    "# take Q2B into account; some values are taken from hep-ex/0207086",
    "#JB PDG07",
    "0.001370  anti-D0  D0  K+                 PHOTOS PHSP;",
    "# (from D-D0K+) JB PDG07",
    "0.001700  anti-D0  D+   K0                PHOTOS PHSP;",
    "0.000430   anti-D0  D0  K*+                PHOTOS PHSP;",
    "0.000430   anti-D0  D+   K*0               PHOTOS PHSP;",
    "# (from D*-D+K0) JB PDG07",
    "0.006500   anti-D*0 D0 K+                  PHOTOS PHSP;",
    "# (from D*-D0K+) JB PDG07",
    "0.003100   anti-D*0 D+ K0                  PHOTOS PHSP;",
    "# JB PDG07",
    "0.004700   anti-D0  D*0  K+                PHOTOS PHSP;",
    "# JB PDG07",
    "0.005200   anti-D0  D*+  K0                PHOTOS PHSP;",
    "0.001721   anti-D*0 D0   K*+               PHOTOS PHSP;",
    "0.000860   anti-D*0 D+   K*0               PHOTOS PHSP;",
    "0.001290   anti-D0  D*0  K*+               PHOTOS PHSP;",
    "0.001290   anti-D0  D*+  K*0               PHOTOS PHSP;",
    "# JB PDG07",
    "0.005300   anti-D*0  D*0  K+               PHOTOS PHSP;",
    "# JB PDG07",
    "0.007800   anti-D*0  D*+  K0               PHOTOS PHSP;",
    "0.002237   anti-D*0  D*0  K*+              PHOTOS PHSP;",
    "0.003011   anti-D*0  D*+  K*0              PHOTOS PHSP;",
    "0.000112   anti-D_10 D0  K+                 PHOTOS PHSP;",
    "0.000112   anti-D_10 D+   K0                PHOTOS PHSP;",
    "0.000112   anti-D0  D_10  K+                PHOTOS PHSP;",
    "0.000112   anti-D0  D_1+  K0                PHOTOS PHSP;",
    "0.000336   anti-D_10  D*0  K+               PHOTOS PHSP;",
    "0.000336   anti-D_10  D*+  K0               PHOTOS PHSP;",
    "0.000336   anti-D*0  D_10  K+               PHOTOS PHSP;",
    "0.000336   anti-D*0  D_1+  K0               PHOTOS PHSP;",
    "0.000112   anti-D_2*0 D0  K+                 PHOTOS PHSP;",
    "0.000112   anti-D_2*0 D+   K0                PHOTOS PHSP;",
    "0.000112   anti-D0  D_2*0  K+                PHOTOS PHSP;",
    "0.000112   anti-D0  D_2*+  K0                PHOTOS PHSP;",
    "0.000336   anti-D_2*0  D*0  K+               PHOTOS PHSP;",
    "0.000336   anti-D_2*0  D*+  K0               PHOTOS PHSP;",
    "0.000336   anti-D*0  D_2*0  K+               PHOTOS PHSP;",
    "0.000336   anti-D*0  D_2*+  K0               PHOTOS PHSP;",
    "0.000086   anti-D0  D0  K+  pi0               PHOTOS PHSP;",
    "0.000086   D-  D0  K+  pi+               PHOTOS PHSP;",
    "0.000086   anti-D0  D0  K0  pi+               PHOTOS PHSP;",
    "0.000086   anti-D0  D+  K+  pi-               PHOTOS PHSP;",
    "0.000086   anti-D0  D+  K0  pi0               PHOTOS PHSP;",
    "0.000086   D-  D+  K0  pi+               PHOTOS PHSP;",
    "0.000172   anti-D*0  D0  K+  pi0                  PHOTOS PHSP;",
    "0.000172   D*-  D0  K+  pi+                  PHOTOS PHSP;",
    "0.000172   anti-D*0  D0  K0  pi+                  PHOTOS PHSP;",
    "0.000172   anti-D*0  D+  K+  pi-                  PHOTOS PHSP;",
    "0.000172   anti-D*0  D+  K0  pi0                  PHOTOS PHSP;",
    "0.000172   D*-  D+  K0  pi+                  PHOTOS PHSP;",
    "0.000172   anti-D0  D*0  K+  pi0                  PHOTOS PHSP;",
    "0.000172   D-  D*0  K+  pi+                  PHOTOS PHSP;",
    "0.000172   anti-D0  D*0  K0  pi+                  PHOTOS PHSP;",
    "0.000172   anti-D0  D*+  K+  pi-                  PHOTOS PHSP;",
    "0.000172   anti-D0  D*+  K0  pi0                  PHOTOS PHSP;",
    "0.000172   D-  D*+  K0  pi+                  PHOTOS PHSP;",
    "0.000308   anti-D*0  D*0  K+  pi0                  PHOTOS PHSP;",
    "0.000309   D*-  D*0  K+  pi+                  PHOTOS PHSP;",
    "0.000308   anti-D*0  D*0  K0  pi+                  PHOTOS PHSP;",
    "0.000309   anti-D*0  D*+  K+  pi-                  PHOTOS PHSP;",
    "0.000308   anti-D*0  D*+  K0  pi0                  PHOTOS PHSP;",
    "0.000309   D*-  D*+  K0  pi+                  PHOTOS PHSP;",
    "0.000400   anti-D0  D0  K*+  pi0               PHOTOS PHSP;",
    "0.000400   D-  D0  K*+  pi+               PHOTOS PHSP;",
    "0.000400   anti-D0  D0  K*0  pi+               PHOTOS PHSP;",
    "0.000400   anti-D0  D+  K*+  pi-               PHOTOS PHSP;",
    "0.000400   anti-D0  D+  K*0  pi0               PHOTOS PHSP;",
    "0.000400   D-  D+  K*0  pi+               PHOTOS PHSP;",
    "0.000400   anti-D0  D0  K+  rho0               PHOTOS PHSP;",
    "0.000400   D-  D0  K+  rho+               PHOTOS PHSP;",
    "0.000400   anti-D0  D0  K0  rho+               PHOTOS PHSP;",
    "0.000400   anti-D0  D+  K+  rho-               PHOTOS PHSP;",
    "0.000400   anti-D0  D+  K0  rho0               PHOTOS PHSP;",
    "0.000400   D-  D+  K0  rho+               PHOTOS PHSP;",
    "# color suppressed mode",
    "0.000344   D-  D+  K+                         PHOTOS PHSP;",
    "0.000086   D-  D+  K*+                        PHOTOS PHSP;",
    "# JB PDG07",
    "0.001500   D*-  D+  K+                        PHOTOS PHSP;",
    "0.000344   D*-  D+  K*+                       PHOTOS PHSP;",
    "# (from D*-D+K+) JB PDG07",
    "0.001500   D-  D*+  K+                        PHOTOS PHSP;",
    "0.000344   D-  D*+  K*+                       PHOTOS PHSP;",
    "0.002409   D*-  D*+  K+                       PHOTOS PHSP;",
    "0.000774   D*-  D*+  K*+                      PHOTOS PHSP;",
    "0.000086   D_1-  D+  K+                         PHOTOS PHSP;",
    "0.000086   D-  D_1+  K+                         PHOTOS PHSP;",
    "0.000258   D_1-  D*+  K+                        PHOTOS PHSP;",
    "0.000258   D*-  D_1+  K+                        PHOTOS PHSP;",
    "0.000086   D_2*-  D+  K+                         PHOTOS PHSP;",
    "0.000086   D-  D_2*+  K+                         PHOTOS PHSP;",
    "0.000258   D_2*-  D*+  K+                        PHOTOS PHSP;",
    "0.000258   D*-  D_2*+  K+                        PHOTOS PHSP;",
    "# B->DX",
    "#    updated by Horii 20080131",
    "#  B+ -> anti-D0 n(pi)",
    "### PDG 2007",
    "0.00484  anti-D0   pi+                      PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0005   anti-D0  pi0  pi+                  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0134   rho+   anti-D0                     PHOTOS SVS;",
    "### PDG 2007 (non-resonanct)",
    "0.0046   anti-D0  pi-  pi+  pi+             PHOTOS PHSP;",
    "### PDG 2007 - assume excludes D0 a_1-",
    "0.0039   anti-D0  rho0 pi+                  PHOTOS PHSP;",
    "### PDG 2007",
    "0.00365   a_1+  anti-D0                      PHOTOS SVS;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   b_1+  anti-D0                      PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0000   anti-D0  pi-  pi+  pi+  pi0        PHOTOS PHSP;",
    "### PDG 2007",
    "0.0041   anti-D0  omega  pi+                PHOTOS PHSP;",
    "#  B+ -> anti-D*0 n(pi)",
    "### PDG 2007",
    "0.0052   anti-D*0 pi+                       PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0005   anti-D*0 pi0  pi+                  PHOTOS PHSP;",
    "### HELAMP parameters taken from ICHEP 98-852 (unchanged)",
    "### BF and HELAMP parameters updated in  PRD 67:112002,2003 (CLEO)",
    "### not yet in PDG - leave the same for now",
    "0.0155   anti-D*0  rho+               PHOTOS SVV_HELAMP 0.228 0.95 0.932 0 0.283 1.13;",
    "### PDG 2007 (assume non-resonant)",
    "0.0103   anti-D*0 pi-  pi+  pi+             PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0000   anti-D*0 rho0 pi+                  PHOTOS PHSP;",
    "### PDG 2007 - helicity amplitudes from factorization",
    "0.0190   anti-D*0  a_1+               PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   anti-D*0  b_1+               PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007",
    "0.0005   anti-D*0  pi+  pi0  pi0            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0005   anti-D*0  rho+  pi0                PHOTOS PHSP;",
    "### PDG 2007 (assume non-resonant)",
    "0.0180   anti-D*0  pi-  pi+  pi+  pi0       PHOTOS PHSP;",
    "### PDG 2007",
    "0.0045   anti-D*0  omega  pi+               PHOTOS PHSP;",
    "#  B+ -> D(*)- n(pi)",
    "### PDG 2003 is confusing - leave same for now",
    "0.0017  D-  pi+  pi+                  PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D-  pi0   pi+  pi+            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D-  rho+  pi+                 PHOTOS PHSP;",
    "### PDG 2003 is confusing - leave same for now",
    "0.00001  D*- pi+  pi+                  PHOTOS PHSP;",
    "### PDG 2007 (assume non-resonant)",
    "0.0150   D*- pi0   pi+  pi+            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D*- rho+  pi+                 PHOTOS PHSP;",
    "#  B+ -> anti-D**0 pi+",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (5.5 +- 0.9 10^-4)/(B(D_0*0 -> D- pi+) ~ 2/3)",
    "0.00083   anti-D_0*0 pi+                      PHOTOS PHSP;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (6.2 +- 1.2 10^-4)/(B(D_10 -> D*+ pi+) ~ 2/3)",
    "0.00093   anti-D_10 pi+                       PHOTOS SVS;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (4.1 +- 0.9 10^-4)/(B(D'_10 -> D*+ pi+) ~ 2/3)",
    "0.00062   anti-D'_10 pi+                      PHOTOS SVS;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (3.5 +- 0.6 10^-4)/(B(D_2*0 -> D-  pi+) ~ 0.46)",
    "##                 (2.0 +- 0.6 10^-4)/(B(D_2*0 -> D*+ pi+) ~ 0.21)",
    "## take average (7.6 + 9.5) 10^-4/2",
    "0.00085   anti-D_2*0 pi+                      PHOTOS STS;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00083   rho+ anti-D_0*0                     PHOTOS SVS;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00093   anti-D_10 rho+                      PHOTOS PHSP;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00062   anti-D'_10 rho+                     PHOTOS PHSP;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00085   anti-D_2*0 rho+                     PHOTOS PHSP;",
    "#  B+ -> anti-D(*)0 K(*)-",
    "### PDG 2007",
    "0.00040  anti-D0  K+                          PHOTOS PHSP;",
    "### PDG 2007",
    "0.00053  K*+ anti-D0                          PHOTOS SVS;",
    "### PDG 2007",
    "0.00042  anti-D*0 K+                          PHOTOS SVS;",
    "### PDG 2007 and helicity amplitudes same as D*0 rho-",
    "0.00081  anti-D*0 K*+                        PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### 2005 update (same as 2003)",
    "###  - assume same iso-spin violation in b -> c (s ubar) as b -> c (d ubar)",
    "###  - make b -> c (s ubar) sum to 2.44% (2.5% * 0.9764)",
    "###  - sum of b -> c (s ubar) BFs is 0.00206 (2005 -> same)",
    "###  - use PYTHIA 48 as per previous comments",
    "## Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.02234  u  anti-s  anti-c  u            PHOTOS PYTHIA 48;",
    "# mopping up",
    "### 2005 update (same as 2003)",
    "###  - make b -> c (d ubar) + b -> c (s ubar) sum to 44.4% (2005 -> 47.6%)",
    "###  - sum of BFs is 16.232% (2005 -> 15.194%)",
    "###  - use PYTHIA 48 as per previous comments",
    "## Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.31234  u  anti-d  anti-c  u            PHOTOS PYTHIA 48;",
    "#0.32406  u  anti-d  anti-c  u            PHOTOS PYTHIA 48;",
    "#0.31006  u  anti-d  anti-c  u            PHOTOS PYTHIA 48;",
    "# B->BaryonX",
    "# B- -> charmed baryons, total 6.1%",
    "#  Sigma_c X",
    "#  two body decay",
    "0.000020       anti-Lambda_c- Delta++            PHOTOS PHSP;",
    "0.000020       anti-Sigma_c-  Delta++            PHOTOS PHSP;",
    "0.000010       anti-Sigma_c0  Delta+             PHOTOS PHSP;",
    "0.000010       anti-Sigma_c0  p+                 PHOTOS PHSP;",
    "0.034000       anti-cd_1       uu_1         PHOTOS PYTHIA 23;",
    "0.012668       anti-cd_1       uu_1         PHOTOS PYTHIA 24;",
    "#  double charmed baryon Xi_c Sigma_c",
    "#    two body decay",
    "0.001000       anti-Xi_c-     Sigma_c++          PHOTOS PHSP;",
    "0.001000       anti-Xi_c0     Lambda_c+          PHOTOS PHSP;",
    "0.001000       anti-Xi_c0     Sigma_c+           PHOTOS PHSP;",
    "0.008500       anti-cs_0       cu_0         PHOTOS PYTHIA 23;",
    "#  cabibbo suppressed  Xi_c N",
    "#    two body decay",
    "0.000001       anti-Xi_c-     Delta++           PHOTOS PHSP;",
    "0.000001       anti-Xi_c0     p+                PHOTOS PHSP;",
    "0.001800       anti-cs_1       uu_1        PHOTOS PYTHIA 23;",
    "0.000450       anti-cs_1       uu_1        PHOTOS PYTHIA 24;",
    "#  cabibbo suppressed  Sigma_c Sigma_c",
    "#    two body decay",
    "0.000040        anti-Lambda_c-  Sigma_c++     PHOTOS PHSP;",
    "0.000040        anti-Sigma_c-   Sigma_c++     PHOTOS PHSP;",
    "0.000040        anti-Sigma_c0   Sigma_c+     PHOTOS PHSP;",
    "0.000400         anti-cd_0       cu_0         PHOTOS PYTHIA 23;",
    "Enddecay",
]  # type: List[str]

original_b_minus_decay_lines = [
    "Decay B-",
    "# B->DXlnu",
    "#     B- -> X_c l nu, sum = 11.35% (using: tau_B+/tau_B0 = 1.09)",
    "#     2005/04/02 H.Kakuno, Update using the following numbers:",
    "#       B->Xlnu     Br=10.9%               (HFAG summer 2004)",
    "#       B0->Dlnu    Br=2.14%,  rho_D=1.15  (HFAG summer 2004)",
    "#       B0->D*lnu   Br=5.33%               (HFAG summer 2004),",
    "#                    rho_A1=1.3            (hepex/0408027),",
    "#                    R1=1.18,  R2=0.71     (PRL76,3893)",
    "#       B->D**lnu   Br=3.04%          (CERN-EP/2001-050)",
    "#       B->D_1 lnu  Br=0.74%          (PDG 2004)",
    "#       B->D_2* lnu Br<0.65%          (PDG 2004)",
    "#       B->D pi lnu Br=1.54%          (PLB475, 407)",
    "#       B->D*pi lnu Br=1.86%          (PLB475, 407)",
    "#       referred also D0 preliminary result from CKM2005 for D**lnu",
    "0.0579   D*0       e-   anti-nu_e   PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0231   D0        e-   anti-nu_e   PHOTOS  HQET2 1.15;",
    "0.0081   D_10      e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0091   D_0*0     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0081   D'_10     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0039   D_2*0     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0002   D(2S)0    e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0005   D*(2S)0   e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0003   D*0  pi0  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0007   D*+  pi-  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0005   D0   pi0  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0011   D+   pi-  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0579   D*0       mu-  anti-nu_mu  PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0231   D0        mu-  anti-nu_mu  PHOTOS  HQET2 1.15;",
    "0.0081   D_10      mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0091   D_0*0     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0081   D'_10     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0039   D_2*0     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0002   D(2S)0    mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0005   D*(2S)0   mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0003   D*0  pi0  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0007   D*+  pi-  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0005   D0   pi0  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0011   D+   pi-  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "# B->DXtaunu",
    "#                    B- -> X_c tau nu, sum = 2.6% (using: tau_B+/tau_B0 = 1.09)",
    "#                    H.Kakuno 2004/04/02",
    "0.0154   D*0       tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0077   D0        tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0013   D_10      tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0016   D_0*0     tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0012   D'_10     tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0008   D_2*0     tau- anti-nu_tau         PHOTOS ISGW2;",
    "# B->ccX",
    "#	BF[%]		B0	B+",
    "#	eta_c		0.5034	0.5455",
    "#	eta_c(2S)	0.2015	0.2190",
    "#	J/psi		0.7539	0.8118",
    "#	psi(2S)		0.2967	0.3152",
    "#	chi_c0		0.2127	0.2291",
    "#	chi_c1		0.3228	0.3476",
    "#	chi_c2		0.1595	0.1718",
    "#	h_c		0.1345	0.1449",
    "#	X(3872)		0.0080	0.0080",
    "#	Y(3940)		0.0070	0.0071",
    "# -------------------------------------",
    "#	sum		2.6000	2.8000",
    "# B- --> charmonium X,  2.8000%",
    "# B -> cc= s",
    "# eta_c inclusive",
    "# PDG 2005: < 0.009",
    "# Guess 0.005455",
    "# B- --> eta_c K-",
    "# PDG 2005: 0.00103 +/- 0.00021",
    "0.001030   eta_c K- PHOTOS    PHSP;",
    "# B- --> K*- eta_c",
    "# Guess 0.001700",
    "0.001700   K*- eta_c  PHOTOS  SVS;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   eta_c anti-K0 pi- PHOTOS     PHSP;",
    "0.000000   eta_c K- pi0      PHOTOS     PHSP;",
    "0.000457   eta_c K- pi+ pi-   PHOTOS    PHSP;",
    "0.000457   eta_c K- pi0 pi0   PHOTOS    PHSP;",
    "0.000457   eta_c anti-K0 pi- pi0 PHOTOS PHSP;",
    "# baryonic decays",
    "0.000003  eta_c Lambda0 anti-p-      PHOTOS PHSP;",
    "0.000003  eta_c Sigma- anti-n0  PHOTOS PHSP;",
    "# Fill up by string fragmentation",
    "0.001348  eta_c anti-Xsu PHSP;",
    "# Sum of eta_c: 0.005455",
    "# eta_c(2S) Inclusive",
    "# guess 0.002190 (40% of eta_c)",
    "# B- --> eta_c(2S) K-",
    "# Guess 0.00036",
    "0.000360   eta_c(2S) K-  PHOTOS  PHSP;",
    "# B- --> eta_c(2S) K*-",
    "# Guess 0.00068",
    "0.000680   K*- eta_c(2S) PHOTOS   SVS;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   eta_c(2S) anti-K0 pi- PHOTOS     PHSP;",
    "0.000000   eta_c(2S) K- pi0      PHOTOS     PHSP;",
    "0.000202   eta_c(2S) K- pi+ pi-   PHOTOS    PHSP;",
    "0.000202   eta_c(2S) K- pi0 pi0   PHOTOS    PHSP;",
    "0.000202   eta_c(2S) anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.000544   eta_c(2S)  anti-Xsu  PHSP;",
    "# Sum of eta_c(2S): 0.002190",
    "# J/psi inclusive (direct)",
    "# PDG 2005:0.008118",
    "# B- --> J/psi K-",
    "# PDG 2003:  0.00099 +/- 0.00004",
    "0.000990   J/psi K-  PHOTOS     SVS;",
    "# B- --> J/psi K*-",
    "# PDG 2005:  0.00135 +/- 0.00010",
    "0.001350   J/psi K*-  PHOTOS    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B- --> J/psi K(1270)-",
    "# PDG 2005:  0.0018  +/- 0.0005",
    "0.001800   J/psi K_1-  PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B- --> J/psi K(1400)-",
    "# PDG 2005:  < 0.0005",
    "0.000000   J/psi K'_1-  PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# This is not part of K1(1270)?",
    "# B- --> J/psi K- pi+ pi-",
    "# PDG 2005:  0.00077 +/- 0.00020",
    "0.000770   J/psi K- pi+ pi- PHOTOS  PHSP;",
    "# Omit B- --> J/psi eta K-, because it comes from Psi(2S) K-.",
    "# B- --> J/psi phi K-",
    "# PDG 2005:  0.000052 +/- 0.000017",
    "0.000052   J/psi phi K- PHOTOS  PHSP;",
    "# B- --> J/psi pi-",
    "# PDG 2005:  0.000048 +/- 0.000006",
    "0.000048   J/psi pi-   PHOTOS   SVS;",
    "# B- --> J/psi rho-",
    "# PDG 2005:  < 0.00077",
    "# Guess 0.000017",
    "0.000017   J/psi rho-   PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B- --> J/psi a1(1260)-",
    "# PDG 2005: < 0.0012",
    "0.000000   J/psi a_1-   PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Belle 0.000050   Belle Nitesh K0(1430) (Kpi [93+-10]%)",
    "0.000050   J/psi K_0*-  PHOTOS  SVS;",
    "# Belle 0.000260   Belle Nitesh K2*(1430) (Kpi [49.9+-1.2]%)  STV not implemented?",
    "0.000260   J/psi K_2*-  PHOTOS  PHSP;",
    "# Belle ?   Belle Nitesh K*(1680) (Kpi [38.7 +- 2.5]%)",
    "0.000005   J/psi K''*-  PHOTOS  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Fill up inclusive rate",
    "# kpipi replaced by K1,",
    "# little evidence for Kpi other than K*, K*(1430)",
    "0.000000   J/psi anti-K0 pi- PHOTOS     PHSP;",
    "0.000000   J/psi K- pi0      PHOTOS     PHSP;",
    "0.001120   J/psi K- pi0 pi0   PHOTOS    PHSP;",
    "0.001120   J/psi anti-K0 pi- pi0 PHOTOS PHSP;",
    "# baryonic decays",
    "0.000012  J/psi Lambda0 anti-p-      PHOTOS PHSP;",
    "0.000012  J/psi Sigma- anti-n0  PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.000512   J/psi anti-Xsu PHSP;",
    "# Sum of J/psi: 0.008118",
    "# Psi(2S) inclusive",
    "# PDG 2005: 0.003152",
    "# B- --> Psi(2S) K-",
    "# PDG 2005: 0.00068 +/- 0.00004",
    "0.000680   psi(2S) K-   PHOTOS  SVS;",
    "# B- --> Psi(2S) K*(892)-",
    "# PDG 2003: 0.00092 +/- 0.00022",
    "# Belle-Conf 0349 (LP03): 0.000813 +/- 0.000077 +/- 0.000089",
    "0.000920   psi(2S) K*-  PHOTOS  SVV_HELAMP  PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# B- --> Psi(2S) K- pi+ pi-",
    "# PDG 2005: (1.9+-1.2)*10^-3, adjusted to match inclusive Br.",
    "0.001552   psi(2S) K- pi+ pi- PHOTOS PHSP;",
    "# Omit multi-body decays.",
    "#0.000000   psi(2S) anti-K0 pi-  PHOTOS PHSP;",
    "#0.000000   psi(2S) K- pi0 PHOTOS  PHSP;",
    "#0.000532   psi(2S) K- pi0 pi0 PHOTOS PHSP;",
    "#0.000533   psi(2S) anti-K0 pi- pi0 PHOTOS PHSP;",
    "# No room to put string fragmenataion.",
    "# Sum of Psi(2S): 0.003152",
    "# chi_c0 inclusive",
    "# Guess 0.2291",
    "# B- --> chi_c0 K-",
    "# PDG 2005: 0.000300 +- 0.000060",
    "0.000300   chi_c0 K-  PHOTOS  PHSP;",
    "# B- --> chi_c0 K*-",
    "# Guess 0.000700",
    "0.000700   K*- chi_c0  PHOTOS SVS;",
    "# Guess 0.001291  Fill up inclusive rate with low momentum",
    "#0.000118   chi_c0 anti-K0 pi- PHOTOS     PHSP;",
    "#0.000118   chi_c0 K- pi0       PHOTOS    PHSP;",
    "#0.000351   chi_c0 K- pi+ pi-    PHOTOS   PHSP;",
    "#0.000352   chi_c0 K- pi0 pi0    PHOTOS   PHSP;",
    "#0.000352   chi_c0 anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001291  chi_c0 anti-Xsu PHSP;",
    "# Sum of chi_c0: 0.2291",
    "# chi_c1 inclusive (direct)",
    "# PDG 2005: 0.003476",
    "# B- --> chi_c1 K-",
    "# PDG 2005: 0.00068 +/- 0.00011",
    "0.000680   chi_c1 K-  PHOTOS  SVS;",
    "# B- --> chi_c1 K*-",
    "# PDG 2005: < 0.0021",
    "# Weighted average Belle(PLB634(255)155) and BaBar(PRL94(2005)141801) is",
    "# (5.67+-0.97)*10^-4.",
    "0.000567   chi_c1 K*- PHOTOS   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Belle chi_c1 pi- (2.21+-0.42+-0.32)*10^-5 by Rajeev(BN876).",
    "0.000022   chi_c1 pi-  PHOTOS  SVS;",
    "# Fill up inclusive rate with low momentum  0.002207",
    "#0.000000   chi_c1 anti-K0 pi- PHOTOS     PHSP;",
    "#0.000000   chi_c1 K- pi0     PHOTOS      PHSP;",
    "#0.000735   chi_c1 K- pi+ pi-  PHOTOS     PHSP;",
    "#0.000736   chi_c1 K- pi0 pi0   PHOTOS    PHSP;",
    "#0.000736   chi_c1 anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation",
    "0.002207   chi_c1  anti-Xsu PHSP;",
    "# Sum of chi_c1: 0.003476",
    "# chi_c2 inclusive (direct)",
    "# PDG 2005: 0.001718",
    "# No real evidence for these decays",
    "# B- --> chi_c2 K-",
    "# Guess 0.000084",
    "0.000084   chi_c2 K-        PHOTOS       STS;",
    "# B- --> chi_c2 K*-",
    "# Guess 0.000126",
    "0.000126   chi_c2 K*-       PHOTOS       PHSP;",
    "# Fill up inclusive rate with low momentum",
    "#0.000000   chi_c2 anti-K0 pi-  PHOTOS    PHSP;",
    "#0.000000   chi_c2 K- pi0      PHOTOS     PHSP;",
    "#0.000503   chi_c2 K- pi+ pi-   PHOTOS    PHSP;",
    "#0.000502   chi_c2 K- pi0 pi0     PHOTOS  PHSP;",
    "#0.000503   chi_c2 anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001508    chi_c2 anti-Xsu  PHSP;",
    "# Sum of chi_c2: 0.001718",
    "# h_c Inclusive",
    "# Guess 0.001449",
    "# Guess 0.000200",
    "0.000200   h_c K-        PHOTOS       SVS;",
    "# Guess 0.000240",
    "0.000229   h_c K*-        PHOTOS      PHSP;",
    "# Guess 0.001020",
    "#0.000000   h_c anti-K0 pi-  PHOTOS    PHSP;",
    "#0.000000   h_c K- pi0       PHOTOS    PHSP;",
    "#0.000340   h_c K- pi+ pi-   PHOTOS    PHSP;",
    "#0.000340   h_c K- pi0 pi0    PHOTOS   PHSP;",
    "#0.000340   h_c anti-K0 pi- pi0 PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.001020   h_c anti-Xsu PHSP;",
    "# Sum of h_c: 0.001449",
    "# X(3872), register B- -> X(3872) K- only.",
    "# ?",
    "0.000080   X(3872) K- PHOTOS SVS;",
    "# Omit B -> X(3872) K*, because no significant signal (Karim's work BAM200509)",
    "# 0.000130   X(3872) K*- PHOTOS SVV_HELAMP PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# sum of X(3872) 0.000080",
    "# Register  B- -> Y(3940) K-",
    "0.000071  Y(3940)  K-  PHOTOS PHSP;",
    "# Sum of B- --> charmonium X: 0.0280",
    "# B->DsX",
    "#   b -> c (sc=) -> D Ds X    Sum =  8.9%",
    "#   BF(Ds->phipi)=4.38(Babar2005+PDG) is assumed.",
    "#   Updated on 29/01/2008",
    "#   based on PDG update (Ref: W.-M. Yao et al., J. Phys. G 33, 1 (2006) and 2007 partial update for the 2008)",
    "#   by Nikhil J Joshi (njoshi@bpost)",
    "## Upper Vertex Ds",
    "0.01000   D0   D_s-              	PHOTOS PHSP;",
    "0.00820   D*0  D_s-              	PHOTOS SVS;",
    "0.00760   D_s*- D0              	PHOTOS SVS;",
    "0.01710   D_s*- D*0                  	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00033   D_s- D_0*0                 	PHOTOS PHSP;",
    "0.00067   D_s*- D_0*0                 	PHOTOS SVS;",
    "0.00034   D'_10  D_s-                	PHOTOS SVS;",
    "0.00068   D'_10  D_s*-                	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00068   D_10  D_s-                  	PHOTOS SVS;",
    "0.00135   D_10  D_s*-                 	PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00181   D_2*0  D_s-                 	PHOTOS STS;",
    "0.00363   D_2*0  D_s*-                	PHOTOS PHSP;",
    "# DsJ(2317)=D_s0*-  and  DsJ(2457)=D'_s1-",
    "0.00073   D0 D_s0*-                       PHOTOS PHSP;",
    "0.00090   D*0 D_s0*-                      PHOTOS SVS;",
    "0.00310   D'_s1- D0                       PHOTOS SVS;",
    "0.01200   D'_s1- D*0                      PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00100    D_s-  D+  pi-                  PHOTOS PHSP;",
    "0.00200    D_s-  D0  pi0                  PHOTOS PHSP;",
    "0.00100    D_s*- D+  pi-                  PHOTOS PHSP;",
    "0.00200    D_s*- D0  pi0                  PHOTOS PHSP;",
    "0.00160    D_s-  D+  pi-  pi0             PHOTOS PHSP;",
    "0.00160    D_s-  D0  pi-  pi+             PHOTOS PHSP;",
    "0.00070    D_s-  D0  pi0  pi0             PHOTOS PHSP;",
    "0.00160    D_s*- D+  pi-  pi0             PHOTOS PHSP;",
    "0.00160    D_s*- D0  pi-  pi+             PHOTOS PHSP;",
    "0.00070    D_s*- D0  pi0  pi0             PHOTOS PHSP;",
    "## End of upper vertex",
    "## Lower vertex Ds",
    "0.00070   D_s+  pi-  K-                    PHOTOS PHSP;",
    "0.00096   D_s*+ pi-  K-                    PHOTOS PHSP;",
    "0.00100   D_s+  pi-  K-  pi0               PHOTOS PHSP;",
    "0.00100   D_s*+ pi-  K-  pi0               PHOTOS PHSP;",
    "0.00067  D_s+  pi-  anti-K0  pi-               PHOTOS PHSP;",
    "0.00067  D_s*+ pi-  anti-K0  pi-               PHOTOS PHSP;",
    "0.00018   D_s+  pi-  K-  pi0 pi0           PHOTOS PHSP;",
    "0.00018   D_s*+ pi-  K-  pi0 pi0           PHOTOS PHSP;",
    "0.00018   D_s+  pi-  K-  pi+ pi-           PHOTOS PHSP;",
    "0.00018   D_s*+ pi-  K-  pi+ pi-           PHOTOS PHSP;",
    "0.00018   D_s+  pi-  anti-K0  pi- pi0           PHOTOS PHSP;",
    "0.00018   D_s*+ pi-  anti-K0  pi- pi0           PHOTOS PHSP;",
    "# B- ->DDX",
    "# updated by H.Miyake 20021006",
    "# updated by H.Miyake 20031019",
    "# updated by J.Brodzicka (JB) 20080130 basing on PDG07",
    "# b-> c (dc=)",
    "# JB PDG07",
    "0.00042      D-   D0           PHOTOS PHSP;",
    "# JB PDG07",
    "0.00063      D*0  D-           PHOTOS SVS;",
    "# JB PDG07",
    "0.00039      D*-  D0           PHOTOS SVS;",
    "# JB PDG07",
    "0.00081      D*-  D*0          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "# fudge factor to tune DDX: ~0.860298",
    "# guess from neutral mode",
    "0.000018   D'_1-  D0           PHOTOS SVS;",
    "0.000018   D'_10  D-           PHOTOS SVS;",
    "0.000037   D'_1-  D*0          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000037   D'_10  D*-          PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000037   D_1-  D0            PHOTOS SVS;",
    "0.000037   D_10  D-            PHOTOS SVS;",
    "0.000075   D_1-  D*0           PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000075   D_10  D*-           PHOTOS SVV_HELAMP 0.616 0.0 0.743 0.0 0.262 0.0;",
    "0.000122   D_2*-  D0           PHOTOS STS;",
    "0.000122   D_2*0  D-           PHOTOS STS;",
    "0.000244   D_2*-  D*0          PHOTOS PHSP;",
    "0.000244   D_2*0  D*-          PHOTOS PHSP;",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00045   D_s1- D0                       PHOTOS SVS;",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00094   D_s1- D*0                      PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00036   D_s2*- D0                      PHOTOS STS;",
    "0.00036   D_s2*- D*0                     PHOTOS PHSP;",
    "# JB hep-ex/0707.3491",
    "0.00113   D_sj(2700)- D0                 PHOTOS SVS;",
    "0.00036   D_sj(2700)- D*0                PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "# JB PDG07",
    "0.00049   psi(3770) K-   PHOTOS SVS;",
    "0.00068   psi(3770) K*-  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.00064   psi(3770) anti-K0 pi- PHOTOS PHSP;",
    "0.00031   psi(3770) K- pi0 PHOTOS PHSP;",
    "0.00093   psi(3770) K- pi0 pi0 PHOTOS PHSP;",
    "0.00093   psi(3770) anti-K0 pi- pi0 PHOTOS PHSP;",
    "# JB fudge factor to tune DDX: ~0.737246",
    "0.001548    psi(4040) K-   PHOTOS SVS;",
    "0.002322    psi(4040) K*-  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000811   psi(4040) anti-K0 pi- PHOTOS PHSP;",
    "0.000405   psi(4040) K- pi0 PHOTOS PHSP;",
    "0.001548    psi(4160) K-   PHOTOS SVS;",
    "0.002322    psi(4160) K*-  PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000811   psi(4160) anti-K0 pi- PHOTOS PHSP;",
    "0.000405   psi(4160) K- pi0 PHOTOS PHSP;",
    "#not included in qq98",
    "0.000026     D-   D0   pi0     PHOTOS PHSP;",
    "0.000026     D-   D+   pi-     PHOTOS PHSP;",
    "0.000026     anti-D0   D0   pi-     PHOTOS PHSP;",
    "0.000086     D*-   D0   pi0     PHOTOS PHSP;",
    "0.000086     D*-   D+   pi-     PHOTOS PHSP;",
    "0.000086     anti-D*0   D0   pi-     PHOTOS PHSP;",
    "0.000086     D-   D*0   pi0     PHOTOS PHSP;",
    "0.000086     D-   D*+   pi-     PHOTOS PHSP;",
    "0.000086     anti-D0   D*0   pi-     PHOTOS PHSP;",
    "0.000258     D*-   D*0   pi0     PHOTOS PHSP;",
    "0.000258     D*-   D*+   pi-     PHOTOS PHSP;",
    "0.000258     anti-D*0   D*0   pi-     PHOTOS PHSP;",
    "0.000009     D-   D0   rho0     PHOTOS PHSP;",
    "0.000009     D-   D+   rho-     PHOTOS PHSP;",
    "0.000009     anti-D0   D0   rho-     PHOTOS PHSP;",
    "0.000026     D*-   D0   rho0     PHOTOS PHSP;",
    "0.000026     D*-   D+   rho-     PHOTOS PHSP;",
    "0.000026     anti-D*0   D0   rho-     PHOTOS PHSP;",
    "0.000026     D-   D*0   rho0     PHOTOS PHSP;",
    "0.000026     D-   D*+   rho-     PHOTOS PHSP;",
    "0.000026     anti-D0   D*0   rho-     PHOTOS PHSP;",
    "0.000086     D*-   D*0   rho0     PHOTOS PHSP;",
    "0.000086     D*-   D*+   rho-     PHOTOS PHSP;",
    "0.000086     anti-D*0   D*0   rho-     PHOTOS PHSP;",
    "0.000009     D-   D0   pi0  pi0    PHOTOS PHSP;",
    "0.000009     D-   D0   pi+  pi-    PHOTOS PHSP;",
    "0.000009     D-   D+   pi-  pi0    PHOTOS PHSP;",
    "0.000009     anti-D0   D0   pi-  pi0    PHOTOS PHSP;",
    "0.000009     anti-D0   D+   pi-  pi-    PHOTOS PHSP;",
    "0.000026     D*-   D0   pi0  pi0    PHOTOS PHSP;",
    "0.000026     D*-   D0   pi+  pi-    PHOTOS PHSP;",
    "0.000026     D*-   D+   pi-  pi0    PHOTOS PHSP;",
    "0.000026     anti-D*0   D0   pi-  pi0    PHOTOS PHSP;",
    "0.000026     anti-D*0   D+   pi-  pi-    PHOTOS PHSP;",
    "0.000026     D-   D*0   pi0  pi0    PHOTOS PHSP;",
    "0.000026     D-   D*0   pi+  pi-    PHOTOS PHSP;",
    "0.000026     D-   D*+   pi-  pi0    PHOTOS PHSP;",
    "0.000026     anti-D0   D*0   pi-  pi0    PHOTOS PHSP;",
    "0.000026     anti-D0   D*+   pi-  pi-    PHOTOS PHSP;",
    "0.000077     D*-   D*0   pi0  pi0    PHOTOS PHSP;",
    "0.000077     D*-   D*0   pi+  pi-    PHOTOS PHSP;",
    "0.000077     D*-   D*+   pi-  pi0    PHOTOS PHSP;",
    "0.000077     anti-D*0   D*0   pi-  pi0    PHOTOS PHSP;",
    "0.000077     anti-D*0   D*+   pi-  pi-    PHOTOS PHSP;",
    "#  b -> c (sc=) -> D D= K X    Sum =  %",
    "# take Q2B into account; some values are taken from hep-ex/0207086",
    "#JB PDG07",
    "0.001370   D0  anti-D0  K-                 PHOTOS PHSP;",
    "# (from D+D0BK-) JB PDG07",
    "0.001700   D0  D-   anti-K0                PHOTOS PHSP;",
    "0.000430   D0  anti-D0  K*-                PHOTOS PHSP;",
    "0.000430   D0  D-   anti-K*0               PHOTOS PHSP;",
    "# (from D*+D-K0) JB PDG07",
    "0.006500   D*0 anti-D0  K-                 PHOTOS PHSP;",
    "# (from D*+D0BK-) JB PDG07",
    "0.003100   D*0 D-   anti-K0                PHOTOS PHSP;",
    "# JB PDG07",
    "0.004700   D0  anti-D*0  K-                PHOTOS PHSP;",
    "# JB PDG07",
    "0.005200   D0  D*-  anti-K0                PHOTOS PHSP;",
    "0.001721   D*0 anti-D0  K*-                PHOTOS PHSP;",
    "0.000860   D*0 D-   anti-K*0               PHOTOS PHSP;",
    "0.001290   D0  anti-D*0  K*-               PHOTOS PHSP;",
    "0.001290   D0  D*-  anti-K*0               PHOTOS PHSP;",
    "# JB PDG07",
    "0.005300   D*0  anti-D*0  K-               PHOTOS PHSP;",
    "# JB PDG07",
    "0.007800   D*0  D*-  anti-K0               PHOTOS PHSP;",
    "0.002237   D*0  anti-D*0  K*-              PHOTOS PHSP;",
    "0.003011   D*0  D*-  anti-K*0              PHOTOS PHSP;",
    "0.000112   D_10 anti-D0  K-                 PHOTOS PHSP;",
    "0.000112   D_10 D-   anti-K0                PHOTOS PHSP;",
    "0.000112   D0  anti-D_10  K-                PHOTOS PHSP;",
    "0.000112   D0  D_1-  anti-K0                PHOTOS PHSP;",
    "0.000336   D_10  anti-D*0  K-               PHOTOS PHSP;",
    "0.000336   D_10  D*-  anti-K0               PHOTOS PHSP;",
    "0.000336   D*0  anti-D_10  K-               PHOTOS PHSP;",
    "0.000336   D*0  D_1-  anti-K0               PHOTOS PHSP;",
    "0.000112   D_2*0 anti-D0  K-                 PHOTOS PHSP;",
    "0.000112   D_2*0 D-   anti-K0                PHOTOS PHSP;",
    "0.000112   D0  anti-D_2*0  K-                PHOTOS PHSP;",
    "0.000112   D0  D_2*-  anti-K0                PHOTOS PHSP;",
    "0.000336   D_2*0  anti-D*0  K-               PHOTOS PHSP;",
    "0.000336   D_2*0  D*-  anti-K0               PHOTOS PHSP;",
    "0.000336   D*0  anti-D_2*0  K-               PHOTOS PHSP;",
    "0.000336   D*0  D_2*-  anti-K0               PHOTOS PHSP;",
    "0.000086   D0  anti-D0  K-  pi0               PHOTOS PHSP;",
    "0.000086   D+  anti-D0  K-  pi-               PHOTOS PHSP;",
    "0.000086   D0  anti-D0  anti-K0  pi-               PHOTOS PHSP;",
    "0.000086   D0  D-  K-  pi+               PHOTOS PHSP;",
    "0.000086   D0  D-  anti-K0  pi0               PHOTOS PHSP;",
    "0.000086   D+  D-  anti-K0  pi-               PHOTOS PHSP;",
    "0.000172   D*0  anti-D0  K-  pi0               PHOTOS PHSP;",
    "0.000172   D*+  anti-D0  K-  pi-               PHOTOS PHSP;",
    "0.000172   D*0  anti-D0  anti-K0  pi-               PHOTOS PHSP;",
    "0.000172   D*0  D-  K-  pi+               PHOTOS PHSP;",
    "0.000172   D*0  D-  anti-K0  pi0               PHOTOS PHSP;",
    "0.000172   D*+  D-  anti-K0  pi-               PHOTOS PHSP;",
    "0.000172   D0  anti-D*0  K-  pi0               PHOTOS PHSP;",
    "0.000172   D+  anti-D*0  K-  pi-               PHOTOS PHSP;",
    "0.000172   D0  anti-D*0  anti-K0  pi-               PHOTOS PHSP;",
    "0.000172   D0  D*-  K-  pi+               PHOTOS PHSP;",
    "0.000172   D0  D*-  anti-K0  pi0               PHOTOS PHSP;",
    "0.000172   D+  D*-  anti-K0  pi-               PHOTOS PHSP;",
    "0.000308   D*0  anti-D*0  K-  pi0               PHOTOS PHSP;",
    "0.000309   D*+  anti-D*0  K-  pi-               PHOTOS PHSP;",
    "0.000308   D*0  anti-D*0  anti-K0  pi-               PHOTOS PHSP;",
    "0.000309   D*0  D*-  K-  pi+               PHOTOS PHSP;",
    "0.000308   D*0  D*-  anti-K0  pi0               PHOTOS PHSP;",
    "0.000309   D*+  D*-  anti-K0  pi-               PHOTOS PHSP;",
    "0.000400   D0  anti-D0  K*-  pi0               PHOTOS PHSP;",
    "0.000400   D+  anti-D0  K*-  pi-               PHOTOS PHSP;",
    "0.000400   D0  anti-D0  anti-K*0  pi-               PHOTOS PHSP;",
    "0.000400   D0  D-  K*-  pi+               PHOTOS PHSP;",
    "0.000400   D0  D-  anti-K*0  pi0               PHOTOS PHSP;",
    "0.000400   D+  D-  anti-K*0  pi-               PHOTOS PHSP;",
    "0.000400   D0  anti-D0  K-  rho0               PHOTOS PHSP;",
    "0.000400   D+  anti-D0  K-  rho-               PHOTOS PHSP;",
    "0.000400   D0  anti-D0  anti-K0  rho-               PHOTOS PHSP;",
    "0.000400   D0  D-  K-  rho+               PHOTOS PHSP;",
    "0.000400   D0  D-  anti-K0  rho0               PHOTOS PHSP;",
    "0.000400   D+  D-  anti-K0  rho-               PHOTOS PHSP;",
    "# color suppressed mode",
    "0.000344   D+  D-  K-                         PHOTOS PHSP;",
    "0.000086   D+  D-  K*-                        PHOTOS PHSP;",
    "# JB PDG07",
    "0.001500   D*+  D-  K-                        PHOTOS PHSP;",
    "0.000344   D*+  D-  K*-                       PHOTOS PHSP;",
    "# (from D*-D+K+) JB PDG07",
    "0.001500   D+  D*-  K-                        PHOTOS PHSP;",
    "0.000344   D+  D*-  K*-                       PHOTOS PHSP;",
    "0.002409   D*+  D*-  K-                       PHOTOS PHSP;",
    "0.000774   D*+  D*-  K*-                      PHOTOS PHSP;",
    "0.000086   D_1+  D-  K-                         PHOTOS PHSP;",
    "0.000086   D+  D_1-  K-                         PHOTOS PHSP;",
    "0.000258   D_1+  D*-  K-                        PHOTOS PHSP;",
    "0.000258   D*+  D_1-  K-                        PHOTOS PHSP;",
    "0.000086   D_2*+  D-  K-                         PHOTOS PHSP;",
    "0.000086   D+  D_2*-  K-                         PHOTOS PHSP;",
    "0.000258   D_2*+  D*-  K-                        PHOTOS PHSP;",
    "0.000258   D*+  D_2*-  K-                        PHOTOS PHSP;",
    "# B->DX",
    "#    updated by Horii 20080131",
    "#  B- -> D0 n(pi)",
    "### PDG 2007",
    "0.00484  D0   pi-                      PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0005   D0  pi0  pi-                  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0134   rho-   D0                     PHOTOS SVS;",
    "### PDG 2007 (non-resonanct)",
    "0.0046   D0  pi+  pi-  pi-             PHOTOS PHSP;",
    "### PDG 2007 - assume excludes D0 a_1-",
    "0.0039   D0  rho0 pi-                  PHOTOS PHSP;",
    "### PDG 2007",
    "0.00365   a_1- D0                       PHOTOS SVS;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   b_1- D0                       PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0000   D0   pi+  pi-  pi-  pi0       PHOTOS PHSP;",
    "### PDG 2007",
    "0.0041   D0  omega  pi-                PHOTOS PHSP;",
    "#  B- -> D*0 n(pi)",
    "### PDG 2007",
    "0.0052   D*0 pi-                       PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0005   D*0 pi0  pi-                  PHOTOS PHSP;",
    "### HELAMP parameters taken from ICHEP 98-852 (unchanged)",
    "### BF and HELAMP parameters updated in  PRD 67:112002,2003 (CLEO)",
    "### not yet in PDG - leave the same for now",
    "0.0155   D*0  rho-               PHOTOS SVV_HELAMP 0.228 0.95 0.932 0 0.283 1.13;",
    "### PDG 2007 (assume non-resonant)",
    "0.0103   D*0 pi+  pi-  pi-             PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0000   D*0 rho0 pi-                  PHOTOS PHSP;",
    "### PDG 2007 - helicity amplitudes from factorization",
    "0.0190   D*0 a_1-                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   D*0 b_1-                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007",
    "0.0005   D*0 pi-   pi0  pi0           PHOTOS  PHSP;",
    "### not in PDG 2007",
    "0.0005   D*0 rho-  pi0                 PHOTOS PHSP;",
    "### PDG 2007 (assume non-resonant)",
    "0.0180   D*0  pi+  pi-  pi-  pi0       PHOTOS PHSP;",
    "### PDG 2007",
    "0.0045   D*0  omega  pi-               PHOTOS PHSP;",
    "#  B- -> D(*)+ n(pi)",
    "### PDG 2003 is confusing - leave same for now",
    "0.0017  D+  pi-  pi-                  PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D+  pi0   pi-  pi-            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D+  rho-  pi-                 PHOTOS PHSP;",
    "### PDG 2003 is confusing - leave same for now",
    "0.00001  D*+ pi-  pi-                  PHOTOS PHSP;",
    "### PDG 2007 (assume non-resonant)",
    "0.0150   D*+ pi0   pi-  pi-            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0020   D*+ rho-  pi-                 PHOTOS PHSP;",
    "#  B- -> D**0 pi-",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (5.5 +- 0.9 10^-4)/(B(D_0*0 -> D+ pi-) ~ 2/3)",
    "0.00083   D_0*0 pi-                      PHOTOS PHSP;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (6.2 +- 1.2 10^-4)/(B(D_10 -> D*+ pi-) ~ 2/3)",
    "0.00093   D_10 pi-                       PHOTOS SVS;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (4.1 +- 0.9 10^-4)/(B(D'_10 -> D*+ pi-) ~ 2/3)",
    "0.00062   D'_10 pi-                      PHOTOS SVS;",
    "### not in PDG 2003 - leave same",
    "## BELLE-CONF-0235 (3.5 +- 0.6 10^-4)/(B(D_2*0 -> D+  pi-) ~ 0.46)",
    "##                 (2.0 +- 0.6 10^-4)/(B(D_2*0 -> D*+ pi-) ~ 0.21)",
    "## take average (7.6 + 9.5) 10^-4/2",
    "0.00085   D_2*0 pi-                      PHOTOS STS;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00083   rho- D_0*0                     PHOTOS SVS;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00093   D_10 rho-                      PHOTOS PHSP;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00062   D'_10 rho-                     PHOTOS PHSP;",
    "### not in PDG 2003 - take same value as D**0 pi-",
    "0.00085   D_2*0 rho-                     PHOTOS PHSP;",
    "#  B- -> D(*)0 K(*)-",
    "### PDG 2007",
    "0.00040  D0  K-                          PHOTOS PHSP;",
    "### PDG 2007",
    "0.00053  K*- D0                          PHOTOS SVS;",
    "### PDG 2007",
    "0.00042  D*0 K-                          PHOTOS SVS;",
    "### PDG 2007 and helicity amplitudes same as D*0 rho-",
    "0.00081  D*0 K*-                        PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### 2005 update (same as 2003)",
    "###  - assume same iso-spin violation in b -> c (s ubar) as b -> c (d ubar)",
    "###  - make b -> c (s ubar) sum to 2.44% (2.5% * 0.9764)",
    "###  - sum of b -> c (s ubar) BFs is 0.00206 (2005 -> same)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.02234  anti-u  s  c  anti-u            PHOTOS PYTHIA 48;",
    "# mopping up",
    "### 2005 update (same as 2003)",
    "###  - include iso-spin violating effect",
    "###  - make b -> c (d ubar) + b -> c (s ubar) sum to 44.4% (2005 -> 47.6%)",
    "###  - sum of BFs is 16.232% (2005 -> 15.194%)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.31234  anti-u  d  c  anti-u            PHOTOS PYTHIA 48;",
    "#0.32406  anti-u  d  c  anti-u            PHOTOS PYTHIA 48;",
    "#0.31006  anti-u  d  c  anti-u            PHOTOS PYTHIA 48;",
    "# B->BaryonX",
    "# B- -> charmed baryons, total 6.1%",
    "#  Sigma_c X",
    "#  two body decay",
    "0.000020       Lambda_c+ anti-Delta--            PHOTOS PHSP;",
    "0.000020       Sigma_c+  anti-Delta--            PHOTOS PHSP;",
    "0.000010       Sigma_c0  anti-Delta-             PHOTOS PHSP;",
    "0.000010       Sigma_c0  anti-p-                 PHOTOS PHSP;",
    "0.034000       cd_1       anti-uu_1         PHOTOS PYTHIA 23;",
    "0.012668       cd_1       anti-uu_1         PHOTOS PYTHIA 24;",
    "#  double charmed baryon Xi_c Sigma_c",
    "#    two body decay",
    "0.001000       Xi_c+     anti-Sigma_c--          PHOTOS PHSP;",
    "0.001000       Xi_c0     anti-Lambda_c-          PHOTOS PHSP;",
    "0.001000       Xi_c0     anti-Sigma_c-          PHOTOS PHSP;",
    "0.008500       cs_0       anti-cu_0         PHOTOS PYTHIA 23;",
    "#  cabibbo suppressed  Xi_c N",
    "#    two body decay",
    "0.000001       Xi_c+     anti-Delta--           PHOTOS PHSP;",
    "0.000001       Xi_c0     anti-p-                PHOTOS PHSP;",
    "0.001800       cs_1       anti-uu_1        PHOTOS PYTHIA 23;",
    "0.000450       cs_1       anti-uu_1        PHOTOS PYTHIA 24;",
    "#  cabibbo suppressed  Sigma_c Sigma_c",
    "#    two body decay",
    "0.000040        Lambda_c+   anti-Sigma_c--     PHOTOS PHSP;",
    "0.000040        Sigma_c+   anti-Sigma_c--     PHOTOS PHSP;",
    "0.000040        Sigma_c0    anti-Sigma_c-     PHOTOS PHSP;",
    "0.000400         cd_0       anti-cu_0         PHOTOS PYTHIA 23;",
    "Enddecay",
]  # type: List[str]
